

#include "iml_file.h"
#include "iml_rofn.h"
#include "mvl_macv.h"

		

//============================================================================
//
//============================================================================
#ifdef _Windows
int run_experiment(void) {
#else
int main () {
#endif

//======== EXAMPLE  CODE FOR CREATING AND INITIALIZING mac_video_manager

	
try {  
	// ALL THE CODE HERE IS SET FOR CATCHING EXCEPTIONS 
	// WITH AN EMPTY catch() AT THE BOTTOM.  

	// MAKE A VIDEO MANAGER OBJECT AND INITIALIZE IT
	// I'LL PROVIDE AND USE MORE ERROR CHECKING LATER . . .
	IML_USINT MaxSequenceCnt = 32;
	IML_USINT MaxBlitCnt = 64;
	IML_USINT GroupCount = 8;
	IML_ULINT ExitOnErr = TRUE;
	mvl_video_manager MacVideo(MaxSequenceCnt,MaxBlitCnt, GroupCount, ExitOnErr);
	
	MacVideo.initialize_screen(0);
//	MacVideo.use_dgp_calibration();
	MacVideo.set_mvl_calibration(MacVideo.MaxDac, 2, 0, 100);
	MacVideo.set_attenuator_red_steps(0.0);
	MacVideo.use_mvl_calibration();
	MacVideo.set_display_CLUT(0.205,1.0);

	IML_LINT FrameCount=8;
	IML_LINT  XSize= 256,  YSize= 256;
	IML_USINT Duration=8; // number of "DwellTime" frames to show
	IML_USINT DwellTime = 10;
	IML_FLOAT ContrastRatio = 1.0;
	IML_FLOAT ModulationContrast = 0.4;
	
	IML_USINT CosPeriod = 64; // in pixels
	IML_USINT TexturePeriod = CosPeriod/2; // in pixels
	
	
	IML_FLOAT TextureContrast = 0.9;
	
	IML_USINT RED_CLUT_ENTRY = 1;
	IML_USINT GREEN_CLUT_ENTRY = 254;
	// GRAB A COPY OF THE CURRENT CLUT
	mvl_clut &CLUTable = *MacVideo.get_clut_copy();
	
	// SET RED AND GREEN BAR COLOR ENTRIES
	IML_LINT Green=127;
	IML_LINT Red = 210;
	CLUTable.set_clut_entry(GREEN_CLUT_ENTRY, 0, Green, 0);	
	CLUTable.set_clut_entry(RED_CLUT_ENTRY, Red,   0, 0);	
	MacVideo.make_default_clut(CLUTable);
	

	// MAKE THE 3D IMAGE SEQUENCE
	iml_ubyte_image MotionImage( iml_ld_index(XSize, YSize, FrameCount+1) );
	if( !MotionImage.valid_memory() ) {
		iml_error("Attentional Motion","Could not allocate MotionImage!");
		return(0);
		} // end if

	// WE USE THE LAST FRAME AS TEMPORARY MEMORY BEFORE SETTING IT UP AS A NOISE MASK.
	// MAKE AN ALIAS TO THE THREE-D IMAGE; WE HAVE TO CHEAT SO WE CAN GET SIGNED BYTES. :-)
	iml_byte_image NoiseFrame( *((iml_byte_image*)&MotionImage) );	
	NoiseFrame.set_hyperplane_size(2);			// MAKE IT TWO-D
	NoiseFrame.select_hyperplane(FrameCount);	// POINT TO LAST FRAME

	mvl_clock Timer;
		
	uniform_rv Uniform; // this is for the carrier
	Uniform.seed(Timer.seed());
		
	iml_ld_index TSize(XSize);
	iml_float_image Buffer( TSize );
	if( !Buffer.valid_memory() ) {
		iml_error("Attentional Motion","Could not allocate Buffer!");
		return(0);
		} // end if
	
	
	iml_cos_rofn Cos(1);	// a one D cosine
	Cos.set_center(IML_2D_X, XSize/2.0 + 0.5);	
	Cos.set_phase(0);    // in degrees
	Cos.set_period(IML_2D_X, CosPeriod);   // pixels/cycle
	
	// NOW THE COS IN THE DRIFTER
	Cos.set_magnitude(ModulationContrast);
	
	// WE WILL CREATE THE IMAGE ONE SLICE AT TIME, SO RESTRICT TO 2D SUBSPACE
	MotionImage.set_hyperplane_size(2);
	
	Timer.start();
	// NOW, ONE SLICE AT A TIME . . .
	for(IML_ULINT K=0; K<FrameCount; K++) {

		// SET CURRENT SLICE
		MotionImage.select_hyperplane(K);

		if( K&0x01 ) { // IF THIS IS AN ODD NUMBERED FRAME
		
			// MAKE A FRESH FRAME OF ZERO-MEAN CARRIER NOISE (RANDOM - ALL PIXELS);
			for(IML_LINT Q=0; Q<(XSize*(long)YSize); Q++) {
				IML_LINT TempJunk = (Uniform.sample()>0.5) ? -126 : 126;
				NoiseFrame(Q) = TempJunk;
				}
				
			// construct one-d pedestal plus drifter
			Cos.set_phase(90 + 90*K);
			Buffer  = 0.5;
			Buffer += Cos;
			
			// MODULATE THE NOISE IN THE FRAME AND STORE IN APPROPRIATE PLACE.
			for(IML_LINT I=0; I<(XSize*(long)YSize); I++) {
				MotionImage(I) = 127 + NoiseFrame(I) * Buffer(I%XSize);
				}
				
//MacVideo.display_image(	MotionImage, iml_screen_position(50,50));
//Timer.wait(0.5);
		} else { // an even numbered frame

/***** FOR LUMINANCE TEXTURE
			// MAKE A FRESH FRAME OF Texture
			for(IML_LINT Y=0; Y < YSize; Y++) {
				for(IML_LINT X=0; X < XSize; X++) {
					IML_FLOAT Sign = ( ((X/TexturePeriod) + (K/2)&0x01) & 0x01 ) ? -1.0 : 1.0;
//					IML_FLOAT Sign = ( ((X/TexturePeriod) + 1) & 0x01 ) ? -1.0 : 1.0;
					MotionImage(X,Y) = 127 + Sign*TextureContrast*127*(Uniform.sample()>0.5);
					} // end for
				} // end for
*******/

		// MAKE A FRESH FRAME OF BARS
		for(IML_LINT Y=0; Y < YSize; Y++) {
			for(IML_LINT X=0; X < XSize; X++) {
				IML_UBYTE Color = ( ((X/TexturePeriod) + (K/2)&0x01) & 0x01 ) ? RED_CLUT_ENTRY : GREEN_CLUT_ENTRY;
				MotionImage(X,Y) = Color;
				} // end for
			} // end for


//MacVideo.display_image(	MotionImage, iml_screen_position(50,50+YSize));
//Timer.wait(0.5);
			} // end else

		} // end for
	
	// FINALLY, SET UP THE LAST FRAME AS A NOISE MASK
	for(IML_LINT Q=0; Q<(XSize*(long)YSize); Q++) {
		IML_FLOAT Temp24 = (Uniform.sample() > 0.5) ? 2 : 253;
		NoiseFrame(Q) = Temp24;
		}

	IML_FLOAT Time = Timer.time();
	fprintf(stderr,"frame generation %f\n",Time/FrameCount);

	// RESET SUBSPACE SIZE TO 3D FOR MAC VIDEO STORAGE
	MotionImage.set_hyperplane_size(3);
	
	// STORE THE IMAGE
	IML_USINT ImSeqID = MacVideo.store_image_sequence(MotionImage);
	
	IML_LINT TotalDisplayFrames = Duration*DwellTime;
	iml_imageframe_sequence Sequence(TotalDisplayFrames + 1); // ADD ONE FOR MASK
	for(long K=0; K< TotalDisplayFrames; K++) {
		long Temp = (K/DwellTime)%FrameCount;
		Sequence(K) = Temp;
		}
	// REMEMBER TO DISPLAY THE NOISE MASK.
	Sequence(TotalDisplayFrames) = FrameCount; 
	
	// DISPLAY THE MOVIE IN THE MIDDLE OF THE SCREEN.
	mvl_screen_position Position(	(MacVideo.XResolution-XSize)/2,
								 	(MacVideo.YResolution-YSize)/2 );
	
	// MAKE THE BLIT SEQUENCE
	IML_USINT BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, Position, Sequence);

	MacVideo.play_blit_sequence(BlitSeqID);
	char Buff[128]; IML_LINT Delta=5;
	do {

		MacVideo.play_blit_sequence(BlitSeqID);
		
		fprintf(stderr,"Type 'q' to quit; anything else to repeat. >"); fflush(stderr); scanf("%s", Buff);
		} while(Buff[0] != 'q');

	} // end try
	
catch ( ... ) {
// THIS IS JUST TO ALLOW TERMINATION OF THE PROGRAM AND CLEAN UP OF STUFF.
	}

	return(0);

} // end main


