
// INCLUDE THE PROPER FILES
//#include "iml_file.h"
//#include "iml_rofn.h"
#include "mvl_macv.h"

// FUNCTION PROTOTYPES
void filter_dynamic_noise(mvl_video_manager & MacVideo,
		IML_LINT  XSize, IML_LINT YSize, IML_LINT SeqLength, 
		IML_USINT Repeats, mvl_screen_position IML_REF Position);
		
//============================================================================
//
//============================================================================
int main () {
	
try {  
	// ALL THE CODE HERE IS SET FOR CATCHING EXCEPTIONS 
	// WITH AN EMPTY catch() AT THE BOTTOM.  

	// MAKE A VIDEO MANAGER OBJECT AND INITIALIZE IT
	// I'LL PROVIDE AND USE MORE ERROR CHECKING LATER . . .
	IML_USINT MaxSequenceCnt = 32;
	IML_USINT MaxBlitCnt = 64;
	IML_USINT GroupCount = 8;
	IML_ULINT ExitOnErr = TRUE;
	mvl_video_manager MacVideo(MaxSequenceCnt,MaxBlitCnt, GroupCount, ExitOnErr);
	
	MacVideo.initialize_screen(0);
//	MacVideo.use_dgp_calibration();
	MacVideo.set_mvl_calibration(MacVideo.MaxDac, 2, 0, 100);
	MacVideo.set_attenuator_red_steps(0.0);
	MacVideo.use_mvl_calibration();
	MacVideo.set_display_CLUT(0.5,1.0);

	IML_LINT SeqLength = 32;
	IML_LINT  XSize= 100,  YSize= 100;
	IML_USINT Repeats=20;
	filter_dynamic_noise(MacVideo, XSize, YSize, SeqLength, Repeats, 
			mvl_screen_position( (MacVideo.XResolution-XSize)/2,
								 (MacVideo.YResolution-YSize)/2   )  
			);		
		
	} // end try
	
catch ( ... ) {
// THIS IS JUST TO ALLOW TERMINATION OF THE PROGRAM AND CLEAN UP OF STUFF.
	}

	return(0);
} // end run_experiment


//====================================================================
//
//======================================================================
void filter_dynamic_noise(mvl_video_manager & MacVideo,
		IML_LINT XSize, IML_LINT YSize, IML_LINT SeqLength, 
		IML_USINT Repeats, mvl_screen_position IML_REF Position) {		

	// MAKE THE 3D IMAGE SEQUENCE
	iml_ld_index ImageShape(XSize, YSize ,SeqLength);
	iml_fcomplex_image ComplexImage( ImageShape );
	iml_ubyte_image ByteImage( ImageShape );
	if( !ComplexImage.valid_memory() || !ByteImage.valid_memory() ) {
		iml_error("filter_dynamic_noise","Could not allocate image!");
		return;
		} // end if
	
	uniform_rv Uniform( 1, 255 - 1e-10 );// OUR CLUT'S GO FROM 1 TO 254
	mvl_clock Timer;
	Uniform.seed(Timer.seed());
	
	Timer.start();
	// THIS IS THE SLOW BUT SAFE WAY;
/****
	for(IML_ULINT S=0; S<SeqLength; S++) {
		for(IML_LINT X=0; X<XSize; X++) {
			for(IML_LINT Y=0; Y<YSize; Y++) {
				ComplexImage(X,Y,S).Real = Uniform.sample(); 
				ComplexImage(X,Y,S).Imag = 0;
				}
			}
		} // end for
****/
	fprintf(stderr,"fill the complex image with noise\n");fflush(stderr);
	fprintf(stderr,"fill the complex image with noise\n");
	// the fast, unsafe way; it is 15 times faster than the above
	for(IML_ULINT Index = 0; Index < ComplexImage.ImageSpec.ElementCount; Index++) {
		ComplexImage[Index].Real = Uniform.sample();
		ComplexImage[Index].Imag = 0.0;
		ByteImage[Index] = ComplexImage[Index].Real;
		}

	IML_SINT NoiseID = MacVideo.store_image_sequence(ByteImage);

	double Min, Max;
	ComplexImage.real().min_max_element(Min,Max);
	ComplexImage.real().scale_elements(1,254);
	ComplexImage.magnitude().min_max_element(Min,Max);

	// make sure the image knows that it is in the fourier domain
	ComplexImage.non_fourier_domain();
	
	fprintf(stderr,"FFT the complex image\n");
	// true means change indices center axes in image
	ComplexImage.forward_FFT(TRUE);
	
	float XC=3, YC=3, SC=2.5, SD=2.0;
	
	// NOW WE FILTER; INDICES ARE IN *FREQUENCY*!!
	for(IML_LINT S=-SeqLength/2; S<=SeqLength/2-1; S++) {
		for(IML_LINT X=-XSize/2; X<=XSize/2-1; X++) {
			for(IML_LINT Y=-YSize/2; Y<=YSize/2-1; Y++) {
				if( iml_abs(S)>0  &&  iml_abs(S)<5 && 
					iml_abs(X)>0  &&  iml_abs(X)<6 && 
					iml_abs(Y)>0  &&  iml_abs(Y)<6 &&
					( (S>0&& X>0 && Y>0) || (S<0&& X<0 && Y<0) )
					)
					{ 
					float fX=iml_abs(X), fY=iml_abs(Y), fS=iml_abs(S);
					float	Mag  = ( (fX-XC)*(fX-XC)/(SD*SD) );
							Mag += ( (fY-YC)*(fY-YC)/(SD*SD) );
							Mag += ( (fS-SC)*(fS-SC)/(SD*SD) );
							Mag = exp(-0.5*Mag);
					ComplexImage(X,Y,S).Real *= Mag; // or whatever 
					ComplexImage(X,Y,S).Imag *= Mag; // or whatever
					} else {
					ComplexImage(X,Y,S).Real = 0.0; // or whatever 
					ComplexImage(X,Y,S).Imag = 0.0; // or whatever
					}
				}
			}
		} // end for
	
	fprintf(stderr,"inverse FFT the complex image\n");
	// GO BACK TO SPACE-TIME DOMAIN FOR DISPLAY
	ComplexImage.inverse_FFT(TRUE);
	
	// normalize magnitudes to max of 255
	ComplexImage.real().min_max_element(Min,Max);
	ComplexImage.imaginary().min_max_element(Min,Max);
	ComplexImage.magnitude().min_max_element(Min,Max);
	ComplexImage.real().scale_elements(1,254);
	
	fprintf(stderr,"scale the image\n");
	// the fast, unsafe way; it is 15 times faster than the index method
	for(IML_ULINT Index = 0; Index < ComplexImage.ImageSpec.ElementCount; Index++) {
		ByteImage[Index] = ComplexImage[Index].Real;
		}
	
	// just to see how long this takes
	float Time = Timer.time();
	
	IML_SINT FilteredID = MacVideo.store_image_sequence(ByteImage);
	
	iml_imageframe_sequence SinglePassSeq(SeqLength);
	SinglePassSeq.generate_linear_sequence(0,1);
		
	iml_imageframe_sequence MultiPassSeq(SeqLength*Repeats);
	MultiPassSeq = SinglePassSeq;
	
	// HERE WE STORE AND PLAY THE TWO SEQUENCES.
	//========================================================================
	// CREATE A PAIR OF OFFSET POSITIONS
	mvl_screen_position NoisePos(Position);		NoisePos.XPos    -= XSize/2 + 1;
	mvl_screen_position FilteredPos(Position);	FilteredPos.XPos += XSize/2 + 1;

// THIS IS TEST CODE
/****
for(long int KK=0; KK<32; KK++) {
	MacVideo.display_frame(NoiseID, KK, NoisePos);
	MacVideo.display_frame(FilteredID, KK, FilteredPos);
	char Buff[128]; scanf("%s",Buff);
	}
***/

	// STORE TWO BLIT SEQUENCES
	IML_USINT NoiseSeqID = MacVideo.create_blit_sequence(NoiseID, NoisePos, MultiPassSeq);
	IML_USINT FilteredSeqID = MacVideo.create_blit_sequence(FilteredID, FilteredPos, MultiPassSeq);
	
	// ADD THE TWO SEQUENCES TO A GROUP
	const int GroupZero=0;
	MacVideo.add_to_group(NoiseSeqID, GroupZero);
	MacVideo.add_to_group(FilteredSeqID, GroupZero);

	// PLAY BOTH SEQUENCES SIMULTANEOUSLY
	IML_FLOAT PlayingTime = MacVideo.play_blit_group(GroupZero);
	
	fprintf(stderr,"calculation time is %f seconds, or %f sec/pixel \n", 
			Time, Time/(float)(XSize*YSize*SeqLength));
	fprintf(stderr,"playing time is %f (FPS=%f)\n", 
			PlayingTime, SeqLength*Repeats/PlayingTime);
char Buff[128];
do {
	fprintf(stderr,"Do you want to show that again? ");fflush(stderr);
	scanf("%s", Buff);
	if(Buff[0] == 'y' || Buff[0] == 'Y')
		PlayingTime = MacVideo.play_blit_group(GroupZero);
	} while(Buff[0] != 'q');

} // end filter_dynamic_noise


