
//#define IML_NO_SAFETY

// INCLUDE THE APPLICATION INDEPENDENT HELPER ROUTINES
// NOT DOING THIS FOR MAC YET . . .
//#include "imlapp.h"


#include "iml_file.h"
#include "iml_rofn.h"

// THESE REPRESENT TWO LEVELS OF SAFTEY FOR INDEXING.
// TURN AT LEAST ONE ON TO PREVENT BAD INDEX VALUES.
// THIS MAY SLOW YOUR PROGRAM.  NOTE THAT CORRECTIONS ARE
// SILENT.  BUT IF YOU ARE GETTING CRASHES WITH THIS OFF,
// THEN TURN THEM ON TO FIND OUT IF IT IS BECAUSE YOU ARE
// INDEXING IMAGES INCORRECTLY.
#include "mvl_macv.h"


void gabor_FFT_xform(mvl_video_manager & MacVideo, 
	IML_USINT XSize, IML_USINT YSize, IML_FLOAT SpatialPeriod,
	IML_FLOAT SDev, IML_LINT WriteFiles=FALSE);
	
void drifting_gabor(mvl_video_manager & MacVideo, 
			IML_USINT XSize, IML_USINT YSize, IML_LINT SeqLength, 
				IML_USINT RepeatCount, IML_FLOAT SpatialPeriod, IML_FLOAT SDev,
				IML_FLOAT DriftFrequency);		

void make_gabor_size_stimulus(void);

//============================================================================
//
//============================================================================
#ifdef _Windows
int run_experiment(void) {
#else
int main () {
#endif

//======== EXAMPLE  CODE FOR CREATING AND INITIALIZING mvl_video_manager

	
try {  
	// ALL THE CODE HERE IS SET FOR CATCHING EXCEPTIONS 
	// WITH AN EMPTY catch() AT THE BOTTOM.  

	// MAKE A VIDEO MANAGER OBJECT AND INITIALIZE IT
	// I'LL PROVIDE AND USE MORE ERROR CHECKING LATER . . .
	IML_USINT MaxSequenceCnt = 32;
	IML_USINT MaxBlitCnt = 64;
	IML_USINT GroupCount = 8;
	IML_ULINT ExitOnErr = TRUE;
	mvl_video_manager MacVideo(MaxSequenceCnt,MaxBlitCnt, GroupCount, ExitOnErr);
	
	MacVideo.initialize_screen(0);
//	MacVideo.use_dgp_calibration();
	MacVideo.set_mvl_calibration(MacVideo.MaxDac, 2, 0, 100);
	MacVideo.set_attenuator_red_steps(0.0);
	MacVideo.use_mvl_calibration();
	
	MacVideo.set_mean_luminance(0.5);
	MacVideo.set_max_contrast(0.5);
	MacVideo.set_display_CLUT(0.5,1.0);

enum {
		DRIFTING_GABOR, 
		GABOR_SIZE_STIMULUS, 
		GABOR_FFT, 
		OPTION_COUNT
		};

	IML_LINT DemoSelector = DRIFTING_GABOR;

	switch(DemoSelector) {
	
	
		case GABOR_SIZE_STIMULUS: {
			make_gabor_size_stimulus();
			}; break;
	
		case DRIFTING_GABOR: {
			IML_LINT  XSize= 256,  YSize= 256;
			IML_FLOAT SpatialPeriod=XSize/6, SDev = XSize/6;
			IML_FLOAT DriftFrequency = 1; // cycles per display period (Seqlength)
			IML_LINT SeqLength = 16;
			IML_USINT RepeatCount=1;
			drifting_gabor(MacVideo, XSize, YSize, SeqLength, RepeatCount, 
						SpatialPeriod, SDev, DriftFrequency);
			}; break;
	
		case GABOR_FFT: {
			MacVideo.set_display_CLUT(0.5,1.0);
			IML_LINT  XSize= 128,  YSize= 128;
			IML_FLOAT SpatialPeriod=8, SDev = SpatialPeriod;
			IML_LINT WriteFiles=FALSE;
			gabor_FFT_xform(MacVideo, XSize, YSize, 
						SpatialPeriod, SDev, WriteFiles);
			}; break;
	
		} // end switch
		
	} // end try
	
catch ( ... ) {
// THIS IS JUST TO ALLOW TERMINATION OF THE PROGRAM AND CLEAN UP OF STUFF.
	}

	return(0);
} // end run_experiment


//======================================================================
//
//======================================================================
void make_gabor_size_stimulus(void) {

	int  XSize= 512,  YSize= 512;
	iml_ld_index ImSize(XSize, YSize);

	iml_ubyte_image Image(ImSize);

	if( !Image.valid_memory() ) {
		iml_error("make_gabor_size_stimulus","Could not allocate image!");
		return;
		} // end if

	iml_gabor_rofn Gabor(2);

	Gabor.set_center(IML_2D_X, 64);
	Gabor.set_center(IML_2D_Y, 64);
	Gabor.set_sdev(IML_2D_X, 15);
	Gabor.set_sdev(IML_2D_Y, 15);
	Gabor.set_magnitude(255);
	Gabor.Carrier.set_center(IML_2D_X, 64);
	Gabor.Carrier.set_center(IML_2D_Y, 64);
	Gabor.Carrier.set_phase(0);    // in degrees
	Gabor.Carrier.set_period(IML_2D_X, 24.0);   // cycles per pixel

	IML_LINT DSteps = 3, CSteps = 3;
	IML_DOUBLE CMax = 128*0.87, CDecF = 0.24;
	IML_DOUBLE DMax = 28, DDec = 5;
	IML_LINT DelX=XSize/CSteps, DelY=YSize/DSteps;
	IML_LINT SX = DelX/2, SY = DelY/2;

	// mean luminance level
	Image = 127;
	for(int D=0; D<DSteps; D++) {
		// re-init contrast -- vary it across the image
		IML_LINT Con = CMax;

		for(int C=0; C<CSteps; C++) {
			printf("Center<%d, %d> of mag<%f> with SD<%f>\n",
				 SX+C*DelX, SY+D*DelY, Con, DMax);

			Gabor.set_center(SX+C*DelX, SY+D*DelY);
			Gabor.Carrier.set_center(SX+C*DelX, SY+D*DelY);
			Gabor.set_sdev(DMax,DMax);
			Gabor.set_magnitude(Con);
			Gabor.set_domain_sd(IML_2D_X); // restrict assignment to smaller area
			Gabor.set_domain_sd(IML_2D_Y); // restrict assignment to smaller area
			Con *= CDecF;

			Image += Gabor; // PUT THE IMAGE INTO THE THING
			} // end for
		// vary the sdev down theimage
		DMax -= DDec;
		} // end for

	iml_image_file  OIFile;
	OIFile.create_file("gaborsize.raw");
	OIFile.write_image(&Image, IML_RAW_FORMAT);
//	OIFile.write_image(&Image); // the default is IML_IML_FORMAT
	OIFile.close_file();

} // end make_gabor_size_stimulus




//======================================================================
//
//======================================================================
void drifting_gabor(mvl_video_manager & MacVideo, 
				IML_USINT XSize, IML_USINT YSize, IML_LINT SeqLength, 
				IML_USINT RepeatCount, IML_FLOAT SpatialPeriod, IML_FLOAT SDev,
				IML_FLOAT DriftFrequency) {

	iml_ld_index Index(XSize, YSize, SeqLength);

	// MAKE A TEMPORAL WINDOW CONTROL THINGY AND LOAD IT WITH A 
	// RAISED-COSINE-EDGE.  COSINUSOIDAL EDGES MUST BE <= HALF SEQUENCE.
	IML_USINT HalfPeriod = (10 < SeqLength/2 ? 10 : SeqLength/2);
	mvl_temporal_window TWindow(SeqLength); 
	
	TWindow = 1.0; // MAX CONTRAST IN THE MIDDLE
	for(IML_LINT I=0;I<HalfPeriod;I++) // MAKE THE EDGES
		TWindow[I] = TWindow[SeqLength-I]= 
					0.5+0.5*iml_sin(IML_PI*I/HalfPeriod - IML_2PI/4);

	// CREATE THE CLUT TABLES REQUIRED FOR DISPLAY
	MacVideo.initialize_CLUT_tables(TWindow);
	MacVideo.use_CLUT_sequence(TWindow);
	
	// USE A TWO-D GABOR TO CREATE A SPATIALLY-WINDOWED DRIFTING GRATING.
	iml_gabor_rofn Gabor(2);
	// SET GABOR DEFAULTS
	//IML_FLOAT SpatialPeriod = 16.0; // pixels
	IML_FLOAT XCenter = (XSize-1.0)/2.0, YCenter = (YSize-1.0)/2.0;
	Gabor.set_center(IML_2D_X, XCenter);
	Gabor.set_center(IML_2D_Y, YCenter );
	Gabor.set_sdev(IML_2D_X, SDev);
	Gabor.set_sdev(IML_2D_Y, SDev);
	Gabor.set_magnitude(63);
	Gabor.Carrier.set_center(IML_2D_X, XCenter);
	Gabor.Carrier.set_center(IML_2D_Y, YCenter);
	Gabor.Carrier.set_phase(0);    // in degrees
	Gabor.Carrier.set_period(IML_2D_X, SpatialPeriod);   // pixels/cycle
	Gabor.Carrier.set_period(IML_2D_Y, 0.0);   // zero flags zero frequency
	
	// CREATE AN UNSIGNED BYTE IMAGE OF THIS SIZE
	iml_ubyte_image UByteImage(Index);
	if( !UByteImage.valid_memory() ) {
		iml_error("drifting_gabor","Could not allocate image!");
		return;
		}

	UByteImage = 127;
	// RESTRICT IMAGE OPERATIONS TO 2D SUBSPACE
	UByteImage.set_hyperplane_size(2);
	
	// SET INDEXER TO START OF IMAGE
	Index.restart(); // FOR CLARITY
	
	// NUMBER OF TOTAL
	for(IML_LINT I=0; I<SeqLength; I++) {
	
		// SELECT A 2D IMAGE SLICE; 
		// [A MORE COMPLICATED(?) METHOD OF SELECTING A SUBSPACE THAN IN
		// mckee_welch_randots ABOVE.]
		Index.set(IML_2ST_T,I);
		UByteImage.select_hyperplane(Index);
		
		// SET CARRIER PHASE;
		Gabor.Carrier.set_phase(360*DriftFrequency*(I/(float)SeqLength));    // in degrees
		// save the image
		UByteImage += Gabor;
		} // end for
		
	// GO BACK TO 3D FORMAT FOR DISPLAY
	UByteImage.set_hyperplane_size(3);
	// now the image should contain a spatially-gaussian-windowed drifting sine wave.

	IML_SINT ImSeqID = MacVideo.store_image_sequence(UByteImage);
	if( ImSeqID < 0 ) {
		iml_error("main","FAILED ON IMAGE STORE");
		return;
		} // end if

	// MAKE A SEQUENCE TO PLAY THE SEQUENCE FORWARD
	iml_imageframe_sequence FrameSequence(SeqLength);
	FrameSequence.generate_linear_sequence(/*from*/ 0, /*increment*/ 1);
	//FrameSequence.shuffle();
	//FrameSequence.sin( SeqLength/2.0, SeqLength/2.0-0.00001, SeqLength, 90.0);
	//FrameSequence.reverse();
//	FrameSequence.rotate(-12);

	IML_LINT PlayFrames = SeqLength*RepeatCount;
	iml_imageframe_sequence LongSequence(PlayFrames);
	LongSequence = FrameSequence;
	
	// AT THIS POSITION
	mvl_screen_position ImageDest(200,200);
	
	// CREATE THE BLIT SEQUENCE
	IML_SINT BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, ImageDest, LongSequence);

// DISPLAY A FEW FRAMES FOR FUN
//	MacVideo.set_display_CLUT(0.5, 1.0);
//	MacVideo.display_frame(ImSeqID, 0, mvl_screen_position(0,0) );
//	MacVideo.display_frame(ImSeqID, 20, mvl_screen_position(300,0) );
//	MacVideo.display_frame(ImSeqID, 40, mvl_screen_position(0,330) );
//	MacVideo.set_display_CLUT(0.5, 0.5);
//	MacVideo.set_display_CLUT(0.5, 0.2);
//	MacVideo.set_display_CLUT(0.5, 0.0);

	IML_FLOAT Acc = 1e-10;
	IML_LINT NumSteps = 10;
	mvl_clock Timer;
	for(IML_LINT Count=0; Count < NumSteps; Count++) {
		// PLAY THE BLIT SEQUENCE
		IML_FLOAT PlayingTime = MacVideo.play_blit_sequence(BlitSeqID);
		// ALTER THE CONTRAST!
		MacVideo.set_max_contrast(MacVideo.get_max_contrast()*0.7);
		MacVideo.initialize_CLUT_tables(TWindow);
		Acc += PlayingTime;
		printf("playing time is %f\n",PlayingTime);
		Timer.wait(0.5);
		} // end for

		printf("Average playing time is %f or %f frames/sec\n",
				Acc/NumSteps, NumSteps*PlayFrames/Acc);

} // end drifting_gabor



//======================================================================
//
//======================================================================
void gabor_FFT_xform(mvl_video_manager & MacVideo, 
	IML_USINT XSize, IML_USINT YSize, IML_FLOAT SpatialPeriod,
	IML_FLOAT SDev, IML_LINT WriteFiles) { // begin block

	// USE A TWO-D GABOR TO CREATE A SPATIALLY-WINDOWED GRATING.
	iml_gabor_rofn Gabor(2);
	// CONFIGURE GABOR
	Gabor.set_center(IML_2D_X, (XSize-1.0)/2.0);
	Gabor.set_center(IML_2D_Y, (YSize-1.0)/2.0);
	Gabor.set_sdev(IML_2D_X, SDev);
	Gabor.set_sdev(IML_2D_Y, SDev);
	Gabor.set_magnitude(126);
	Gabor.Carrier.set_center(IML_2D_X, (XSize-1.0)/2.0);
	Gabor.Carrier.set_center(IML_2D_Y, (YSize-1.0)/2.0 );
	Gabor.Carrier.set_phase(0);    // in degrees
	Gabor.Carrier.set_period(IML_2D_X, SpatialPeriod);   // pixels/cycle

	// MAKE AN IMAGE FILE MANAGER
	iml_image_file  OIFile;
	
	// MAKE A COUPLE OF IMAGES
	iml_fcomplex_image FCImage(iml_ld_index(XSize,YSize));
	iml_ubyte_image BImage(iml_ld_index(XSize,YSize));
	if( !FCImage.valid_memory() || !BImage.valid_memory() ) {
		iml_error("gabor_FFT_xform","Could not allocate images!");
		return;
		}
	
	// MAKE A PLAIN GABOR AND SCALE FOR DISPLAY
	FCImage.real() = Gabor;
	FCImage.real().scale_elements(1,254);
	BImage = FCImage.real();

	// PLUNK THE IMAGE DIRECTLY TO THE SCREEN
	MacVideo.display_image( BImage, mvl_screen_position(0,50) );

	if(WriteFiles) {
		OIFile.create_file("gabor.raw");
		OIFile.write_image(&BImage, IML_RAW_FORMAT);
		OIFile.close_file();
		} // end if

	// THIS IS HOW THE SHIFTING WORKS IF REQUESTED AT FFT TIME
	FCImage.real() = Gabor;
	// MODULATE BY HIGHEST POSSIBLE FREQUENCY
	FCImage.modulate();
	FCImage.real().scale_elements(1,254);
	BImage = FCImage.real();

	// PLUNK THE IMAGE DIRECTLY TO THE SCREEN
	MacVideo.display_image( BImage, mvl_screen_position(XSize+10,50) );

	if(WriteFiles) {
		OIFile.create_file("gabor-modulated.raw");
		OIFile.write_image(&BImage, IML_RAW_FORMAT);
		OIFile.close_file();
		} // end if

	FCImage.real() = Gabor;

	// TRUE MEANS MOVE DC TO IMAGE CENTER, 
	// FALSE MEANS LEAVE IN FUNKY WRAPPED STATE
	FCImage.forward_FFT(TRUE); 

	// LET'S LOOK AT THE MAGNITUDE SPECTRUM
	FCImage.magnitude().scale_elements(2,253);
	// THIS ASSIGNMENT ONLY COPIES ONE QUARTER OF THE IMAGE
	// BECAUSE THE INDICES HAVE CHANGED AFTER FFT. THAT IS, THE 
	// ASSIGNMENT ONLY OCCURS OVER THE COMMON SUBSPACE DOMAIN.
	// THE FFT'ED IMAGE SUBSPACE IS (-Size/2, +Size/2 - 1)
	// WHILE THE BYTE IMAGE IS (0, SIZE)
	// BImage = FCImage.magnitude();
	
	// NOTE THAT THIS METHOD IS FAST AND AVOIDS THE SUBSPACE-ORIENTED 
	// ASSIGNMENT OPERATION, BUT BY-PASSES INDEX CHECKING.
	for(IML_LINT I=0; I<=FCImage.Subspace.ElementCount; I++)
		BImage[I] = FCImage[I].get_mag();

	// PLUNK THE IMAGE DIRECTLY TO THE SCREEN
	MacVideo.display_image( BImage, mvl_screen_position(2*(XSize+10),50) );

	if(WriteFiles) {
		OIFile.create_file("gabor_fft_mag.raw");
		OIFile.write_image(&BImage, IML_RAW_FORMAT);
		OIFile.close_file();
		} // end if

	mvl_clock Timer;
	Timer.wait(20);

} // end gabor_FFT_xform

