
//#define IML_NO_SAFETY

// INCLUDE THE APPLICATION INDEPENDENT HELPER ROUTINES
// NOT DOING THIS FOR MAC YET . . .
//#include "imlapp.h"


#include "iml_file.h"
#include "iml_rofn.h"

// THESE REPRESENT TWO LEVELS OF SAFTEY FOR INDEXING.
// TURN AT LEAST ONE ON TO PREVENT BAD INDEX VALUES.
// THIS MAY SLOW YOUR PROGRAM.  NOTE THAT CORRECTIONS ARE
// SILENT.  BUT IF YOU ARE GETTING CRASHES WITH THIS OFF,
// THEN TURN THEM ON TO FIND OUT IF IT IS BECAUSE YOU ARE
// INDEXING IMAGES INCORRECTLY.
#include "mvl_macv.h"

		
#include "dac_quest.h"

IML_USINT RED_CLUT_ENTRY = 1;
IML_USINT GREEN_CLUT_ENTRY = 254;
	
void get_red_green_mm_point(mvl_video_manager & MacVideo, mvl_clut &CLUTable,
				IML_USINT TrialCount,  IML_USINT DeltaDAC, 
				IML_USINT StartDac,IML_SINT GreenOrWhite, IML_USINT BlitSeqID);

struct match_results {
//public:
//	IML_DOUBLE LumValue;
	IML_DOUBLE OptimumValue;
	IML_DOUBLE BestLogLikelihood;
	IML_DOUBLE BestMean;
	IML_DOUBLE SDofBestMean;
	IML_DOUBLE BestSDev;
	IML_LINT DOF;
	IML_DOUBLE ModelMean;
	IML_DOUBLE squared_error(void) 
		{ return( (ModelMean-BestMean)*(ModelMean-BestMean)/(BestSDev*BestSDev)); }
		
	match_results(	IML_DOUBLE LumVal, IML_DOUBLE BMean, IML_DOUBLE BSDev, IML_DOUBLE ModelM) {
//			LumValue = LumVal;
			BestMean=BMean;
			BestSDev=BSDev;
			ModelMean = ModelM;
			OptimumValue = 0;
			BestLogLikelihood=0;
			SDofBestMean=1;
			DOF=0;
			};
	match_results(void) {
//		LumValue = 1;
		OptimumValue = 0;
		BestLogLikelihood=0;
		BestMean=0;
		SDofBestMean=1;
		BestSDev=1;
		ModelMean = 0;
		DOF=0;
		};
};
match_results OP;


//============================================================================
//
//============================================================================
#ifdef _Windows
int run_experiment(void) {
#else
int main () {
#endif

//======== EXAMPLE  CODE FOR CREATING AND INITIALIZING mvl_video_manager

	
try {  
	// ALL THE CODE HERE IS SET FOR CATCHING EXCEPTIONS 
	// WITH AN EMPTY catch() AT THE BOTTOM.  

	// MAKE A VIDEO MANAGER OBJECT AND INITIALIZE IT
	// I'LL PROVIDE AND USE MORE ERROR CHECKING LATER . . .
	IML_USINT MaxSequenceCnt = 32;
	IML_USINT MaxBlitCnt = 64;
	IML_USINT GroupCount = 8;
	IML_ULINT ExitOnErr = TRUE;
	mvl_video_manager MacVideo(MaxSequenceCnt,MaxBlitCnt, GroupCount, ExitOnErr);
	
	MacVideo.initialize_screen(0);
//	MacVideo.use_dgp_calibration();
	MacVideo.set_mvl_calibration(MacVideo.MaxDac, 2.08, 0, 100);
	MacVideo.set_attenuator_red_steps(0.0);
	MacVideo.use_mvl_calibration();
	MacVideo.set_display_CLUT(0.5,1.0);

	IML_LINT FrameCount=8; // TOTAL NUMBER OF REAL FRAMES THAT NEED TO BE MADE
	IML_LINT  XSize= 128,  YSize= 128;
	IML_USINT Duration=FrameCount*5; // number of "DwellTime" frames to show
	IML_USINT DwellTime = 1;
	
	IML_SINT GreenOrWhite = FALSE;
	
	IML_USINT BarWidth = 16; // in pixels
	
	// GRAB A COPY OF THE CURRENT CLUT
	mvl_clut &CLUTable = *MacVideo.get_clut_copy();
	
	// SET RED AND GREEN BAR COLOR ENTRIES
	IML_LINT Green=127;
	IML_LINT Red = 210;
	if(GreenOrWhite)
		CLUTable.set_clut_entry(GREEN_CLUT_ENTRY, 0, Green, 0);
	else
		CLUTable.set_clut_entry(GREEN_CLUT_ENTRY, Green, Green, Green);
	
	CLUTable.set_clut_entry(RED_CLUT_ENTRY, Red,   0, 0);	
	MacVideo.make_default_clut(CLUTable);

	// MAKE THE 3D IMAGE SEQUENCE
	iml_ubyte_image MotionImage( iml_ld_index(XSize, YSize, FrameCount + 1) );
	if( !MotionImage.valid_memory() ) {
		iml_error("Attentional Motion","Could not allocate MotionImage!");
		return(0);
		} // end if

	// USE THE LAST FRAME OF THE 3D SEQUENCE FOR A MASK.
	iml_ubyte_image NoiseMaskFrame( MotionImage );	// MAKE AN ALIAS TO THE THREE-D IMAGE
	NoiseMaskFrame.set_hyperplane_size(2);				// MAKE IT TWO-D
	NoiseMaskFrame.select_hyperplane(FrameCount);	// POINT TO LAST FRAME

	mvl_clock Timer;
		
	uniform_rv Uniform; // this is for the carrier
	Uniform.seed(Timer.seed());
	
	// FILL THE MASK WITH NOISE
	for(IML_LINT Q=0; Q<(XSize*(long)YSize); Q++) {
		IML_FLOAT Temp24 = (Uniform.sample() > 0.5) ? 2 : 253;
		NoiseMaskFrame(Q) = Temp24;
		}

MacVideo.display_image(	NoiseMaskFrame, mvl_screen_position(50,50) );

	// WE WILL CREATE THE IMAGE ONE SLICE AT TIME, SO RESTRICT TO 2D SUBSPACE
	MotionImage.set_hyperplane_size(2);
	
	Timer.start();
	// NOW, ONE SLICE AT A TIME . . .
	for(IML_ULINT K=0; K<FrameCount; K++) {

		// SET CURRENT SLICE
		MotionImage.select_hyperplane(K);

		// CALCULATE DELTA-X SIZES
		IML_LINT Kmod4 = (K/2)%4;
		IML_LINT StepSize = BarWidth/2;
		IML_LINT GOffset = (-Kmod4*StepSize + BarWidth*2) % (BarWidth*2); // reverse direction
		IML_LINT ROffset =   Kmod4*StepSize;
		fprintf(stderr,"<G,R>Offset <%.3d, %.3d)\n", GOffset, ROffset); fflush(stderr);

		// WHICH COLOR ARE THE BARS ON THIS FRAME?
		IML_UBYTE Color = (K&0x01) ? GREEN_CLUT_ENTRY	: RED_CLUT_ENTRY;
		IML_LINT Offset = (K&0x01) ? GOffset 				: ROffset;
		
		// MAKE A FRESH FRAME OF BARS
		for(IML_LINT Y=0; Y < YSize; Y++) {
			for(IML_LINT X=0; X < XSize; X++) {
				IML_UBYTE BlackOrColor = ( ( (X+Offset)/BarWidth) & 0x01 ) ? 255 : Color;
				MotionImage(X,Y) = BlackOrColor;
				} // end for
			} // end for

MacVideo.display_image(	MotionImage, mvl_screen_position(50,50+(K&0x01)*YSize));
//Timer.wait(0.5);
		} // end for
	
	IML_FLOAT Time = Timer.time();
	fprintf(stderr,"frame generation %f\n",Time/FrameCount);

	// RESET SUBSPACE SIZE TO 3D FOR MAC VIDEO STORAGE
	MotionImage.set_hyperplane_size(3);
	IML_USINT ImSeqID = MacVideo.store_image_sequence(MotionImage);
	
	IML_LINT TotalDisplayFrames = Duration*DwellTime;
	iml_imageframe_sequence Sequence(TotalDisplayFrames + 1);
	for(long K=0; K< TotalDisplayFrames; K++) {
		long Temp = (K/DwellTime)%FrameCount;
		Sequence(K) = Temp;
		}
	Sequence(TotalDisplayFrames) = FrameCount; // last, we want to display the noise mask
	
	mvl_screen_position Position( (MacVideo.XResolution-XSize)/2,
								 			(MacVideo.YResolution-YSize)/2 );
	
	IML_USINT BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, Position, Sequence);


get_red_green_mm_point(MacVideo, CLUTable, /*TrialCnt*/150,  /*DeltaDAC*/ 7, Green, FALSE, BlitSeqID);

return(0);

	char Buff[128]; IML_LINT Delta=5;
	do {

		MacVideo.play_blit_sequence(BlitSeqID);
//		MacVideo.display_image(NoiseImage, Position);
		
		fprintf(stderr,"r(%d)g(%d)? :",Red,Green); fflush(stderr); scanf("%s", Buff);
		switch(Buff[0]) {
			case 'g': Green-=Delta; break;
			case 'G': Green+=Delta; break;
			case 'r': Red-=Delta;   break;
			case 'R': Red+=Delta;   break;
			default: {
				sscanf(Buff,"%d",&Delta);
				}; break;
			} // end switch
		CLUTable.set_clut_entry(GREEN_CLUT_ENTRY, 0, Green, 0);	
		CLUTable.set_clut_entry(RED_CLUT_ENTRY, Red,   0, 0);	
		MacVideo.make_default_clut(CLUTable);
		} while(Buff[0] != 'q');

	} // end try
	
catch ( ... ) {
// THIS IS JUST TO ALLOW TERMINATION OF THE PROGRAM AND CLEAN UP OF STUFF.
	}

	return(0);

} // end main


//============================================================================
//
//============================================================================
void get_red_green_mm_point(mvl_video_manager & MacVideo, mvl_clut &CLUTable,
		IML_USINT TrialCount,  IML_USINT DeltaDAC, 
		IML_USINT StartDac, IML_SINT GreenOrWhite, IML_USINT BlitSeqID) {


	mvl_clock Clock;
	
	uniform_rv UniformRV(1e-6, 1 - 1e-6);
	UniformRV.seed(Clock.seed());
	
	IML_USINT SmallStep = DeltaDAC;
	IML_USINT LargeStep = DeltaDAC * 5;
	
	IML_SINT CategoryCnt = 20;
	
	mvl_ColorSpec Gun; 
	if(GreenOrWhite)
		Gun = mvl_rgb(0, StartDac, 0);
	else
		Gun = mvl_rgb(StartDac, StartDac, StartDac);

	IML_USINT &DACValue = Gun.rgb.green;
	
	char Buff[128];
	do {
		CLUTable.set_clut_entry(GREEN_CLUT_ENTRY, Gun);
		MacVideo.make_default_clut(CLUTable);
		MacVideo.play_blit_sequence(BlitSeqID);
		
		IML_LINT StepSize=0;
		scanf("%s",Buff);
		switch(Buff[0]) {
			case 'j': StepSize = SmallStep; break;
			case 'J': StepSize = LargeStep; break;
			case 'k': StepSize = -SmallStep; break;
			case 'K': StepSize = -LargeStep; break;
			default:  StepSize = 0;
			} // end switch
			
		if(GreenOrWhite)
			Gun.rgb.green += StepSize;
		else
			Gun.rgb = (Gun.rgb.green + StepSize);

		fprintf(stderr,"Value=%d\n", DACValue);
		} while(Buff[0] != 'q');

	// ALLOCATE A FITTER FOR CategoryCnt CATEGORIES
	dac_curve_quest MatchEstimator(CategoryCnt,MacVideo.MaxDac);
	
	// GUESS THE PARAMETERS FOR FITTING
	MatchEstimator.configure_parameter(0, "CNorm Mean",	0,		255,	DACValue,	TRUE);
	MatchEstimator.configure_parameter(1, "CNorm SDev",	0.1,	500,	DeltaDAC,	TRUE);
	
	IML_LINT StartDAC = DACValue-(CategoryCnt*DeltaDAC)/2;
	StartDAC = (StartDAC<0) ? 0 : ((StartDAC > 255) ? 255-CategoryCnt : StartDAC);
	MatchEstimator.initialize(
		StartDAC, 			// start bin value 
		DeltaDAC, 			// delta bin value 
		DACValue, 			// prior Mean;
		4, 					// prior SDev;
		20					// give a weight to the prior;
		);
	
	MatchEstimator.prior_decay( 20 /*Alpha: trials*/, 3 /*Beta: slope*/);
	
	// SET CONVERGENCE CRITERIA:  ANSWER TOLERANCE AND MAX-POWELL ITERATIONS
	MatchEstimator.set_convergence(1e-10, 30);

	while(TrialCount--) {
	fprintf(stderr,"Trial %d:  ",TrialCount);fflush(stderr);

		// SET UP THE ENTRY
		CLUTable.set_clut_entry(GREEN_CLUT_ENTRY, Gun);
		MacVideo.make_default_clut(CLUTable);
		MacVideo.play_blit_sequence(BlitSeqID);
			
		IML_BYTE Answer = -1;
		do {
			scanf("%s",Buff);
			
			// IF BAD KEY PRESS
			if( Buff[0] != 'k' && Buff[0] != 'j' ) continue;
			
			// TASK: WHICH DIRECTION IS IT GOING? (L,R);  STORE: GOING RIGHT (seeing green)
			Answer =  ( Buff[0] == 'k' );
fprintf(stderr,"Ans = %d; ",Answer);fflush(stderr);
			} while(Answer < 0);

		MatchEstimator.add_trial(DACValue, Answer);
	
		// OPTIMUM VALUE SQUARED ERROR -- THATS OUR OBJECTIVE FUNCTION;
		IML_LINT Index = MatchEstimator.next_guess();
fprintf(stderr,"Next: %d;  ",MatchEstimator.Data[Index].DACValue);
		// add random +/- 3 bins
		IML_LINT Select = long(UniformRV.sample()*7.0);
		Index += Select - 3;
		if(Index < 0) Index = 0;
		if(Index > CategoryCnt) Index = CategoryCnt-1;

		if(GreenOrWhite)
			Gun.rgb.green = MatchEstimator.Data[Index].DACValue;
		else
			Gun.rgb = MatchEstimator.Data[Index].DACValue;
//		DACValue = MatchEstimator.Data[Index].DACValue;

		}	 // end while

	// CALCULATE THIS FIRST
	OP.OptimumValue = 	MatchEstimator.fitdata();
	OP.BestLogLikelihood = MatchEstimator.bestLL();

	OP.BestMean = MatchEstimator.parameter_value(0);
	OP.BestSDev = MatchEstimator.parameter_value(1);
	fprintf(stderr,"Optimal Mean=%f; SDev=%f\n", OP.BestMean, OP.BestSDev); fflush(stderr);

	// PRINT OUT DATA
	OP.DOF = -2; // IT IS A TWO PARAMETER FIT
	for(int I=0; I<CategoryCnt; I++) {
		dac_data & DD = MatchEstimator.Data[I];
		if(DD.TrialCount) {
			OP.DOF++;
			fprintf(stderr,"DB %d;  T %f;  NC %f; CN %f\n",
				(long)DD.DACValue, (float)DD.TrialCount, (float)DD.NumberCorrect,
				psycho_cnorm(DD.DACValue, OP.BestMean, OP.BestSDev)*DD.TrialCount
				);
			} // end if
		} // end for
	
	MatchEstimator.calculate_covariance_information();
	OP.SDofBestMean = sqrt( MatchEstimator.covariance(0,0) );
	fprintf(stderr,"95%% Conf on Mean = [%5.5f  :  %5.5f\n",
		OP.BestMean-OP.SDofBestMean*2, OP.BestMean+OP.SDofBestMean*2 );
		
} // end get_match_points

