
//#define IML_NO_SAFETY

// INCLUDE THE APPLICATION INDEPENDENT HELPER ROUTINES
// NOT DOING THIS FOR MAC YET . . .
//#include "imlapp.h"


#include "iml_file.h"
#include "iml_rofn.h"

// THESE REPRESENT TWO LEVELS OF SAFTEY FOR INDEXING.
// TURN AT LEAST ONE ON TO PREVENT BAD INDEX VALUES.
// THIS MAY SLOW YOUR PROGRAM.  NOTE THAT CORRECTIONS ARE
// SILENT.  BUT IF YOU ARE GETTING CRASHES WITH THIS OFF,
// THEN TURN THEM ON TO FIND OUT IF IT IS BECAUSE YOU ARE
// INDEXING IMAGES INCORRECTLY.
#include "mvl_macv.h"


typedef enum {
	VERTICAL_NOISE = -1, 
	TWO_D_NOISE = 0, 
	HORIZONTAL_NOISE = 1
	} noise_type;		
void dynamic_noise(mvl_video_manager & MacVideo,
		IML_LINT  XSize, IML_LINT YSize, IML_LINT SeqLength, 
		IML_USINT Repeats, noise_type NoiseType,
		point_2d & Position);
		
void drifting_noise_fourier_xform(mvl_video_manager & MacVideo,
		IML_USINT XSize, IML_USINT YSize, 
		IML_LINT WithWrap, IML_LINT WithLog10, IML_LINT WriteFiles=FALSE);


//============================================================================
//
//============================================================================
#ifdef _Windows
int run_experiment(void) {
#else
int main () {
#endif

//======== EXAMPLE  CODE FOR CREATING AND INITIALIZING mvl_video_manager

	
try {  
	// ALL THE CODE HERE IS SET FOR CATCHING EXCEPTIONS 
	// WITH AN EMPTY catch() AT THE BOTTOM.  

	// MAKE A VIDEO MANAGER OBJECT AND INITIALIZE IT
	// I'LL PROVIDE AND USE MORE ERROR CHECKING LATER . . .
	IML_USINT MaxSequenceCnt = 32;
	IML_USINT MaxBlitCnt = 64;
	IML_USINT GroupCount = 8;
	IML_ULINT ExitOnErr = TRUE;
	mvl_video_manager MacVideo(MaxSequenceCnt,MaxBlitCnt, GroupCount, ExitOnErr);
	
	MacVideo.initialize_screen(1);
//	MacVideo.use_dgp_calibration();
	MacVideo.set_mvl_calibration(MacVideo.MaxDac, 2.159, 0, 100);
	MacVideo.set_attenuator_red_steps(28.7);
	MacVideo.use_mvl_calibration();
	MacVideo.set_display_CLUT(0.5,1.0);

enum {
		DRIFTING_NOISE_FFT, 
		DYNAMIC_NOISE, 
		OPTION_COUNT
		};

	IML_LINT DemoSelector = DRIFTING_NOISE_FFT;

	switch(DemoSelector) {
	
	
		case DRIFTING_NOISE_FFT: {
			MacVideo.linearize_clut(0.5,1.0);
			IML_LINT  XSize= 256,  YSize= XSize;
			IML_LINT WithWrap=FALSE, WithLog10 = FALSE;
			IML_LINT WriteFiles=FALSE;
			drifting_noise_fourier_xform(MacVideo, XSize, YSize, 
							WithWrap, WithLog10, WriteFiles);
			}; break;
	
		case DYNAMIC_NOISE: {
			IML_LINT SeqLength = 32;
			IML_LINT  XSize= 64,  YSize= 64;
			IML_USINT Repeats=5;
			// VERTICAL_NOISE, TWO_D_NOISE, HORIZONTAL_NOISE
			noise_type NoiseType = HORIZONTAL_NOISE;
			dynamic_noise(MacVideo, XSize, YSize, SeqLength, 
					Repeats, NoiseType, point_2d(0,100) );		
			}; break;
	
		} // end switch
		
	} // end try
	
catch ( ... ) {
// THIS IS JUST TO ALLOW TERMINATION OF THE PROGRAM AND CLEAN UP OF STUFF.
	}

	return(0);
} // end run_experiment


//====================================================================
//
//======================================================================
void dynamic_noise(mvl_video_manager & MacVideo,
		IML_LINT XSize, IML_LINT YSize, IML_LINT SeqLength, 
		IML_USINT Repeats, noise_type NoiseType,
		point_2d IML_REF Position) {		

	// MAKE THE 3D IMAGE SEQUENCE
	iml_ubyte_image MotionImage2( XSize, YSize ,SeqLength );
	if( !MotionImage2.valid_memory() ) {
		iml_error("dynamic_noise","Could not allocate image!");
		return;
		} // end if

	iml_ubyte_image MotionImage( MotionImage2 );
	
	uniform_rv Uniform(1,  255 - 1e-10);	// OUR CLUT'S GO FROM 1 TO 254
	mvl_clock Timer;
	Uniform.seed(Timer.seed());
	
	switch(NoiseType) {
	
		case TWO_D_NOISE: {
			// WE WILL CREATE THE IMAGE ONE SLICE AT TIME, 
			// SO RESTRICT TO 2D SUBSPACE
			MotionImage.set_hyperplane_size(2);

			// NOW, ONE SLICE AT A TIME . . .
			for(IML_ULINT K=0; K<SeqLength; K++) {

				// SET CURRENT SLICE
				MotionImage.select_hyperplane(K);

				// RANDOMIZE THE PIXEL VALUES FOR THIS SLICE
				for(IML_LINT X=0; X<XSize; X++)
					for(IML_LINT Y=0; Y<YSize; Y++)
						MotionImage(X,Y) = Uniform.sample(); 
				} // end for
			// RESET SUBSPACE SIZE TO 3D FOR MAC VIDEO STORAGE
			MotionImage.set_hyperplane_size(3);
			}; break;

		case HORIZONTAL_NOISE: {
			iml_ubyte_image YTSlice( 1,YSize,SeqLength );
			if( !YTSlice.valid_memory() ) {
				iml_error("dynamic_noise","Could not allocate YTSlice!");
				return;
				} // end if
				
			for(IML_LINT T=0; T<SeqLength; T++)
				for(IML_LINT Y=0; Y<YSize; Y++)
					YTSlice(0,Y,T) = Uniform.sample(); // MAKE A Y-T SLICE
				
			for(IML_LINT X=0; X<XSize;X++) { // RE ASSIGN TO ALL X SLICES
				YTSlice.shift_hyperplane_corner_to(IML_2ST_X,X);
				MotionImage = YTSlice;
				} // end for
			}; break;

		case VERTICAL_NOISE: {
			iml_ubyte_image XTSlice( XSize,1,SeqLength );
			if( !XTSlice.valid_memory() ) {
				iml_error("dynamic_noise","Could not allocate XTSlice!");
				return;
				} // end if
				
			for(IML_LINT T=0; T<SeqLength; T++)
				for(IML_LINT X=0; X<XSize; X++)
					XTSlice(X,0,T) = Uniform.sample(); // MAKE A X-T SLICE
//			for(IML_LINT I=0; I<XTSlice.Subspace.ElementCount; I++)
//				XTSlice(I) = Uniform.sample(); // MAKE A X-T SLICE
				
			for(IML_LINT Y=0; Y<YSize;Y++) { // RE ASSIGN TO ALL Y SLICES
				XTSlice.shift_hyperplane_corner_to(IML_2ST_Y,Y);
				MotionImage = XTSlice;
				} // end for		
			}; break;
			
		} // end switch
	
	IML_SINT ImSeqID = MacVideo.store_image_sequence(MotionImage);
	
	iml_imageframe_sequence SinglePassSeq(SeqLength);
	SinglePassSeq.generate_linear_sequence(0,1);
	
	//SinglePassSeq.sin( SeqLength/2.0, SeqLength/2.0-0.00001, SeqLength, 0.0);
	
	iml_imageframe_sequence MultiPassSeq(SeqLength*Repeats);
	MultiPassSeq = SinglePassSeq;
	
	IML_USINT BlitSeqID;
	IML_FLOAT PlayingTime;

	// HERE WE TEST STOING AND PLAYING GROUPS OF BLITS.

	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(0,500), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,0);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(100,500), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,1);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(200,500), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,2);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(300,500), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,3);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(0,600), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,0);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(100,600), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,1);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(200,600), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,2);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(300,600), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,3);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(0,300), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,0);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(100,300), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,1);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(200,300), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,2);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(300,300), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,3);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(0,400), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,0);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(100,400), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,1);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(200,400), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,2);
	BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, point_2d(300,400), MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID,3);

for(int Q=0;Q<12;Q++) MacVideo.add_to_group(Q,4);

	PlayingTime = MacVideo.play_blit_group(0);
	fprintf(stderr,"playing time is %f (FPS=%f)\n", PlayingTime, SeqLength*Repeats/PlayingTime);
	PlayingTime = MacVideo.play_blit_group(1);
	fprintf(stderr,"playing time is %f (FPS=%f)\n", PlayingTime, SeqLength*Repeats/PlayingTime);
	PlayingTime = MacVideo.play_blit_group(2);
	fprintf(stderr,"playing time is %f (FPS=%f)\n", PlayingTime, SeqLength*Repeats/PlayingTime);
	PlayingTime = MacVideo.play_blit_group(3);
	fprintf(stderr,"playing time is %f (FPS=%f)\n", PlayingTime, SeqLength*Repeats/PlayingTime);
	PlayingTime = MacVideo.play_blit_group(4);
	fprintf(stderr,"playing time is %f (FPS=%f)\n", PlayingTime, SeqLength*Repeats/PlayingTime);


} // end dynamic_noise


//======================================================================
//
//======================================================================
void drifting_noise_fourier_xform(mvl_video_manager & MacVideo,
		IML_USINT XSize, IML_USINT YSize, 
		IML_LINT WithWrap, IML_LINT WithLog10, IML_LINT WriteFiles) {

	// ALLOCATE AN IMAGE FILE MANAGER
	iml_image_file  OIFile;

	// MAKE A COUPLE OF IMAGES
	iml_fcomplex_image FCImage(iml_ld_index(XSize,YSize));
	iml_ubyte_image BImage(iml_ld_index(XSize,YSize));
	if( !FCImage.valid_memory() || !BImage.valid_memory() ) {
		iml_error("drifting_noise_fourier_xform","Could not allocate images!");
		return;
		}

	// US A UNIFORM ONE FOR KICKS (AND MAXIMUM RMS CONTRAST)
	uniform_rv Uniform;
	mvl_clock Timer;
	Uniform.seed(Timer.seed());
	Uniform.set_min_max(-1,1);
	
	// MAKE THE DRIFTING NOISE DOTS
	IML_LINT BufferSize = 2*XSize;
	if(WithWrap)BufferSize = XSize;
	
	IML_BYTE *Buffer = new IML_BYTE[BufferSize];
	for(long I=0; I< BufferSize; I++) 
		Buffer[I] = ((Uniform.sample()<0) ? -1 : 1);
	
	// MAKE A SPACE-TIME 2D IMAGE FOR THE DRIFTING NOISE
	BImage = 127;
	float XF=32, YF=32;
	for(long J=0; J< YSize; J++) {
		for(long I=0; I< XSize; I++)  {
			BImage(I,J) += Buffer[ (I+J)%BufferSize ]*123;
			FCImage(I,J).Real = Buffer[ (I+J)%BufferSize ];
			FCImage(I,J).Imag = 0.0;
			} // end for
		} // end for
		
	// PLUNK THE IMAGE DIRECTLY TO THE SCREEN
	MacVideo.display_image( BImage, point_2d(10,200) );

	// TRUE MEANS MOVE DC TO THE CENTER OF THE IMAGE
	FCImage.forward_FFT(TRUE);
	FCImage.goto_polar();
	if(WithLog10) {
		for(long K=0; K< FCImage.element_count(); K++)
			FCImage[K].Real = log10( FCImage[K].Real + 0.1);//1e-100;
		} // end if
	FCImage.real().scale_elements(2,253);
	
	// HERE IS A CHEAT TO GET AROUND THE SUB-SPACE INDEXING PROBLEM. :-)
	// IT WORKS BECAUSE THE IMAGES ARE THE SAME SIZE.  IF YOUR
	// IMAGES ARE OF DIFFERENT SIZE, THE RESULT IS UNPREDICTABLE.
	//BImage.Subspace = FCImage.Subspace;
	//BImage = FCImage.magnitude();

	// NOTE THAT THIS METHOD IS FAST, BUT BY-PASSES INDEX CHECKING
	// BUT IT AVOIDS THE SUBSPACE-ORIENTED ASSIGNMENT OPERATION.
	for(IML_LINT I=0; I<=FCImage.Subspace.ElementCount; I++)
		BImage[I] = FCImage[I].Real;
	
	// PLUNK THE IMAGE DIRECTLY TO THE SCREEN
	MacVideo.display_image( BImage, point_2d(XSize + 20,200) );

	if(WriteFiles) {
		OIFile.create_file("noise-fft-mag.raw");
		OIFile.write_image(&BImage, IML_RAW_FORMAT);
		OIFile.close_file();
		} // end if

	char TrashBuf[123];
	scanf("%s", TrashBuf);
} // end drifting_noise_fourier_xform

