
//#define IML_NO_SAFETY

// INCLUDE THE APPLICATION INDEPENDENT HELPER ROUTINES
// NOT DOING THIS FOR MAC YET . . .
//#include "imlapp.h"


#include "iml_file.h"
#include "iml_rofn.h"

// THESE REPRESENT TWO LEVELS OF SAFTEY FOR INDEXING.
// TURN AT LEAST ONE ON TO PREVENT BAD INDEX VALUES.
// THIS MAY SLOW YOUR PROGRAM.  NOTE THAT CORRECTIONS ARE
// SILENT.  BUT IF YOU ARE GETTING CRASHES WITH THIS OFF,
// THEN TURN THEM ON TO FIND OUT IF IT IS BECAUSE YOU ARE
// INDEXING IMAGES INCORRECTLY.
#include "mvl_macv.h"



//============================================================================
//
//============================================================================
int main () {

//======== EXAMPLE  CODE FOR CREATING AND INITIALIZING mac_video_manager

	
try {  
	// ALL THE CODE HERE IS SET FOR CATCHING EXCEPTIONS 
	// WITH AN EMPTY catch() AT THE BOTTOM.  

	// MAKE A VIDEO MANAGER OBJECT AND INITIALIZE IT
	// I'LL PROVIDE AND USE MORE ERROR CHECKING LATER . . .
	IML_USINT MaxSequenceCnt = 32;
	IML_USINT MaxBlitCnt = 64;
	IML_USINT GroupCount = 8;
	IML_ULINT ExitOnErr = TRUE;
	mvl_video_manager MacVideo(MaxSequenceCnt,MaxBlitCnt, GroupCount, ExitOnErr);
	
	MacVideo.initialize_screen(0);

char Junk1[128];
fprintf(stderr,"Move this window to where you want and then type a character<CR>\n");
scanf("%s",Junk1);

//	MacVideo.use_dgp_calibration();
	MacVideo.set_mvl_calibration(MacVideo.MaxDac, 2.08, 0, 100);
	MacVideo.set_attenuator_red_steps(0);
	MacVideo.use_mvl_calibration();
//	MacVideo.set_display_CLUT(0.5, 1.0);

	IML_LINT SeqLength = 16;
	IML_LINT  XSize= 200,  YSize= 100;
	IML_USINT Repeats=5;
	IML_FLOAT PeriodDelta = 1.0/SeqLength;
	IML_FLOAT ContrastRatio = 2.0;
	IML_FLOAT MaxContrast = 0.03;
	IML_FLOAT SpatialPeriod = XSize/5.0;
	IML_USINT DwellTime = 2;
	
	MacVideo.set_display_CLUT(0.5, MaxContrast);
		
	mvl_screen_position Position( 
		(MacVideo.XResolution-XSize)/2.0, (MacVideo.XResolution-XSize)/2.0 
		);

	// MAKE THE 3D IMAGE SEQUENCE
	iml_ubyte_image MotionImage( iml_ld_index(XSize, YSize ,SeqLength) );
	if( !MotionImage.valid_memory() ) {
		iml_error("example_motion","Could not allocate image!");
		return(-1);
		} // end if
	iml_ld_index TSize(XSize);
	iml_float_image Temp( TSize );
	
	uniform_rv Uniform;
	mvl_clock Timer;
	Uniform.seed(Timer.seed());
		
	iml_cos_rofn Cos(1);
	IML_FLOAT MaxDelta = 124.0;
	IML_FLOAT DriftMax = MaxDelta/(1+ContrastRatio);
	IML_FLOAT PedestalMax = MaxDelta-DriftMax;
	
	// SET UP THE PEDESTAL
	Cos.set_magnitude(PedestalMax);
	Cos.set_center(IML_2D_X, XSize/2.0 + 0.5);
	Cos.set_phase(Uniform.sample()*360);    // in degrees
	Cos.set_period(IML_2D_X, SpatialPeriod);   // pixels/cycle
	
	// SAVE A 1-D COPY
	Temp = Cos;
	
	// SET UP THE MEAN LUMINANCE AND ADD IN THE PEDESTAL
	MotionImage = 127; // set the mean luminance
	MotionImage += Temp;
	
	// SET UP DRIFTER'S MAGNITUDE
	Cos.set_magnitude(DriftMax);
	
	IML_FLOAT StartPhase = 0;//Uniform.sample()*360;
	IML_FLOAT IncPhase = 360*PeriodDelta;

	// WE WILL CREATE THE IMAGE ONE SLICE AT TIME, SO RESTRICT TO 2D SUBSPACE
	MotionImage.set_hyperplane_size(2);
	
	// DO THIS TO SHOW THE STATIC IMAGE
//	MacVideo.display_image( MotionImage, Position );

	// NOW, ONE SLICE AT A TIME . . .
	for(IML_ULINT K=0; K<SeqLength; K++) {

		// SET CURRENT SLICE
		MotionImage.select_hyperplane(K);

		Cos.set_phase(StartPhase+IncPhase*K);
		Temp = Cos;
		MotionImage += Temp;

	// TEMPORARY FOR CHECKING
	MacVideo.display_image( MotionImage, Position );
//	MacVideo.start_timer();
//	while(MacVideo.get_timer() < 0.1) {};
		
		} // end for
	
	// RESET SUBSPACE SIZE TO 3D FOR MAC VIDEO STORAGE
	MotionImage.set_hyperplane_size(3);
	IML_USINT ImSeqID = MacVideo.store_image_sequence(MotionImage);
	
	// MAKE A ONE-PERIOD DISPLAY
	IML_LINT TotalFrames = SeqLength*DwellTime;
	iml_imageframe_sequence SinglePassSeq(TotalFrames);
	for(IML_LINT L=0;L<TotalFrames;L++) SinglePassSeq(L)=L/DwellTime;
	
	iml_imageframe_sequence MultiPassSeq(TotalFrames*Repeats);
	MultiPassSeq = SinglePassSeq;
	
	Position.YPos-=YSize/2;
	IML_USINT BlitSeqID1 = MacVideo.create_blit_sequence(ImSeqID, Position, MultiPassSeq);
	MultiPassSeq.reverse();
	Position.YPos+=YSize;
	IML_USINT BlitSeqID2 = MacVideo.create_blit_sequence(ImSeqID, Position, MultiPassSeq);
	MacVideo.add_to_group(BlitSeqID1,0);
	MacVideo.add_to_group(BlitSeqID2,0);


	char Junk2[128];
	
	do {
//		if(Junk2[0] = 'j')
//			MacVideo.play_blit_sequence(BlitSeqID1);
//		else
//			MacVideo.play_blit_sequence(BlitSeqID2);
		MacVideo.play_blit_group(0);

		fprintf(stderr,"Do you want to see that again?\n");
		scanf("%s",Junk2);
	} while(Junk2[0] != 'q');


	} // end try
	
catch ( ... ) {
// THIS IS JUST TO ALLOW TERMINATION OF THE PROGRAM AND CLEAN UP OF STUFF.
	}

	return(0);
} // end run_experiment


