
//#define IML_NO_SAFETY

// INCLUDE THE APPLICATION INDEPENDENT HELPER ROUTINES
// NOT DOING THIS FOR MAC YET . . .
//#include "imlapp.h"


#include "iml_file.h"
#include "iml_rofn.h"

// THESE REPRESENT TWO LEVELS OF SAFTEY FOR INDEXING.
// TURN AT LEAST ONE ON TO PREVENT BAD INDEX VALUES.
// THIS MAY SLOW YOUR PROGRAM.  NOTE THAT CORRECTIONS ARE
// SILENT.  BUT IF YOU ARE GETTING CRASHES WITH THIS OFF,
// THEN TURN THEM ON TO FIND OUT IF IT IS BECAUSE YOU ARE
// INDEXING IMAGES INCORRECTLY.
#include "mvl_macv.h"


void SecondOrder_pedestal_motion(mvl_video_manager & MacVideo,
		IML_LINT  XSize, IML_LINT YSize, IML_LINT SeqLength, 
		IML_FLOAT PeriodDelta, IML_USINT Repeats,
		IML_FLOAT ContrastRatio, IML_FLOAT MaxContrast,
		mvl_screen_position IML_REF Position);	
		

//============================================================================
//
//============================================================================
#ifdef _Windows
int run_experiment(void) {
#else
int main () {
#endif

//======== EXAMPLE  CODE FOR CREATING AND INITIALIZING mvl_video_manager

	
try {  
	// ALL THE CODE HERE IS SET FOR CATCHING EXCEPTIONS 
	// WITH AN EMPTY catch() AT THE BOTTOM.  

	// MAKE A VIDEO MANAGER OBJECT AND INITIALIZE IT
	// I'LL PROVIDE AND USE MORE ERROR CHECKING LATER . . .
	IML_USINT MaxSequenceCnt = 32;
	IML_USINT MaxBlitCnt = 64;
	IML_USINT GroupCount = 8;
	IML_ULINT ExitOnErr = TRUE;
	mvl_video_manager MacVideo(MaxSequenceCnt,MaxBlitCnt, GroupCount, ExitOnErr);
	
	MacVideo.initialize_screen(0);
//	MacVideo.use_dgp_calibration();
	MacVideo.set_mvl_calibration(MacVideo.MaxDac, 2, 0, 100);
	MacVideo.set_attenuator_red_steps(0.0);
	MacVideo.use_mvl_calibration();
	MacVideo.set_display_CLUT(0.5,1.0);

	IML_LINT SeqLength = 16;
	IML_LINT  XSize= 200,  YSize= 200;
	IML_USINT Repeats=20;
	IML_FLOAT PeriodDelta = 1.0/(SeqLength-1);
	IML_FLOAT ContrastRatio = 3.0;
	IML_FLOAT MaxContrast = 0.5;
	
	MacVideo.set_display_CLUT(0.5,MaxContrast);

	SecondOrder_pedestal_motion(MacVideo, XSize, YSize, SeqLength, 
		PeriodDelta, Repeats, ContrastRatio, MaxContrast,
		mvl_screen_position( (MacVideo.XResolution-XSize)/2,
			(MacVideo.YResolution-YSize)/2) );
		
	} // end try
	
catch ( ... ) {
// THIS IS JUST TO ALLOW TERMINATION OF THE PROGRAM AND CLEAN UP OF STUFF.
	}

	return(0);
} // end run_experiment


//====================================================================
//
//======================================================================
void SecondOrder_pedestal_motion(mvl_video_manager & MacVideo,
		IML_LINT  XSize, IML_LINT YSize, IML_LINT SeqLength, 
		IML_FLOAT PeriodDelta, IML_USINT Repeats,
		IML_FLOAT ContrastRatio, IML_FLOAT MaxContrast,
		mvl_screen_position IML_REF Position) {		

	// MAKE THE 3D IMAGE SEQUENCE
	iml_ubyte_image MotionImage( iml_ld_index(XSize, YSize ,SeqLength) );
	if( !MotionImage.valid_memory() ) {
		iml_error("SecondOrder_pedestal_motion","Could not allocate MotionImage!");
		return;
		} // end if

	iml_float_image NoiseFrame( iml_ld_index(XSize, YSize) );
	if( !NoiseFrame.valid_memory() ) {
		iml_error("SecondOrder_pedestal_motion","Could not allocate NoiseFrame!");
		return;
		} // end if

	iml_ld_index TSize(XSize);
	iml_float_image Pedestal( TSize );
	iml_float_image Buffer( TSize );
	if( !Pedestal.valid_memory()  ||  !Buffer.valid_memory() ) {
		iml_error("SecondOrder_pedestal_motion","Could not allocate Pedestal or Buffer!");
		return;
		} // end if
	
	mvl_clock Timer;
		
	uniform_rv Uniform; // this is for the carrier
	Uniform.seed(Timer.seed());
	
	iml_cos_rofn Cos(1);	
	Cos.set_center(IML_2D_X, XSize/2.0 + 0.5);	
	Cos.set_phase(Uniform.sample()*360);    // in degrees
	Cos.set_period(IML_2D_X, 64);   // pixels/cycle


	// ARBITRARY PEDESTAL MAGNITUDE RIGHT NOW
	Cos.set_magnitude(MaxContrast*(ContrastRatio/(ContrastRatio+1))); 
	
	// MAKE A ONE-D ARRAY FOR THE PEDESTAL
	Pedestal = Cos;
	
	// NOW THE COS IN THE DRIFTER
	Cos.set_magnitude(MaxContrast/(ContrastRatio+1));
	
	IML_FLOAT StartPhase = Uniform.sample()*360;
	IML_FLOAT IncPhase1 = 360*PeriodDelta;

	// MAKE ONE FRAME OF CARRIER NOISE (RANDOM - ALL PIXELS);
	for(IML_LINT I=0; I<NoiseFrame.ImageSpec.ElementCount; I++) {
		IML_FLOAT Temp = -120 + 240 * (Uniform.sample()>0.5);
		NoiseFrame[I] = Temp;
		}
		
	// WE WILL CREATE THE IMAGE ONE SLICE AT TIME, SO RESTRICT TO 2D SUBSPACE
	MotionImage.set_hyperplane_size(2);
	
	Timer.start();
	// NOW, ONE SLICE AT A TIME . . .
	for(IML_ULINT K=0; K<SeqLength; K++) {

		// SET CURRENT SLICE
		MotionImage.select_hyperplane(K);

		// construct one-d pedestal plus drifter
		Cos.set_phase(StartPhase+IncPhase1*K);
		Buffer  = 0.5;
		Buffer += Cos;
		Buffer += Pedestal;
		
		// MODULATE THE NOISE IN THE FRAME.
		for(IML_LINT I=0; I<NoiseFrame.ImageSpec.ElementCount; I++) {
			MotionImage(I) = 127 + NoiseFrame(I) * Buffer(I%XSize);
			}
				
		} // end for
	
	IML_FLOAT Time = Timer.time();
	fprintf(stderr,"frame generation %f\n",Time/SeqLength);

	// RESET SUBSPACE SIZE TO 3D FOR MAC VIDEO STORAGE
	MotionImage.set_hyperplane_size(3);
	IML_USINT ImSeqID = MacVideo.store_image_sequence(MotionImage);
	
	iml_imageframe_sequence SinglePassSeq(SeqLength);
	SinglePassSeq.generate_linear_sequence(0,1);
	
	//SinglePassSeq.sin( SeqLength/2.0, SeqLength/2.0-0.00001, SeqLength, 0.0);
	
	iml_imageframe_sequence MultiPassSeq(SeqLength*Repeats);
	MultiPassSeq = SinglePassSeq;
	
	
	IML_USINT BlitSeqID1 = MacVideo.create_blit_sequence(ImSeqID, Position, MultiPassSeq);
	MultiPassSeq.reverse();
	IML_USINT BlitSeqID2 = MacVideo.create_blit_sequence(ImSeqID, Position, MultiPassSeq);

	MacVideo.play_blit_sequence(BlitSeqID1);
	Timer.wait( 3);
	MacVideo.play_blit_sequence(BlitSeqID2);


} // end SecondOrder_pedestal_motion


