
//#define IML_NO_SAFETY

// INCLUDE THE APPLICATION INDEPENDENT HELPER ROUTINES
// NOT DOING THIS FOR MAC YET . . .
//#include "imlapp.h"


#include "iml_file.h"
#include "iml_rofn.h"

// THESE REPRESENT TWO LEVELS OF SAFTEY FOR INDEXING.
// TURN AT LEAST ONE ON TO PREVENT BAD INDEX VALUES.
// THIS MAY SLOW YOUR PROGRAM.  NOTE THAT CORRECTIONS ARE
// SILENT.  BUT IF YOU ARE GETTING CRASHES WITH THIS OFF,
// THEN TURN THEM ON TO FIND OUT IF IT IS BECAUSE YOU ARE
// INDEXING IMAGES INCORRECTLY.
#include "mvl_macv.h"


void drifting_radial_gabor(mvl_video_manager & MacVideo, 
			IML_USINT XSize, IML_USINT YSize, IML_LINT SeqLength, 
				IML_USINT RepeatCount, IML_FLOAT SpatialPeriod, IML_FLOAT SDev,
				IML_FLOAT DriftFrequency);
			

//============================================================================
//
//============================================================================
#ifdef _Windows
int run_experiment(void) {
#else
int main () {
#endif
//#pragma argsused
//int  run_experiment(char *CmdLineString) {


//======== EXAMPLE  CODE FOR CREATING AND INITIALIZING mvl_video_manager


try {  
	// ALL THE CODE HERE IS SET FOR CATCHING EXCEPTIONS 
	// WITH AN EMPTY catch() AT THE BOTTOM.  

	// MAKE A VIDEO MANAGER OBJECT AND INITIALIZE IT
	// I'LL PROVIDE AND USE MORE ERROR CHECKING LATER . . .
	IML_USINT MaxSequenceCnt = 32;
	IML_USINT MaxBlitCnt = 64;
	IML_USINT GroupCount = 8;
	IML_ULINT ExitOnErr = TRUE;
	mvl_video_manager MacVideo(MaxSequenceCnt,MaxBlitCnt, GroupCount, ExitOnErr);
	
	MacVideo.initialize_screen(0);
//	MacVideo.use_dgp_calibration();
	MacVideo.set_mvl_calibration(MacVideo.MaxDac, 2, 0, 100);
	MacVideo.set_attenuator_red_steps(0.0);
	MacVideo.use_mvl_calibration();
	
	MacVideo.set_mean_luminance(0.5);
	MacVideo.set_max_contrast(1.0);
	MacVideo.set_display_CLUT(0.5,1.0);
	
	IML_LINT  XSize= 128,  YSize= 128;
	IML_FLOAT SpatialPeriod=XSize/8, SDev = XSize/5;
	IML_FLOAT DriftFrequency = 1; // cycles per display period (Seqlength)
	IML_LINT SeqLength = 16;
	IML_USINT RepeatCount=20;
	drifting_radial_gabor(MacVideo, XSize, YSize, SeqLength, RepeatCount, 
				SpatialPeriod, SDev, DriftFrequency);
				
	} // end try
	
catch ( ... ) {
// THIS IS JUST TO ALLOW TERMINATION OF THE PROGRAM AND CLEAN UP OF STUFF.
	}

	return(0);
} // end run_experiment


//======================================================================
//
//======================================================================
void drifting_radial_gabor(mvl_video_manager & MacVideo, 
				IML_USINT XSize, IML_USINT YSize, IML_LINT SeqLength, 
				IML_USINT RepeatCount, IML_FLOAT SpatialPeriod, IML_FLOAT SDev,
				IML_FLOAT DriftFrequency) {

/*
	// MAKE A TEMPORAL WINDOW CONTROL THINGY AND LOAD IT WITH A 
	// RAISED-COSINE-EDGE.  COSINUSOIDAL EDGES MUST BE <= HALF SEQUENCE.
	IML_USINT HalfPeriod = (10 < SeqLength/2 ? 10 : SeqLength/2);
	iml_temporal_window TWindow(SeqLength); 
	
	TWindow = 1.0; // MAX CONTRAST IN THE MIDDLE
	for(IML_LINT I=0;I<HalfPeriod;I++) // MAKE THE EDGES
		TWindow[I] = TWindow[SeqLength-I]= 
					0.5+0.5*iml_sin(IML_PI*I/HalfPeriod - IML_2PI/4);

	// CREATE THE CLUT TABLES REQUIRED FOR DISPLAY
	MacVideo.initialize_CLUT_tables(TWindow);
	MacVideo.use_CLUT_sequence(TWindow);
*/	
	
	IML_FLOAT XCenter = (XSize-1.0)/2.0, YCenter = (YSize-1.0)/2.0;
	iml_polar_cos_rofn PCos(2);
	PCos.set_magnitude(100);
	PCos.set_center(IML_2D_X, XCenter);
	PCos.set_center(IML_2D_Y, YCenter);
	PCos.set_phase(0);    // in degrees
	PCos.set_period(IML_2D_X, SpatialPeriod);   // pixels/cycle
//	PCos.set_period(IML_2D_X, 0);   // pixels/cycle
		// FOR POLAR, X1 (Y) frequency IS IN CYCLES/2pi; period is 2pi/cycles
	PCos.set_period(IML_2D_Y, IML_2PI/8);  
	 // zero flags zero frequency	
	
	// CREATE AN UNSIGNED BYTE IMAGE OF THIS SIZE
	iml_ld_index Index(XSize, YSize, SeqLength);
	iml_ubyte_image UByteImage(Index);
	if( !UByteImage.valid_memory() ) {
		iml_error("drifting_radial_gabor","Could not allocate image!");
		return;
		}

//	UByteImage = 127;
	// RESTRICT IMAGE OPERATIONS TO 2D SUBSPACE
	UByteImage.set_hyperplane_size(2);
	
	iml_ld_index TDesc(XSize,YSize);
	iml_ubyte_image H(TDesc);
	
	mvl_clock Clock1, Clock2;
	// NUMBER OF TOTAL
	Clock1.start();
	for(IML_LINT I=0; I<SeqLength; I++) {
	
		// SELECT A 2D IMAGE SLICE; 
		// [A MORE COMPLICATED(?) METHOD OF SELECTING A SUBSPACE THAN IN
		// mckee_welch_randots ABOVE.]
		UByteImage.select_hyperplane(I);
		
		PCos.set_phase(360*DriftFrequency*(I/(float)SeqLength));    // in degrees
		// save the image
//		H = 127; 
		memset(&H(0), 127, XSize*(long)YSize);
		H += PCos;
		
 		//this one is by far the fastest
//		for(IML_LINT YY=0;YY<YSize;YY++)
//				memcpy(&UByteImage(0,YY), &H(0), XSize);

		// SET CARRIER PHASE;
		UByteImage = H; // runs at ~223,000 assignments/sec (no optimization)
//		UByteImage += PCos; // runs at ~60,000 assignments/sec (no optimization)
//		MacVideo.display_image(UByteImage, mvl_screen_position(100,100) );
		
//		Clock2.wait(0.5);
		} // end for
	IML_DOUBLE Time = Clock1.time();
	fprintf(stderr,"PCos ops/sec : %f\n",(XSize * YSize *(double) SeqLength)/Time);
		
	// GO BACK TO 3D FORMAT FOR DISPLAY
	UByteImage.set_hyperplane_size(3);
	// now the image should contain a spatially-gaussian-windowed drifting sine wave.


	IML_SINT ImSeqID = MacVideo.store_image_sequence(UByteImage);
	if( ImSeqID < 0 ) {
		iml_error("drifting_radial_gabor","FAILED ON IMAGE STORE");
		return;
		} // end if

	// MAKE A SEQUENCE TO PLAY THE SEQUENCE FORWARD
	iml_imageframe_sequence FrameSequence(SeqLength);
	FrameSequence.generate_linear_sequence(/*from*/ 0, /*increment*/ 1);
	//FrameSequence.shuffle();
	//FrameSequence.sin( SeqLength/2.0, SeqLength/2.0-0.00001, SeqLength, 90.0);
	//FrameSequence.reverse();
//	FrameSequence.rotate(-12);

	IML_LINT PlayFrames = SeqLength*RepeatCount;
	iml_imageframe_sequence LongSequence(PlayFrames);
	LongSequence = FrameSequence;
	
	// AT THIS POSITION
	mvl_screen_position ImageDest(200,200);
	
	// CREATE THE BLIT SEQUENCE
	IML_SINT BlitSeqID = MacVideo.create_blit_sequence(ImSeqID, ImageDest, LongSequence);

// DISPLAY A FEW FRAMES FOR FUN

//	MacVideo.display_frame(ImSeqID, 0, mvl_screen_position(0,0) );
//	MacVideo.display_frame(ImSeqID, 1, mvl_screen_position(300,0) );
//	MacVideo.display_frame(ImSeqID, 2, mvl_screen_position(0,330) );

		// PLAY THE BLIT SEQUENCE
	IML_FLOAT PlayingTime = MacVideo.play_blit_sequence(BlitSeqID);

	printf("Playing time is %f or %f frames/sec\n",
				PlayingTime, PlayFrames/PlayingTime);
	Clock1.wait(5);

} // end drifting_radial_gabor
