

// INCLUDE THE APPLICATION INDEPENDENT HELPER ROUTINES
// NOT DOING THIS FOR MAC YET . . .
//#include "imlapp.h"


#include "iml_file.h"
#include "iml_rofn.h"

// THESE REPRESENT TWO LEVELS OF SAFTEY FOR INDEXING.
// TURN AT LEAST ONE ON TO PREVENT BAD INDEX VALUES.
// THIS MAY SLOW YOUR PROGRAM.  NOTE THAT CORRECTIONS ARE
// SILENT.  BUT IF YOU ARE GETTING CRASHES WITH THIS OFF,
// THEN TURN THEM ON TO FIND OUT IF IT IS BECAUSE YOU ARE
// INDEXING IMAGES INCORRECTLY.
#include "mvl_macv.h"

void SDLT_random_pixel_array(mvl_video_manager & MacVideo,
		IML_LINT  XSize, IML_LINT YSize, IML_LINT SeqLength, IML_USINT Repeats, 
		IML_FLOAT LSNR, IML_FLOAT RMSContrast,
		vector_2d & Delta, 
		vector_2d & Position);

void SDLT_random_pixel_array_fft(mvl_video_manager & MacVideo,
		IML_LINT  XSize, IML_LINT YSize, IML_LINT SeqLength,
		IML_FLOAT LSNR, IML_FLOAT RMSContrast,
		vector_2d & Delta,
		vector_2d & Position);
		


char get_key_stroke(void);


char get_key_stroke(void) {
static	EventRecord AnyEvent;
	FlushEvents(everyEvent,0);
	while( !GetNextEvent( keyDownMask | autoKeyMask, &AnyEvent ) );
	return(AnyEvent.message&0x000000FF);
}

//============================================================================
//
//============================================================================
int main () {

//======== EXAMPLE  CODE FOR CREATING AND INITIALIZING mvl_video_manager

	
try {  
	// ALL THE CODE HERE IS SET FOR CATCHING EXCEPTIONS 
	// WITH AN EMPTY catch() AT THE BOTTOM.  

	// MAKE A VIDEO MANAGER OBJECT AND INITIALIZE IT
	// I'LL PROVIDE AND USE MORE ERROR CHECKING LATER . . .
	IML_USINT MaxSequenceCnt = 32;
	IML_USINT MaxBlitCnt = 64;
	IML_USINT GroupCount = 8;
	IML_ULINT ExitOnErr = TRUE;
	mvl_video_manager MacVideo(MaxSequenceCnt,MaxBlitCnt, GroupCount, ExitOnErr);
	
	MacVideo.initialize_screen(0);
	MacVideo.use_mvl_calibration();
	MacVideo.set_mvl_calibration(MacVideo.MaxDac, 2, 0, 100);
	MacVideo.set_attenuator_red_steps(0.0);
	MacVideo.linearize_clut(0.5,1.0);

enum {
		SDLT_RPA,
		SDLT_RPA_FFT,
		OPTION_COUNT
		};

	IML_LINT DemoSelector = SDLT_RPA_FFT;

	switch(DemoSelector) {
	
		
		case SDLT_RPA: {
			MacVideo.set_display_CLUT(0.5,1.0);
			IML_LINT  XSize= 128,  YSize= XSize, SeqLength = 16;
			IML_LINT Repeats = 40;
			IML_FLOAT LSNR = 0.1, RMSContrast = 0.5;
			vector_2d Delta(2,2);
			vector_2d Position(MacVideo.get_screen_position(mvl_video_manager::CENTER,XSize,YSize));

			SDLT_random_pixel_array(MacVideo, XSize, YSize, SeqLength,
				Repeats, LSNR, RMSContrast, Delta, Position);

			}; break;
	
		case SDLT_RPA_FFT: {
			MacVideo.linearize_clut(0.5,1.0);
			IML_LINT  XSize= 128,  YSize= XSize, SeqLength = 64;
			IML_FLOAT LSNR = 10000, RMSContrast = 0.3;
			vector_2d Delta(0,0);
			vector_2d Position(100,100);

			SDLT_random_pixel_array_fft(MacVideo, XSize, YSize, SeqLength,
				LSNR, RMSContrast, Delta, Position);

			// WAIT A BIT SO WE CAN SEE THE IMAGES
			get_key_stroke();
			}; break;
		} // end switch
	
	}
	
catch ( ... ) {
// THIS IS JUST TO ALLOW TERMINATION OF THE PROGRAM AND CLEAN UP OF STUFF.
	}

	return(0);
} // end run_experiment


//======================================================================
//
//======================================================================
void SDLT_random_pixel_array(mvl_video_manager & MacVideo,
		IML_LINT  XSize, IML_LINT YSize, IML_LINT SeqLength, IML_USINT Repeats,
		IML_FLOAT LSNR, IML_FLOAT RMSContrast,
		vector_2d IML_REF Delta,
		vector_2d IML_REF Position) {

	// CREATE THE SPACE-TIME IMAGE SEQUENCE
//	iml_ld_index ThreeDSize();	
	iml_ubyte_image RPImage(XSize,YSize,SeqLength);
	if( !RPImage.valid_memory() ) {
		iml_error("SDLT_random_pixel_array","Could not allocate DotImage image!");
		return;
		} // end if


	// THESE VALUES ARE CALCULATED AS A FRACTION OF MAXIMUM CONTRAST
	IML_FLOAT DeltaN = RMSContrast*iml_sqrt(1.0/(1.0+LSNR));
	IML_FLOAT DeltaSComposite = RMSContrast*iml_sqrt(LSNR/(1.0+LSNR));
	IML_FLOAT DeltaS = DeltaSComposite/IML_SQRT2;
	
	// THESE ARE FOR GENERATING THE PIXEL VALUES USING DeltaN, DeltaS
	IML_UBYTE MeanLuminance = 127;
	IML_UBYTE MaxDelta = 126;
	
	// THESE ARE THE DISCRETE DELTA'S FOR SIGNAL AND NOISE
	IML_BYTE SignalDelta = MaxDelta*DeltaS;
	IML_BYTE NoiseDelta  = MaxDelta*DeltaN;
	
	// BUFFER SIZE IS IMAGE WIDTH PLUS JUMP SIZE
	iml_ld_index 
		BufferSize( XSize + iml_abs(Delta.X), YSize + iml_abs(Delta.Y));
	
	// MAKE THE BUFFERS
	iml_byte_image Signal1(BufferSize);
	iml_byte_image Signal2(BufferSize);
	if( !Signal1.valid_memory() || !Signal2.valid_memory() ) {
		iml_error("SDLT_random_pixel_array_fft","Could not allocate Signal images!");
		return;
		} // end if
			
	// SET BUFFER INDICES SO THAT WE CAN DO THE DISPLACEMENT THING
	if(Delta.X > 0) {
		Signal1.Subspace.shift_corner_by(IML_2D_X, -Delta.X);
		Signal2.Subspace.shift_corner_by(IML_2D_X, -Delta.X);
		} // end if
				
	if(Delta.Y > 0) {
		Signal1.Subspace.shift_corner_by(IML_2D_Y, -Delta.Y);
		Signal2.Subspace.shift_corner_by(IML_2D_Y, -Delta.Y);
		} // end if
	
	// MAKE A COPY OF THE IMAGE CORNER START
	iml_index Start(Signal1.Subspace);
	
	mvl_clock Timer;
	uniform_rv Uniform;
	Uniform.seed(Timer.seed());
	Uniform.set_min_max(-1,1);

	// SET INITIAL Pixels
	for(IML_ULINT I=0; I<BufferSize.ElementCount; I++) {
		Signal1[I] = ((Uniform.sample()<0) ? -SignalDelta : SignalDelta);
		Signal2[I] = ((Uniform.sample()<0) ? -SignalDelta : SignalDelta);
		} // end for

	RPImage = MeanLuminance;
	// ADD NOISE FOR ALL FRAMES AT ONCE (BECAUSE IT IS EASY)
	for(IML_ULINT I=0; I<RPImage.Subspace.ElementCount; I++) {
		RPImage[I] += ((Uniform.sample()<0) ? -NoiseDelta  : NoiseDelta);
		} // end for

	// PREPARE TO MAKE ONE SLICE AT A TIME
	RPImage.set_hyperplane_size(2);

	for(IML_ULINT K=0; K<SeqLength; K++) {
					
		// PREPARE TO MAKE ONE SLICE AT A TIME
		RPImage.select_hyperplane(K);

		// UPDATE SIGNAL DOT POSITIONS (SELF INDEXING FOR SPEED)
		for(IML_LINT J=0; J<YSize; J++)
			for(IML_LINT I=0; I<XSize; I++)
				RPImage(I,J) += Signal1(I,J) + Signal2(I,J);
				
		MacVideo.display_image(RPImage, Position);
		// WAIT A BIT SO WE CAN SEE THE IMAGES
		Timer.wait(0.1);

		if( K%2 ) {
		
			// SHIFT THE SIGNAL COMPONENT WHICH WILL BE DISPLACED NEXT FRAME
			Signal1.Subspace.shift_corner_by(IML_2D_X, Delta.X);
			Signal1.Subspace.shift_corner_by(IML_2D_Y, Delta.Y);
			
			// RESTART THE REFRESHED COMPONENT
			Signal2.Subspace = Start;			
			// RANDOMIZE SIGNAL DOT POSITIONS
			for(IML_ULINT I=0; I<BufferSize.ElementCount; I++)
				Signal2[I] = ((Uniform.sample()<0) ? -SignalDelta : SignalDelta);
		} else {
			// SHIFT THE SIGNAL COMPONENT WHICH WILL BE DISPLACED NEXT FRAME
			Signal2.Subspace.shift_corner_by(IML_2D_X, Delta.X);
			Signal2.Subspace.shift_corner_by(IML_2D_Y, Delta.Y);
			
			// RESTART THE REFRESHED COMPONENT
			Signal1.Subspace = Start;			
			// RANDOMIZE SIGNAL DOT POSITIONS
			for(IML_ULINT I=0; I<BufferSize.ElementCount; I++)
				Signal1[I] = ((Uniform.sample()<0) ? -SignalDelta : SignalDelta);
			} // end else
						
		} // end for
		
	// RESET SUBSPACE SIZE TO 3D FOR MAC VIDEO STORAGE
	RPImage.set_hyperplane_size(3);
	IML_USINT ImSeqID = MacVideo.store_image_sequence(RPImage);
	
	iml_imageframe_sequence SinglePassSeq(SeqLength);
	SinglePassSeq.generate_linear_sequence(0,1);
	
	//SinglePassSeq.sin( SeqLength/2.0, SeqLength/2.0-0.00001, SeqLength, 0.0);
	
	iml_imageframe_sequence MultiPassSeq(SeqLength*Repeats);
	MultiPassSeq = SinglePassSeq;
	
	
	IML_USINT BlitSeqID1 = MacVideo.create_blit_sequence(ImSeqID, Position, MultiPassSeq);
	MultiPassSeq.reverse();
	IML_USINT BlitSeqID2 = MacVideo.create_blit_sequence(ImSeqID, Position, MultiPassSeq);

	MacVideo.play_blit_sequence(BlitSeqID1);
	Timer.wait(3);
	MacVideo.play_blit_sequence(BlitSeqID2);



} // end SDLT_random_pixel_array


//======================================================================
//
//======================================================================
void SDLT_random_pixel_array_fft(mvl_video_manager & MacVideo,
		IML_LINT  XSize, IML_LINT YSize, IML_LINT SeqLength,
		IML_FLOAT LSNR, IML_FLOAT RMSContrast,
		vector_2d IML_REF Delta,
		vector_2d IML_REF Position) {

IML_LINT Amplitude = TRUE;
IML_LINT MakeAxes  = TRUE;

	iml_ubyte_image RPImage(XSize,YSize);
	if( !RPImage.valid_memory() ) {
		iml_error("SDLT_random_pixel_array_fft","Could not allocate DotImage image!");
		return;
		} // end if
	iml_fcomplex_image FImage(XSize,YSize);
	if( !FImage.valid_memory() ) {
		iml_error("SDLT_random_pixel_array_fft","Could not allocate FImage image!");
		return;
		} // end if
	iml_double_image Averager(XSize,YSize);
	if( !Averager.valid_memory() ) {
		iml_error("SDLT_random_pixel_array_fft","Could not allocate Averager image!");
		return;
		} // end if

	iml_double_image OneDeeAvg(YSize);
	if( !OneDeeAvg.valid_memory() ) {
		iml_error("SDLT_random_pixel_array_fft","Could not allocate OneDeeAvg image!");
		return;
		} // end if


	// THESE VALUES ARE CALCULATED AS A FRACTION OF MAXIMUM CONTRAST
	IML_FLOAT DeltaN = RMSContrast*iml_sqrt(1.0/(1.0+LSNR));
	IML_FLOAT DeltaSComposite = RMSContrast*iml_sqrt(LSNR/(1.0+LSNR));
	IML_FLOAT DeltaS = DeltaSComposite/IML_SQRT2;
	
	// THESE ARE FOR GENERATING THE PIXEL VALUES USING DeltaN, DeltaS
	IML_UBYTE MeanLuminance = 127;
	IML_UBYTE MaxDelta = 126;
	
	// THESE ARE THE DISCRETE DELTA'S FOR SIGNAL AND NOISE
	IML_BYTE SignalDelta = MaxDelta*DeltaS;
	IML_BYTE NoiseDelta  = MaxDelta*DeltaN;
	
	// BUFFER SIZE IS IMAGE WIDTH PLUS JUMP SIZE
	IML_LINT BufferSize = XSize + Delta.X;
	
	// MAKE THE BUFFERS
	iml_byte_sequence Signal1(BufferSize);
	iml_byte_sequence Signal2(BufferSize);
	iml_byte_sequence Noise(BufferSize);
	
	mvl_clock Timer;
	uniform_rv Uniform;
	Uniform.seed(Timer.seed());
	Uniform.set_min_max(-1,1);

	// SET INITIAL Pixels
	for(IML_ULINT I=0; I<BufferSize; I++) {
		Signal1[I] = ((Uniform.sample()<0) ? -SignalDelta : SignalDelta);
		Signal2[I] = ((Uniform.sample()<0) ? -SignalDelta : SignalDelta);
		Noise[I]   = ((Uniform.sample()<0) ? -NoiseDelta  : NoiseDelta);
		} // end for


	iml_fcomplex Zero(0,0);
	Averager = 0.0;
	IML_DOUBLE Divisor = (SeqLength*iml_sqrt(XSize*(float)YSize));
	for(IML_ULINT K=0; K<SeqLength; K++) {

		RPImage=MeanLuminance;
		FImage = Zero;
		FImage.non_fourier_domain();
		// generate an image
		for(IML_ULINT J=0; J<YSize; J++) {
			
			// ADD SIGNALS & NOISE TO IMAGE
			for(IML_ULINT I=0; I<XSize; I++) { 
				FImage(I, J).Real += Signal1[I] + Signal2[I] + Noise[I];
				RPImage(I, J) += Signal1[I] + Signal2[I] + Noise[I];
				} // END FOR
			
			if( J%2 ) {
				// UPDATE SIGNAL DOT POSITIONS
				Signal1.rotate(Delta.X);
				// RANDOMIZE SIGNAL DOT POSITIONS
				for(IML_ULINT I=0; I<BufferSize; I++)
					Signal2[I] = ((Uniform.sample()<0) ? -SignalDelta : SignalDelta);
			} else {
				// UPDATE SIGNAL DOT POSITIONS
				Signal2.rotate(Delta.X);
				// RANDOMIZE SIGNAL DOT POSITIONS
				for(IML_ULINT I=0; I<BufferSize; I++)
					Signal1[I] = ((Uniform.sample()<0) ? -SignalDelta : SignalDelta);
				} // end else
							
			// RANDOMIZE NOISE DOT POSITIONS
			for(IML_ULINT I=0; I<BufferSize; I++)
				Noise[I] = ((Uniform.sample()<0) ? -NoiseDelta  : NoiseDelta);
			
			} // end for
		
		MacVideo.display_image(RPImage, Position);				
		FImage.forward_FFT(TRUE);


		// ADD NEW INFO TO AVERAGER
		if(Amplitude) {
			for(IML_LINT I=0; I<=Averager.Subspace.ElementCount; I++) {
				float Sign = FImage[I].Real > 0 ? 1 : -1;
				Averager[I] += Sign * FImage[I].get_mag()/Divisor;
				}
		} else {
			for(IML_LINT I=0; I<=Averager.Subspace.ElementCount; I++)
				Averager[I] += FImage[I].get_mag()/Divisor;
				} // end else

		// IT IS NOT NECESSARY TO ZERO OUT THE DC COMPONENTS
		//for(IML_LINT I=0; I<=Averager.Subspace.ElementCount; I++)
		//	Averager[I] += FImage[I].get_mag()/Divisor;

		} // end for
		
	IML_DOUBLE Min, Max;
	Averager.min_max_element(Min,Max);
	fprintf(stderr,"min -- max: %g -- %g\n",Min,Max);
	Averager.scale_elements(10,200);
	for(IML_ULINT I=0; I<Averager.Subspace.ElementCount; I++)
			RPImage[I] = Averager[I];
			
{
OneDeeAvg = 0.0;
for(long YY=0;YY<YSize; YY++) {
	for(long XX=0;XX<XSize;XX++) {
		OneDeeAvg[YY] += Averager[YY*XSize + XX];
		}
	}
OneDeeAvg *= 1.0/SeqLength;
for(long K=0; K<YSize; K++) 
	fprintf(stderr,"%d	%f\n", K, OneDeeAvg[K] );
}

	// FORCE THE DC VALUES (INCLUDING ALL FLICKER) TO MID-RANGE FOR
	// DISPLAY PURPOSES (MAXIMIZE THE CONTRAST OF THE SINC() LOBES)
	if(MakeAxes) {
		IML_BYTE AxisValue = 255;
		iml_index IML_REF Indexer = Averager.get_indexer();
		for(IML_LINT I=Indexer.min(IML_2D_X); I<=Indexer.max(IML_2D_X); I++)
			RPImage(I,YSize/2) = AxisValue;
		for(IML_LINT I=Indexer.min(IML_2D_Y); I<=Indexer.max(IML_2D_Y); I++)
			RPImage(XSize/2,I) = AxisValue;
		} // end if			
	
	Position.Y+=YSize;
	MacVideo.display_image( RPImage, Position );

} // end SDLT_random_pixel_array_fft

