//====================================================================
//
// This is a sample program.  A "hello world" if you like. Its basic
// function is to 
//     1) create a single instance of a macintosh video manager
//        object, propertly initialized
//     2) read in some information to decide what kind of window
//        clearing to use
//
//  This program assumes you already know the linearization parameters 
// (built directly in source code below)
//====================================================================


// INCLUDE THE PROPER FILES
#include "mvl_macv.h"  //header file for mac video manager, has class defs etc


//============================================================================
//
//============================================================================
int main () {


try { // an exception handling block;  see the "catch" at the bottom .... a runtime facility

	// MAKE A VIDEO MANAGER OBJECT AND INITIALIZE IT
	// I'LL PROVIDE AND USE MORE ERROR CHECKING LATER . . .
	IML_USINT MaxSequenceCnt = 32;  // maximum no of images sequences storable in video mgr
	IML_USINT MaxBlitCnt = 64;  // total no of blit sequences (frame sequence)
	IML_USINT GroupCount = 8;   //  group is collection of simultaneous blit sequences
	IML_ULINT ExitOnErr = TRUE;  // obvious
	
	// allocate a video manager, one per monitor (at most)
	// TAKE A LOOK AT THE CONSTRUCTOR FOR mvl_video_manager AND SEE THE DEFAULT VALUES FOR THESE ARGS
	mvl_video_manager MacVideo(MaxSequenceCnt,MaxBlitCnt, GroupCount, ExitOnErr);
	
	// NOTE: IN THE MACINTOSH UNIVERSE, MULTPLE MONITORS ARE DESIGNATED BY NUMBER.  

	// THESE OBJECTS ARE USED FOR LINEARIZING A MONITOR;  LOOK AT THE INFO ON
	// WHAT THESE PARAMETERS REPRESNT (class mvl_calibration_polynomial)
	// ln(DAC) = A + B ln(Lum) +C ln(Lum)^2 ...     
	//	mvl_calibration_polynomial Monitor_1(A, B, C, D) ;  
    //anywhere from 2 to 6 coefficients OK, A==ln(255.)
	mvl_calibration_polynomial Monitor_1(log(255.0), 0.314571, 0.0110987, 0.000845069);
	mvl_calibration_polynomial Monitor_0(log(255.0), 0.450019, 0.0384390, 0.00587792);
	
	// ONCE WE CHOOSE WHICH MONITOR TO USE, SET THAT ONE UP CORRECTLY.
	//Display usual 0 or 1 to indicate monitor, could be long or short or any integer
	long Display = 0;    
	
	// read what this does (depth planes, resolution, not Hz, ...
	MacVideo.initialize_screen(Display);	
	
	// you have a choice between Pelli or MVL style linearization. sets a flag
	MacVideo.use_mvl_calibration();	
	
	switch(Display) {
	
		default:
		case 0:	
			// if you DON'T have an attenuator attached use 0
			MacVideo.set_attenuator_red_steps(0.0); 
			// read what this does (send polynomial to video mgr)
			MacVideo.set_mvl_calibration(Monitor_0, 0, 100); 
			break;
			
		case 1:	
			// if you HAVE an attenuator attached .. define ratio
			MacVideo.set_attenuator_red_steps(23.0); 
			// calibration for monitor 1
			MacVideo.set_mvl_calibration(Monitor_1, 0, 100);  
			break;
		} // end switch
		

	// this fills the default clut with a luminance ramp given the above
	// param1 = mean luminance (scale 0-1), param2 = +- fraction of min residual range
	MacVideo.linearize_clut(0.5,1.0); 
	
	// parameter for random image-fill process, typical values 64-128
	long Speed = 100;   

	// read integer number from keyboard + CR 
	//--incorreect entry results in default of previous number & no msg
	char Buff[128];
	fprintf(stderr,"What speed clearing? ");fflush(stderr);
	fgets(Buff,128,stdin);
	sscanf(Buff, "%d", &Speed);

	//  use my window filling method instead of the operating systems
	MacVideo.use_mvl_window_filling(Speed);

	// use absolute DAC value, 127 is mean luminance, param1 (above)
	MacVideo.fill_window_absolute(127);

	fprintf(stderr,"DONE! ");fflush(stderr); //all fprint -> console window of monitor 0
	fgets(Buff,128,stdin);  // waits for CR to terminate

	} // end try
	
catch ( ... ) {
// THIS IS JUST TO ALLOW TERMINATION OF THE PROGRAM AND CLEAN UP OF STUFF.
	}

} // end main

