/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/


//========================================================
// gml_gfit.cpp  A SET OF CLASSES FOR FITTING MULTIDIMENSIOAL
// MODELS TO MULTIDIMENSIONAL DATA.  THE MODELS CAN BE NONLINEAR.
//========================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "gml_gfit.h"


//extern "C" { // COMPILED AS C++

	float dfridr(float (*func)(float), float x, float h, float *err);
	double solve(double (*f)(double x), double a, double b, double k, double eps);
	float gammq(float a, float x);
//	float dfridr(float (*func)(float), float x, float h, float *err);
	void gaussj(float **a, int n, float **b, int m);
	void lubksb(float **a, int n, int *indx, float b[]);
	void ludcmp(float **a, int n, int *indx, float *d);
	int *ivector(long nl, long nh);
	float **matrix(long nrl, long nrh, long ncl, long nch);
//};

// UGLY GLOBAL VARIABLES FOR CALCULATING HESSIAN, ETC.
gml_generic_fitter *GlobalFitterPtr;
gml_double_vec *GlobalSamplePoint;
gml_double_vec *GlobalMaxParms;
int  FirstParameter = 0;  // stupid defaults
int  SecondParameter = 0; // stupid defaults
float Global_H_Fraction=0.2;


float error_surface_value(float PValue); // PValue is the value of hte first parameter
float first_partial(float PValue);  // PValue is the value of the second parameter
//float second_partial(int Par_1, int Par_2, float & Error);



//=================================================================================
float error_surface_value(float PValue) { // PValue is the value of hte first parameter

	(*GlobalSamplePoint)[FirstParameter]=PValue;
	float Value = GlobalFitterPtr->evaluate( *GlobalSamplePoint );
	return( Value );

} // end error_surface_value

//=================================================================================
float first_partial(float PValue) {  // PValue is the value of the second parameter

	*GlobalSamplePoint = *GlobalMaxParms;
	(*GlobalSamplePoint)[SecondParameter] = PValue;
	float Par1Value = (*GlobalSamplePoint)[FirstParameter];
	float HSize = Par1Value*Global_H_Fraction;

	// USE THE NRC FUNCTION
	float Error1;
	float Value = dfridr(error_surface_value, Par1Value, HSize, &Error1);
	return(Value);

} // end first_partial


//================================================================================
gml_generic_fitter::gml_generic_fitter(IML_USINT PCount)
	: 	OptimizeFlags(PCount), OptimizedParms(PCount), 
		MinParmV(PCount), MaxParmV(PCount) , 
		SamplePoint(PCount), TestParms(PCount)  {

	DataCount = 0;
	ParmCount = PCount;
	CovMatrixRank = PCount;
	FittedParmCnt = PCount;
	H_Fraction = 0.2; // default to 20%
	Hessian 				= NULL;
	Covariance 				= NULL;
	Covariance_LU_DComp 	= NULL;
	Correlation			 	= NULL;
	Error 					= NULL;
	ParameterIDs 			= NULL;
	
	// TO KEEP TRACK OF TOTAL EVALUATIONS AND USED ITERATIONS
	Evaluations = 0;
	LineMinimizations = 0;
	PowellIterations = 0;
	
	// GLOBAL VARIABLES TO CONTROL 'Powell()' BEHAVIOR
	MaxIters = 30;
	Epsilon = 0.00001;
	
	ObjectiveFunctionMinimum = 0;
	
	// USE THIS TO KEEP TRACK OF PARAMETER "NAMES"
	ParmName = new ptr_to_const_char[ParmCount];
	// SET SOME DEFAULTS
	for(IML_USINT I=0; I<ParmCount; I++) {
		ParmName[I]="";
		MinParmV[I] = -4294967295U; // negative of maximum unsigned 32-bit int
		MaxParmV[I] =  4294967295U; // maximum unsigned 32-bit int
		OptimizeFlags[I] = 0.0;
		} // end for

} // end constructor
			
//================================================================================
void gml_generic_fitter::calculate_covariance_information(void) {


	CovMatrixRank = count_fitted_parms(OptimizeFlags);

		// STORE LIST OF FITTED PARAMETER IDS
	ParameterIDs = new int[CovMatrixRank];
	for(int I=0, ID=0; I<ParmCount; I++) {
		if( OptimizeFlags[I] ) ParameterIDs[ID++]=I;
		} // end for

		// allocate matrix memory
	Hessian 					= matrix(1,CovMatrixRank, 1,CovMatrixRank);
	Covariance 				= matrix(1,CovMatrixRank, 1,CovMatrixRank);
	Covariance_LU_DComp 	= matrix(1,CovMatrixRank, 1,CovMatrixRank);
	Correlation			 	= matrix(1,CovMatrixRank, 1,CovMatrixRank);
	Error 					= matrix(1,CovMatrixRank, 1,CovMatrixRank);

	float **Junk 			= matrix(1,CovMatrixRank, 1,1); // for satisfying gaussj
	int *Index 				= ivector(1,CovMatrixRank);  // FOR LU-DECOMPOSITION

/*************

			// MAKE THE RIGHT SIZED VECTOR AND STORE GLOBALLY FOR LATER
		gml_generic_fitter::ParameterIDs = new iml_sint_vec(CovMatrixRank);
		iml_sint_vec &ParameterIDs = *gml_generic_fitter::ParameterIDs;
		
			// STORE LIST OF FITTED PARAMETER IDS
		for(int I=0, ID=0; I<ParmCount; I++) {
			if( OptimizedParms[I] )  ParameterIDs[ID++]=I;
			} // end for

		// ADD SPACE PADDING KLUDGE BECAUSE OF NRC'S FORTRAN-STYLE INDEXING; UGH!!
		IML_LINT ArraySize = CovMatrixRank+1;
			// MAKE THE RIGHT SIZED MATRICES AND STORE GLOBALLY FOR LATER
		if(gml_generic_fitter::Hessian) delete gml_generic_fitter::Hessian;
		gml_generic_fitter::Hessian = new iml_float_mat(ArraySize, ArraySize);
		iml_float_mat &Hessian = *gml_generic_fitter::Hessian;

		if(gml_generic_fitter::Covariance) delete gml_generic_fitter::Covariance;
		gml_generic_fitter::Covariance = new iml_float_mat(ArraySize, ArraySize);
		iml_float_mat &Covariance = *gml_generic_fitter::Covariance;

		if(gml_generic_fitter::Covariance_LU_DComp) delete gml_generic_fitter::Covariance_LU_DComp;
		gml_generic_fitter::Covariance_LU_DComp = new iml_float_mat(ArraySize, ArraySize);
		iml_float_mat &Covariance_LU_DComp = *gml_generic_fitter::Covariance_LU_DComp;

		if(gml_generic_fitter::Correlation) delete gml_generic_fitter::Correlation;
		gml_generic_fitter::Correlation = new iml_float_mat(ArraySize, ArraySize);
		iml_float_mat &Correlation = *gml_generic_fitter::Correlation;

		if(gml_generic_fitter::Error) delete gml_generic_fitter::Error;
		gml_generic_fitter::Error = new iml_float_mat(ArraySize, ArraySize);
		iml_float_mat &Error = *gml_generic_fitter::Error;

		// local variables
		iml_float_mat Junk(CovMatrixRank, 1); // for satisfying gaussj
		int *Index 				= new int[CovMatrixRank];  // FOR LU-DECOMPOSITION

**********/

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		// CALCULATE THE MIXED PARTIALS FOR ALL FITTED PARAMETERS

	// DO THIS TO COMMUNICATE TO THE NON-MEMBER FUNCTIONS; WHAT AN UGLY KLUDGE
	GlobalFitterPtr = this;
	GlobalSamplePoint = &SamplePoint;
	GlobalMaxParms = &TestParms;
	for(IML_LINT I=0; I<ParmCount; I++) TestParms[I]=OptimizedParms[I];
	Global_H_Fraction = H_Fraction;
	
	for(int J=CovMatrixRank; J>=1; J--) {
		// fill in the junk vector to give some stupid value
		Junk[J][1] = J;
		for(int K=J; K<=CovMatrixRank; K++) {

			// DO THIS TO COMMUNICATE TO THE NON-MEMBER FUNCTIONS; WHAT AN UGLY KLUDGE
			FirstParameter = ParameterIDs[J-1];
			SecondParameter = ParameterIDs[K-1];
			float Par2Value = OptimizedParms[SecondParameter];
			float HSize = Par2Value*H_Fraction;
			float EstError;
			float Curvature = dfridr(first_partial, Par2Value, HSize, &EstError);
			
			Hessian[J][K] = Hessian[K][J] = Curvature;
			Error[J][K] = Error[K][J] = EstError;
			// INITIALIZE THE COVARIANCE MATRIX IN PREP FOR FINDING INVERSE OF HESSIAN
			Covariance[J][K] = Covariance[K][J] = Curvature;
			} // end for
		} // end for
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// METHOD FOR FINDING THE INVERSE:  USE GAUSS-JORDAN ELIMINATION
	// TURN THE HESSIAN DATA IN THE COV MATRIX INTO THE COVARIANCE VALUES by inverting
	gaussj(Covariance, CovMatrixRank, Junk, 1);
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// USE LU DECOMPOSITION BECAUSE WE CAN EASILY GET THE DETERMINANT OF THE
	// LU MATRIX AND CALCULATE THE "OCCAM FACTOR" AS A MEASURE OF THE GOODNESS
	// OF THE MODEL!!!!!  (COOL . . hehe he he hehe)

		// TRANSFER THE INVERTED MATRIX
	for(int J=CovMatrixRank; J>=1; J--)
		for(int K=1; K<=CovMatrixRank; K++)
			Covariance_LU_DComp[J][K] = Covariance[J][K];
		// GET THE LU-DECOMPOSITION
	float DeterminantSign;
	ludcmp(Covariance_LU_DComp, CovMatrixRank, Index, &DeterminantSign);

		// CALCULATE THE DETERMINANT AS THE PRODUCT OF THE DIAGONAL ELEMENTS OF THE LU MATRIX
		// USE A DOUBLE TO ACCULATE BECAUSE OF POSSIBLE OVERFLOW/UNDERFLOW
	Determinant = DeterminantSign; // GET SIGN BECAUSE OF ROW PERMUTATIONS FROM ludcmp()
	for(int d=1; d<=CovMatrixRank; d++) Determinant *= Covariance_LU_DComp[d][d];

		// HEY!! NOW WE CAN CALCULATE THAT INFAMOUS 'OCCAM FACTOR' THING. :-)
	PseudoOccamFactor = pow(2*3.141592654, CovMatrixRank/2.0);
	PseudoOccamFactor /= sqrt(fabs(Determinant));
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


	// IF THE DETERMINANT IS NEGATIVE, THEN WE ARE IN BIG TROUBLE.  MAKE
	// THE DATA SUCH THAT IT WILL BE NOTICED!!
	if(Determinant<0.0) {  // BAD SIGN!! SINGULAR MATRIX:  GIVE BAD FIT
		for(int J=CovMatrixRank; J>=1; J--)
			for(int K=1; K<=CovMatrixRank; K++)
				Covariance[J][K] = (J==K ? 10e-20 : 0.0);
		fprintf(stderr,"\007");
		} // end if


	// CALCULATE CORRELATION MATRIX
	for( int J=CovMatrixRank; J>=1; J--) {
		for(int K=J; K<=CovMatrixRank; K++) {
			Correlation[J][K] = Correlation[K][J] =
					Covariance[J][K]/sqrt( fabs(double(Covariance[J][J])*double(Covariance[K][K])) );
			} // end for
		} // end for

} // end calculate_covariance_information


//=============================================================================
// calculates and returns goodness of fit estimates
//=============================================================================
	// PASS IN VARIABLE HANDLES
void gml_generic_fitter::calculate_fit_statistics(void) {

	// we must first calculate the model values
	evaluate(OptimizedParms);

	// local registers for calculations
	struct stats {
		IML_DOUBLE SObsCal2, SWObsCal2;
		IML_DOUBLE SObs2, SWObs2, SObsMObs2, SWObsWMObs2;
		IML_DOUBLE SObs, SWObs, SW;
		} S = {0,0,0,0,0,0,0,0,0};

	register IML_DOUBLE Temp=0; // used by SQUARE_IT macro
	for (int I=0; I<DataCount; I++) {
			// square of difference of observed and calculated
		S.SObsCal2  += SQUARE_IT(model_error(I));
			// same but weighted
		S.SWObsCal2 += SQUARE_IT(model_error(I)) * weight(I);
			// square of observations
		S.SObs2  += SQUARE_IT(data(I));
			// weighted squares of observations
		S.SWObs2 += SQUARE_IT(data(I)) * weight(I);
			// Sum of observations
		S.SObs   += data(I);
			// sum of weighted observations
		S.SWObs  += data(I) * weight(I);
			// sum of weights
		S.SW     += weight(I);
		} // end for

	AlgebraicMean = S.SObs/(IML_DOUBLE)DataCount;
	WeightedMean = S.SWObs/(IML_DOUBLE)S.SW;

	StandardDeviation = sqrt(S.SObsCal2/(IML_DOUBLE)DegreesOfFreedom);
	WStandardDeviation = sqrt(S.SWObsCal2/(IML_DOUBLE)DegreesOfFreedom);

		// calculate bits requireing the means
	for ( int I=0; I<DataCount; I++) {
//		DClass IML_REF DP = Data[I];
		S.SObsMObs2   +=  SQUARE_IT(data(I)-AlgebraicMean);
		S.SWObsWMObs2 +=  SQUARE_IT(data(I)-WeightedMean) * weight(I);
		} // end for

		// calculate weigted and unweighted versions
	CoeffOfDetermination  = (S.SObsMObs2-S.SObsCal2)    / S.SObsMObs2;
	WCoeffOfDetermination = (S.SWObsWMObs2-S.SWObsCal2) / S.SWObsWMObs2;

	RSquared   = fabs( S.SObs2 - S.SObsCal2 )   / S.SObs2;
	WRSquared  = fabs( S.SWObs2 - S.SWObsCal2 ) / S.SWObs2;
		// note  that this is the natural log
	MSC  = log(  S.SObsMObs2/S.SObsCal2 )    -  (2*ParmCount)/DataCount;
	WMSC = log(  S.SWObsWMObs2/S.SWObsCal2 ) -  (2*ParmCount)/DataCount;

} // end calculate_fit_statistics



//====================================================================
// ROUTINES FOR THE FITTER
//====================================================================
inline void gml_generic_fitter::check_limits(gml_double_vec &Parm) {
	for(int I=0; I<ParmCount; I++) {
		Parm[I] = GML_PFLOOR(Parm[I],MinParmV[I]);
		Parm[I] = GML_PCEIL(Parm[I],MaxParmV[I]);
		} // end for
} // end iml_parameter_vector


//====================================================================
inline void gml_generic_fitter::scale_parms(IML_DOUBLE F, gml_double_vec &Parm) {
	for(int I=0; I<ParmCount; I++) Parm[I] *= F;
	} // end iml_parameter_vector


//====================================================================
IML_DOUBLE gml_generic_fitter::fitdata(void) {

	// work out the best fit of the model
	Evaluations = 0;
	LineMinimizations = 0;
	PowellIterations = 0;
	powell(OptimizedParms, OptimizeFlags);
	
	// KEEP THE MINIMUM FOR ERROR SURFACE SEARCHING PURPOSES
	ObjectiveFunctionMinimum =  evaluate(OptimizedParms);
	return(ObjectiveFunctionMinimum);
} // fit data





/* Prototypes =========================================================== */
/* Copy these prototypes to the program using these routines.             */

//double LineMin(int n, gml_double_vec &XStart, gml_lint_vec &Subspace, gml_double_vec &Direction,
//               double (*f)(double *XStart));
//int Powell(int n, gml_double_vec &XStart, gml_lint_vec &Subspace, double (*f)(double *XStart) );


/* Documentation of prototypes ===========================================

   Powell
   ------
   This is used to minimse a function where neither 1st nor 2nd derivatives
   are supplied. The parameters are:
   
   n : dimension of the parameter vector 
   XStart : the parameter vector, initialised to the start point of the minimisation,
       returns the minimum if the routine converges.
   Subspace : is 1 if this dimension is to be minimised, 0 if it is to be
       left alone.
   f : the function to be minimised (which must be implicitly aware of the 
       dimension n); returns its value at XStart

   The routine returns the number of iterations, so convergence can be 
   checked.
   
   
*/

/* Includes ==========================================================*/

//#include <stdlib.h>
//#include <math.h>
//#include <stdio.h>


/* Powell's method -------------------------------------------------------*/

#define GoldenRatio 1.618


//=============================================================================================
//  Performs n-dimensional line minimization of evaluate() along the 
//  line through XStart in the given Direction. This function is used
//  in powell(). The value of XStart is set to that value which minimizes the function,
//  and line_minimization() returns evaluate(XStart) at the minimum 
//=============================================================================================
IML_DOUBLE gml_generic_fitter::line_minimization( 
gml_double_vec &XStart, gml_lint_vec &Subspace, gml_double_vec &Direction) {
	IML_LINT i;

	LineMinimizations++;
	
  	// XStart is our initial parameter vector; NextX is our temporary variable
	gml_double_vec NextX(ParmCount);
  
	//Bracket the minimum. This finds three numbers a,b,c such that
	// 1. The ratio (b-a):(c-b) = 1:1.618. (ie a golden ratio ).
	// 2. For the three vectors xa = x+a.Direction, xb=x+b.Direction, and 
	//    xc = x+c.Direction, we have fb<fa,fc; thus a local minimum must
	//    exist between a and c.
	//Note that only the components specified by the Subspace are varied.  
  
	// initialise the interval with Xa=0, Xb=1
	IML_DOUBLE Xa = 0; 
	check_limits(XStart);
	IML_DOUBLE FofXa = evaluate(XStart);
	IML_DOUBLE Xb = 1; 
	
	// CALCULATE NEXT X VALUE
	for (i=0; i<ParmCount; i++) 
		NextX[i] = (Subspace[i])?(XStart[i]+Xb*Direction[i]):XStart[i];
  	
  	check_limits(NextX);
	IML_DOUBLE FofXb = evaluate(NextX);
	// check that going from Xa to Xb reduces function F, ie going downhill. If not,
	// reverse the values of Xa and Xb 
	if (FofXb > FofXa) { 
		IML_DOUBLE temp=Xa; Xa=Xb; Xb=temp; 
		temp=FofXa; FofXa=FofXb; FofXb=temp; 
		} // end if
		
	// extend the search in the Direction a->b using a golden-ratio multiple
	IML_DOUBLE Xc = Xb + (Xb-Xa)*GoldenRatio;
	
	for (i=0; i<ParmCount; i++) 
		NextX[i] = (Subspace[i])?(XStart[i]+Xc*Direction[i]):XStart[i];
		
  	check_limits(NextX);
	IML_DOUBLE FofXc = evaluate(NextX);
	
	// now keep going along the downhill Direction Xa->Xb->Xc until FofXc>FofXb
	while( (FofXb > FofXa) || (FofXb > FofXc) ) {
		Xa = Xb;	FofXa = FofXb; 
		Xb = Xc;	FofXb = FofXc; 
		Xc = Xb + (Xb-Xa)*GoldenRatio;
     
     	for (i=0; i<ParmCount; i++) 
     		NextX[i] = (Subspace[i])?(XStart[i]+Xc*Direction[i]):XStart[i];
     	
		check_limits(NextX);
		FofXc = evaluate(NextX); 
		} // end while
     
	// A bracketing interval has now been found 
  
  
	// Now use golden section search to find minimum. The search terminates after
    // maxIters, or when the average difference between FofXb (the smallest value in
    // the bracketing interval) and FofXa,FofXc, is less than epsilon of the average of
    // FofXa,FofXc
	IML_LINT nIters = 0;
	while ( (nIters<MaxIters) && 
			(fabs(FofXb-(FofXa+FofXc)/2) > Epsilon*fabs(FofXa+FofXc)/2) ) {
			
		IML_DOUBLE NextStep = Xa + (Xb-Xa)*GoldenRatio; 
		// NextStep is the next position in the interval to test. It is between Xb, Xc because
		//the gap Xb->Xc is larger than the gap Xa->Xb 
		
		for (i=0; i<ParmCount; i++) 
     		NextX[i] = (Subspace[i])?(XStart[i]+NextStep*Direction[i]):XStart[i];
     	
  		check_limits(NextX);
		IML_DOUBLE FofNextStep = evaluate(NextX);
     
		// next minimum, replace Xb, and ensure that Xb->Xc remains the largest gap 
		if (FofNextStep>FofXb) {
			Xc = Xa; 		FofXc = FofXa; 
			Xa = NextStep;	FofXa = FofNextStep; 
		}  else { // replace one of the endpoints of the bracketing interval
			Xa = Xb; 		FofXa = FofXb; 
			Xb = NextStep;	FofXb = FofNextStep; 
			} // end else
		nIters++;
		
		} // end while
   
  
	// set XStart to the minimum
	for (i=0; i<ParmCount; i++) 
		XStart[i] = (Subspace[i])?(XStart[i]+Xb*Direction[i]):XStart[i];

  	return(FofXb);

} // end line_minimization


//=============================================================================================
//=============================================================================================
unsigned short gml_generic_fitter::count_fitted_parms(gml_lint_vec &Subspace ) {
	long Count = 0; 
	// count subspace dimensions used for optimization
	for( IML_LINT i=0; i<ParmCount; i++)  if( Subspace[i] ) Count++;
	return(Count);
} // end

//=============================================================================================
//  This is a very simple version of powell, which discards the Direction of greatest
//  increase. This is not always the best strategy for Powell's method, and to
//  overcome potential problems, the directions of search are frequently reinitialised.
//=============================================================================================
//template <class DClass>  
void gml_generic_fitter::powell(gml_double_vec &XStart, gml_lint_vec &Subspace ) {

	//  Create the the matrix of search directions. The number of search directions is
	//  equal to the dimension of the subspace (ssdim). Initially, the directions
	//  are the coordinate vectors of the subspace
	FittedParmCnt = count_fitted_parms(Subspace); 
	
	DegreesOfFreedom = DataCount - FittedParmCnt;

	// make a temporary matrix for holding search dimensions
	gml_double_mat Direction(FittedParmCnt,ParmCount); 

	// fill the matrix with coordinate vectors
	for( IML_LINT i=0, ss=0; i<ParmCount; i++) {
		if( Subspace[i]) {
			for( IML_LINT j=0; j<ParmCount; j++)  Direction[ss][j] = 0; 
			Direction[ss][i] = 1; 
			ss++;
			} // end if
		} // end for
  
  
	// some initialisations 
  	check_limits(XStart);
	IML_DOUBLE FofX = evaluate(XStart); 
	IML_DOUBLE OldFofX = 100*FofX; 
	if( FofX<0 )	OldFofX -= 100000; 
	else 			OldFofX += 100000;
	
	// allocate scratch vectors
	gml_double_vec Change(FittedParmCnt);
	gml_double_vec OldX(ParmCount);

	PowellIterations = 0;
	// do powells minimisation
	while( (PowellIterations<MaxIters) &&
			!( fabs(OldFofX-FofX) < Epsilon*(fabs(OldFofX)+fabs(FofX)+0.0001) ) ) {
	// do a cycle of Powell's method: this is a line minimisation along
	// each of the search directions in the matrix. 

		// OldX is the value of XStart at start of cycle
		// and OldFofX is the function value at start
		for( IML_LINT i=0;i<ParmCount; i++ ) OldX[i] = XStart[i];  
		OldFofX = FofX;
		
		// NOTE: this assumes local linarity of the surface.  Otherwise
		// the line "change[i] = fcycle-FofX;" would be "change[i] = OldFofX-FofX;"
		IML_DOUBLE fcycle;
		for( IML_LINT i=0; i<FittedParmCnt; i++ ) {
			fcycle = FofX;  
			// fcycle is the function value at the start of the line min.
			FofX = line_minimization(XStart, Subspace, Direction[i]);
			// change[i] is the incremental change generated by minimisation along Direction i 
			Change[i] = fcycle-FofX;  // this way Change[] is a positive number
       		} // end for
       
		// discard Direction of greatest change & replace with the Direction XStart-oldx,
        // which is the total change in XStart over the cycle of minimisation. The logic
        // of this is 1. XStart-oldx is a good estimate of the local gradient, and 2. 
        // XStart-oldx probably has a large component due to the Direction of greatest
        // decrease. 
      
  // REF: feb 12 1997;  
  // THIS LOOKS WRONG; THE < BELOW SHOULD BE A > INSTEAD.  I'LL CHANGE THAT AND
  // SEE WHAT HAPPENS.
		IML_DOUBLE GreatestDeltaFofX = 0; 
		IML_LINT GreatestChangeIndex = 0;
		for( IML_LINT i=0; i<FittedParmCnt; i++ ) {
			// if( change[i] < greatest )  
			if( Change[i] > GreatestDeltaFofX ) { 
				GreatestChangeIndex = i; 
				GreatestDeltaFofX = Change[i];
				} // end if 
			} // end for
			
		for( IML_LINT i=0; i<ParmCount; i++ ) 
			Direction[GreatestChangeIndex][i] = XStart[i] - OldX[i];
     
		//  reinitialise the directions: this is done every ssdim cycles of Powell
		//  i.e. after possibly all of the original directions are replaced by the
		//  above procedure. 
		if( (PowellIterations+1)%FittedParmCnt == 0 ) {
			for( IML_LINT i=0, ss=0; i<ParmCount; i++ ) {
				if( Subspace[i]) {
					for( IML_LINT j=0; j<ParmCount; j++ ) 
						Direction[ss][j] = 0; 
					Direction[ss][i] = 1; 
					ss++;
					} // end if
				} // end for
			} // end if
			
		PowellIterations++;
	  
	} // end while
 
} // end powell



//=============================================================================
double gml_generic_fitter::find_parameter_CI(IML_LINT UpOrDown, IML_DOUBLE Contour, IML_LINT TargetParm,
				IML_DOUBLE LowBracket, IML_DOUBLE HighBracket, IML_DOUBLE Tolerance, IML_LINT & MaxSteps) {
// sign is +1 or -1 for lower or upper CI limit 

	// WE SEARCH THE SPACE AS INDICATED GLOBALLY, EXCEPT FOR THE
	// TARGET PARAMETER DIRECTION.  MAKE SURE TO ALLOCATE THE NEW ARRAY!
	gml_lint_vec SearchSpace(ParmCount);
	SearchSpace = OptimizeFlags;
	
	// SET THE TARGET PARAMETER DIRECTION INDICATOR TO ZERO.
	SearchSpace[TargetParm] = 0;

	// GET A LOCAL COPY OF THE BEST PARAMETER SET.    
	// MAKE SURE TO ALLOCATE THE NEW ARRAY!
	gml_double_vec SearchParms(ParmCount);
	SearchParms = OptimizedParms;

	// SAVE REFERENCES TO THE MAX AND ADJUSTED PARAMETER VALUES
	// FOR EASE OF ACCESS WITHIN THE LOOP BELOW.
	IML_DOUBLE &ChiP  = SearchParms[TargetParm];
	IML_DOUBLE &BestP = OptimizedParms[TargetParm];

	// SET THE CONVERGENCE PARAMETERS FOR POWELL().
	set_convergence(0.0001,15);

	// NOW FIND A NEW TARGET PARAMETER VALUE FOR THE REQUIRED
	// CHI LIMIT USING powell AND THE BISECTION ROOT FINDING METHOD.  THAT IS,
	// solve CHANGES THE TARGET PARAMETER UNTIL A VALUE IS FOUND THAT
	// PRODUCES A FIT THAT IS DIFFERENT FROM THE MAXIMUM
	// AS GIVEN BY THE Contour VALUE.  REPEATING THIS WITH powell()
	// WALKS ALONG THE BOUNDARY OF THE HYPER-ELLIPSE DEFINING THE CONFIDENCE
	// REGION UNTIL WE GET TO THE PROPER "TIP" OF THAT ELLIPSE.  [OR UNTIL WE
	// RUN OUT OF ITERATIONS; OR UNTIL WE STALL BECAUSE THE REGION IS NOT CONVEX]

	// TEMPORARY VARIABLE FOR USE WITHIN THE LOOP BELOW.
	IML_DOUBLE LastPValue;
	do {
		// SAVE THE CURRENT VALUE OF THE TARGET PARAMETER
		LastPValue = ChiP;

		// SEARCH THROUGH THE SPACE TO FIND THE BEST POSITION GIVEN
		// THE CURRENT TARGET PARAMETER VALUE.
		powell(SearchParms, SearchSpace);

		IML_LINT Steps;
		// FIND THE PARAMETER VALUE AT Contour USING THE BISECTION METHOD
		if(UpOrDown < 0) {
			Steps = find_contour(SearchParms, TargetParm, LowBracket, BestP, Contour, 0.001);
			// CHECK FOR A WEIRD FAILURE MODE; INCORRECT DIRECTION CHANGE OF THE PARAMETER.
			if( ChiP > BestP ) ChiP = LastPValue;
		} else {
			Steps = find_contour(SearchParms, TargetParm, BestP, HighBracket, Contour, 0.001);
			// CHECK FOR A WEIRD FAILURE MODE; INCORRECT DIRECTION CHANGE OF THE PARAMETER.
			if( ChiP < BestP ) ChiP = LastPValue;
			} // end else

		// CHECK TO MAKE SURE THAT SOLVE DID NOT TAKE US BEYOND THE LIMITS
		check_limits(SearchParms);

			// WHILE OUR TOLERANCE IS TOO BIG, AND WHILE WE HAVE NOT MADE ENOUGH STEPS.
		}  while ( (fabs(LastPValue-ChiP)>Tolerance) && (--MaxSteps) );

		// RETURN THE VALUE OF THE PARAMETER THAT
	// MEETS THE REQUIRED CONFIDENCE INTERVAL.
	return(ChiP);

} // end parmaeter_CI_finder



//======================================================================================
//   This solves the function evaluate(Parms)==k by adjusting ParmID; returns that value 
//   of ParmID. It uses bisection, halting after 100 bisections or when the interval
//   is smaller than k. The values a,b are the initial estimate of an interval
//   which brackets the solution. If it doesn't, solve returns immediately.
//======================================================================================
int gml_generic_fitter::find_contour( 
	gml_double_vec &Parms, IML_LINT ParmID, IML_DOUBLE a, IML_DOUBLE b, IML_DOUBLE k, IML_DOUBLE eps) {

IML_DOUBLE fa,fb,fmid;
IML_LINT Iterations;

	// GRAB REFERENCE FOR REQUESTED PARAMETER FOR EASY CHANGING
	IML_DOUBLE &Answer = Parms[ParmID];
	
	// EVALUATE AT BRACKET POINTS
	Answer = a; fa = evaluate(Parms)-k; 
	Answer = b; fb = evaluate(Parms)-k;
	
	// CHECK IF WE DON'T BRACKET A SOLUTION
	if (fa*fb > 0.0) return(-1); // solution outside interval 

	Iterations = 100;  
	// JUST IN CASE WE LANDED RIGHT ON IT
	if (fa==0.0) { Answer = a; return(Iterations); }
	if (fb==0.0) { Answer = b; return(Iterations); }
	
	Answer = (a+b)/2; // in case input brackets are too small
	while ( (fabs(b-a)>eps) && Iterations-- ) {
		fmid = evaluate(Parms)-k;
		if (fmid == 0.0) return(Iterations);
		if (fa*fmid>0.0)	{ a = Answer; fa = fmid; }
        else  				{ b = Answer; fb = fmid; }
        Answer = (a+b)/2;
     	} // end while

   return(Iterations);
}
              


