/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "type.h"

#ifdef __I_DO_WINDOWS__
#include <windows.h>
#endif

#include "iml_app.h"

#include <stdarg.h>
#include <time.h>
#include <stdlib.h>

IML_LINT IML_Max_Errors  = 5;
static IML_LINT IML_Error_Count = 0;


//=======================================================================
// A BUFFER FOR BUILDING MESSAGES
//=======================================================================
char IMLAppMsgBuf[512];

// Global variable for application use
class iml_app_shell IML_PTR IMLAppShell = NULL;

//=======================================================================
// DELAY FUNCTION FOR SOUND PLAYING...
//=======================================================================
void iml_app_shell::delay(IML_ULINT WaitTime) {
	clock_t Start, End;
	Start = clock();
	IML_ULINT Elapsed;
	do {
		End = clock();
//MACMOD -- OCT/14/96 REF
#ifdef __I_DO_WINDOWS__
		Elapsed = 1000*(End-Start)/CLK_TCK;

#else // assume we are on the mac for now
		Elapsed = 1000*(End-Start)/CLOCKS_PER_SEC;
#endif

	} while( Elapsed < WaitTime );
} // end delay

//=======================================================================
// FOR CONVENIENCE, ADD A VARIABLE ARGUMENT LIST PRINTING FUNCTION...
//=======================================================================
void iml_app_shell::printf(char *FormatStr, ...) {

	va_list ArgPtr;
	va_start(ArgPtr,FormatStr);
	vsprintf(IMLAppMsgBuf, FormatStr, ArgPtr);
	va_end(ArgPtr);
	print(IMLAppMsgBuf);

} // end print(format, var_arg list)


//=======================================================================
// FOR CONVENIENCE, ADD A VARIABLE ARGUMENT LIST PRINTING FUNCTION...
//=======================================================================
void iml_app_shell::printf(IML_USINT Row, IML_USINT Col, char *FormatStr, ...) {

	va_list ArgPtr;
	va_start(ArgPtr,FormatStr);
	vsprintf(IMLAppMsgBuf, FormatStr, ArgPtr);
	va_end(ArgPtr);
	print(Row,Col,IMLAppMsgBuf);

} // end print(format, var_arg list)

//=======================================================================
// ADD SOUNDS FOR EXPERIMENT CUES
//=======================================================================
void iml_app_shell::sound(IML_SINT SoundNumber) {

#ifdef __I_DO_WINDOWS__
	switch(SoundNumber) {
		case 0:
			sound(20); delay(100);
			sound(400); delay(100);
			break;

		case 1:
			sound(1000); delay(200);
			break;

		case 2:
			sound(1000); delay(100);
			sound(100); delay(100);
			sound(800); delay(100);
			break;

		default:
			sound(60); delay(400);
			break;
		} // end case

#endif // __I_DO_WINDOWS__

} // end sound



//=======================================================================
// DEFINE DEFAULT MESSAGING FUNCITONS FOR THE LIBRARY
//=======================================================================
void iml_error(char *Prefix, char *Msg) {
	IML_Error_Count++;
	sprintf(IMLAppMsgBuf, "ERROR:%.125s:::%.125s",Prefix,Msg );
	if(IMLAppShell) IMLAppShell->print(IMLAppMsgBuf);
#ifdef __I_DO_WINDOWS__
#ifdef IML_EASY_WIN
	else fprintf(stderr,"%s\n",IMLAppMsgBuf);
#else
	else {}//MessageBox(NULL,"SOMETHING IS WRONG!!","!!!HEY!!!",MB_OK);
#endif
#else // THIS CATCHES THE MACINTOSH AND UNIX PLATFORMS FOR NOW . . .
	else fprintf(stderr,"%s\n",IMLAppMsgBuf);
//	else fprintf(stderr,"!!!HEY!!! SOMETHING IS WRONG!!!\n");
#endif
	if(IML_Error_Count >= IML_Max_Errors) 
		throw("Ugh!  Too many errors . . . quitting.");
} // end iml_error



void iml_warning(char *Prefix, char *Msg) {
	sprintf(IMLAppMsgBuf, "WARNING:%.125s:::%.125s",Prefix,Msg );
	if(IMLAppShell) IMLAppShell->print(IMLAppMsgBuf);
#ifdef __I_DO_WINDOWS__
#ifdef IML_EASY_WIN
	else fprintf(stderr,"%s\n",IMLAppMsgBuf);
#else
	else {}//MessageBox(NULL,"SOMETHING IS WRONG!!","!!!HEY!!!",MB_OK);
#endif
#else // THIS CATCHES THE MACINTOSH AND UNIX PLATFORMS FOR NOW . . .
//	else fprintf(stderr,"!!!HEY!!! SOMETHING IS WRONG!!!\n");
	else fprintf(stderr,"%s\n",IMLAppMsgBuf);
#endif
} // end iml_warning

void iml_message(char *Prefix, char *Msg) {
	sprintf(IMLAppMsgBuf, "NOTE:%.125s:::%.125s",Prefix,Msg );
	if(IMLAppShell) IMLAppShell->print(IMLAppMsgBuf);
#ifdef __I_DO_WINDOWS__
#ifdef IML_EASY_WIN
	else fprintf(stderr,"%s\n",IMLAppMsgBuf);
#else
	else {}//MessageBox(NULL,"SOMETHING IS WRONG!!","!!!HEY!!!",MB_OK);
#endif
#else // THIS CATCHES THE MACINTOSH AND UNIX PLATFORMS FOR NOW . . .
//	else fprintf(stderr,"!!!HEY!!! SOMETHING IS WRONG!!!\n");
	else fprintf(stderr,"%s\n",IMLAppMsgBuf);
#endif
} // end iml_message


