#ifndef __IML_APP_H__
#define __IML_APP_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include <stdio.h>

#include "iml_type.h"
#include "iml_err.h"
#ifdef __I_DO_WINDOWS__
#include "iml_snds.h"
#else
#define IML_DEFAULT_SOUND 0
#endif

#define	NO_BUTTON			0x0000
#define	LEFT_BUTTON			0x0001
#define	RIGHT_BUTTON		0x0002
#define	MIDDLE_BUTTON		0x0004
#define	SHIFT_KEY			0x0008
#define	CONTROL_KEY			0x0010

extern IML_LINT IML_Max_Errors;


// PROTOTYPE FOR THE SYSTEM INDEPENDENT APPLICATION CODE
int run_experiment(char *CmdLineString);

// THE MODELS IMPLEMENTING WINDOWS/DOS APPLICATION SHELLS SHOULD
// POINT THIS AT THEIR RESPECTIVE DEFINED SHELL OBJECTS
extern class iml_app_shell IML_PTR IMLAppShell;


//======================================================================
// BUTTON PRESS INFO RETURN
//======================================================================
class button_info {
	public:
		button_info(unsigned short int Clk = NO_BUTTON,
						unsigned short int Mod = NO_BUTTON) {
			Clicked = Clk;
			Modifiers = Mod;
			};

		unsigned short int Clicked;
		unsigned short int Modifiers;

		operator==(button_info Button) {
			return(	(Button.Clicked   == Clicked) &&
						(Button.Modifiers == Modifiers) );
         };
		operator!=(button_info Button) { return(	!operator==(Button) ); };

}; // end class


//=====================================================================
// BASE CLASS FOR DOS/WINDOWS APPLICATION SHELL
//=====================================================================
class iml_app_shell {
	private:
		button_info Safety;

	protected:

		void delay(IML_ULINT WaitTime);

      IML_SINT MouseSound;
		IML_BYTE MouseClickEnabled;
		IML_BYTE SoundsEnabled;

	public:
		iml_app_shell(void) {
			SoundsEnabled = TRUE;
			MouseClickEnabled = TRUE;
			MouseSound = IML_DEFAULT_SOUND;
			};

		virtual button_info get_button_press(void) {
			iml_error("iml_app_shell", "Call to virtual get_button_press();");
			return(Safety);
			};

		virtual void print(char *String) {
			iml_error("iml_app_shell", "Call to virtual print(string);");
			};

		virtual void print(IML_USINT Row, IML_USINT Col, char *String) {
			iml_error("iml_app_shell", "Call to virtual print(string,r,c);");
			};

			// THESE TWO CALLS ARE NOT *VIRTUAL* ON PURPOSE!
			// THEY CREATE A STRING AND THEN CALL THE TWO
			// VIRTUAL FUNCTIONS ABOVE, WHICH SHOULD BE REPLACED BY THE
			// ENVIRONMENT-SPECIFIC APPLICATION CLASS
		void printf(char *FormatStr, ...);
		void printf(IML_USINT Row, IML_USINT Col, char *FormatStr, ...);

		virtual void clear(void) {
			iml_error("iml_app_shell", "Call to virtual clear();");
			};

		virtual void copy_output_to(const char *FileNAme) {
			iml_error("iml_app_shell", "Call to virtual copy_output_to(filename);");
			};
		virtual void end_copy_output(void) {
			iml_error("iml_app_shell", "Call to virtual end_copy_output();");
			};

		virtual void sound(IML_SINT SoundNumber);
		
		void enable_mouse_click(IML_BYTE Flag) { MouseClickEnabled = Flag; };
		void set_mouse_sound(IML_SINT MSound) { MouseSound = MSound; };
		void enable_sounds(IML_BYTE Flag) { SoundsEnabled = Flag; };

}; // end class



#endif // __IML_APP_H__
