#ifndef __IML_CPLX_H__
#define __IML_CPLX_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "iml_type.h"
#include "iml_math.h"

/** This class provides a minimal complex number type.
There are much more complete versions of this, but none that
were publicly available when I first wrote this code.  This
might get replaced by something that is public domain, eventually.
</p> */
template <class FLPTType> class iml_complex {

	public:
   
   	/**@name A: Public data. */
   	//@{
   
   	/// The real part of the complex number.
	FLPTType Real;
   	/// The imaginary part of the complex number.
	FLPTType Imag;
	//@}
	
   	/**@name B: Constructors and destructors. */
   	//@{
	
	/// Default constructor necessary for array allocation.
	iml_complex() {Real = 0.0; Imag = 0.0;};
	/// The simple constructor.
	iml_complex(FLPTType Re, FLPTType Im)
		{Real = Re; Imag = Im;};
	/// A copy constructor.
	iml_complex(iml_complex<FLPTType> IML_REF CVal)
		{Real = CVal.Real; Imag = CVal.Imag;};
	// A useless do-nothing destructor.
	~iml_complex() { /* NO NEED TO DO ANYTHING */};
	//@}
	
   	/** @name C: Mathematical operators. */
   	//@{

	/// Multiplication by a constant.
	void operator*= (IML_DOUBLE DVal) { Real *=  DVal; Imag *= DVal; };

	/// Assignment of a like complex.
	void operator= (iml_complex<FLPTType> IML_REF CVal)
		{Real =  CVal.Real; Imag =  CVal.Imag; };

	/// Addition of a like complex.
	void operator+=(iml_complex<FLPTType> IML_REF CVal)
		{Real += CVal.Real; Imag += CVal.Imag; };

	/// Subtraction of a like complex.
	void operator-=(iml_complex<FLPTType> IML_REF CVal)
		{Real -= CVal.Real; Imag -= CVal.Imag; };

	/// Multiplication by a like complex.
	void operator*=(iml_complex<FLPTType> IML_REF CVal) {
		FLPTType TNewReal = Real*CVal.Real - Imag*CVal.Imag;
		Imag 		= Real*CVal.Imag + Imag*CVal.Real;
		Real 		= TNewReal;
		};

		
	/// Get the magnitude: sqrt(r*r+i*i).
	FLPTType get_mag()     // MAGNITUDE
		{ return(iml_sqrt(Real*Real + Imag*Imag)); };

	/// Get the phase: atan(i/r). (PRESERVING QUADRANT)
	FLPTType get_phase() {  
		// IF PHASE IS UNDEFINED...
		if( Imag == 0.0 && Real == 0.0) return(0.0);
		// OTHERWISE RETURN VALUE
		return( iml_atan2(Imag, Real) );
		};

	/// Set the phase keeping the current magnitude.
	void set_phase(FLPTType Phase) {	// PHASE IN DEGREES
		FLPTType Mag = get_mag();
		Real = Mag*iml_cos(Phase);
		Imag = Mag*iml_sin(Phase);
		};

	/// Set the magnitude keeping the current phase.
	void set_mag(FLPTType Mag) {
		IML_FLOAT Phase = get_phase();	// PHASE IN DEGREES
		// SET THE MAGNITUDE ONLY. IF SIGNED, THEN PHASE IS AFFECTED.
		FLPTType Temp = iml_abs(Mag);
		Real = Mag*iml_cos(Phase * IML_PI/180.0);
		Imag = Mag*iml_sin(Phase * IML_PI/180.0);
		};

	//@}
	
}; // end template class iml_complex

/** 32-bit floating point complex number. </p> */
typedef iml_complex<IML_FLOAT>	iml_fcomplex;
/** 64-bit floating point complex number. </p> */
typedef iml_complex<IML_DOUBLE>	iml_dcomplex;
/** 80-bit floating point complex number, if IEEE long doubles are suported. </p> */
typedef iml_complex<IML_LDOUBLE>	iml_ldcomplex;


#endif // __IML_CPLX_H__
