/*
15. Appendix: Sample CRC Code

The following sample code represents a practical implementation 
of the CRC (Cyclic Redundancy Check) employed in PNG chunks. 
(See also ISO 3309 [ISO-3309] or ITU-T V.42 [ITU-V42] 
for a formal specification.) 

The sample code is in the ANSI C programming language. Non C users 
may find it easier to read with these hints: 

&		Bitwise AND operator. 
^		Bitwise exclusive-OR operator. (Caution: elsewhere in this 
			document, ^ represents exponentiation.) 
>>		Bitwise right shift operator. When applied to an unsigned 
			quantity, as here, right shift inserts zeroes at the left. 
!		Logical NOT operator. 
++		"n++" increments the variable n. 
0xNNN	0x introduces a hexadecimal (base 16) constant. Suffix L indicates 
			a long value (at least 32 bits). 

*/

#include "iml_crc.h"

unsigned long crc_calculator::crc_table[256];

int crc_calculator::crc_table_computed = 0;

//=========================================================================
// Initialize the table for a fast CRC. 
//=========================================================================
void crc_calculator::make_crc_table(void) {
	unsigned long c;
	int n, k;

	for (n = 0; n < 256; n++) {
		c = (unsigned long) n;
		for (k = 0; k < 8; k++) {
			if (c & 1)	c = 0xedb88320L ^ (c >> 1);
			else		c = c >> 1;
			} // end for
		crc_table[n] = c;
		} // end for
	crc_table_computed = 1;
} // end make_crc_table(void)



//=========================================================================
// Update a running CRC with the bytes buf[0..len-1]--the CRC
// should be initialized to all 1's, and the transmitted value
// is the 1's complement of the final running CRC (see the
// crc() routine below)). */
//=========================================================================
unsigned long crc_calculator::
	update_crc(unsigned long crc, unsigned char *buf, int len) {
	unsigned long c = crc;
	int n;

	for (n = 0; n < len; n++)  c = crc_table[(c ^ buf[n]) & 0xff] ^ (c >> 8);

	return c;
} // end update_crc
   
