/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1993-99 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.

  
----------------------------------------------------------------------*/

#include <string.h>
#include <limits.h>
#include <time.h>

#include "iml_type.h"
#include "iml_file.h"
#include "iml_arch.h"


// HERE ARE WINDOW'S SPECIFIC INCLUDES
#ifdef __I_DO_WINDOWS__
//#include <io.h>
#else
#define __IML_MWERKS_TEMPLATES__
#endif

//#define IMFILE_MINENTRIES	4



static char ErrBuf[100];

// I'VE MODIFIED THE FORMAT FOR N-DIMENSIONAL IMAGES
const IML_ULINT ImageFormatVersion = 0x0002;

//=====================================================================
//
//=====================================================================
void image_file_header::fix_byte_order(IML_LINT TransformationFlag) {
	if(TransformationFlag == IML_BIG_LITTLE_ENDIAN_XFORM){
		iml_2byte_reverse( DirEntryCnt );
		}
} // end fix_byte_order

//=====================================================================
//
//=====================================================================
void image_file_directory_entry::fix_byte_order(IML_LINT TransformationFlag) {
	if(TransformationFlag == IML_BIG_LITTLE_ENDIAN_XFORM) {
		iml_4byte_reverse( Type );
		iml_4byte_reverse( Location );
		iml_4byte_reverse( Value );
		} // end if
} // end fix_byte_order

//=====================================================================
// USE THESE FUNCTIONS JUST SO THAT WE CAN COMPILE FOR MS WINDOWS 3.0
// RATHER THAN MS WINDOWS 3.1; THIS LET'S US WRITE CHUNKS > 64kb
//=====================================================================
IML_LINT  iml_image_file::huge_write(void IML_PTR Data, IML_LINT Size) {


	IML_LINT Status;
	// THERE IS A 64 KB LIMIT UNDER WINDOWS 3.0 (AND 3.1, THOUGH THEY CLAIM NOT) 
	// FOR A READ/WRITE
	const IML_ULINT ChunkSize = MAX_CHUNK_SIZE; 
	IML_LINT Written = 0;
	unsigned char IML_PTR ChunkPtr = (unsigned char IML_PTR)Data;

	while(Size > 0) {
		IML_LINT WriteSize = (Size > ChunkSize) ? ChunkSize : Size;
		Size -= WriteSize;

		#ifdef __I_DO_WINDOWS__
		// THIS ONE IS FOR WINDOWS
		Status = _lwrite(hDiskFile, (const char*)ChunkPtr, (unsigned int)WriteSize);
		#else
		// THIS SHOULD WORK FOR UNIX OR THE MAC (UNTIL/UNLESS WE GET FANCY ON THE MAC)
		Status = fwrite((void *)ChunkPtr, sizeof(unsigned char), WriteSize, hDiskFile);
		#endif

		// CHECK FOR ERROR CONDITIONS
		if( Status != WriteSize ) return(Status);

		Written += WriteSize;
		ChunkPtr += WriteSize;
		} // end while

	return(Written);
} // end huge_write


//=====================================================================
// USE THESE FUNCTIONS JUST SO THAT WE CAN COMPILE FOR MS WINDOWS 3.0
// RATHER THAN MS WINDOWS 3.1; THIS LET'S US READ CHUNKS > 64kb
//=====================================================================
IML_LINT  iml_image_file::huge_read( void IML_PTR Data, IML_LINT Size) {

	IML_LINT Status;
	const IML_LINT ChunkSize = MAX_CHUNK_SIZE; // < 64 KB LIMIT; ERR VAL IS -1
	IML_LINT Read = 0;
	unsigned char IML_PTR ChunkPtr = (unsigned char IML_PTR)Data;
//char Buffer[256];
	while(Size > 0) {
		IML_LINT ReadSize = (Size > ChunkSize) ? ChunkSize : Size;
		Size -= ReadSize;

		#ifdef __I_DO_WINDOWS__
		// THIS ONE IS FOR WINDOWS
		Status = _lread(hDiskFile, ChunkPtr, ReadSize);
		#else
		// THIS SHOULD WORK FOR UNIX OR THE MAC (UNTIL/UNLESS WE GET FANCY ON THE MAC)
		Status = fread((void *)ChunkPtr, sizeof(unsigned char), ReadSize, hDiskFile);
		#endif

		// CHECK FOR ERROR CONDITIONS
		if( Status != ReadSize ) return(Status);
		Read += ReadSize;
		ChunkPtr += ReadSize;
	   } // end while

	return(Read);
} // end huge_read



//==============================================================
//
//==============================================================
iml_image_file::iml_image_file(const char *FileName, IML_USINT FileOp) {


	hDiskFile = NULL;

	switch(FileOp) {

		case IML_FILE_OPEN:
			open_file(FileName);
			break;

		case IML_FILE_CREATE:
			create_file(FileName);
			break;

		default:
			iml_error("image_file","Unknown file operation in constructor");
         break;

		} // end switch
}; // end iml_image_file(char *, IML_USINT)



//==============================================================
//
//==============================================================
IML_SINT iml_image_file::close_file(void) {

	if(hDiskFile) {
	long int Status;
	
		#ifdef __I_DO_WINDOWS__
		Status = _lclose( hDiskFile );
		#else
		Status = ((fclose( hDiskFile )) ? HFILE_ERROR : NULL);
		#endif

		if(Status == HFILE_ERROR) {
			iml_error("image_file","Error on closing file");
			return(FALSE);
			} // end if

      //  CLEAN UP
      hDiskFile = NULL;
		} // end if

	return( TRUE );
} // end close_file



//==============================================================
//
//==============================================================
IML_SINT iml_image_file::open_file(const char *FileName) {

	// JUST IN CASE WE HAVE SOMETHING ELSE OPEN; I.E. WE CAN ONLY HAVE ONE
   // FILE OPEN AT A TIME FOR EACH INSTANCE OF iml_image_file
	close_file();
	
	long int Status;

	#ifdef __I_DO_WINDOWS__
	Status = hDiskFile = _lopen((LPCSTR)FileName, OF_READWRITE | OF_SHARE_EXCLUSIVE);
	#else
   hDiskFile = fopen(FileName,"r+");
	Status =  (hDiskFile) ? NULL : HFILE_ERROR;
	#endif
   
   if( Status == HFILE_ERROR) {
   
		sprintf(ErrBuf,"Error on opening file <%s>",FileName);
		iml_error("image_file",ErrBuf);

      // CLEAN UP THE MESS
      hDiskFile = NULL;
      } // end if

	return(Status != HFILE_ERROR);
} // end open_file



//==============================================================
//
//==============================================================
IML_SINT iml_image_file::create_file(const char *FileName) {

	// JUST IN CASE WE HAVE SOMETHING ELSE OPEN; I.E. WE CAN ONLY HAVE ONE
   // FILE OPEN AT A TIME FOR EACH INSTANCE OF iml_image_file
	close_file();

	//hDiskFile = _lcreat((LPCSTR)FileName, READ_WRITE | OF_SHARE_EXCLUSIVE);
	long int Status;
	#ifdef __I_DO_WINDOWS__
	Status = hDiskFile = _lcreat((LPCSTR)FileName, 0);
	#else
   hDiskFile = fopen(FileName,"w+");
	Status =  (hDiskFile) ? NULL : HFILE_ERROR;
	#endif

	if(Status == HFILE_ERROR) {
		sprintf(ErrBuf,"Error on creating file <%s>",FileName);
		iml_error("image_file",ErrBuf);

      // CLEAN UP THE MESS
      hDiskFile = NULL;
      } // end if

	return(Status != HFILE_ERROR);
} // end create_file


//==============================================================
//
//==============================================================
/****
IML_LINT iml_image_file::get_location(void) {

#ifdef __I_DO_WINDOWS__
	// The microsoft docs say that this _llseek() function
	//    returns the new location after the seek.
	// This just seeks zero bytes from the current location
	//   and should return the current location in the file;
	return( _llseek(hDiskFile, 0, FILE_CURRENT) ); 

#else
	return(ftell(hDiskFile));
#endif

}
*****/

//==============================================================
//
//==============================================================
IML_LINT iml_image_file::set_location(IML_ULINT Loc) {
	#ifdef __I_DO_WINDOWS__		// ZERO MEANS FROM FILE START
	return( _llseek(hDiskFile, Loc, 0) ); 
	#else									// SEEK_SET MEANS FROM FILE STARTS
	return( ((fseek(hDiskFile, Loc, SEEK_SET)) ? HFILE_ERROR : NULL) );
	#endif
};


//==============================================================
//  UNDER DOS/WINDOWS THERE IS A DIRECT CALL FOR filelength(FILE *)
//  BUT FOR COMPATIBILITY ISSUES, I WRITE ONE DIRECTLY HERE;  THAT WAY
//  WE SHOULD WORK ON MAC AN UNIX BOXES TOO . . .
//==============================================================
IML_LINT iml_image_file::file_length(void) {

	// this is some sort of thing in the SDK that I don't know how to
	// use yet.
	//LARGE_INTEGER LargeInt;
	//bool Success = GetFileSizeEx(hDiskFile, &LargeInt);

	// this function worked under dos/win3.1/win3.11 
	//  -- it does not seem to compute under NT because it has been
	// changed to _filelength(int Handle)
	//return(filelength(hDiskFile));

#ifdef __I_DO_WINDOWS__
	IML_LINT CurrentPosition = _llseek(hDiskFile, 0, FILE_CURRENT);
	_llseek(hDiskFile, 0, FILE_END);
	IML_LINT FileLength = _llseek(hDiskFile, 0, FILE_CURRENT);
	_llseek(hDiskFile, CurrentPosition, FILE_BEGIN);
	return(FileLength);
#else

	IML_LINT CurrentPosition = ftell(hDiskFile);
	fseek(hDiskFile, 0, SEEK_END);
	IML_LINT FileLength = ftell(hDiskFile);
	fseek(hDiskFile, CurrentPosition, SEEK_SET);
	return(FileLength);
#endif

} // end iml_image_file::file_length





//==============================================================
//
//==============================================================
IML_SINT iml_image_file::write_image(iml_image IML_PTR Image, IML_LINT Format) {
IML_ULINT WriteSize;

	if(!hDiskFile) {
		iml_error("image_file","Cannot write_image until a file is open");
		return(FALSE);
		} // end if

	if(!Image) {
		iml_error("image_file","Null ptr passed to write_image");
		return(FALSE);
		} // end if

	IML_LINT DataStart;
	// WRITE THE PREFIX INFORMATION FOR THE REQUESTED IMAGE FORMAT
	switch(Format) {
		case IML_IML_FORMAT:
			DataStart = write_iml_fmt_prefix(Image);
			break;

		case IML_HIPS_FORMAT:
			DataStart = write_hips_fmt_prefix(Image);
			break;

		case IML_RAW_FORMAT:
			DataStart = 0;
			break; // NO PREFIX FOR THESE FORMATS

 		default:
			iml_error("image_file","Unknown format specifier passed to write_image");
			return(FALSE);
		} // end switch


	// CALCULATE THE TOTAL NUMBER OF BYTES OF IMAGE DATA
//	WriteSize = Image->Subspace.unclipped_count();
	WriteSize = Image->subspace_size_in_bytes();

	// MOVE THE FILE POINTER TO THE DATA LOCATION
	if(set_location(DataStart) == HFILE_ERROR) {
		iml_error("image_file", "File pointer error 2 in write_image");
		return(FALSE);
		} // end if

	// WRITE THE DATA TO THE FILE
	if( huge_write(Image->subspace_chunk_address(), WriteSize) != WriteSize) {
		iml_error("image_file","Error on writing image file data");
		return(FALSE);
		} // end if


	// WRITE THE SUFFIX INFORMATION FOR THE REQUESTED IMAGE FORMAT
	switch(Format) {
		case IML_HIPS_FORMAT:
			write_hips_fmt_suffix(Image);
			break;

		case IML_IML_FORMAT:
		case IML_RAW_FORMAT:
			DataStart = 0;
			break; // NO PREFIX FOR THESE FORMATS

 		default:
			iml_error("image_file","***Unknown format specifier passed to write_image");
			return(FALSE);
		} // end switch


	// WE COULD CLOSE THE FILE HERE, BUT FOR NOW WE ASSUME THAT THE
	// CALLER IS RESPONSIBLE FOR THAT.  TO BE CAREFUL, WE CALL close_file
	// IN THE DESTRUCTOR.
	
	return(TRUE);
} // end write_image


//==============================================================
//
//==============================================================
IML_LINT iml_image_file::write_hips_fmt_prefix(iml_image IML_PTR Image) {

	// SEARCH TO THE BEGINNING OF THE FILE
	if(set_location(0) == HFILE_ERROR) {
		iml_error("write_hips_fmt_prefix", "File pointer error 1");
		return(0);
		} // end if

	IML_LINT WriteSize;
	IML_ULINT DataStart = 0;

	// WRITE THE HEADER INFORMATION
	char *OrigName = "IML Image Processing\n";
	WriteSize = strlen(OrigName);
	DataStart += WriteSize;
	if( huge_write(OrigName, WriteSize) != WriteSize ) {
		iml_error("write_hips_fmt_prefix","Error on writing OrigName");
		return(0);
		} // end if

	char *SeqName = "\n";
	WriteSize = strlen(SeqName);
	DataStart += WriteSize;
	if( huge_write(SeqName, WriteSize) != WriteSize ) {
		iml_error("write_hips_fmt_prefix","Error on writing SeqName");
		return(0);
		} // end if

	char *NumFrame = "1\n";
	WriteSize = strlen(NumFrame);
	DataStart += WriteSize;
	if( huge_write(NumFrame, WriteSize) != WriteSize ) {
		iml_error("write_hips_fmt_prefix","Error on writing NumFrame");
		return(0);
		} // end if

	time_t t;
	time(&t);
	char *Date = ctime(&t);
	WriteSize = strlen(Date);
	DataStart += WriteSize;
	if( huge_write(Date, WriteSize) != WriteSize ) {
		iml_error("write_hips_fmt_prefix","Error on writing Date");
		return(0);
		} // end if

	char RCBuf[32];                 // rows and cols
	sprintf(RCBuf,"%d\n%d\n",
				Image->Subspace.Axes[IML_2D_Y].Size,
				Image->Subspace.Axes[IML_2D_X].Size
				);
	WriteSize = strlen(RCBuf);
	DataStart += WriteSize;
	if( huge_write(RCBuf, WriteSize) != WriteSize ) {
		iml_error("write_hips_fmt_prefix","Error on writing RCBuf");
		return(0);
		} // end if

	IML_LINT BPPixel;
	IML_LINT HIPSType;   // values gathered from HIPS V1 hipl_formtat.h
	switch(Image->image_type()) {
		case IML_UBYTE_IMAGE:
		case IML_BYTE_IMAGE: 		BPPixel = 8; 	HIPSType = 0; break;

		case IML_USINT_IMAGE:
		case IML_SINT_IMAGE: 		BPPixel = 16; 	HIPSType = 1; break;

		case IML_ULINT_IMAGE:
		case IML_LINT_IMAGE:			BPPixel = 32; 	HIPSType = 2; break;

		case IML_FLOAT_IMAGE:		BPPixel = 32; 	HIPSType = 3; break;

		case IML_DOUBLE_IMAGE:		BPPixel = 64; 	HIPSType = 6; break;

		case IML_FCOMPLEX_IMAGE:	BPPixel = 64; 	HIPSType = 4; break;

		case IML_DCOMPLEX_IMAGE:	BPPixel = 128; HIPSType = 7; break;

		default:
			iml_error("write_hips_fmt_prefix","Image type does not exist in HIPS");
			return(0);

		} // end switch

	char HTBuf[64];                 // rows and cols
	sprintf(HTBuf,"%d\n%d\n%d\n",BPPixel,0,HIPSType);
	WriteSize = strlen(RCBuf);
	DataStart += WriteSize;
	if( huge_write(HTBuf, WriteSize) != WriteSize ) {
		iml_error("write_hips_fmt_prefix","Error on writing HTBuf");
		return(0);
		} // end if

	char *SeqHist = "Created by IML Image Processing on above date :-)\n\n";
	WriteSize = strlen(SeqHist);
	DataStart += WriteSize;
	if( huge_write(SeqHist, WriteSize) != WriteSize ) {
		iml_error("write_hips_fmt_prefix","Error on writing SeqHist");
		return(0);
		} // end if

		
	IML_LINT CRLines = (DataStart + 2) % 4;
	// PAD WITH CR'S SO WE GET THE "HEADER" LENGTH DIVISIBLE BY 4 (THEIR FORMAT)
	while(CRLines) huge_write("\n", 1);
	huge_write(".\n",2);
	DataStart += CRLines + 2;

	// WRITE THE FLAG THAT SAYS WE WON'T USE ANY MATRIX/VECTOR TRANSOFMS
	char B = 0;
	huge_write(&B,1);
	DataStart++;

	return(DataStart);

} // end write_hips_fmt_prefix


//==============================================================
//
//==============================================================
#ifdef __BORLAND_CPP__
#pragma argsused
#endif
void iml_image_file::write_hips_fmt_suffix(iml_image IML_PTR Image) {

	char B = 'e';
	huge_write(&B, 1); // PAD WITH CR'S

} // end write_hips_fmt_suffix


//==============================================================
//
//==============================================================
IML_LINT iml_image_file::write_iml_fmt_prefix(iml_image IML_PTR Image) {
IML_LINT WriteSize;

	// SAVE SIMPLE ID INFO
	IML_ULINT ByteOrderSignature = calculate_byte_order_signature();
	char *Bytes = (char*)&ByteOrderSignature;
	sprintf(FileHeader.IDString, "%c%c%c%c IMLI  \n\n\n\n\n\n\n\n\n\n",
			Bytes[0]+'0', Bytes[1]+'0', Bytes[2]+'0', Bytes[3]+'0');

		// KEEP THESE FIRST BECAUSE WE DON'T KNOW THEIR VALUES AND WE HAVE TO
		// INDEX THIS STRUCTURE LATER TO FILL THEM IN.  THIS WAY IF WE ADD ENTRIES
		// WHICH ARE KNOWN AT INITIALIZATION TIME I DON'T HAVE TO RE-ADJUST THE
		// INDEX VALUES.
	const short int DIM_INFO_START_ENTRY=0;
	const short int ENTRY_DATABITS_ENTRY=1;
	
	// CREATE THE FILE DIRECTORY
	image_file_directory_entry Directory[] =  {
		// REMEMBER, THESE TWO FIRST; ALL PRE-KNOWN VALUES FOLLOW TO-BE-CALCUALTED ONES.
		{IML_DIR_DIM_INFO_START,		IML_IN_DIRECTORY, 0},
		{IML_DIR_ENTRY_DATABITS,		IML_IN_DIRECTORY, 0},
		// THESE ARE KNOWN, AND THEIR ORDER IS NOT IMPORTANT
		{IML_DIR_FORMAT_VERSION, 		IML_IN_DIRECTORY, ImageFormatVersion},
		{IML_DIR_ENTRY_IMAGE_TYPE, 	IML_IN_DIRECTORY, Image->image_type()},
		{IML_DIR_IMAGE_DIMENSION,		IML_IN_DIRECTORY, Image->Subspace.Dimensionality},
		{31, 31, 31} /* place holder for debugging */
	};

	// SET UP THE DIRECTORY ENTRY COUNT;  THE IDSTRING SHOULD ALREADY
	// BE SET WHEN WE OPEN THE FILE FOR WRITING.
	FileHeader.DirEntryCnt = sizeof(Directory)/sizeof(image_file_directory_entry);

	// ESTIMATE WHAT THE MINIMUM POSITION OF THE VARIABLE DATA WILL BE
	IML_ULINT MinPos = sizeof(image_file_header) + sizeof(Directory) + 1;

	// STORE DIMENSION INFO BEHIND THE DIRECTORY ON A 16-BYTE BOUNDARY
	IML_ULINT DimInfoStart = (MinPos | 0x000fL) + 1;

	// SET FILE LOCATION OF DIMENSION INFO INTO THE DIRECTORY
	Directory[DIM_INFO_START_ENTRY].Value = DimInfoStart;

	// CALCULATE SIZE OF THE DIMENSION INFO
	IML_ULINT DimInfoSize = sizeof(iml_axis)*Image->Subspace.Dimensionality;

	// START DATA ON A 256-BYTE BOUNDARY AFTER THE DIMENSION INFO
	IML_ULINT DataStart = ((DimInfoStart + DimInfoSize) | 0x01ffL) + 1;

	// SET FILE LOCATION OF DATA INTO THE DIRECTORY
	Directory[ENTRY_DATABITS_ENTRY].Value = DataStart;

	// SEARCH TO THE BEGINNING OF THE FILE
	if(set_location(0) == HFILE_ERROR) {
		iml_error("write_iml_fmt_prefix", "File pointer error 1 in write_image");
		return(0);
		} // end if


	// WRITE THE HEADER INFORMATION
	WriteSize = sizeof(FileHeader);
	if( huge_write(&FileHeader, WriteSize) != WriteSize ) {
		iml_error("write_iml_fmt_prefix","Error on writing image file header");
		return(0);
		} // end if

	// WRITE THE DIRECTORY INFORMATION
	WriteSize = sizeof(Directory);
	if( huge_write(&Directory, WriteSize) != WriteSize ) {
		iml_error("write_iml_fmt_prefix","Error on writing image file directory");
		return(0);
		} // end if

	// SET THE POINTER FOR WRITING THE DIMENSION INFO
	if(set_location(DimInfoStart) == HFILE_ERROR) {
		iml_error("write_iml_fmt_prefix", "Error on seeking image dimension info");
		return(0);
		} // end if

	// WRITE THE IMAGE DIMENSIONALITY INFORMATION
	WriteSize = DimInfoSize;
	
	// THIS IS TEMPORARY BOOK KEEPING REQUIRED TO GET THE NUMBER OF
	// PIXELS IN THE FILE, AND THE DIMENSION INFO, TO MATCH!  WE NEED
	// THIS BECAUSE THE ORIGINAL IMAGE SIZE, AND THE SAVED IMAGE
	// SIZE AND DIMENSION INFORMATION CAN BE DIFFERENT;  WE CAN SAVE SUB-SPACE
	// CHUNKS OF THE ORIGINAL IMAGE.
	iml_index IML_REF ISS = Image->Subspace;			// GRAB THE SUBSPACE
	iml_index TSS(ISS.Dimensionality);					// CREATE A TEMPORARY COPY
	for(int I=0; I<TSS.Dimensionality; I++){ 			// SET THE MIN/MAX/SIZE CORRECTLY
		TSS.set_min_max(I, ISS.amin(I), ISS.amax(I) );
		 }
	TSS.house_keeping();				// THIS FIXES UP Stride, ElementCount, AND PTRS.
	
	if( huge_write(TSS.Axes, WriteSize) != WriteSize ) {
		iml_error("write_iml_fmt_prefix","Error on writing image file directory");
		return(0);
		} // end if

	// MUST RETURN FILE POSITION FOR STARTING THE IMAGE-DATA-WRITE
	return(DataStart);
} // end write_iml_fmt_prefix





#ifdef __BORLAND_CPP__
#pragma argsused
#endif

//==============================================================
//  LOAD RAW FILE BYTES INTO PIXELS
//==============================================================
IML_SINT iml_image_file::load_raw_pixels(iml_image *Image, IML_LINT SkipHeader) {
IML_LINT ReadSize;

	if(!hDiskFile) {
		iml_error("image_file","Cannot load_raw_pixels until a file is open");
		return(FALSE);
		} // end if

   // DO A SUPERFICIAL CHECK OF FILE SIZE
	IML_LINT FileSize = file_length();
	if(FileSize < 0) {
		iml_error("image_file","Error reading file size in load_raw_pixels");
		return(FALSE);
      } // end if
	
	// FIND OUT HOW MANY bytes TO LOAD
   ReadSize = Image->subspace_size_in_bytes();
   
	long StartLocation = 0;
	if(SkipHeader == iml_image_file::IML_IMAGE_AT_EOF)
		StartLocation = FileSize - ReadSize;
	else if(SkipHeader > 0)
		StartLocation = SkipHeader;
		
   if(ReadSize != (FileSize-StartLocation)) {
   	iml_warning("image_file", "Mismatch of file size, SkipHeader and image size in load_raw_pixels");
   	} // end if

	// RESET THE LOCATION TO ZERO, JUST IN CASE MULTIPLE WRITES/READS
	// HAVE TAKEN PLACE TO THE FILE.
	if(set_location(StartLocation) == HFILE_ERROR) {
		iml_error("image_file", "File pointer error 1 in load_raw_pixels");
		return(FALSE);
		} // end if

	// READ THE RAW PIXELS
	if( huge_read(Image->subspace_chunk_address(), ReadSize) != ReadSize) {
		iml_warning("image_file", "Error loading raw pixels in load_raw_pixels");
		return(FALSE);
		} // end if
		
	return(TRUE); // WE SEEM TO HAVE BEEN SUCCESSFUL
	
} // end

//==============================================================
//  AT THE MOMENT THIS DOES NOT READ ANY FORMAT BUT IML-FORMAT
//==============================================================
iml_image *iml_image_file::read_image() {
IML_LINT ReadSize;

	if(!hDiskFile) {
		iml_error("image_file","Cannot read_image until a file is open");
		return(NULL);
		} // end if


   // DO A SUPERFICIAL CHECK OF FILE SIZE
	IML_LINT FileSize = file_length();
	if(FileSize < 0) {
		iml_error("image_file","Error reading file size in read_image");
		return(NULL);
      } // end if

	// RESET THE LOCATION TO ZERO, JUST IN CASE MULTIPLE WRITES/READS
	// HAVE TAKEN PLACE TO THE FILE.
	if(set_location(0) == HFILE_ERROR) {
		iml_error("image_file", "File pointer error 1 in read_image");
		return(NULL);
		} // end if

	// READ THE IMAGE FILE HEADER INFORMATION
   ReadSize = sizeof(FileHeader);
	if( huge_read(&FileHeader, ReadSize) != ReadSize) {
		iml_error("image_file", "Error on read_image of file header");
		return(NULL);
		} // end if

	// CHECK FOR THE IML-IMAGE TAG LETTERS THAT SHOULD APPEAR IN THE
	// IDSTRING. strstr RETURNS 0 IF S2 IS NOT A SUBSTRING OF S1
	if( !strstr(FileHeader.IDString, "IMLI") ) {
		iml_error("image_file","File appears to be the wrong type");
		return(NULL);
      } // end if

	IML_ULINT CurrentByteOrderSignature = calculate_byte_order_signature();
	FileHeader.IDString[0] -= '0';	FileHeader.IDString[1] -= '0';
	FileHeader.IDString[2] -= '0';	FileHeader.IDString[3] -= '0';
	IML_ULINT ByteOrderSignatureOnDisk = (*((IML_ULINT *)FileHeader.IDString));
	IML_LINT ByteOrderConflict = (CurrentByteOrderSignature != ByteOrderSignatureOnDisk);
	IML_LINT TransformationFlag = CurrentByteOrderSignature | ByteOrderSignatureOnDisk;

	if(ByteOrderConflict) FileHeader.fix_byte_order(TransformationFlag);


	// CHECK THE DIRECTORY ENTRY COUNT FOR SUSPICIOUS VALUES
	if( FileHeader.DirEntryCnt < 0 || FileHeader.DirEntryCnt > 256 ) {
		iml_error("image_file",
				"The directory entry count seems wrong...aborting...");
		return(NULL);
      } // end if

   // CALCULATE DIRECTORY SIZE
	IML_ULINT DirectorySize =
			FileHeader.DirEntryCnt * sizeof(image_file_directory_entry);

	// ALLOCATE DIRECTORY SPACE
	image_file_directory_entry *Directory =
					new image_file_directory_entry[FileHeader.DirEntryCnt];


	// READ THE IMAGE FILE DIRECTORY INFORMATION
   ReadSize = DirectorySize;
	if( huge_read(Directory, ReadSize) != ReadSize) {
		iml_error("image_file", "Error on read_image of file header");
      delete [] Directory;
		return(NULL);
		} // end if

	// I DID NOT STORE THE VERSION IN THE FIRST EDITION; IF IT IS NOT
	// IN THE FILE, THEN IT IS THE FIRST VERSION.
	IML_ULINT StoredImFormat = 0x0001;
	IML_ULINT Dimensionality;

	// GRAB THE REQUIRED DATA FROM THE DIRECTORY
	IML_LINT XWidth = 0, YHeight = 0;		// FOR READING OLD IMAGE FORMATS.
   IML_LINT ImageType = -1;
   IML_LINT DimInfoPosition = 0;
	IML_LINT DataPosition = 0;
   int RequiredDataCount = 0;
	for(int Index = 0; Index < FileHeader.DirEntryCnt; Index++) {

		if(ByteOrderConflict) Directory[Index].fix_byte_order(TransformationFlag);

		switch( Directory[Index].Type ) {
			case IML_DIR_ENTRY_XWIDTH:
				XWidth = Directory[Index].Value;
            RequiredDataCount++;
				break;

			case IML_DIR_ENTRY_YHEIGHT:
				YHeight = Directory[Index].Value;
				RequiredDataCount++;
				break;

			case IML_DIR_ENTRY_IMAGE_TYPE:
				ImageType = Directory[Index].Value;
				RequiredDataCount++;
				break;
				
			case IML_DIR_ENTRY_DATABITS:
				DataPosition = Directory[Index].Value;
				RequiredDataCount++;
				break;
				
			case IML_DIR_FORMAT_VERSION:
				StoredImFormat = Directory[Index].Value;
				RequiredDataCount++;
				break;
				
			case IML_DIR_IMAGE_DIMENSION:
				Dimensionality = Directory[Index].Value;
				RequiredDataCount++;
				break;
				
			case IML_DIR_DIM_INFO_START:
				DimInfoPosition = Directory[Index].Value;
				RequiredDataCount++;
				break;

			default: // THERE COULD BE NEW ONES THAT I DON'T NEED...
				break;

			} // end switch

      } // end for


	if(Dimensionality==0) {
		iml_error("iml_image_file","Error on image dimensionality");
		return(NULL);
	} // error condition

	// WE ALLOCATE AND INITIALIZE THE IMAGE DIMENSION OBJECT IN THIS SWITCH;
	iml_index *DimInfo;
	switch(StoredImFormat) {
	
			// ONLY DID 2-D IMAGES IN THE FIRST VERSION;
		case 0x0001: {
			DimInfo = new iml_ld_index(XWidth, YHeight);
			Dimensionality = 2;
			if( !(DimInfo && DimInfo->Axes) ) {
				iml_error("image_file", "Error in read_image: dimension object allocation");
				return(NULL);
				} // end if
			}; break;
	
		case 0x0002: {
			DimInfo = new iml_index(Dimensionality);
			if( !(DimInfo && DimInfo->Axes) ) {
				iml_error("image_file", "Error in read_image: dimension object allocation");
				return(NULL);
				} // end if

			// MOVE THE FILE POINTER TO THE DIMENSIONALITY INFORMATION POSITION
			if(set_location(DimInfoPosition) == HFILE_ERROR) {
				iml_error("image_file", "File pointer error finding dimension information");
				delete DimInfo;
				return(NULL);
				} // end if

			// LOAD IN THE DIMINFO DATA
			ReadSize = sizeof(iml_axis)*Dimensionality;
			if( huge_read(DimInfo->Axes, ReadSize) != ReadSize) {
				iml_error("image_file", "Error in read_image: reading dimension info");
				return(NULL);
				} // end if
				
			DimInfo->house_keeping();
			DimInfo->restart();
				
			}; break;
		
		}; // END switch(StoredImFormat)

	// CALCULATE THE MINIMUM IMAGE SIZE IF WE ARE TO HAVE ANY DATA
	IML_ULINT MinDataPos = sizeof(image_file_header) + DirectorySize;

	// CHECK FOR DATA ERRORS
//	if( (RequiredDataCount != 4) ||
//			(XWidth <= 0) || (YHeight <= 0) ||
//			(DataPosition <= MinDataPos) ) {
	if( DataPosition <= MinDataPos ) {
		iml_error( "image_file", "Image information not valid for image creation");
		delete DimInfo;
		delete [] Directory;
      return(NULL);
		} // end if

	// CREATE THE 2D IMAGE
	iml_image *Image;
	switch(ImageType) {
		case IML_BYTE_IMAGE:
			Image = new iml_byte_image( *DimInfo );
			break;
		case IML_UBYTE_IMAGE:
			Image = new iml_ubyte_image( *DimInfo );
			break;
		case IML_SINT_IMAGE:
			Image = new iml_sint_image( *DimInfo );
			break;
		case IML_USINT_IMAGE:
			Image = new iml_usint_image( *DimInfo );
			break;
		case IML_LINT_IMAGE:
			Image = new iml_lint_image( *DimInfo );
			break;
		case IML_ULINT_IMAGE:
			Image = new iml_ulint_image( *DimInfo );
			break;
		case IML_FLOAT_IMAGE:
			Image = new iml_float_image( *DimInfo );
			break;
		case IML_DOUBLE_IMAGE:
			Image = new iml_double_image( *DimInfo );
			break;
		case IML_FCOMPLEX_IMAGE:
			Image = new iml_fcomplex_image( *DimInfo );
			break;
		case IML_DCOMPLEX_IMAGE:
			Image = new iml_dcomplex_image( *DimInfo );
			break;
		case IML_LDCOMPLEX_IMAGE:
			Image = new iml_ldcomplex_image( *DimInfo );
			break;
		default:
			iml_error("image_file",
				"Unknown image type found in the file.  Aborting file read.");
			delete [] Directory;
			delete DimInfo;
			return(NULL);
		}; // end switch

	// CHECK FOR GOOD IMAGE CREATION
	if( !Image->valid_memory() ) {
		iml_error("image_file",
				"Must have run out of memory: could not create image");
		delete [] Directory;
		delete Image;
		delete DimInfo;
		return(NULL);
		} // end if

	// CALCULATE IMAGE DATA SIZE
	IML_ULINT DataBytes = Image->subspace_size_in_bytes();

	// CALCULATE THE MINIMUM FILE SIZE IF WE ARE TO HAVE ANY DATA
	IML_ULINT MinFileSize = DataPosition + DataBytes;

	// IF THE FILE IS TOO SMALL....
	if(FileSize < MinFileSize) {
		iml_error("image_file", "File appears to be too small...aborting...");
		delete [] Directory;
		delete Image;
		delete DimInfo;
		return(NULL);
		} // end if


	// MOVE THE FILE POINTER TO THE IMAGE DATA POSITION
	if(set_location(DataPosition) == HFILE_ERROR) {
		iml_error("image_file", "File pointer error 2 in read_image");
		delete [] Directory;
		delete Image;
		delete DimInfo;
		return(NULL);
		} // end if

	// LOAD IN THE IMAGE DATA
	ReadSize = DataBytes;
	if( huge_read(Image->subspace_chunk_address(), ReadSize) != ReadSize) {
		iml_error("image_file", "Error on read_image of image data");
		delete [] Directory;
		delete Image;
		delete DimInfo;
		return(NULL);
		} // end if

	// NOW WE CAN SAFELY GET RID OF THE DIRECTORY INFORMATION
	delete [] Directory;
	delete DimInfo;
	
	if(ByteOrderConflict) {
		IML_LINT Success = fix_byte_order(
				Image, ImageType, CurrentByteOrderSignature | ByteOrderSignatureOnDisk);
		if(!Success) iml_warning("image_file","Image byte order may be wrong!");
		} // end if
	
	// WE HAVE NO WAY TO KNOW IF THE IMAGE DATA IS GOOD,
	// SO WE JUST RETURN IT.  FREEING UP THE IMAGE IS UP TO THE CALLER.
	return(Image);


} // end read_image


//============================================================
IML_LINT iml_image_file::fix_byte_order(iml_image IML_PTR Image, IML_LINT ImageType,
			IML_LINT TransformationFlag) {


	IML_LINT Result = FALSE;
	switch( TransformationFlag ) {
	
		case(IML_BIG_LITTLE_ENDIAN_XFORM): {
		
			switch(ImageType) {
				case IML_BYTE_IMAGE: case IML_UBYTE_IMAGE: {
					Result = TRUE;
					} break;
				
				case IML_SINT_IMAGE: case IML_USINT_IMAGE: {
					register IML_USINT IML_PTR Ptr = (IML_USINT IML_PTR)Image->subspace_chunk_address();
					register IML_ULINT Cnt = Image->subspace_size_in_bytes()/sizeof(IML_USINT);
					while(Cnt--) iml_2byte_reverse( *Ptr++ );
					Result = TRUE;
					} break;
					
				case IML_LINT_IMAGE: case IML_ULINT_IMAGE: case IML_FLOAT_IMAGE: {
					register IML_ULINT IML_PTR Ptr = (IML_ULINT IML_PTR)Image->subspace_chunk_address();
					register IML_ULINT Cnt = Image->subspace_size_in_bytes()/sizeof(IML_ULINT);
					while(Cnt--) iml_4byte_reverse( *Ptr++ );
					Result = TRUE;
					} break;

				case IML_DOUBLE_IMAGE: {
					register IML_DOUBLE IML_PTR Ptr = (IML_DOUBLE IML_PTR)Image->subspace_chunk_address();
					register IML_ULINT Cnt = Image->subspace_size_in_bytes()/sizeof(IML_DOUBLE);
					while(Cnt--) iml_8byte_reverse( *Ptr++ );
					Result = TRUE;
					} break;

				case IML_FCOMPLEX_IMAGE: {
					register IML_FCOMPLEX IML_PTR Ptr = (IML_FCOMPLEX IML_PTR)Image->subspace_chunk_address();
					register IML_ULINT Cnt = Image->subspace_size_in_bytes()/sizeof(IML_FCOMPLEX);
					while(Cnt--) iml_2x4byte_reverse( *Ptr++ );
					Result = TRUE;
					} break;

				case IML_DCOMPLEX_IMAGE: {
					register IML_DCOMPLEX IML_PTR Ptr = (IML_DCOMPLEX IML_PTR)Image->subspace_chunk_address();
					register IML_ULINT Cnt = Image->subspace_size_in_bytes()/sizeof(IML_DCOMPLEX);
					while(Cnt--) iml_2x8byte_reverse( *Ptr++ );
					Result = TRUE;
					} break;

				case IML_LDCOMPLEX_IMAGE: {
		#if __HAVE_LDOUBLE__
					register IML_LDCOMPLEX IML_PTR Ptr = (IML_LDCOMPLEX IML_PTR)Image->subspace_chunk_address();
					register IML_ULINT Cnt = Image->subspace_size_in_bytes()/sizeof(IML_LDCOMPLEX);
					while(Cnt--) iml_ld_byte_reverse( *Ptr++ );
		#else
					register IML_DCOMPLEX IML_PTR Ptr = (IML_DCOMPLEX IML_PTR)Image->subspace_chunk_address();
					register IML_ULINT Cnt = Image->subspace_size_in_bytes()/sizeof(IML_DCOMPLEX);
					while(Cnt--) iml_2x8byte_reverse( *Ptr++ );
		#endif
					Result = TRUE;
					} break;		
				} // end switch
			} // end case
	
		} // end switch

	return(Result);

} // end fix_byte_order

