#ifndef __IM_FILE_H__
#define __IM_FILE_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1993-99 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "iml_type.h"
#include "iml_err.h"

#include "iml_ndim.h"

#include <stdio.h>

	// THIS IS A KLUDGE TO ALLOW A COMMON ERROR MESSAGE
	// FOR FILE OPERATIONS; ORIGINALLY THIS NAME CAME FROM
	// THE MS WINDOWS ROUTINES
#ifdef _Windows
#else
#ifdef __INTEL__
#else
#ifndef HFILE
#define HFILE long
#endif
#ifndef HFILE_ERROR
#define HFILE_ERROR			((HFILE)-1)
#endif
#endif
#endif

#define IML_IDSTRING_SIZE				256
#define IML_FILE_CREATE					0x0001
#define IML_FILE_OPEN					0x0002
#define IML_IML_FORMAT					0x0100
#define IML_RAW_FORMAT					0x0200
#define IML_HIPS_FORMAT					0x0400

#define IML_IN_DIRECTORY				0x0100

enum {
		IML_DIR_ENTRY_XWIDTH ,
		IML_DIR_ENTRY_YHEIGHT,
		IML_DIR_ENTRY_IMAGE_TYPE,
		IML_DIR_ENTRY_DATABITS,
		IML_DIR_FORMAT_VERSION,
		IML_DIR_IMAGE_DIMENSION,
		IML_DIR_DIM_INFO_START,
		IML_LAST_DIR_ENTRY
		}; // end enum

//#define IML_MAX_ENTRY_COUNT			4

class image_file_directory_entry {
public:
	IML_LINT 	Type;			// ENTRY TYPE
	IML_ULINT   Location;	// OFFSET INTO THE FILE
	IML_LINT		Value;		// value OF THINGS
	void	fix_byte_order(IML_LINT TransformationFlag);
};

class image_file_header {
public:
	char 									IDString[IML_IDSTRING_SIZE];
	IML_SINT  							DirEntryCnt;
	void	fix_byte_order(IML_LINT TransformationFlag);
	
		// THERE SHOULD BE A DIRECTORY RIGHT AFTER THE HEADER INFO.
		// READ THE HEADER SO THAT WE KNOW HOW BIT HTE DIRECTORY IS,
      // THEN READ THE DIRECTORY
  //	image_file_directory_entry		Directory[?];  // there should be a
};


/** This is the class used to manage n-dimensional images stored
on disk.  It provides byte-order independent, hyperplane aware
storage and retrieval of the image classes defined in iml_ndim.h.
The image file manager only manages one file at a time.  If you wish
to have multiple open image files, then either open them sequentially,
loading them one at a time, or use multiple <em>iml_image_file</em>
instances. 
*/
class iml_image_file {

	protected:
#ifdef __I_DO_WINDOWS__
		HFILE hDiskFile;
#else
		FILE *hDiskFile;
#endif

		image_file_header FileHeader;

		IML_LINT	set_location(IML_ULINT Loc);
		IML_LINT	file_length(void);
      IML_LINT	huge_write(void IML_PTR Data, IML_LINT Size);
		IML_LINT	huge_read(void IML_PTR Data, IML_LINT Size);

		IML_LINT	write_hips_fmt_prefix(iml_image IML_PTR Image);
		void 		write_hips_fmt_suffix(iml_image IML_PTR Image);
		IML_LINT	write_iml_fmt_prefix(iml_image IML_PTR Image);
		IML_LINT fix_byte_order(iml_image IML_PTR Image,IML_LINT ImageType,
			IML_LINT TransformationFlag);

	public:
		/* Constructor used when we already know the name of the 
		file we wish to open.
		*/
		iml_image_file(const char *FileName, IML_USINT FileOp = IML_FILE_OPEN);

		/** A no-argument constructor.
		*/
		iml_image_file(void) {  hDiskFile = NULL;  };

		~iml_image_file(void) { close_file(); };

		/** Explicitly close the file that is currently open.  If no file is
		open no action is taken. Returns TRUE
		if the close process was successful, FALSE if it was not.
		*/
		IML_SINT close_file(void);

		/** Explicitly attempt to open a file that already exists on the disk.
		If another file is already open that file is first closed. Returns TRUE
		if the open process was successful, FALSE if it was not.
		*/
		IML_SINT open_file(const char *FileName);

		/** Explicitly attempt to create a new image file on the disk.  If
		another file is already open that file is first closed. Returns TRUE
		if the creation process was successful, FALSE if it was not.
		*/
		IML_SINT create_file(const char *FileName);

		/** Write an image to disk.  Assumes that a file has already been opened.
		The default format is the native image format (a header format defined
		in this library, using a data directory to store image attributes and 
		pixel values).  The image can also be stored as IML_RAW_FORMAT format,
		or just raw pixel values with no header information. Returns TRUE
		if the write process was successful, FALSE if it was not.
		*/
		IML_SINT write_image(iml_image IML_REF Image, IML_LINT Format=IML_IML_FORMAT)
			{ return(write_image(&Image, Format)); };

		IML_SINT write_image(iml_image IML_PTR Image, IML_LINT Format=IML_IML_FORMAT);
		
		/** Read an image from disk and allocate an image of the appropriate
		type and size.  The function returns a pointer to a base image class,
		<em>iml_image</em>.  The user can  determine the image type with the
		image class function <em>image_type()</em> and then cast the returned
		pointer to the correct type.  Normally, however, the user will know 
		what type of image is supposed to be obtained from the file and only use 
		<em>image_type()</em> to make sure that the image is of the correct type.
		Returns a NULL pointer if the image could not be read from the file.
		*/
		iml_image *read_image(void);
		
		enum {IML_IMAGE_AT_EOF = -1};
		
		/** Read raw image-pixel values into an image from disk. The user
		allocates an image of the appropriate type and size in the program,
		then uses this function to read the raw pixels into that image's
		memory space.  
		The function returns a non-zero number if the read went without a hitch.
		This function is primarily to allow portability of other images into
		and out of the image file format native to this library.
		*/
		IML_SINT load_raw_pixels(iml_image IML_REF Image, IML_LINT SkipHeader=0)
			{ return(load_raw_pixels(&Image, SkipHeader)); };

		IML_SINT load_raw_pixels(iml_image IML_PTR Image, IML_LINT SkipHeader=0);
		
}; // end class iml_image_file


#endif // __IMFILE_H__
