/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/


#include "iml_indx.h"

	
IML_LINT iml_index::set_offset(IML_LINT NewICOffset) {
	IML_LINT Temp = ImageCornerOffset;
	ImageCornerOffset = NewICOffset;
	return(Temp);
	}
	
	
iml_index::iml_index(IML_ULINT DimCnt) {
	ImageCornerOffset = 0;
	Dimensionality = DimCnt;
	Axes = new iml_axis[Dimensionality];
	house_keeping();
	}// end constructor


iml_index::iml_index(const iml_index IML_REF Example) {
	ImageCornerOffset = 0;
	Dimensionality = Example.Dimensionality;
	Axes = new iml_axis[Dimensionality];
	for(int I=0; I<Dimensionality; I++) Axes[I] = Example.Axes[I];
	// do this just in case; we don't know that the example is safe.
	house_keeping();
	}// end constructor


void iml_index::resize(IML_ULINT NewDim) {
	if( Axes ) {
		iml_axis *NewAxes = new iml_axis[NewDim];
		if( !NewAxes ) return;  // memory problems
		IML_ULINT MinDim = (NewDim < Dimensionality) ? NewDim : Dimensionality;
		for(IML_USINT I = 0; I < MinDim; I++) NewAxes[I] = Axes[I];
		delete [] Axes;
		Axes = NewAxes;
		Dimensionality = NewDim;
		}
	// else assume memory problems.
	}

		
iml_index IML_REF iml_index::set(IML_ULINT Axis, IML_AXIS_TYPE V) { 

	// KEEP THE AXIS WITHIN THE CORRECT BOUNDS
	IML_CLIPIT(Axis, 0, Axis, Dimensionality-1);

	// KEEP INDEX WITHIN RANGE
	IML_CLIPIT(Axes[Axis].AxisV, amin(Axis), V, amax(Axis) );

		// this routine fixes up the linear index
	set_pointer();

	return( *this );
	} // end set_axis_index
		
		
void iml_index::restart(void) {
	for(int I=0; I<Dimensionality; I++)  Axes[I].AxisV = Axes[I].MinV;
	set_pointer();
	}

// IN THIS ASSIGNMENT OPERATOR WE JUST DO THE COMMON SUBSPACE
const iml_index IML_REF iml_index::operator=( const iml_index IML_REF Source) {
	IML_ULINT MinDim = (Dimensionality < Source.Dimensionality) ? Dimensionality : Source.Dimensionality;
	for(int I=0; I<MinDim; I++) Axes[I] = Source.Axes[I];
	// do this just in case; we don't know that the example is safe.
	house_keeping();
		
	return( *this );
	} 

	// THIS IS A *POSTFIX* INCREMENTOR! ;  THAT IS THE 'INT' KLUDGE
const iml_index IML_REF iml_index::operator++(int) {

	// INCREMENT THE LEAST AXIS
	Axes[0].AxisV++;
	// SEE ABOUT WRAP AND CARRY
	int I=0; 
	while(  (Axes[I].AxisV > Axes[I].MaxV) && (I < Dimensionality) ) {
			Axes[I].AxisV = Axes[I].MinV;
			I++;
				//	MAKE SURE NOT TO RE-INCREMENT THE LOWEST DIMENSION;
				// THE IF IS SLOW . . THERE MUST BE A BETTER WAY
			if( I < Dimensionality ) Axes[I].AxisV++;
		} // end while
	set_pointer();
	
	return( *this );
}


void iml_index::set_pointer(void) {
	// CALCULATE OFFSET INTO THE LINEAR ARRAY
	ElementPtr = 
		AxesOrigin +
			ImageCornerOffset;
	for(int I=0; I< Dimensionality; I++) {
		ElementPtr += (IML_LINT)(Axes[I].Stride*Axes[I].AxisV);
		} // end for
	
	IML_LINT MinVal = ImageCornerOffset;
	IML_LINT MaxVal = MinVal + ElementCount-1;
	// MAKE SURE THAT THE POINTER INTO THE IMAGE STAYS WITHIN THE IMAGE
	IML_CLIPIT_NW(ElementPtr, MinVal, ElementPtr, MaxVal);

	} // end set_pointer
	
	
void iml_index::set_min_max(IML_ULINT Dim, IML_AXIS_TYPE MinV, IML_AXIS_TYPE MaxV) { 
	Axes[Dim].set_min_max(MinV,MaxV); 

	// MAKE SURE THAT EACH INDEX STAYS WITHIN RANGE
	IML_CLIPIT_NW(Axes[Dim].AxisV, amin(Dim), axis(Dim), amax(Dim) );

		// this routine fixes up stride values and the linear index
	house_keeping();
	}
		
void iml_index::shift_corner_by(IML_ULINT Dim, IML_AXIS_TYPE V) { 
	Axes[Dim].set_min_max(amin(Dim)+V, amax(Dim)+V); 

	// MAKE SURE THAT EACH INDEX STAYS WITHIN RANGE
	IML_CLIPIT_NW(Axes[Dim].AxisV, amin(Dim), axis(Dim), amax(Dim) );

		// this routine fixes up stride values and the linear index
	house_keeping();
	}
	
void iml_index::shift_corner_to(IML_ULINT Dim, IML_AXIS_TYPE V) {
	Axes[Dim].set_min_max(V, V+size(Dim)-1); 
		// this routine fixes up stride values and the linear index

	// MAKE SURE THAT EACH INDEX STAYS WITHIN RANGE
	IML_CLIPIT_NW(Axes[Dim].AxisV, amin(Dim), axis(Dim), amax(Dim) );

	house_keeping();
	}

// NOTICE THAT THIS FUNCTION DOES NOT RESTRICT THE ELEMENT COUNT TO BE REDUCED
// TO THAT OF THE OTHER DOMAIN.  THAT MEANS THAT IN ANY LATER ASSIGNMENT OPERATIONS
// WHICH USE THE RETURNED INDEX ELEMENT COUNT, THE LOWER DIMENSIONAL OBJECT BEGIN ASSIGNED
// TO A HIGHER DIMENSIONAL OBJECT WILL BE REPEATED FOR ALL SUBSPACE SLICES.
// THIS WAS A "FEATURE" THAT I FAILED TO THINK ABOUT, BUT LIKE AFTER HAVING SEEN IT
// HAPPEN.  IF YOU WANT THE ASSIGNMENT TO BE *ONLY* A SUBSPACE, THEN RESTRICT THE
// DESTINATION IMAGES SUBSPACE BEFORE HAND!
void iml_index::clip_against(const iml_index IML_REF OD) {
 	set_offset(0);
	IML_ULINT DimCnt =  (OD.Dimensionality < Dimensionality) ? OD.Dimensionality : Dimensionality;
 	// DO THE CLIPPING
 	for(IML_USINT I = 0; I < DimCnt; I++) {
 		Axes[I].set_min_max( MAX_OF(Axes[I].MinV, OD.Axes[I].MinV), 
 									MIN_OF(Axes[I].MaxV, OD.Axes[I].MaxV) );
 		} // end for
 	}

// THIS IS USED IN CONJUNCTION WITH clip_against() FOR ASSIGNMENT OPERATIONS;
// OTHERWISE WE WOULD DO DUPLICATED WORK.
IML_ULINT iml_index::unclipped_count(void) {
 	IML_ULINT UCCnt = 1;
 	for(IML_ULINT I = 0; I < Dimensionality; I++)  UCCnt *= (IML_ULINT)Axes[I].Size;
 	return(UCCnt);
 	}
		
		
void iml_index::house_keeping(void) {
	// CALCULATE LINEAR MEMORY STRIDE SIZES 
	// AND ORIGIN OFFSET INTO THE LINEAR ARRAY
	Axes[0].Stride = 1;
	AxesOrigin = (IML_LINT)-Axes[0].MinV;
	for(int I=1; I< Dimensionality; I++) {
		Axes[I].Stride = Axes[I-1].Stride*Axes[I-1].Size;
//  9/12/97  add this to try to get (X,1,Z) dimensional images to work right
		if(Axes[I].Stride == 0) Axes[I].Stride = 1;
		AxesOrigin -= (IML_LINT)(Axes[I].Stride*Axes[I].MinV);
		} // end for
	ElementCount = (IML_LINT)(Axes[Dimensionality-1].Stride*Axes[Dimensionality-1].Size);
	set_pointer();
	} // end house_keeping

