#ifndef __IML_INDX_H__
#define __IML_INDX_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/


//#include "iml_type.h"
#include "iml_cplx.h"
#include "iml_err.h"

// DEFINE SOME DEFAULT INDICES FOR LATER USE WITH IMAGES. :-)
typedef enum { IML_2D_X, IML_2D_Y, IML_2D_DIM_CNT, 
					IML_2D_COL = IML_2D_X, IML_2D_ROW = IML_2D_Y 
					} iml_2D_axes;

typedef enum { IML_2ST_X, IML_2ST_Y, IML_2ST_T, 
				IML_2ST_DIM_CNT } iml_2ST_axes;


typedef enum { IML_3ST_X, IML_3ST_Y, IML_3ST_Z, IML_3ST_T, 
				IML_3ST_DIM_CNT } iml_3ST_axes;

#define IML_AXIS_TYPE IML_DOUBLE
#define IML_AXIS_MAX	10e100
#define IML_AXIS_MIN	-10e100

// THE USER SHOULD NOT REALLY HAVE TO ACCESS THE iml_axis BY HAND.
class iml_axis {

	public:
		IML_AXIS_TYPE AxisV;
		IML_AXIS_TYPE MaxV;
		IML_AXIS_TYPE MinV;
		IML_AXIS_TYPE Size;
		IML_AXIS_TYPE Stride;
		
		void set_min_max(IML_AXIS_TYPE Min, IML_AXIS_TYPE Max) {
			MaxV = Max; MinV = Min; 
			Size = (Max>=Min) ? MaxV-MinV + 1 : 0;
			}; 
			
			// use some default limits
		iml_axis(void) { 
			AxisV = 0; 
			Stride = 1; 
			set_min_max(IML_AXIS_MIN, IML_AXIS_MAX); 
		};
		
}; // end class iml_axis


//========================================================================
// THIS CLASS IS USED FOR A NUMBER OF THINGS, AND PROVIDES A COMMON
// MEANS OF INDEXING IMAGES, AND INDEXING N-DIMENSIONAL FUNCTIONS.
// THIS WORKS BECAUSE THE AXIS CLASS HAS FLOATING POINT VALUES RATHER
// THAN INTEGRAL VALUES. :-)
//========================================================================
class iml_index {
	private:
		friend class iml_image_file;
		friend class iml_image;

		IML_LINT set_offset(IML_LINT NewICOffset);
		void set_pointer(void);
		void house_keeping(void);
		void resize(IML_ULINT NewDim);

	protected:
		IML_LINT AxesOrigin;
		IML_LINT ImageCornerOffset;
		IML_LINT ElementPtr;
				
	public:
	
//==========================================================================
//  UNLESS YOU WROTE THE LIBRARY, DON'T USE STUFF BELOW THIS LINE! :-)
//==========================================================================
		iml_axis IML_PTR Axes;
		IML_ULINT Dimensionality;
		IML_LINT ElementCount;

	// THESE ARE USED BY THE IMAGE CLASSES.
	void set_ns(IML_ULINT Axis, IML_AXIS_TYPE V) { Axes[Axis].AxisV = V; };
	IML_LINT get_offset(void) { return(ImageCornerOffset); };
	IML_LINT get_origin(void) { return(AxesOrigin); };
		
	// THIS FUNCTION DOES NOT REDUCE THE ELEMENT COUNT TO THE CLIPPED DOMAIN.  
	// BUILT-IN ASSIGNMENT OPERATIONS ARE EFFICIENT USING unclipped_count().
	void clip_against(const iml_index IML_REF OD);
	
	// THIS IS USED IN CONJUNCTION WITH clip_against() FOR ASSIGNMENT OPERATIONS;
	// OTHERWISE WE WOULD DO DUPLICATE WORK.
	IML_ULINT unclipped_count(void);
	
	// MAKE SURE THAT YOUR AXIS SIZE IS NO BIGGER THAN 
	// THAT OF THE IMAGE TO BE INDEXED. :-)
	void set_min_max(IML_ULINT Dim, IML_AXIS_TYPE MinV, IML_AXIS_TYPE MaxV);
				
				
//==========================================================================
//**************** STUFF BELOW IS FOR PUBLIC CONSUMPTION **************
//==========================================================================
	// TWO KINDS OF CONSTRUCTOR, ONE DESTRUCTOR
	explicit iml_index(IML_ULINT DimCnt);
	explicit iml_index(const iml_index IML_REF Example);
	~iml_index() { if(Axes)    delete [] Axes; }

	// FOR CASTING AND USING AS A LINEAR INDEX
//	operator IML_LINT() { return( ElementPtr ); };
	IML_LINT indexize() { return( ElementPtr ); };
	
      // FOR USER ACCESS TO VALUES
	IML_AXIS_TYPE size(IML_ULINT Axis)  { return(Axes[Axis].Size); };
	IML_AXIS_TYPE axis(IML_ULINT Axis) { return(Axes[Axis].AxisV); };
	IML_AXIS_TYPE amin(IML_ULINT Axis)  { return(Axes[Axis].MinV); };
	IML_AXIS_TYPE amax(IML_ULINT Axis)  { return(Axes[Axis].MaxV); };

	// FOR USER MANIPULATION OF AXIS VALUES AND RANGES
	iml_index IML_REF set(IML_ULINT Axis, IML_AXIS_TYPE V);
	
	// RESTART THE INDEX AT THE "CORNER" OF THE SPACE
	void restart(void);

	// IN THIS ASSIGNMENT OPERATOR WE JUST ASSIGN THE COMMON SUBSPACE
	const iml_index IML_REF operator=( const iml_index IML_REF Source); 

		// THIS IS A *POSTFIX* INCREMENTOR! ;  THAT IS THE 'INT' KLUDGE
	const iml_index IML_REF operator++(int);

	void shift_corner_by(IML_ULINT Dim, IML_AXIS_TYPE V);
		
	void shift_corner_to(IML_ULINT Dim, IML_AXIS_TYPE V);

}; // end class


// A SIMPLE CLASS TO SAVE TYPING AND AGRAVATION :-)
class iml_ld_index : public iml_index {

	public:
		iml_ld_index(IML_ULINT XSize) 
					: iml_index(1) {
			set_min_max(IML_2D_X,0,XSize-1);
			};
			
		iml_ld_index(IML_ULINT XSize, IML_ULINT YSize) 
					: iml_index(2) {
			set_min_max(IML_2D_X,0,XSize-1);
			set_min_max(IML_2D_Y,0,YSize-1);
			};
			
		iml_ld_index(IML_ULINT XSize, IML_ULINT YSize, IML_ULINT TSize) 
					: iml_index(3) {
			set_min_max(IML_2ST_X,0,XSize-1);
			set_min_max(IML_2ST_Y,0,YSize-1);
			set_min_max(IML_2ST_T,0,TSize-1);
			};
			
		iml_ld_index(IML_ULINT X0, IML_ULINT X1, IML_ULINT X2, IML_ULINT X3) 
					: iml_index(4) {
			set_min_max(0,0,X0-1);		set_min_max(1,0,X1-1);
			set_min_max(2,0,X2-1);		set_min_max(3,0,X3-1);
			};
			

}; // end iml_ld_index



#endif // __IML_INDX_H__
