#ifndef __IML_MATH_H__
#define __IML_MATH_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

// THIS IS INTENDED TO PROVIDE A PORTABLE SET OF OVERLOADED MATH FUNCTIONS.
// WHEN YOU PORT, JUST MAKE SURE THAT THIS FILE AND THE imltypes.h FILE
// ARE UPDATED CORRECTLY.


#include "iml_type.h"
#include <math.h>
#include <stdio.h>

#define MAX_OF(a,b) ((a)<(b)?(b):(a))
#define MIN_OF(a,b) ((a)>(b)?(b):(a))

//MACMOD -- REF OCT/14/96
#define IML_MAC_PI							pi
//============
#define IML_2PI								(2.0*IML_PI)
#define IML_PI									(3.14159265359)
//#endif
#define IML_SQRT2								(1.414213562373)
#define IML_RAD_PER_DEG						(IML_PI/180)
#define IML_DEG_PER_RAD						(180.0/IML_PI)

inline IML_DOUBLE  iml_sgn(IML_DOUBLE Val) {return(((Val>0)?1:((Val<0)?-1:0))); };

inline IML_DOUBLE  iml_acos(IML_DOUBLE Val)						{ return(acos(Val)); };
inline IML_DOUBLE  iml_asin(IML_DOUBLE Val)						{ return(asin(Val)); };
inline IML_DOUBLE  iml_atan(IML_DOUBLE Val)						{ return(atan(Val)); };
inline IML_DOUBLE  iml_atan2(IML_DOUBLE X, IML_DOUBLE Y) 	{ return(atan2(X,Y)); };

inline IML_DOUBLE  iml_sin(IML_DOUBLE Val)						{ return(sin(Val)); };
inline IML_DOUBLE  iml_cos(IML_DOUBLE Val)						{ return(cos(Val)); };
inline IML_DOUBLE  iml_tan(IML_DOUBLE Val)						{ return(tan(Val)); };

inline IML_DOUBLE  iml_sinh(IML_DOUBLE Val)						{ return(sinh(Val)); };
inline IML_DOUBLE  iml_cosh(IML_DOUBLE Val)						{ return(cosh(Val)); };
inline IML_DOUBLE  iml_tanh(IML_DOUBLE Val)						{ return(tanh(Val)); };

inline IML_DOUBLE  iml_ceil(IML_DOUBLE Val)						{ return(ceil(Val)); };
inline IML_DOUBLE  iml_floor(IML_DOUBLE Val)						{ return(floor(Val)); };

inline IML_DOUBLE  iml_gaussian(IML_DOUBLE C, IML_DOUBLE SC, IML_DOUBLE X)			{
										return(exp( -(X-C)*(X-C)/(SC*SC) ));	};
inline IML_DOUBLE  iml_exp(IML_DOUBLE Val)						{ return(exp(Val)); };
inline IML_DOUBLE  iml_sqrt( IML_DOUBLE Val) 					{ return(sqrt(Val)); };
inline IML_DOUBLE  iml_pow(IML_DOUBLE X, IML_DOUBLE Y)		{ return(pow(X,Y)); };
inline IML_DOUBLE  iml_log(IML_DOUBLE Val)						{ return(log(Val)); };
inline IML_DOUBLE  iml_log10(IML_DOUBLE Val)						{ return(log10(Val)); };

inline IML_DOUBLE  iml_abs(IML_DOUBLE Val)						{ return(fabs(Val)); };
inline IML_DOUBLE  iml_mod(IML_DOUBLE X, IML_DOUBLE Y)		{ return(fmod(X,Y)); };


#if __HAVE_LDOUBLE__  // THEN ADD THESE FOR LONG DOUBLE TYPE
inline IML_LDOUBLE iml_sgn(IML_LDOUBLE Val) {return(((Val>0)?1:((Val<0)?1:0))); };

inline IML_LDOUBLE iml_acos(IML_LDOUBLE Val)						{ return(acosl(Val)); };
inline IML_LDOUBLE iml_asin(IML_LDOUBLE Val)						{ return(asinl(Val)); };
inline IML_LDOUBLE iml_atan(IML_LDOUBLE Val)						{ return(atanl(Val)); };
inline IML_LDOUBLE iml_atan2(IML_LDOUBLE Y, IML_LDOUBLE X) 	{ return(atan2l(Y,X)); };

inline IML_LDOUBLE iml_sin(IML_LDOUBLE Val)						{ return(sinl(Val)); };
inline IML_LDOUBLE iml_cos(IML_LDOUBLE Val)						{ return(cosl(Val)); };
inline IML_LDOUBLE iml_tan(IML_LDOUBLE Val)						{ return(tanl(Val)); };

inline IML_LDOUBLE iml_sinh(IML_LDOUBLE Val)						{ return(sinhl(Val)); };
inline IML_LDOUBLE iml_cosh(IML_LDOUBLE Val)						{ return(coshl(Val)); };
inline IML_LDOUBLE iml_tanh(IML_LDOUBLE Val)						{ return(tanhl(Val)); };

inline IML_LDOUBLE iml_ceil(IML_LDOUBLE Val)						{ return(ceill(Val)); };
inline IML_LDOUBLE iml_floor(IML_LDOUBLE Val)					{ return(floorl(Val)); };

inline IML_DOUBLE  iml_gaussian(IML_LDOUBLE C, IML_LDOUBLE SD, IML_LDOUBLE X)	{
										return(expl( -(X-C)*(X-C)/(SD*SD) ));	};
inline IML_LDOUBLE iml_exp(IML_LDOUBLE Val)						{ return(expl(Val)); };
inline IML_LDOUBLE iml_sqrt( IML_LDOUBLE Val) 					{ return(sqrtl(Val)); };
inline IML_LDOUBLE iml_pow(IML_LDOUBLE X, IML_DOUBLE Y)		{ return(powl(X,Y)); };
inline IML_LDOUBLE iml_log(IML_LDOUBLE Val)						{ return(logl(Val)); };
inline IML_LDOUBLE iml_log10(IML_LDOUBLE Val)					{ return(log10l(Val)); };

inline IML_LDOUBLE iml_abs(IML_LDOUBLE Val)						{ return(fabsl(Val)); };
inline IML_LDOUBLE iml_mod(IML_LDOUBLE X, IML_LDOUBLE Y)		{ return(fmodl(X,Y)); };


#endif // __HAVE_LDOUBLE__

//=====================================================================
//  AN ABSTRACT STIMULUS PARAMETER THAT CONTROLS SOME ASPECT OF THE STIMULUS.
//=====================================================================
class adjustable_parameter {
		static IML_BYTE Buf[100];  // STATIC SO WE GET ONLY ONE FOR THE WHOLE CLASS
		IML_UBYTE	Precision;
		IML_DOUBLE	Value;

	public:                           

		adjustable_parameter(IML_DOUBLE Val=0.0, IML_UBYTE Prec = 10) {
			// MAKE SURE THAT PRECISION IS AT LEAST 1 AND LESS THAN OR EQUAL TO 40 DIGITS
			Precision = (Prec < 1 || Prec > 40) ? 10 : Prec;  // SET THIS FIRST
			// EXPLICIT USE OF THE DEFINED ASSIGNMENT OPERATOR
			operator=(Val);
			};

		IML_DOUBLE	get_dBV(void) { return(20.0*iml_log10(Value)); };

		void set_dBV(IML_DOUBLE PVal) 
			{ operator=((IML_DOUBLE) (iml_pow(10.0,PVal/20.0))); };

		void inc_dBV(IML_DOUBLE PVal) 
			{ operator=((IML_DOUBLE) (Value*iml_pow(10.0,PVal/20.0))); };

		IML_DOUBLE	get_dBW(void) { return(10.0*iml_log10(Value)); };

		void set_dBW(IML_DOUBLE PVal) 
			{ operator=((IML_DOUBLE) (iml_pow(10.0,PVal/10.0))); };

		void inc_dBW(IML_DOUBLE PVal) 
			{ operator=((IML_DOUBLE) (Value*iml_pow(10.0,PVal/10.0))); };

		// A SINGLE CONVERSION OPERATOR SO THAT WE CAN USE THE VARIABLE LIKE A DOUBLE.
		operator IML_DOUBLE() { return(Value); };

		IML_DOUBLE & operator=(IML_DOUBLE Val) {
			sprintf((char*)Buf,"%.*lf",(IML_SINT)Precision,(IML_DOUBLE)Val);
			sscanf((char*)Buf,"%lf",&Value);
			return(Value);
			};


}; // end class


#endif // __IML_MATH_H__
