/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

// THIS FILE CONTAINS THE BODIES TO THE IMAGE TEMPLATE CLASSES

// THIS PRAGMA OPTION CAUSES TEMPLATE FUNCTION BODIES TO BE GENERATTED IN THIS FILE;
// OTHE FILES SHOULD BE COMPILED "TEMPLATE GENERATION EXTERNAL" IN "C++ Options"

#ifdef __BORLAND_CPP__
#pragma option -Jgd  // THIS PRAGMA IS USED BY BORLAND, NOT CODEWARRIOR
#else
//#define __IML_MWERKS_TEMPLATES__
#endif

#define __IML_NDIO_CPP__
#include "iml_ndim.h"
#include "iml_arch.h"
#include <stdlib.h>

//============================================================
// OUR PSEUDO-CONSTRUCTOR
//============================================================
template <class T, iml_image_type IType>
void iml_typed_image< T, IType>::iml_TI_init() {

//	PixelSize = sizeof(T);

	// IF WE ARE AN ALIAS, DO NOTHING
	if(AliasImage) return;

#ifdef __I_DO_WINDOWS__
	IML_ULINT TotalBytes = sizeof(T)*ImageSpec.ElementCount;

	// ALLOCATE MEMORY VIA WINDOWS' CALL TO GlobalAlloc
	// THE 'GPTR' FLAG SAYS TO ZERO INITIZIALIZE AND FLAG AS FIXED
	MemBlockHandle = GlobalAlloc(GPTR, (DWORD)TotalBytes);


	// IF GLOBAL MEMORY ALLOCATION FAILED THEN QUIT NOW
	if(!MemBlockHandle) {
		pData = NULL;
      iml_error("imlimage","GlobalAlloc allocation failure");
		return;
		}
   
	GlobalFix(MemBlockHandle);

	// ELSE LOCK THE MEMORY AND GET A STATIC POINTER...
	pData = (T IML_PTR) GlobalLock(MemBlockHandle);
#else
	pData = new T[ImageSpec.ElementCount];
#endif

	// IF GLOBAL MEMORY ALLOCATION FAILED THEN QUIT NOW
	if(!pData) {
		pData = NULL;
      iml_error("imlimage","Memory allocation failure");
		return;
		}


} // END iml_image::iml_image


//============================================================
//	WHEN WE DESTRUCT WE FREE UP THE GLOBAL MEMORY HANDLE!!
//============================================================
template <class T, iml_image_type IType>
void iml_typed_image<T,IType>::iml_TI_uninit(void) {

	// IF WE ARE AN ALIAS, DO NOTHING
	if(AliasImage) return;

#ifdef __I_DO_WINDOWS__
	if(MemBlockHandle) {  // IF WE HAVE SUCCESSFULLY ALLOCATED MEMORY
		GlobalUnfix(MemBlockHandle);
		GlobalUnlock(MemBlockHandle);  // UNLOCK THE MEMORY
		GlobalFree(MemBlockHandle);    // FREE THE MEMORY
		} // end if
#else // IF DOS OR OTHER STUFF
	if(pData) delete [] pData;
#endif

} // end iml_TI_uninit

//============================================================
template <class T, iml_image_type IType>
IML_BYTE iml_typed_image<T,IType>::valid_memory(void) {
#ifdef __I_DO_WINDOWS__
			return( (MemBlockHandle && pData) ? TRUE : FALSE);
#else
			return( (pData) ? TRUE : FALSE);
#endif
         } // end valid_memory


//============================================================
template <class T, iml_image_type IType>
T IML_REF iml_typed_image<T,IType>
::operator()(IML_LINT Axis0Index, ...) {
//::operator()(IML_LINT Axis0Index, IML_LINT Axis1Index, ...) {

	// WE WILL ASSUME THAT THE ARGUEMENT LIST LENGTH IS THE SAME
	// AS THE CURRENT SUBSPACE DIMENSIONALITY (AT LEAST 2).  "CAVEAT USER!"
	
	// DECLARE A "VARIABLE-ARGUMENT-LIST ARGUMENT POINTER
	va_list ArgPtr;  
	// SET IT TO THE ADDRESS OF THE LAST OBJECT IN THE KNOWN PART OF THE LIST.
	
// use this one when requiring two dimensions
//	Subspace.set(IML_2D_Y, Axis1Index);
//	va_start(ArgPtr,Axis1Index);
//	for(IML_USINT I=2; I<Subspace.Dimensionality; I++) {

	Subspace.set(IML_2D_X, Axis0Index);
	va_start(ArgPtr,Axis0Index);
	for(IML_USINT I=1; I<Subspace.Dimensionality; I++) {
	
		// REMEMBER THAT ALL chars, shorts AND ints ARE STORED ON THE STACK AS
		// THE LARGEST INTEGER:  4-BYTE INTEGER
		Subspace.set(I, va_arg(ArgPtr, IML_LINT) );
		}

	va_end(ArgPtr);	// THIS REALLY DOES NOTHING, BUT WE KEEP IT FOR STYLE POINTS.
	
	// RETURN A REFERENCE TO THE REQUESTED ELEMENT.  REMEMBER THAT THE Subspace
	// OBJECT REMEMBERS FOR US WHICH SLICE OF THE SUPERSPACE WE ARE CURRENTLY IN. :-)
	return( pData[(IML_LINT)Subspace.indexize()] );
} // end


//============================================================
template <class T, iml_image_type IType>
T IML_REF iml_typed_image<T,IType>
::hp_fa(IML_LINT Axis0Index, ...) {

	// WE WILL ASSUME THAT THE ARGUEMENT LIST LENGTH IS THE SAME
	// AS THE CURRENT SUBSPACE DIMENSIONALITY.  "CAVEAT USER!"
	
	// WE NEED TO ACCOUNT FOR THE HYPERPLANE (ImageCornerOffset)
	// AND FOR THE INDEX RANGE (AxesOrigin)
	long FastIndex = 	Subspace.get_origin() + Subspace.get_offset();
	
	// make this step explicit
	FastIndex += Subspace.Axes[0].Stride*Axis0Index;

	// DECLARE A "VARIABLE-ARGUMENT-LIST ARGUMENT POINTER
	va_list ArgPtr;  
	// SET IT TO THE ADDRESS OF THE LAST OBJECT IN THE KNOWN PART OF THE LIST.
	va_start(ArgPtr,Axis0Index);
	for(IML_USINT I=1; I<Subspace.Dimensionality; I++) {
		// REMEMBER THAT ALL chars, shorts AND ints ARE STORED ON THE STACK AS
		// THE LARGEST INTEGER:  4-BYTE INTEGER
		FastIndex += Subspace.Axes[I].Stride*va_arg(ArgPtr, IML_LINT);
		} // end for

	va_end(ArgPtr);	// THIS REALLY DOES NOTHING, BUT WE KEEP IT FOR STYLE POINTS.

	// RETURN A REFERENCE TO THE REQUESTED ELEMENT.  REMEMBER THAT THE Subspace
	// OBJECT REMEMBERS FOR US WHICH SLICE OF THE SUPERSPACE WE ARE CURRENTLY IN. :-)
	return( pData[FastIndex] );
} // end


//============================================================
//	DO UP THE SPECIAL OPERATORS
//============================================================
template <class T, iml_image_type IType>
void iml_scalar_image<T,IType>::operator=(IML_DOUBLE TVal) {
	iml_index Index(get_domain()); Index.restart();
	IML_ULINT SSCount = Index.unclipped_count();
	for(IML_LINT I=0; I < SSCount; I++,Index++)
		hp_raw_fa((const IML_LINT)Index.indexize()) = TVal;
};
template <class T, iml_image_type IType>
void iml_scalar_image<T,IType>::operator+=(IML_DOUBLE TVal) {
	iml_index Index(get_domain()); Index.restart();
	IML_ULINT SSCount = Index.unclipped_count();
	for(IML_LINT I=0; I < SSCount; I++,Index++)
		hp_raw_fa((const IML_LINT)Index.indexize()) += TVal;
};
template <class T, iml_image_type IType>
void iml_scalar_image<T,IType>::operator-=(IML_DOUBLE TVal) {
	iml_index Index(get_domain()); Index.restart();
	IML_ULINT SSCount = Index.unclipped_count();
	for(IML_LINT I=0; I < SSCount; I++,Index++)
		hp_raw_fa((const IML_LINT)Index.indexize()) -= TVal;
};
template <class T, iml_image_type IType>
void iml_scalar_image<T,IType>::operator*=(IML_DOUBLE TVal) {
	iml_index Index(get_domain()); Index.restart();
	IML_ULINT SSCount = Index.unclipped_count();
	for(IML_LINT I=0; I < SSCount; I++,Index++)
		hp_raw_fa((const IML_LINT)Index.indexize()) *= TVal;
};


//======================================================================
// THESE ARE FOR WHOLE IMAGES OF THE SAME TYPE
//======================================================================
template <class T, iml_image_type IType> void iml_typed_image<T,IType>::
	operator=(iml_typed_image<T,IType> IML_REF LikeImage) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	iml_index DstBlock(get_domain());
	DstBlock.clip_against(LikeImage.get_domain());
	DstBlock.restart();
	IML_ULINT UnClippedCount = DstBlock.unclipped_count();
	iml_index SrcBlock(LikeImage.get_domain()); 
	SrcBlock.clip_against(get_domain());
	SrcBlock.restart(); 
	for( IML_LINT I=0; I < UnClippedCount; I++,SrcBlock++,DstBlock++ )
		hp_raw_fa((const IML_LINT)DstBlock.indexize()) =
			 LikeImage.hp_raw_fa((const IML_LINT)SrcBlock.indexize());

};

//======================================================================
// THESE ARE FOR WHOLE IMAGES OF THE SAME TYPE
//======================================================================
template <class T, iml_image_type IType>
void iml_typed_image<T,IType>::operator+=(iml_typed_image<T,IType> IML_REF LikeImage) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	iml_index DstBlock(get_domain());
	DstBlock.clip_against(LikeImage.get_domain());
	DstBlock.restart();
	IML_ULINT UnClippedCount = DstBlock.unclipped_count();
	iml_index SrcBlock(LikeImage.get_domain()); 
	SrcBlock.clip_against(get_domain());
	SrcBlock.restart(); 
	for( IML_LINT I=0; I < UnClippedCount; I++,SrcBlock++,DstBlock++ )
		hp_raw_fa((const IML_LINT)DstBlock.indexize()) += 
			 LikeImage.hp_raw_fa((const IML_LINT)SrcBlock.indexize());

};

//======================================================================
// THESE ARE FOR WHOLE IMAGES OF THE SAME TYPE
//======================================================================
template <class T, iml_image_type IType>
void iml_typed_image<T,IType>::operator-=(iml_typed_image<T,IType> IML_REF LikeImage) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	iml_index DstBlock(get_domain());
	DstBlock.clip_against(LikeImage.get_domain());
	DstBlock.restart();
	IML_ULINT UnClippedCount = DstBlock.unclipped_count();
	iml_index SrcBlock(LikeImage.get_domain()); 
	SrcBlock.clip_against(get_domain());
	SrcBlock.restart(); 
	for( IML_LINT I=0; I < UnClippedCount; I++,SrcBlock++,DstBlock++ )
		hp_raw_fa((const IML_LINT)DstBlock.indexize()) -= 
			 LikeImage.hp_raw_fa((const IML_LINT)SrcBlock.indexize());

};

//======================================================================
// THESE ARE FOR WHOLE IMAGES OF THE SAME TYPE
//======================================================================
template <class T, iml_image_type IType>
void iml_typed_image<T,IType>::operator*=(iml_typed_image<T,IType> IML_REF LikeImage) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	iml_index DstBlock(get_domain());
	DstBlock.clip_against(LikeImage.get_domain());
	DstBlock.restart();
	IML_ULINT UnClippedCount = DstBlock.unclipped_count();
	iml_index SrcBlock(LikeImage.get_domain()); 
	SrcBlock.clip_against(get_domain());
	SrcBlock.restart(); 
	for( IML_LINT I=0; I < UnClippedCount; I++,SrcBlock++,DstBlock++ )
		hp_raw_fa((const IML_LINT)DstBlock.indexize()) *= 
			 LikeImage.hp_raw_fa((const IML_LINT)SrcBlock.indexize());

};



//===============================================================================
//	FIND THE MINIMUM AND MAXIMUM MAXIMUM ELEMENT SIZES
//===============================================================================
template <class T, iml_image_type IType>
void iml_scalar_image<T,IType>::mean_and_sdev(IML_DOUBLE &Mean, IML_DOUBLE &SDev) {

	IML_ULINT I;
	IML_DOUBLE SampleMean = 0;
	IML_DOUBLE SampleSDev = 0;
	T IML_PTR Data = pData+Subspace.get_offset();
	for(I = 0; I < Subspace.ElementCount; I++)	
		SampleMean += Data[I];
	SampleMean /= Subspace.ElementCount;
	for(I = 0; I < Subspace.ElementCount; I++)	
		SampleSDev += (Data[I] - SampleMean)*(Data[I] - SampleMean);
	SampleSDev /= (Subspace.ElementCount - 1);
	SampleSDev = iml_sqrt(SampleSDev);
	Mean = SampleMean;
	SDev = SampleSDev;

};

//===============================================================================
//	FIND THE MINIMUM AND MAXIMUM MAXIMUM ELEMENT SIZES
//===============================================================================
template <class T, iml_image_type IType>
void iml_scalar_image<T,IType>::find_min_max_elements(void) {
IML_DOUBLE TMax= -DBL_MAX,TMin=DBL_MAX;

	T IML_PTR Data = pData+Subspace.get_offset();
	for(IML_LINT I = 0; I < Subspace.ElementCount; I++) {
		double Temp = Data[I];
		if(TMax < Temp) TMax = Temp;
		else if(TMin > Temp) TMin = Temp;
		} // end for
	MaxMag = TMax;
	MinMag = TMin;
};



//===============================================================================
//	SCALE THE RANGE OF ELMEMENTS IN THE IMAGE
//===============================================================================
template <class T, iml_image_type IType>
void iml_scalar_image<T,IType>::scale_elements(IML_DOUBLE Min, IML_DOUBLE Max) {

	find_min_max_elements();

	T IML_PTR Data = pData+Subspace.get_offset();
	IML_DOUBLE Scaler = (Max-Min)/(MaxMag-MinMag);
	IML_DOUBLE Const = Min - MinMag*Scaler;
	for(IML_LINT I = 0; I < Subspace.ElementCount; I++) {
		double NewVal = Data[I] * Scaler + Const;
		Data[I] = NewVal;
		} // end for


};



//============================================================
//	DO UP THE SPECIAL OPERATORS
//============================================================

template <class V, iml_image_type VType>
void iml_vector_image<V,VType>::operator=(V IML_REF Vect) {
	V IML_PTR Data = pData+Subspace.get_offset();
	for(IML_LINT I=0; I < Subspace.ElementCount; I++)  Data[I] = Vect;
};
template <class V, iml_image_type VType>
void iml_vector_image<V,VType>::operator+=(V IML_REF Vect) {
	V IML_PTR Data = pData+Subspace.get_offset();
	for(IML_LINT I=0; I < Subspace.ElementCount; I++)  Data[I] += Vect;
};
template <class V, iml_image_type VType>
void iml_vector_image<V,VType>::operator-=(V IML_REF Vect) {
	V IML_PTR Data = pData+Subspace.get_offset();
	for(IML_LINT I=0; I < Subspace.ElementCount; I++)  Data[I] -= Vect;
};
template <class V, iml_image_type VType>
void iml_vector_image<V,VType>::operator*=(V IML_REF Vect) {
	V IML_PTR Data = pData+Subspace.get_offset();
	for(IML_LINT I=0; I < Subspace.ElementCount; I++)  Data[I] *= Vect;
};




//===============================================================================
//	FIND THE MINIMUM AND MAXIMUM ELEMENT SIZES
//===============================================================================
template <class CC, class CT, iml_image_type IType>
void iml_complex_image<CC, CT, IType>::find_min_max_elements(void) {
IML_DOUBLE TMax= -DBL_MAX, TMin=DBL_MAX;

	// SAVE THE ELEMENT TYPE IN CASE SUBSEQUENT CALLS DEFAULT THE VALUE
	iml_vector_part SavedPart = VectorPart;

	CC IML_PTR Data = pData+Subspace.get_offset();
	IML_LINT I;
	switch(SavedPart) {
		default: 
		case IML_COMPLEX_REAL:
			for( I = 0; I < Subspace.ElementCount; I++) {
				double Temp = Data[I].Real;
				if(TMax < Temp) TMax = Temp;
				else if(TMin > Temp) TMin = Temp;
				} // end for
			break;
		case IML_COMPLEX_IMAG:
			for( I = 0; I < Subspace.ElementCount; I++) {
				double Temp = Data[I].Imag;
				if(TMax < Temp) TMax = Temp;
				else if(TMin > Temp) TMin = Temp;
				} // end for
			break;

		} // end switch


	MaxMag = TMax;
	MinMag = TMin;

	VectorPart = IML_DEFAULT_VECTOR_PART;
};

//======================================================================
// USE THE VectorPart VARIABLE TO SELECT IMAGE PARTS TO SCALE
//======================================================================
template <class CC, class CT, iml_image_type IType>
void iml_complex_image<CC, CT, IType>::scale_elements(
					IML_DOUBLE Min, IML_DOUBLE Max) {

	// SAVE THE ELEMENT TYPE IN CASE SUBSEQUENT CALLS DEFAULT THE VALUE
	iml_vector_part SavedPart = VectorPart;

	find_min_max_elements();

	IML_DOUBLE Scaler = (Max-Min)/(MaxMag-MinMag);
	IML_DOUBLE Const = Min - MinMag*Scaler;

	CC IML_PTR Data = pData+Subspace.get_offset();
	IML_LINT I;
	switch(SavedPart) {
		default: 
		case IML_COMPLEX_REAL:
			for( I = 0; I < Subspace.ElementCount; I++)
				Data[I].Real = (CT)(Data[I].Real * Scaler + Const);
			break;
		case IML_COMPLEX_IMAG:
			for( I = 0; I < Subspace.ElementCount; I++)
				Data[I].Imag = (CT)(Data[I].Imag * Scaler + Const);
			break;
      } // end switch

	VectorPart = IML_DEFAULT_VECTOR_PART;

}// end scale_elements



//======================================================================
// USE THE VectorPart VARIABLE TO SELECT IMAGE PARTS
//======================================================================
template <class CC, class CT, iml_image_type IType>
IML_DOUBLE iml_complex_image<CC, CT, IType>::get_value(
					iml_index IML_REF Index) {

IML_DOUBLE Temp;

	switch(VectorPart) {
		default: 
		case IML_COMPLEX_REAL:	Temp = operator()(Index).Real; 			break;
		case IML_COMPLEX_IMAG:	Temp = operator()(Index).Imag; 			break;

		} // end switch

	return(Temp);
}// end get_2d_value

//======================================================================
// USE THE VectorPart VARIABLE TO SELECT IMAGE PARTS
//======================================================================
template <class CC, class CT, iml_image_type IType>
void iml_complex_image<CC, CT, IType>::set_value(
				iml_index IML_REF Index, IML_DOUBLE Val) {

	switch(VectorPart) {
		default: 
		case IML_COMPLEX_REAL:	operator()(Index).Real = (CT)Val; 		break;
		case IML_COMPLEX_IMAG:	operator()(Index).Imag = (CT)Val; 		break;

		} // end switch

} // end set_2d_value


//======================================================================
// USE THE VectorPart VARIABLE TO SELECT IMAGE PARTS
//======================================================================
template <class CC, class CT, iml_image_type IType>
void iml_complex_image<CC, CT, IType>::increment_value(
				iml_index IML_REF Index, IML_DOUBLE Val) {

	switch(VectorPart) {
		default: 
		case IML_COMPLEX_REAL:	operator()(Index).Real += (CT)Val; 			break;
		case IML_COMPLEX_IMAG:	operator()(Index).Imag += (CT)Val; 			break;
 
		} // end switch

} // end set_2d_value


//======================================================================
// USE THE VectorPart VARIABLE TO SELECT IMAGE PARTS
//======================================================================
template <class CC, class CT, iml_image_type IType>
void iml_complex_image<CC, CT, IType>::multiply_value(
				iml_index IML_REF Index, IML_DOUBLE Val) {

	switch(VectorPart) {
		default: 
		case IML_COMPLEX_REAL:	operator()(Index).Real *= (CT)Val; 			break;
		case IML_COMPLEX_IMAG:	operator()(Index).Imag *= (CT)Val; 			break;

		} // end switch

} // end set_2d_value


//============================================================
//
//============================================================
template <class CC, class CT, iml_image_type IType>
CC IML_REF iml_complex_image<CC, CT, IType>
::frequency(IML_LINT Axis0Index, ...) {

	if( !FourierDomain ) {
       iml_warning("imlimage","Non-Fourier complex image accessed by frequency().");
		 return(pData[0]);
       } // end if

	// WE WILL ASSUME THAT THE ARGUMENT LIST LENGTH IS THE SAME
	// AS THE CURRENT SUBSPACE DIMENSIONALITY (AT LEAST 2).  "CAVEAT USER!"

	Subspace.set(IML_2D_X, translate_from_freq(IML_2D_X, Axis0Index) );
	// DECLARE A "VARIABLE-ARGUMENT-LIST ARGUMENT POINTER
	va_list ArgPtr;  
	// SET IT TO THE ADDRESS OF THE LAST OBJECT IN THE KNOWN PART OF THE LIST.
	va_start(ArgPtr,Axis0Index);
	for(IML_USINT I=1; I<Subspace.Dimensionality; I++) {
		// REMEMBER THAT ALL chars, shorts AND ints ARE STORED ON THE STACK AS
		// THE LARGEST INTEGER:  4-BYTE INTEGER
		Subspace.set(I, translate_from_freq(I, va_arg(ArgPtr, IML_LINT) ) );
		}
		
	va_end(ArgPtr);	// THIS REALLY DOES NOTHING, BUT WE KEEP IT FOR STYLE POINTS.
	
	// RETURN A REFERENCE TO THE REQUESTED ELEMENT.  REMEMBER THAT THE Subspace
	// OBJECT REMEMBERS FOR US WHICH SLICE OF THE SUPERSPACE WE ARE CURRENTLY IN. :-)
	return( pData[Subspace.indexize()] );
} // end


//============================================================
//
//============================================================
template <class CC, class CT, iml_image_type IType>
void iml_complex_image<CC, CT, IType>::initialize_modulation_info(void) {
	switch(ByteOrder) {
	
		default:
		case IML_BIG_ENDIAN_BYTE_ORDER: { // = 0x01020408, 
			SignByte = 0;
			MLSByte = sizeof(CT)-1;
			SignMask1 = 0x7F; SignMask2 = ~SignMask1;
			} break;
			
		case IML_LITTLE_ENDIAN_BYTE_ORDER: { // = 0x08040201,
			SignByte = sizeof(CT)-1; 
			MLSByte = 0;
			SignMask1 = 0x7F; SignMask2 = ~SignMask1;
			} break;
		} // end switch
	} // end initialize_modulation_info


//============================================================
//
//============================================================
template <class CC, class CT, iml_image_type IType>
		void iml_complex_image<CC, CT, IType>::goto_polar(void) {

	iml_index Indexer(Subspace);
	Indexer.restart(); 
	for( IML_LINT I=0; I < Indexer.ElementCount; I++, Indexer++ ) {
		long Q = (const IML_LINT)Indexer.indexize();
		CC & Pixel( operator[]( Q ) );
		CC T(Pixel);
		Pixel.Real = iml_sqrt(T.Real*T.Real+T.Imag*T.Imag);
		Pixel.Imag = iml_atan2(T.Imag, T.Real );
		} // end for

} // end goto_polar

//============================================================
//
//============================================================
template <class CC, class CT, iml_image_type IType>
		void iml_complex_image<CC, CT, IType>::goto_cartesian(void) {

	iml_index Indexer(Subspace);
	Indexer.restart(); 
	for( IML_LINT I=0; I < Indexer.ElementCount; I++, Indexer++ ) {
		long Q = (const IML_LINT)Indexer.indexize();
		CC & Pixel( operator[]( Q ) );
		CC T(Pixel);
		Pixel.Real = T.Real*iml_cos(T.Imag);
		Pixel.Imag = T.Real*iml_sin(T.Imag);
		} // end for

} // end goto_polar

//============================================================
//
//============================================================
template <class CC, class CT, iml_image_type IType>
		void iml_complex_image<CC, CT, IType>::modulate(void) {

//The 32-bit (single precision) IEEE: "s eeeeeeee fffffff ffffffffffffffff"
//The 64-bit (double precision) IEEE: "s eeeeeeeeeee ffff ffffffff....ffffffff"
//typedef union { CT RealValue; IML_UBYTE SignBit : 1;
//	void sign_change(void) { SignBit ^= 0x1; };      } sign_bit_grabber; // bit grabber 

	iml_index Indexer(Subspace);
	Indexer.restart(); 
	for( IML_LINT I=0; I < Indexer.ElementCount; I++, Indexer++ ) {
		IML_UBYTE Odd = mlsb(Indexer.Axes[0].AxisV);
		for(IML_USINT J=1; J<Indexer.Dimensionality; J++) 
			Odd += mlsb(Indexer.Axes[J].AxisV);
		if( Odd & 0x01 ) 
			sign_change((bg IML_REF) operator[]((const IML_LINT)Indexer.indexize()));
		} // end for

} // end modulate

//============================================================
// REVERSE FOURIER TRANSFORM
// RETURNS NON-ZER0 ON SUCCESS
//============================================================
template <class CC, class CT, iml_image_type IType>
IML_BYTE iml_complex_image<CC, CT, IType>::inverse_FFT(IML_SINT RemapFlag) {

	if( !FourierDomain ) {
       iml_warning("imlimage","Non-Fourier complex image in inverse_fourier");
		 return(0);
       } // end if

   // SET UP PARAMETERS FOR THE NUMERICAL_RECIPES FOURIER TRANSFORM FUNCTION
	int NDim=Subspace.Dimensionality;
	int *NN = new int[NDim];
	for(int I=0; I< NDim; I++)  NN[I] = Subspace.size(I);

	// FOR WINDOWS, USE A "huge float *" BECAUSE LARGE IMAGES MUST CROSS 64Kb SEGMENT BOUNDARIES
	CT IML_PTR pReal = (CT IML_PTR)(pData+Subspace.get_offset());
	// THIS FOURIER ROUTINE COMES FROM THE Numerical Recipes
	// STABLE.  IT IS A THOROUGHBRED.  IT WANTS A FLOAT POINTER
	// THAT IS (FORTUNATELY) IN THE FORM OF <REAL,IMAG>
	// PAIRS OF DATA, BUT IT WANTS THEM IN A LINEAR ARRAY
	// OF FLOATS, AND ARRANGED IN ROW-MAJOR ORDER (C-STYLE..)

	// THE -2 IS NEEDED FOR INTERLEAVED DATA
	rcs_fftn( NDim, NN, pReal, pReal+1, -2 );
	delete [] NN;
	
	// SWITCH IMAGE TYPE
	FourierDomain = FALSE;
	if(RemapFlag) {
		// RE-MODULATE BACK TO ORIGINAL
		modulate();
		// SET THE INDICES TO NO LONGER FOLLOW FREQUENCY
		set_indices();
		} // end if

	IML_FLOAT ScaleFactor = 1.0/Subspace.ElementCount;
	switch(ScalingAction) {
		case N_IN: break; // NO ACTION HERE
		case ROOT_N:  ScaleFactor = iml_sqrt(ScaleFactor); // AND FALL THROUGH TO NEXT STEP
		case N_OUT:
			for(IML_LINT Q=0; Q<2*Subspace.ElementCount; Q++) {
				pReal[Q] *= ScaleFactor;
				}
			break;
		} // end switch

	return(1);
} // end iml_image::inverse_FFT


//============================================================
// FORWARD FOURIER TRANSFORM
// RETURNS NON-ZER0 ON SUCCESS
//============================================================
template <class CC, class CT, iml_image_type IType>
IML_BYTE iml_complex_image<CC, CT, IType>::forward_FFT(IML_SINT RemapFlag) {

	if( FourierDomain ) {
       iml_warning("imlimage","Fourier complex image in forward_fourier");
		 return(0);
       } // end if

   // SET UP PARAMETERS TO THE NUMERICAL_RECIPES FOURIER TRANSFORM FUNCTION
	int NDim=Subspace.Dimensionality;
	int *NN = new int[NDim];
	for(int I=0; I< NDim; I++)  NN[I] = Subspace.size(I);

	// SWITCH IMAGE TYPE
	FourierDomain = TRUE;

	if(RemapFlag) {
		// MODULATE WITH HIGHEST FREQUENCE FOR THE FOURIER-DOMAIN SHIFT
		modulate();
		// SET THE INDICES TO FOLLOW FREQUENCY
		set_indices();
		} // end if

	// FOR WINDOWS, USE A "huge float *" BECAUSE LARGE IMAGES MUST CROSS 64Kb SEGMENT BOUNDARIES
	CT IML_PTR pReal = (CT IML_PTR)(pData+Subspace.get_offset());
	// THIS FOURIER ROUTINE COMES FROM THE Numerical Recipes
	// STABLE.  IT IS A THOROUGHBRED.  IT WANTS A FLOAT POINTER
	// THAT IS (FORTUNATELY) IN THE FORM OF <REAL,IMAG>
	// PAIRS OF DATA, BUT IT WANTS THEM IN A LINEAR ARRAY
	// OF FLOATS, AND ARRANGED IN ROW-MAJOR ORDER (C-STYLE..)

	// THE 2 IS NEEDED FOR INTERLEAVED DATA (i.e. REAL/IMAG IN SAME ARRAY )
	rcs_fftn( NDim, NN, pReal, pReal+1, 2 );
	delete [] NN;

	IML_FLOAT ScaleFactor = 1.0/Subspace.ElementCount;
	switch(ScalingAction) {
		case N_OUT: break; // NO ACTION HERE
		case ROOT_N:  ScaleFactor = iml_sqrt(ScaleFactor); // AND FALL THROUGH TO NEXT STEP
		case N_IN:
			for(IML_LINT Q=0; Q<2*Subspace.ElementCount; Q++) {
				pReal[Q] *= ScaleFactor;
				}
			break;
		} // end switch

	return(1);
} // end iml_image::forward_FFT





//====================================================================
//====================================================================
template <class CC, class CT, iml_image_type IType>
void iml_complex_image<CC, CT, IType>::set_indices(void) {

	if(FourierDomain) {
			// WE INDEX THE FOURIER IMAGE BY FREQUENCY
			//	SO WE GO FROM -SIZE/2 TO +SIZE/2 IN EACH DIMENSION.
			// WE NEED TO REMEMBER THAT IF WE ARE SHOWING THE IMAGE,
			// WE SHOULD SHOW THAT EXTRA ROW/COL BUT THE MAX/MIN FREQUENCIES
			// ARE EQUAL, AND ARE REPRESENTED BY ONE VALUE IN THE IMAGE.
			// THAT MEANS WE SHOULD NOT ADD, SUBTRACT OR MULTIPLY IT MORE
			// THAN ONCE; HENCE THE +1 AT THE LOWER INDEX.

			// DO ASSIGNMENTS THIS WAY TO ENSURE WE GET SIGNED
			// VALUES; XWidth AND YHeight ARE UNSIGNED!!
			//	I.E. SO WE DON'T GET BITTEN BY THE TYPE-CONVERSION
			// BECAUSE THE NEGATIVE OF AN UNSIGNED IS JUST A BIT
			// FLIP, NOT A SIGNED NUMBER.
			for(int I=0;I<Subspace.Dimensionality; I++) {
				IML_LINT MaxFreq = (IML_LINT)Subspace.size(I)/2;
				IML_LINT Max =  (MaxFreq) ? MaxFreq - 1 : MaxFreq;
				IML_LINT Min = -MaxFreq;  
				// SO THAT WE CAN DO FILTERING, ETC. EASILY
				Subspace.set_min_max(I, 	Min, Max);
				} // end for

	} else {
			// ALL OTHER TYPES ARE REGULAR INDEXING:
			//   ZERO TO SIZE-1
			for(int I=0;I<Subspace.Dimensionality; I++) {
				IML_LINT Max = (IML_LINT)Subspace.size(I)-1;
				IML_LINT Min = 0;  
				// SO THAT WE CAN DO FILTERING, ETC. EASILY
				Subspace.set_min_max(I, 	Min, Max);
				} // end for

		} // end else

} // end image_indices


/*--------------------------------*-C-*---------------------------------*
 * File:
 *	fftn.c
 *
 *	multivariate complex Fourier transform, computed in place
 *	using mixed-radix Fast Fourier Transform algorithm.
 *
 *	Fortran code by:
 *	    RC Singleton, Stanford Research Institute, Sept. 1968
 *		NIST Guide to Available Math Software.
 *		Source for module FFT from package GO.
 *		Retrieved from NETLIB on Wed Jul  5 11:50:07 1995.
 *	translated by f2c (version 19950721) and with lots of cleanup
 *	to make it resemble C by:
 *	    MJ Olesen, Queen's University at Kingston, 1995-97
 */
/*{{{ Copyright: */
/*
 * Copyright(c)1995,97 Mark Olesen <olesen@me.QueensU.CA>
 *		Queen's Univ at Kingston (Canada)
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose without fee is hereby granted, provided that this
 * entire notice is included in all copies of any software which is
 * or includes a copy or modification of this software and in all
 * copies of the supporting documentation for such software.
 *
 * THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR QUEEN'S
 * UNIVERSITY AT KINGSTON MAKES ANY REPRESENTATION OR WARRANTY OF ANY
 * KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
 * FITNESS FOR ANY PARTICULAR PURPOSE.
 *
 * All of which is to say that you can do what you like with this
 * source code provided you don't try to sell it as your own and you
 * include an unaltered copy of this message (including the
 * copyright).
 *
 * It is also implicitly understood that bug fixes and improvements
 * should make their way back to the general Internet community so
 * that everyone benefits.
 *----------------------------------------------------------------------*/
/*}}}*/
/*{{{ notes: */
/*
 * Public:
 *	rcs_fft_free
 *	fftn
 *	(these are documented in the header file)
 *
 * Private:
 *	rcs_fftradix
 *
 * ----------------------------------------------------------------------*
 * int rcs_fftradix (REAL Re[], REAL Im[], size_t nTotal, size_t nPass,
 *		 size_t nSpan, int iSign, size_t maxFactors,
 *		 size_t maxPerm);
 *
 * RE and IM hold the real and imaginary components of the data, and
 * return the resulting real and imaginary Fourier coefficients.
 * Multidimensional data *must* be allocated contiguously.  There is
 * no limit on the number of dimensions.
 *
 *
 * Although there is no limit on the number of dimensions, rcs_factorize()
 * must be called once for each dimension, but the calls may be in
 * any order.
 *
 * NTOTAL = the total number of complex data values
 * NPASS  = the dimension of the current variable
 * NSPAN/NPASS = the spacing of consecutive data values while indexing
 *	the current variable
 * ISIGN - see above documentation
 *
 * example:
 * tri-variate transform with Re[n1][n2][n3], Im[n1][n2][n3]
 *
 *	rcs_fftradix (Re, Im, n1*n2*n3, n1,       n1, 1, maxf, maxp);
 *	rcs_fftradix (Re, Im, n1*n2*n3, n2,    n1*n2, 1, maxf, maxp);
 *	rcs_fftradix (Re, Im, n1*n2*n3, n3, n1*n2*n3, 1, maxf, maxp);
 *
 * single-variate transform,
 *    NTOTAL = N = NSPAN = (number of complex data values),
 *
 *	rcs_fftradix (Re, Im, n, n, n, 1, maxf, maxp);
 *
 * The data can also be stored in a single array with alternating
 * real and imaginary parts, the magnitude of ISIGN is changed to 2
 * to give correct indexing increment, and data [0] and data [1] used
 * to pass the initial addresses for the sequences of real and
 * imaginary values,
 *
 * example:
 *	REAL data [2*NTOTAL];
 *	rcs_fftradix (&data[0], &data[1], NTOTAL, nPass, nSpan, 2, maxf, maxp);
 *
 * for temporary allocation:
 *
 * MAXFACTORS	>= the maximum prime factor of NPASS
 * MAXPERM	>= the number of prime factors of NPASS.  In addition,
 *	if the square-free portion K of NPASS has two or more prime
 *	factors, then MAXPERM >= (K-1)
 *
 * storage in FACTOR for a maximum of 15 prime factors of NPASS. if
 * NPASS has more than one square-free factor, the product of the
 * square-free factors must be <= 210 array storage for maximum prime
 * factor of 23 the following two constants should agree with the
 * array dimensions.
 * ----------------------------------------------------------------------*/
/*}}}*/
/*{{{ Revisions: */
/*
 * 26 July 95	John Beale
 *	- added maxf and maxp as parameters to rcs_fftradix()
 *
 * 28 July 95	Mark Olesen <olesen@me.QueensU.CA>
 *	- cleaned-up the Fortran 66 goto spaghetti, only 3 labels remain.
 *
 *	- added rcs_fft_free() to provide some measure of control over
 *	  allocation/deallocation.
 *
 *	- added fftn() wrapper for multidimensional FFTs
 *
 *
 *	- revised documentation
 *
 * 8 Jan 95	mj olesen <olesen@me.QueensU.CA>
 *	- fixed typo's, including one that broke scaling for scaling by
 *	  total number of matrix elements or the square root of same
 *	- removed unnecessary casts from allocations
 *
 * 10 Dec 96	mj olesen <olesen@me.QueensU.CA>
 *	- shifted some variables to local scope	(better hints for optimizer)
 *	- made it simpler to pass dimensions for 1D FFT.
 *
 * 23 Feb 97	Mark Olesen <olesen@me.QueensU.CA>
 *	- removed the GNU Public License (see 21 July 1995 entry),
 *	  which should make it clear why I have the right to do so.
 *	- Added copyright notice and submitted to netlib
 *	- Moved documentation for the public functions to the header
 *	  files where is will always be available.
 * ----------------------------------------------------------------------*/
/*}}}*/

//#include <stdio.h>
//#include <stdlib.h>
//#include <math.h>
//#include "fftn.h"

/*{{{ defines/constants */
# define RCS_M_PI	3.14159265358979323846264338327950288
# define SIN60	0.86602540378443865	/* sin(60 deg) */
# define COS72	0.30901699437494742	/* cos(72 deg) */
# define SIN72	0.95105651629515357	/* sin(72 deg) */
/*}}}*/

/*{{{ static parameters - for memory management */

/* temp space, (void *) since both float and double routines use it */
//void * Tmp0 = NULL;	/* temp space for real part */
//void * Tmp1 = NULL;	/* temp space for imaginary part */
//void * Tmp2 = NULL;	/* temp space for Cosine values */
//void * Tmp3 = NULL;	/* temp space for Sine values */
//int  * Perm = NULL;	/* Permutation vector */

#define NFACTOR	11
static int factor [NFACTOR];
/*}}}*/

/*{{{ rcs_fft_free() */
template <class CC, class CT, iml_image_type IType>
void iml_complex_image<CC, CT, IType>::rcs_fft_free (void)
{
   if (Rtmp) { delete [] Rtmp;	Rtmp = NULL; }
   if (Itmp) { delete [] Itmp;	Itmp = NULL; }
   if (Cos)  { delete [] Cos;		Cos = NULL;  }
   if (Sin)  { delete [] Sin;		Sin = NULL;  }
   if (Perm) { delete [] Perm;	Perm = NULL; }
}
/*}}}*/

/* return the number of factors */
template <class CC, class CT, iml_image_type IType>
int iml_complex_image<CC, CT, IType>::rcs_factorize (int nPass, int * kt)
{
   int nFactor = 0;
   int j, jj;

   *kt = 0;
   /* determine the factors of n */
   while ((nPass % 16) == 0)	/* factors of 4 */
     {
	factor [nFactor++] = 4;
	nPass /= 16;
     }
   j = 3; jj = 9;		/* factors of 3, 5, 7, ... */
   do {
      while ((nPass % jj) == 0)
	{
	   factor [nFactor++] = j;
	   nPass /= jj;
	}
      j += 2;
      jj = j * j;
   } while (jj <= nPass);
   if (nPass <= 4)
     {
	*kt = nFactor;
	factor [nFactor] = nPass;
	if (nPass != 1)
	  nFactor++;
     }
   else
     {
	if (nPass - (nPass / 4 << 2) == 0)
	  {
	     factor [nFactor++] = 2;
	     nPass /= 4;
	  }
	*kt = nFactor;
	j = 2;
	do {
	   if ((nPass % j) == 0)
	     {
		factor [nFactor++] = j;
		nPass /= j;
	     }
	   j = ((j + 1) / 2 << 1) + 1;
	} while (j <= nPass);
     }
   if (*kt)
     {
	j = *kt;
	do
	  factor [nFactor++] = factor [--j];
	while (j);
     }

   return nFactor;
}

# define Re_Data(i)	Re[i]
# define Im_Data(i)	Im[i]


/*
 *
 */
template <class CC, class CT, iml_image_type IType>
int iml_complex_image<CC, CT, IType>::rcs_fftn(
		int ndim, const int *dims, CT *Re, CT *Im, int iSign) {

	if (dims == NULL) {
   	fprintf (stderr, "Error: rcs_fftn() - dimension error\n");
   	return(0);
		} // end if
		
   // Count the number of elements in the data array
   // and determine the number of dimensions
	size_t nTotal = 1;
	long maxFactors = 1, maxPerm = 1;
	for(long i = 0; i < ndim; i++)  {
		nTotal *= dims [i];
		// follow John Beale's example, just use the largest dimension and don't
		// worry about excess allocation.  May be someone else will do it?
		// the constants used in the original Fortran code were:
		// maxFactors = 23;
		// maxPerm = 209;
		if (dims [i] > maxFactors) maxFactors = dims [i];
		if (dims [i] > maxPerm) maxPerm = dims [i];
		} // end for
		
   // loop over the dimensions and do an fft for each one 
	size_t nSpan = 1;
	for (int i = 0; i < ndim; i++) {
		nSpan *= dims [i];
		int ret = rcs_fftradix (Re, Im, nTotal, dims [i], nSpan, iSign, maxFactors, maxPerm);
	   // exit, clean-up already done
	   if (ret) return ret;
      } // end for

// REF: add this to free up memory in the case of no error
   rcs_fft_free ();	// free-up memory
   return 0;

} // rcs_fftn

/*----------------------------------------------------------------------*/
/*
 * singleton's mixed radix routine
 *
 * could move allocation out to fftn(), but leave it here so that it's
 * possible to make this a standalone function
 */
template <class CC, class CT, iml_image_type IType>
int iml_complex_image<CC, CT, IType>::rcs_fftradix(
		CT *Re,
		CT *Im,
		size_t nTotal,
		size_t nPass,
		size_t nSpan,
		int iSign,
		int maxFactors,
		int maxPerm)
{
   int ii, nFactor, kspan, ispan, inc;
   int j, jc, jf, jj, k, k1, k3, kk, kt, nn, ns, nt;

   CT radf;
   CT c1, c2, c3, cd;
   CT s1, s2, s3, sd;

   CT s60 = SIN60;		/* sin(60 deg) */
   CT s72 = SIN72;		/* sin(72 deg) */
   CT c72 = COS72;		/* cos(72 deg) */
   CT pi2 = RCS_M_PI;		/* use PI first, 2 PI later */

   /* gcc complains about k3 being uninitialized, but I can't find out where
    * or why ... it looks okay to me.
    *
    * initialize to make gcc happy
    */
   k3 = 0;

   /* gcc complains about c2, c3, s2,s3 being uninitialized, but they're
    * only used for the radix 4 case and only AFTER the (s1 == 0.0) pass
    * through the loop at which point they will have been calculated.
    *
    * initialize to make gcc happy
    */
   c2 = c3 = s2 = s3 = 0.0;

   /* Parameter adjustments, was fortran so fix zero-offset */
   Re--;
   Im--;

   if (nPass < 2)
     return 0;

   /* allocate storage */
   rcs_fft_free();
	Rtmp = new CT[maxFactors];
	Itmp = new CT[maxFactors];
	Cos  = new CT[maxFactors];
	Sin  = new CT[maxFactors];
	Perm = new int[maxPerm];

   if (!Rtmp || !Itmp || !Cos || !Sin || !Perm) goto Memory_Error;

   /*
    * Function Body
    */
   inc = iSign;
   if (iSign < 0)
     {
//#ifndef FFT_RADIX4
	s60 = -s60;
	s72 = -s72;
//#endif
	pi2 = -pi2;
	inc = -inc;		/* absolute value */
     }

   /* adjust for strange increments */
   nt = inc * nTotal;
   ns = inc * nSpan;
   kspan = ns;

   nn = nt - inc;
   jc = ns / nPass;
   radf = pi2 * (double) jc;
   pi2 *= 2.0;			/* use 2 PI from here on */

   ii = 0;
   jf = 0;
   /* determine the factors of n */

   nFactor = rcs_factorize (nPass, &kt);
   /* test that nFactors is in range */
   if (nFactor > NFACTOR)
     {
	fprintf (stderr, "Error: rcs_fftradix() - exceeded number of factors\n");
	goto Memory_Error;
     }

   /* compute fourier transform */
   for (;;) {
      sd = radf / (double) kspan;
      cd = iml_sin (sd);
      cd = 2.0 * cd * cd;
      sd = iml_sin (sd + sd);
      kk = 1;
      ii++;

      switch (factor [ii - 1]) {
       case 2:
	 /* transform for factor of 2 (including rotation factor) */
	 kspan /= 2;
	 k1 = kspan + 2;
	 do {
	    do {
	       CT tmpr;
	       CT tmpi;
	       int k2;

	       k2 = kk + kspan;
	       tmpr = Re_Data (k2);
	       tmpi = Im_Data (k2);
	       Re_Data (k2) = Re_Data (kk) - tmpr;
	       Im_Data (k2) = Im_Data (kk) - tmpi;
	       Re_Data (kk) += tmpr;
	       Im_Data (kk) += tmpi;
	       kk = k2 + kspan;
	    } while (kk <= nn);
	    kk -= nn;
	 } while (kk <= jc);
	 if (kk > kspan)
	   goto Permute_Results;	/* exit infinite loop */
	 do {
	    int k2;

	    c1 = 1.0 - cd;
	    s1 = sd;
	    do {
	       CT tmp;
	       do {
		  do {
		     CT tmpr;
		     CT tmpi;

		     k2 = kk + kspan;
		     tmpr = Re_Data (kk) - Re_Data (k2);
		     tmpi = Im_Data (kk) - Im_Data (k2);
		     Re_Data (kk) += Re_Data (k2);
		     Im_Data (kk) += Im_Data (k2);
		     Re_Data (k2) = c1 * tmpr - s1 * tmpi;
		     Im_Data (k2) = s1 * tmpr + c1 * tmpi;
		     kk = k2 + kspan;
		  } while (kk < nt);
		  k2 = kk - nt;
		  c1 = -c1;
		  kk = k1 - k2;
	       } while (kk > k2);
	       tmp = c1 - (cd * c1 + sd * s1);
	       s1 = sd * c1 - cd * s1 + s1;
	       c1 = 2.0 - (tmp * tmp + s1 * s1);
	       s1 *= c1;
	       c1 *= tmp;
	       kk += jc;
	    } while (kk < k2);
	    k1 += (inc + inc);
	    kk = (k1 - kspan) / 2 + jc;
	 } while (kk <= jc + jc);
	 break;

       case 4:			/* transform for factor of 4 */
	 ispan = kspan;
	 kspan /= 4;

	 do {
	    c1 = 1.0;
	    s1 = 0.0;
	    do {
	       do {
		  CT ajm, ajp, akm, akp;
		  CT bjm, bjp, bkm, bkp;
		  int k2;

		  k1 = kk + kspan;
		  k2 = k1 + kspan;
		  k3 = k2 + kspan;
		  akp = Re_Data (kk) + Re_Data (k2);
		  akm = Re_Data (kk) - Re_Data (k2);

		  ajp = Re_Data (k1) + Re_Data (k3);
		  ajm = Re_Data (k1) - Re_Data (k3);

		  bkp = Im_Data (kk) + Im_Data (k2);
		  bkm = Im_Data (kk) - Im_Data (k2);

		  bjp = Im_Data (k1) + Im_Data (k3);
		  bjm = Im_Data (k1) - Im_Data (k3);

		  Re_Data (kk) = akp + ajp;
		  Im_Data (kk) = bkp + bjp;
		  ajp = akp - ajp;
		  bjp = bkp - bjp;
		  if (iSign < 0)
		    {
		       akp = akm + bjm;
		       bkp = bkm - ajm;
		       akm -= bjm;
		       bkm += ajm;
		    }
		  else
		    {
		       akp = akm - bjm;
		       bkp = bkm + ajm;
		       akm += bjm;
		       bkm -= ajm;
		    }
		  /* avoid useless multiplies */
		  if (s1 == 0.0)
		    {
		       Re_Data (k1) = akp;
		       Re_Data (k2) = ajp;
		       Re_Data (k3) = akm;
		       Im_Data (k1) = bkp;
		       Im_Data (k2) = bjp;
		       Im_Data (k3) = bkm;
		    }
		  else
		    {
		       Re_Data (k1) = akp * c1 - bkp * s1;
		       Re_Data (k2) = ajp * c2 - bjp * s2;
		       Re_Data (k3) = akm * c3 - bkm * s3;
		       Im_Data (k1) = akp * s1 + bkp * c1;
		       Im_Data (k2) = ajp * s2 + bjp * c2;
		       Im_Data (k3) = akm * s3 + bkm * c3;
		    }
		  kk = k3 + kspan;
	       } while (kk <= nt);

	       c2 = c1 - (cd * c1 + sd * s1);
	       s1 = sd * c1 - cd * s1 + s1;
	       c1 = 2.0 - (c2 * c2 + s1 * s1);
	       s1 *= c1;
	       c1 *= c2;
	       /* values of c2, c3, s2, s3 that will get used next time */
	       c2 = c1 * c1 - s1 * s1;
	       s2 = 2.0 * c1 * s1;
	       c3 = c2 * c1 - s2 * s1;
	       s3 = c2 * s1 + s2 * c1;
	       kk = kk - nt + jc;
	    } while (kk <= kspan);
	    kk = kk - kspan + inc;
	 } while (kk <= jc);
	 if (kspan == jc)
	   goto Permute_Results;	/* exit infinite loop */
	 break;

       default:
	 /* transform for odd factors */
//#ifdef FFT_RADIX4
//	 fprintf (stderr, "Error: " FFTRADIXS "(): compiled for radix 2/4 only\n");
//	 rcs_fft_free ();		/* free-up memory */
//	 return -1;
//	 break;
//#else	/* FFT_RADIX4 */
	 ispan = kspan;
	 k = factor [ii - 1];
	 kspan /= factor [ii - 1];

	 switch (factor [ii - 1]) {
	  case 3:	/* transform for factor of 3 (optional code) */
	    do {
	       do {
		  CT aj, tmpr;
		  CT bj, tmpi;
		  int k2;

		  k1 = kk + kspan;
		  k2 = k1 + kspan;
		  tmpr = Re_Data (kk);
		  tmpi = Im_Data (kk);
		  aj = Re_Data (k1) + Re_Data (k2);
		  bj = Im_Data (k1) + Im_Data (k2);
		  Re_Data (kk) = tmpr + aj;
		  Im_Data (kk) = tmpi + bj;
		  tmpr -= 0.5 * aj;
		  tmpi -= 0.5 * bj;
		  aj = (Re_Data (k1) - Re_Data (k2)) * s60;
		  bj = (Im_Data (k1) - Im_Data (k2)) * s60;
		  Re_Data (k1) = tmpr - bj;
		  Re_Data (k2) = tmpr + bj;
		  Im_Data (k1) = tmpi + aj;
		  Im_Data (k2) = tmpi - aj;
		  kk = k2 + kspan;
	       } while (kk < nn);
	       kk -= nn;
	    } while (kk <= kspan);
	    break;

	  case 5:	/* transform for factor of 5 (optional code) */
	    c2 = c72 * c72 - s72 * s72;
	    s2 = 2.0 * c72 * s72;
	    do {
	       do {
		  CT aa, aj, ak, ajm, ajp, akm, akp;
		  CT bb, bj, bk, bjm, bjp, bkm, bkp;
		  int k2, k4;

		  k1 = kk + kspan;
		  k2 = k1 + kspan;
		  k3 = k2 + kspan;
		  k4 = k3 + kspan;
		  akp = Re_Data (k1) + Re_Data (k4);
		  akm = Re_Data (k1) - Re_Data (k4);
		  bkp = Im_Data (k1) + Im_Data (k4);
		  bkm = Im_Data (k1) - Im_Data (k4);
		  ajp = Re_Data (k2) + Re_Data (k3);
		  ajm = Re_Data (k2) - Re_Data (k3);
		  bjp = Im_Data (k2) + Im_Data (k3);
		  bjm = Im_Data (k2) - Im_Data (k3);
		  aa = Re_Data (kk);
		  bb = Im_Data (kk);
		  Re_Data (kk) = aa + akp + ajp;
		  Im_Data (kk) = bb + bkp + bjp;
		  ak = akp * c72 + ajp * c2 + aa;
		  bk = bkp * c72 + bjp * c2 + bb;
		  aj = akm * s72 + ajm * s2;
		  bj = bkm * s72 + bjm * s2;
		  Re_Data (k1) = ak - bj;
		  Re_Data (k4) = ak + bj;
		  Im_Data (k1) = bk + aj;
		  Im_Data (k4) = bk - aj;
		  ak = akp * c2 + ajp * c72 + aa;
		  bk = bkp * c2 + bjp * c72 + bb;
		  aj = akm * s2 - ajm * s72;
		  bj = bkm * s2 - bjm * s72;
		  Re_Data (k2) = ak - bj;
		  Re_Data (k3) = ak + bj;
		  Im_Data (k2) = bk + aj;
		  Im_Data (k3) = bk - aj;
		  kk = k4 + kspan;
	       } while (kk < nn);
	       kk -= nn;
	    } while (kk <= kspan);
	    break;

	  default:
	    k = factor [ii - 1];
	    if (jf != k)
	      {
		 jf = k;
		 s1 = pi2 / (double) jf;
		 c1 = iml_cos (s1);
		 s1 = iml_sin (s1);
		 if (jf > maxFactors)
		   goto Memory_Error;
		 Cos [jf - 1] = 1.0;
		 Sin [jf - 1] = 0.0;
		 j = 1;
		 do {
		    Cos [j - 1] = Cos [k - 1] * c1 + Sin [k - 1] * s1;
		    Sin [j - 1] = Cos [k - 1] * s1 - Sin [k - 1] * c1;
		    k--;
		    Cos [k - 1] = Cos [j - 1];
		    Sin [k - 1] = -Sin [j - 1];
		    j++;
		 } while (j < k);
	      }
	    do {
	       do {
		  CT aa, ak;
		  CT bb, bk;
		  int k2;

		  aa = ak = Re_Data (kk);
		  bb = bk = Im_Data (kk);

		  k1 = kk;
		  k2 = kk + ispan;
		  j = 1;
		  k1 += kspan;
		  do {
		     k2 -= kspan;
		     Rtmp [j] = Re_Data (k1) + Re_Data (k2);
		     ak += Rtmp [j];
		     Itmp [j] = Im_Data (k1) + Im_Data (k2);
		     bk += Itmp [j];
		     j++;
		     Rtmp [j] = Re_Data (k1) - Re_Data (k2);
		     Itmp [j] = Im_Data (k1) - Im_Data (k2);
		     j++;
		     k1 += kspan;
		  } while (k1 < k2);
		  Re_Data (kk) = ak;
		  Im_Data (kk) = bk;

		  k1 = kk;
		  k2 = kk + ispan;
		  j = 1;
		  do {
		     CT aj = 0.0;
		     CT bj = 0.0;

		     k1 += kspan;
		     k2 -= kspan;
		     jj = j;
		     ak = aa;
		     bk = bb;
		     k = 1;
		     do {
			ak += Rtmp [k] * Cos [jj - 1];
			bk += Itmp [k] * Cos [jj - 1];
			k++;
			aj += Rtmp [k] * Sin [jj - 1];
			bj += Itmp [k] * Sin [jj - 1];
			k++;
			jj += j;
			if (jj > jf)
			  jj -= jf;
		     } while (k < jf);
		     k = jf - j;
		     Re_Data (k1) = ak - bj;
		     Im_Data (k1) = bk + aj;
		     Re_Data (k2) = ak + bj;
		     Im_Data (k2) = bk - aj;
		     j++;
		  } while (j < k);
		  kk += ispan;
	       } while (kk <= nn);
	       kk -= nn;
	    } while (kk <= kspan);
	    break;
	 }
	 /* multiply by rotation factor (except for factors of 2 and 4) */
	 if (ii == nFactor)
	   goto Permute_Results;	/* exit infinite loop */
	 kk = jc + 1;
	 do {
	    c2 = 1.0 - cd;
	    s1 = sd;
	    do {
	       c1 = c2;
	       s2 = s1;
	       kk += kspan;
	       do {
		  CT tmp;
		  do {
		     CT ak;
		     ak = Re_Data (kk);
		     Re_Data (kk) = c2 * ak - s2 * Im_Data (kk);
		     Im_Data (kk) = s2 * ak + c2 * Im_Data (kk);
		     kk += ispan;
		  } while (kk <= nt);
		  tmp = s1 * s2;
		  s2 = s1 * c2 + c1 * s2;
		  c2 = c1 * c2 - tmp;
		  kk = kk - nt + kspan;
	       } while (kk <= ispan);
	       c2 = c1 - (cd * c1 + sd * s1);
	       s1 += sd * c1 - cd * s1;
	       c1 = 2.0 - (c2 * c2 + s1 * s1);
	       s1 *= c1;
	       c2 *= c1;
	       kk = kk - ispan + jc;
	    } while (kk <= kspan);
	    kk = kk - kspan + jc + inc;
	 } while (kk <= jc + jc);
	 break;
//#endif	/* FFT_RADIX4 */
      }
   }

/* permute the results to normal order -- done in two stages */
/* permutation for square factors of n */
Permute_Results:
   Perm [0] = ns;
   if (kt)
     {
	int k2;

	k = kt + kt + 1;
	if (k > nFactor)
	  k--;
	Perm [k] = jc;
	j = 1;
	do {
	   Perm [j] = Perm [j - 1] / factor [j - 1];
	   Perm [k - 1] = Perm [k] * factor [j - 1];
	   j++;
	   k--;
	} while (j < k);
	k3 = Perm [k];
	kspan = Perm [1];
	kk = jc + 1;
	k2 = kspan + 1;
	j = 1;
	if (nPass != nTotal)
	  {
	     /* permutation for multivariate transform */
	     Permute_Multi:
	     do {
		do {
		   k = kk + jc;
		   do {
		      /* swap
		       * Re_Data (kk) <> Re_Data (k2)
		       * Im_Data (kk) <> Im_Data (k2)
		       */
		      CT tmp;
		      tmp = Re_Data (kk); Re_Data (kk) = Re_Data (k2); Re_Data (k2) = tmp;
		      tmp = Im_Data (kk); Im_Data (kk) = Im_Data (k2); Im_Data (k2) = tmp;
		      kk += inc;
		      k2 += inc;
		   } while (kk < k);
		   kk += (ns - jc);
		   k2 += (ns - jc);
		} while (kk < nt);
		k2 = k2 - nt + kspan;
		kk = kk - nt + jc;
	     } while (k2 < ns);
	     do {
		do {
		   k2 -= Perm [j - 1];
		   j++;
		   k2 = Perm [j] + k2;
		} while (k2 > Perm [j - 1]);
		j = 1;
		do {
		   if (kk < k2)
		     goto Permute_Multi;
		   kk += jc;
		   k2 += kspan;
		} while (k2 < ns);
	     } while (kk < ns);
	  }
	else
	  {
	     /* permutation for single-variate transform (optional code) */
Permute_Single:
	     do {
		/* swap
		 * Re_Data (kk) <> Re_Data (k2)
		 * Im_Data (kk) <> Im_Data (k2)
		 */
		CT t;
		t = Re_Data (kk); Re_Data (kk) = Re_Data (k2); Re_Data (k2) = t;
		t = Im_Data (kk); Im_Data (kk) = Im_Data (k2); Im_Data (k2) = t;
		kk += inc;
		k2 += kspan;
	     } while (k2 < ns);
	     do {
		do {
		   k2 -= Perm [j - 1];
		   j++;
		   k2 = Perm [j] + k2;
		} while (k2 > Perm [j - 1]);
		j = 1;
		do {
		   if (kk < k2)
		     goto Permute_Single;
		   kk += inc;
		   k2 += kspan;
		} while (k2 < ns);
	     } while (kk < ns);
	  }
	jc = k3;
     }

   if ((kt << 1) + 1 >= nFactor)
     return 0;
   ispan = Perm [kt];

   /* permutation for square-free factors of n */
   j = nFactor - kt;
   factor [j] = 1;
   do {
      factor [j - 1] *= factor [j];
      j--;
   } while (j != kt);
   nn = factor [kt] - 1;
   kt++;
   if (nn > maxPerm)
     goto Memory_Error;

   j = jj = 0;
   for (;;) {
      int k2;

      k = kt + 1;
      k2 = factor [kt - 1];
      kk = factor [k - 1];
      j++;
      if (j > nn)
	break;				/* exit infinite loop */
      jj += kk;
      while (jj >= k2)
	{
	   jj -= k2;
	   k2 = kk;
	   kk = factor [k++];
	   jj += kk;
	}
      Perm [j - 1] = jj;
   }
   /* determine the permutation cycles of length greater than 1 */
   j = 0;
   for (;;) {
      do {
	 kk = Perm [j++];
      } while (kk < 0);
      if (kk != j)
	{
	   do {
	      k = kk;
	      kk = Perm [k - 1];
	      Perm [k - 1] = -kk;
	   } while (kk != j);
	   k3 = kk;
	}
      else
	{
	   Perm [j - 1] = -j;
	   if (j == nn)
	     break;		/* exit infinite loop */
	}
   }

   maxFactors *= inc;

   /* reorder a and b, following the permutation cycles */
   for (;;) {
      j = k3 + 1;
      nt -= ispan;
      ii = nt - inc + 1;
      if (nt < 0)
	break;			/* exit infinite loop */
      do {
	 do {
	    j--;
	 } while (Perm [j - 1] < 0);
	 jj = jc;
	 do {
	    int k2;

	    if (jj < maxFactors) kspan = jj; else kspan = maxFactors;

	    jj -= kspan;
	    k = Perm [j - 1];
	    kk = jc * k + ii + jj;

	    k1 = kk + kspan;
	    k2 = 0;
	    do {
	       Rtmp [k2] = Re_Data (k1);
	       Itmp [k2] = Im_Data (k1);
	       k2++;
	       k1 -= inc;
	    } while (k1 != kk);

	    do {
	       k1 = kk + kspan;
	       k2 = k1 - jc * (k + Perm [k - 1]);
	       k = -Perm [k - 1];
	       do {
		  Re_Data (k1) = Re_Data (k2);
		  Im_Data (k1) = Im_Data (k2);
		  k1 -= inc;
		  k2 -= inc;
	       } while (k1 != kk);
	       kk = k2;
	    } while (k != j);

	    k1 = kk + kspan;
	    k2 = 0;
	    do {
	       Re_Data (k1) = Rtmp [k2];
	       Im_Data (k1) = Itmp [k2];
	       k2++;
	       k1 -= inc;
	    } while (k1 != kk);
	 } while (jj);
      } while (j != 1);
   }
   return 0;			/* exit point here */

   /* alloc or other problem, do some clean-up */
Memory_Error:
   fprintf (stderr, "Error: rcs_fftradix() - insufficient memory.\n");
   rcs_fft_free ();			/* free-up memory */
   return -1;
}
/*----------------------- end-of-file (C source) -----------------------*/
