/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-97 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
 	The accompanying software is free; you may use it in your 
 	research  and give it away to others, with the following 
 	restrictions. Any copy  you give away must include this 
 	Legal Stuff(tm) section, unmodified.  Any file that you have 
 	changed must include a note at the top of the file giving your 
 	name, the date, and a description of the changes. This software
   may not be sold, whether in source or compiled form, without 
   my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/


#define IML_VERBOSE_SAFETY

#include "iml_ndim.h"
#include "iml_rofn.h"
#include "iml_math.h"
#include <stdio.h>

//============================================================
IML_ULINT calculate_byte_order_signature(void) {

	IML_ULINT Test = 0x01020408;
	IML_UBYTE *BP = (IML_UBYTE *)&Test;
	IML_ULINT Result =	(((IML_ULINT)BP[0])<<24) | (((IML_ULINT)BP[1])<<16) | 
								(((IML_ULINT)BP[2])<<8) | (((IML_ULINT)BP[3])<<0) ;

	return(Result);

} // end calculate_byte_order_signature


//====================================================================
void iml_image::set_hyperplane_size(IML_ULINT OSSpace) {
	// ERROR CHECKING HERE;
	IML_CLIPIT(OSSpace, 1, OSSpace, ImageSpec.Dimensionality);
	// RESIZE THE SUBSPACE (AND DOMAIN)
	Subspace.resize(OSSpace);
	// RELOAD THE AXIS INFO FROM THE PERMANENT DESCRIPTION;
	Subspace = ImageSpec;
	// SET THE POINTER TO THE BEGINNING;
	Subspace.restart();
	// POINT THE SUBSPACE AT THE START OF THE IMAGE DATA CHUNK;
	Subspace.set_offset(ImageSpec.get_offset());  
} // end


//====================================================================
void iml_image::select_hyperplane(iml_index IML_REF Index) {
	if(Index.Dimensionality != ImageSpec.Dimensionality) 
		{ iml_error("iml_image","Must select subspace with same size index!");
		return; } // end if
	Index.set_pointer();
	Subspace.set_offset( Index.indexize() );
} // end


//====================================================================
iml_index IML_REF iml_image::get_indexer(void) {
	iml_index IML_PTR I= new iml_index(Subspace);
	// set it to the start of the cube
	I->restart();
	return( *I );
} // end


//====================================================================
void iml_image::select_hyperplane(IML_LINT FirstCospaceIndex, ...) {
	if(Subspace.Dimensionality == ImageSpec.Dimensionality) {
		iml_error("iml_image","Image already at full size in select_hyperplane()!");
		return;
		} // end if
	
	// A LOCAL, TEMPORARY VARIABLE
	iml_index Index(ImageSpec);
	Index.restart(); // ALTHOUGH IT SHOULD ALREADY BE RESTARTED
	
	IML_LINT FirstCospaceAxis = Subspace.Dimensionality;
	IML_LINT LastCospaceAxis = ImageSpec.Dimensionality;
	Index.set(FirstCospaceAxis, FirstCospaceIndex);
	// DECLARE A "VARIABLE-ARGUMENT-LIST ARGUMENT POINTER
	va_list ArgPtr;  
	// SET IT TO THE ADDRESS OF THE LAST OBJECT IN THE KNOWN PART OF THE LIST.
	va_start(ArgPtr,FirstCospaceIndex);
	for(IML_ULINT I = FirstCospaceAxis+1; I<LastCospaceAxis; I++) {
		Index.set(I, va_arg(ArgPtr, IML_LINT) );
		} // end for
	va_end(ArgPtr);	// THIS REALLY DOES NOTHING, BUT WE KEEP IT FOR STYLE POINTS.

	Subspace.set_offset( Index.indexize() );
} // end


//====================================================================
// = read-write
void iml_image::operator=(iml_image IML_REF RW2DFn) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	iml_index DstBlock(get_domain());
	DstBlock.clip_against(RW2DFn.get_domain());
	DstBlock.restart();
	IML_ULINT UnClippedCount = DstBlock.unclipped_count();
	iml_index SrcBlock(RW2DFn.get_domain()); 
	SrcBlock.clip_against(get_domain());
	SrcBlock.restart(); 
	for( IML_LINT I=0; I < UnClippedCount; I++,SrcBlock++,DstBlock++ )
		set_value( DstBlock, RW2DFn.get_value(SrcBlock) );
} // end


//====================================================================
// *= read-write
void iml_image::operator+=(iml_image IML_REF RW2DFn) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	iml_index DstBlock(get_domain());
	DstBlock.clip_against(RW2DFn.get_domain());
	DstBlock.restart();
	IML_ULINT UnClippedCount = DstBlock.unclipped_count();
	iml_index SrcBlock(RW2DFn.get_domain()); 
	SrcBlock.clip_against(get_domain());
	SrcBlock.restart(); 
	for( IML_LINT I=0; I < UnClippedCount; I++,SrcBlock++,DstBlock++ )
		increment_value( DstBlock, RW2DFn.get_value(SrcBlock) );
} // end


//====================================================================
// -= read-write
void iml_image::operator-=(iml_image IML_REF RW2DFn) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	iml_index DstBlock(get_domain());
	DstBlock.clip_against(RW2DFn.get_domain());
	DstBlock.restart();
	IML_ULINT UnClippedCount = DstBlock.unclipped_count();
	iml_index SrcBlock(RW2DFn.get_domain()); 
	SrcBlock.clip_against(get_domain());
	SrcBlock.restart(); 
	for( IML_LINT I=0; I < UnClippedCount; I++,SrcBlock++,DstBlock++ )
		increment_value( DstBlock, -RW2DFn.get_value(SrcBlock) );
} // end


//====================================================================
// *= read-write
void iml_image::operator*=(iml_image IML_REF RW2DFn) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	iml_index DstBlock(get_domain());
	DstBlock.clip_against(RW2DFn.get_domain());
	DstBlock.restart();
	IML_ULINT UnClippedCount = DstBlock.unclipped_count();
	iml_index SrcBlock(RW2DFn.get_domain()); 
	SrcBlock.clip_against(get_domain());
	SrcBlock.restart(); 
	for( IML_LINT I=0; I < UnClippedCount; I++,SrcBlock++,DstBlock++ )
		multiply_value( DstBlock, RW2DFn.get_value(SrcBlock) );
} // end



//====================================================================
// = read
void iml_image::operator=(iml_rofn IML_REF R2DFn) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	// FOR NOW, ASSUME THAT A R2DFn IS DEFINED FOR ALL INPUTS; :-(
	iml_index Block(get_domain());   Block.restart();
	for( IML_LINT I=0; I < Block.ElementCount; I++,Block++)
		set_value( Block, R2DFn.get_value(Block) );
} // end


//====================================================================
// *= read
void iml_image::operator*=(iml_rofn IML_REF R2DFn) {


	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	// FOR NOW, ASSUME THAT A R2DFn IS DEFINED FOR ALL INPUTS; :-(
	iml_index Block(get_domain());   Block.restart();
	for( IML_LINT I=0; I < Block.ElementCount; I++,Block++)
		multiply_value( Block, R2DFn.get_value(Block) );
} // end


//====================================================================
// -= read
void iml_image::operator-=(iml_rofn IML_REF R2DFn) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	// FOR NOW, ASSUME THAT A R2DFn IS DEFINED FOR ALL INPUTS; :-(
	iml_index Block(get_domain());   Block.restart();
	for( IML_LINT I=0; I < Block.ElementCount; I++,Block++)
		increment_value( Block, -R2DFn.get_value(Block) );
} // end


//====================================================================
// += read
void iml_image::operator+=(iml_rofn IML_REF R2DFn) {

	// THIS IS UGLY FORMATTING, BUT THE TEXT IS COMPACT . . . :-(
	// FOR NOW, ASSUME THAT A R2DFn IS DEFINED FOR ALL INPUTS; :-(
	iml_index Block(get_domain());   Block.restart();
	for( IML_LINT I=0; I < Block.ElementCount; I++,Block++)
		increment_value( Block, R2DFn.get_value(Block) );
} // end




//====================================================================
// = constant
void iml_image::operator=(double Constant) {

	iml_index Block(get_domain());   Block.restart();
	for( IML_LINT I=0; I < Block.ElementCount; I++, Block++ )
		set_value( Block, Constant );
} // end


//====================================================================
// *= constant
void iml_image::operator*=(double Constant) {

	iml_index Block(get_domain());   Block.restart();
	for( IML_LINT I=0; I < Block.ElementCount; I++, Block++ )
		multiply_value( Block, Constant );
} // end


//====================================================================
// -= constant
void iml_image::operator-=(double Constant) {

	iml_index Block(get_domain());   Block.restart();
	for( IML_LINT I=0; I < Block.ElementCount; I++, Block++ )
		increment_value( Block, -Constant );
} // end


//====================================================================
// += constant
void iml_image::operator+=(double Constant) {

	iml_index Block(get_domain());   Block.restart();
	for( IML_LINT I=0; I < Block.ElementCount; I++, Block++ )
		increment_value( Block, Constant );
} // end

//====================================================================
void iml_image::sgn() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();	
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_sgn(get_value(Block)) );
} // end

//====================================================================
void iml_image::acos() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_acos(get_value(Block)) );
}

//====================================================================
void iml_image::asin() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_asin(get_value(Block)) );
}

//====================================================================
void iml_image::atan() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_atan(get_value(Block)) );
}

//====================================================================
void iml_image::sin() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_sin(get_value(Block)) );
};

//====================================================================
void iml_image::cos() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_cos(get_value(Block)) );
}

//====================================================================
void iml_image::tan() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_tan(get_value(Block)) );
}

//====================================================================
void iml_image::sinh() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_sinh(get_value(Block)) );
}

//====================================================================
void iml_image::cosh() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_cosh(get_value(Block)) );
}

//====================================================================
void iml_image::tanh() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_tanh(get_value(Block)) );
}

//====================================================================
void iml_image::ceil() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_ceil(get_value(Block)) );
}

//====================================================================
void iml_image::floor() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_floor(get_value(Block)) );
}

//====================================================================
void iml_image::exp() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_exp(get_value(Block)) );
}

//====================================================================
void iml_image::sqrt() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_sqrt(get_value(Block)) );
}

//====================================================================
void iml_image::pow(IML_DOUBLE Y) {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_pow(get_value(Block), Y) );
}

//====================================================================
void iml_image::log() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_log(get_value(Block)) );
}

//====================================================================
void iml_image::log10() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_log10(get_value(Block)) );
}

//====================================================================
void iml_image::abs() {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_abs(get_value(Block)) );
}

//====================================================================
void iml_image::mod(IML_DOUBLE Y) {

	iml_index IML_REF Block = (iml_index IML_REF)get_domain();
	Block.restart();
	for(  IML_LINT I=0; I < Block.ElementCount; I++,Block++ )
		set_value( Block, iml_mod(get_value(Block), Y) );
}
