#ifndef __IML_NDIM_H__
#define __IML_NDIM_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

// THIS CLASS IS DESIGNED TO IMPLEMENT IMAGE MANAGEMENT
// USING THE WINDOWS GLOBAL MEMORY ALLOCATION SCHEME.
// THE PROBLEM UNDER DOS IS THAT LARGE IMAGES TAKE UP
// TOO MUCH MEMORY.  WITH WINDOWS WE GET "FREE" MEMORY
// MANAGEMENT AND A FORM OF VIRTUAL MEMORY (IF YOUR WINDOWS
// STUFF IS SET UP RIGHT :-) )

#include "iml_type.h"

#ifdef __I_DO_WINDOWS__
#ifndef STRICT
#define STRICT
#endif

#include <windows.h>
#endif

#include "iml_cplx.h"
#include "iml_indx.h"
#include "iml_math.h"
#include "iml_domn.h"
#include "iml_err.h"
#include <stdio.h>
#include <stdarg.h>

//=======================================================================
// THESE FOUR ARE SPECIFIC TO COMPLEX IMAGES
typedef enum { IML_DEFAULT_VECTOR_PART,  // GENERIC DEFAULT
//					IML_COMPLEX_MAG,
//					IML_COMPLEX_PHASE,
					IML_COMPLEX_REAL,
					IML_COMPLEX_IMAG
} iml_vector_part;


//=======================================================================
typedef enum {
					IML_UBYTE_IMAGE,
					IML_BYTE_IMAGE,
					IML_USINT_IMAGE,
					IML_SINT_IMAGE,
					IML_ULINT_IMAGE,
					IML_LINT_IMAGE,
					IML_FLOAT_IMAGE,
					IML_DOUBLE_IMAGE,
					IML_LDOUBLE_IMAGE,

					IML_FCOMPLEX_IMAGE,
					IML_DCOMPLEX_IMAGE,
					IML_LDCOMPLEX_IMAGE
} iml_image_type;

class iml_rofn;

IML_ULINT calculate_byte_order_signature(void);

//=======================================================================
/** A readable and writable base-class image for typed images.  Provides
functions for the derived classes, particularly hyperplane restriction and
selection.
*/
//=======================================================================
class iml_image : public iml_domain {
	protected:
		friend class iml_image_file;

			// ALLOW ALIAS IMAGES: POINT TO THE SAME MEMORY CHUNK
			// AND HAVE THE SAME IMAGESPEC.
		IML_LINT AliasImage;

		IML_ULINT ByteOrder;
		
//		IML_ULINT PixelSize;
		
	public:
      // THESE TWO FUNCTIONS ARE USED FOR READING AND WRITING IMAGES TO DISK
		virtual IML_ULINT subspace_size_in_bytes(void) 
			{ iml_error("iml_image","Call to virtual base function subspace_chunk_size_in_bytes!!!");
			return(0); };
		virtual void IML_PTR subspace_chunk_address(void) 
			{ iml_error("iml_image","Call to virtual base function subspace_chunk_address!!!");
			return(NULL); };

	
		iml_index ImageSpec;
		iml_index IML_REF Subspace;

		iml_image(const iml_index IML_REF ImSpec) 
				: iml_domain(ImSpec.Dimensionality) , 
				// INITIALIZE Subspace LOCAL REFERENCE WITH "DOMAIN" ADDRESS.
				// THAT IS, Subspace IS JUST AN ALIAS.
				  Subspace(Domain), 
				  ImageSpec(ImSpec) {
				ByteOrder = calculate_byte_order_signature();
				
				AliasImage = FALSE; // THIS IMAGE IS AN ORIGINAL, NOT AN ALIAS.

				// BY DOING THIS WE CAN HAVE A "DOMAIN" AT THE SAME TIME WE HAVE
				// AN IMAGE Subspace, AND THE OLD LIBRARY METHOD WILL STILL WORK.
				Subspace   = ImSpec;
				};
				
		iml_image(const iml_image IML_REF TrueImage) 
				: iml_domain(TrueImage.ImageSpec.Dimensionality) , 
				// INITIALIZE Subspace LOCAL REFERENCE WITH "DOMAIN" ADDRESS.
				// THAT IS, Subspace IS JUST AN ALIAS.
				  Subspace(Domain), ImageSpec(TrueImage.ImageSpec) {
				  
				AliasImage = TRUE; // THIS IMAGE IS AN ALIAS.

				// MIGHT AS WELL ASSUME IT TAKES ON THE SAME SUBSPACE RESTRICTION
				// EXTANT AT THE TIME OF CONSTRUCTION.
				Subspace   = TrueImage.Subspace;
				};
				
		~iml_image() {};

		virtual IML_DOUBLE get_value( iml_index IML_REF Index) 
			{ iml_error("iml_image", "Call to get_value: should be replaced.");
			return(1.0); };
		virtual void set_value( iml_index IML_REF Index, IML_DOUBLE Val) 
			{ iml_error("iml_image", "Call to set_value: should be replaced."); };
		virtual void increment_value( iml_index IML_REF Index, IML_DOUBLE Val) 
			{ iml_error("iml_image", "Call to increment_value: should be replaced."); };
		virtual void multiply_value( iml_index IML_REF Index, IML_DOUBLE Val) 
			{ iml_error("iml_image", "Call to multiply_value: should be replaced."); };
		virtual void min_max_element(IML_DOUBLE &Min, IML_DOUBLE &Max)
			{iml_error("iml_image","Min-Max search does not exist for this image type.");};
		virtual void mean_and_sdev(IML_DOUBLE &Mean, IML_DOUBLE &SDev)
			{iml_error("iml_image","Statistics do not exist for this image type.");};
		virtual void scale_elements(IML_DOUBLE Min, IML_DOUBLE Max) 
			{iml_error("iml_image","'scale_elements' called for type-less image.");};
		virtual iml_image_type image_type() = 0;  // IMAGE TYPE
		virtual IML_BYTE valid_memory(void) { return(FALSE); };

		// GET A COPY OF THE SUBSPACE FOR USE IN INDEXING
		iml_index IML_REF get_indexer(void); 

		void operator=(iml_rofn IML_REF R2DFn);
		void operator=(iml_image IML_REF R2DFn);
		void operator=(IML_DOUBLE Constant);

		void operator+=(iml_rofn IML_REF R2DFn);
		void operator+=(iml_image IML_REF R2DFn);
		void operator+=(IML_DOUBLE Constant);

		void operator-=(iml_rofn  IML_REF R2DFn);
		void operator-=(iml_image  IML_REF R2DFn);
		void operator-=(IML_DOUBLE Constant);

		void operator*=(iml_rofn  IML_REF R2DFn);
		void operator*=(iml_image  IML_REF R2DFn);
		void operator*=(IML_DOUBLE Constant);

		void sgn();		void abs();
		void acos();	void asin();	void atan();
		void sin();		void cos();		void tan();
		void sinh();	void cosh();	void tanh();
		void ceil();	void floor();
		void sqrt();	void mod(IML_DOUBLE Y);
		void exp();		void pow(IML_DOUBLE Y);
		void log();		void log10();

		/** Find the byte-order signature for the current machine.
		Returns an unsigned long integer indicating the byte ordering used by the
		machine upon which this softare is running.  The integer 0x01020408 is placed
		into an unsigned long integer and then bytes zero through three are
		turned into a characteristic long integer with
		<pre>
			IML_ULINT Test = 0x01020408;
			IML_UBYTE *BP = (IML_UBYTE *)&Test;
			IML_ULINT Result =	(((IML_ULINT)BP[0])<<24) | (((IML_ULINT)BP[1])<<16) | 
								(((IML_ULINT)BP[2])<<8) | (((IML_ULINT)BP[3])<<0) ;
		</pre>
		
		Big-Endian machines have the unsigned long signature value 16,909,320.  Little endian
		machines have the unsigned long signature value 134,480,385.  Other (esoteric) byte orders
		are not currently supported and have the unsigned long signature value of zero.
		These are provide in macros:
		<pre>
		
			IML_BIG_ENDIAN_BYTE_ORDER
			IML_LITTLE_ENDIAN_BYTE_ORDER
			IML_UNSPECIFIED_BYTE_ORDER
			
		</pre>
		*/
		IML_ULINT byte_order_signature(void) {return(ByteOrder); };

		/** Returns the number of image dimensions. 
		   */
		IML_ULINT dimensionality(void) 
			{ return(ImageSpec.Dimensionality); };
			
		/** Returns the total number of elements (pixels) in the image, disregarding
		   the hyperplane dimensionality. 
		   */
		IML_ULINT element_count(void) 
			{ return(ImageSpec.ElementCount); };
		
		/** Returns the size of each image dimension. em>Dim</em> goes from 0 (zero) to
		   N-1 if the image has N dimensions.  If <em>Dim</em> is out of range, a
		   value of zero is returned. 
		   */
		IML_LINT dimension_size(IML_USINT Dim) {
			if(Dim >= ImageSpec.Dimensionality) return(0);
			return((IML_LINT)ImageSpec.size(Dim));
			};
		/** Returns the starting index of each image dimension (should be 0 (zero) because
		   image dimensions can't change). em>Dim</em> goes from 0 (zero) to
		   N-1 if the image has N dimensions.  If <em>Dim</em> is out of range, a
		   value of zero is returned.
		   */
		IML_LINT dimension_start(IML_USINT Dim) {
			if(Dim >= ImageSpec.Dimensionality) return(0);
			return((IML_LINT)ImageSpec.amin(Dim));
			};
			
		/** Returns the number of dimensions in the hyperplane. 
		   */
		IML_ULINT hyperplane_dimensionality(void) 
			{ return(Subspace.Dimensionality); };
			
			
		/** Returns the total number of elements (pixels) in the hyperplane. 
		   */
		IML_ULINT hyperplane_element_count(void) 
			// We *have* to use this special approach because ElementCount will
			// be for the entire image, not the subspace.  unclipped_count() calculates
			// the number of elements in the sub(hyper)space.
			{ return(Subspace.unclipped_count()); };
//			{ return(Subspace.ElementCount); };
		
		/** Returns the size of each hyperplane dimension. <em>Dim</em> goes from 0 (zero) to
		   H-1 if the hyperplane has H dimensions.  If <em>Dim</em> is out of range, a
		   value of zero is returned. 
		   */
		IML_LINT hyperplane_dimension_size(IML_USINT Dim) {
			if(Dim >= Subspace.Dimensionality) return(0);
			return((IML_LINT)Subspace.size(Dim));
			};
			
		/** Returns the starting index of each hyperplane dimension (because hyperplane dimension 
		index ranges can be moved about arbitrarily). <em>Dim</em> goes from 0 (zero) to
		   H-1 if the hyperplane has H dimensions.  If <em>Dim</em> is out of range, a
		   value of zero is returned. 
		   */
		IML_LINT hyperplane_dimension_start(IML_USINT Dim) {
			if(Dim >= Subspace.Dimensionality) return(0);
			return((IML_LINT)Subspace.amin(Dim));
			};
			
		/** Move the starting index of the hyperplane. <em>Dim</em> goes from 0 (zero) to
		   H-1 if the hyperplane has H dimensions, and <em>V</em> is the increment/decrement value.  
		   */
		void shift_hyperplane_corner_by(IML_USINT Dim, IML_AXIS_TYPE V)
			{Subspace.shift_corner_by(Dim,V); }
		
		/** Set the starting index of the hyperplane. <em>Dim</em> goes from 0 (zero) to
		   H-1 if the hyperplane has H dimensions, and <em>V</em> is the new starting value.  
		   */
		void shift_hyperplane_corner_to(IML_USINT Dim, IML_AXIS_TYPE V)
			{Subspace.shift_corner_to(Dim,V); }

			
		/** Restrict the image hyperplane to the first <em>OSSpace</em> dimensions. 
		*/
		void set_hyperplane_size(IML_ULINT OSSpace); 

		/** Find out how many dimensions to which the image hyperplane is restricted. 
		*/
		IML_ULINT get_hyperplane_size(void) 
			{ return(Subspace.Dimensionality); };

		/* CURRENTLY USED INTERNALLY:  
		Select one of the image hyperplanes using an index operator. */
		void select_hyperplane(iml_index IML_REF Index);

		/** Select one of the image hyperplanes determined by the call to 
		<em>set_hyperplane_size()</em>.  E.g. for an (m+1) dimensional image of dimensions
		{N1, ... Nk, ... , Nm}
		with <b>set_hyperplane_size(k);</b>, the number of selectable hyperplanes is
		{N(k+1)*N(k+2)*...*Nm} with indices ranging from 0 to {N(k+1)*N(k+2)*...*Nm - 1}
		<p><p><b>NOTE</b>:  You must make sure to have the argument count right!  I have no way
		of checking that for you.
		*/
		void select_hyperplane(IML_LINT FirstSuperAxisIndex, ...);

};// end class iml_image




//=======================================================================
/** An image base-class image for scalar and vector images.  Provides
functions for the derived classes, particularly various methods for
accessing image pixels. Scalar and vector images each implement
the following mathematical operations:
<p>

<ul>
<li>  = OBJECT;
<li> += OBJECT;
<li> -= OBJECT;
<li> *= OBJECT;
<li> /= OBJECT;
</ul>

<p>
where <b>OBJECT</b> can be

<ul>
<li> a scalar: signed or usigned integral, or floating point 
<li> any object derived from the class <em>iml_rofn</em>
<li> an image of a different (pixel) type
<li> an image of identical (pixel) type
</ul>

<p>
There are speed concerns involved with each of these. 

<p>
<b>Scalars</b>
<p>
The {=,+=,-=,*=,/=} operations on scalars are fast because they
are implemented directly.  All scalars are promoted to double
before the operation is executed. Generically, the operational
result is
<pre>

	Image[index] = (pixel type)((double)OBJECT)

or

	Image[index] = (pixel type)( ((double)Image[index]) <b>OPERATOR</b> ((double)OBJECT) )
	
</pre>
where <b>OPERATOR</b> is one of {+,-,*,/}.

<p>
<b>Images of Identical Type</b>
<p>
The {=,+=,-=,*=,/=} operations
are fastest when the images are of identical type because the operations
are implemented directly.

<p>
<b>Images of Different Type</b>
<p>
The {=,+=,-=,*=,/=} operations between
images of unlike type are much slower; they
are implemented using virtual function calls to avoid
combinatorial explosion of template declarations (i.e. typecount * typecount
different combinations!).
<p>
All scalars are promoted to double
before the operation is executed. Generically, the operational
result is
<pre>

	Image[index] = (pixel type)((double)OtherImage[index])

or

	Image[index] = (pixel type)( ((double)Image[index]) <b>OPERATOR</b> ((double)OtherImage[index]) )
	
</pre>
where <b>OPERATOR</b> is one of {+,-,*,/}.


<p>
<b>Objects of <em>class iml_rofn</em></b>
<p>
The {=,+=,-=,*=,/=} operations with <em>iml_rofn</em> objects can be
very slow, depending on the function implemented. Generically, the operational
result is
<pre>

	Image[index] = (pixel type)(OBJECT(index))
	
</pre>

<p>
When operations are performed bewtween vector and scalar images, the default,
or most recently requested vector component is employed in the operation. 
For complex images vector component is either a default, or is specified by
a previous call to one of
<p>
<ul>
<li> magnitude()
<li> phase()
<li> real()
<li> imaginary()
</ul>

These function calls indicate which vector component is to be used in the
next operation.
After the mathematical operation is completed the default component indicator is
restored.  That is, the call above only applies for a single operation.

<p>
See the complex image class documentation for details.

*/
//=======================================================================
template <class T, iml_image_type IType>
class iml_typed_image : public iml_image {
	protected:
	
		#ifdef __I_DO_WINDOWS__
		HGLOBAL MemBlockHandle;
		#endif
		T IML_PTR pData;

			// THESE TWO FUNCTIONS ARE USED BY THE IMAGE FILE MANAGER (ONLY?)
			// WE USE "Subspace" HERE SO THAT WE CAN SAVE SUBSPACE IMAGES INDIVIDUALLY
			// IF WE WANT.

		// FOR USE IN THE CONSTRUCTORS/DESTRUCTORS
		void iml_TI_init(void);
		void iml_TI_uninit(void);

	public:
		/* Returns the number of bytes in the current subspace */
		IML_ULINT subspace_size_in_bytes(void) 
			{ return( sizeof(T)*Subspace.unclipped_count() );};
			
		/* Returns the memory address of the start of the current subspace */
		void IML_PTR subspace_chunk_address(void) 
			{return((void IML_PTR)(pData+Subspace.get_offset()));};

		iml_typed_image(const iml_index IML_REF Dimensions) 
				: iml_image(Dimensions) { 
				iml_TI_init(); };
				
		iml_typed_image(const iml_typed_image<T,IType> IML_REF TrueImage) 
				: iml_image(TrueImage) { 
				pData = TrueImage.pData; };
							
		~iml_typed_image() { iml_TI_uninit(); };

		/** Returns zero if the memory allocation in the constructor failed,
		non-zero otherwise. 
		*/
		IML_BYTE valid_memory(void);

		/** Returns the type of the image.  Use this when you have only a 
		handle on a base-class of the image (e.g. <em>iml_image</em>).  
		@return Return value is one of
		<pre>
		
			IML_UBYTE_IMAGE			IML_BYTE_IMAGE
			IML_USINT_IMAGE			IML_SINT_IMAGE
			IML_ULINT_IMAGE			IML_LINT_IMAGE
			IML_FLOAT_IMAGE			IML_DOUBLE_IMAGE		IML_LDOUBLE_IMAGE,
			IML_FCOMPLEX_IMAGE		IML_DCOMPLEX_IMAGE		IML_LDCOMPLEX_IMAGE

		</pre>
		*/
		iml_image_type image_type(void) {return((iml_image_type)IType); };  // IMAGE TYPE

		void operator=( iml_typed_image<T,IType> IML_REF LikeImage);
		void operator+=( iml_typed_image<T,IType> IML_REF LikeImage);
		void operator-=( iml_typed_image<T,IType> IML_REF LikeImage);
		void operator*=( iml_typed_image<T,IType> IML_REF LikeImage);

		/** <IMG SRC=../images/caution.gif><p>
		Use square bracket's for the fastest access, <b>without built-in index-bounds
		checking</b>.  This accessor also ignores hyperplane limits.  It accesses the
		raw pixels, with indices from zero to {N1*N2*...*Nm - 1} for an <em>m</em> 
		dimensional image.
		@memo  <IMG SRC=../images/endcaution.gif> Use square bracket's for the fastest access.
		@return Returns a reference to the requested element/pixel.
		*/
		T IML_REF operator[](IML_LINT Index) { return(pData[Index]); };
			
		/** <IMG SRC=../images/caution.gif><p>
		Fastest access to raw elements/pixels <b>without built-in index-bounds
		checking</b>.  By <em>raw</em> I mean index ranges start at zero. This function 
		assumes the image is at least one dimensional.
		
		@memo  <IMG SRC=../images/endcaution.gif> Fastest access to 
		<em>raw elements/pixels</em> for 1D images.
		@return Returns a reference to the requested element/pixel.
		*/
		T IML_REF raw_fa(IML_LINT Axis0Index) {
			unsigned long Index =	Subspace.Axes[0].Stride*Axis0Index;
			return( pData[ Index ] );
			};
	
		/** <IMG SRC=../images/caution.gif><p>
		Fastest access to raw elements/pixels <b>without built-in index-bounds
		checking</b>.  By <em>raw</em> I mean index ranges start at zero. This function 
		assumes the image is at least two dimensional.  If your image is <em>less</em>
		than two dimensions, then you might be stomping
		on memory and your program <em>will</em> can do unexpected things.
		
		@memo  <IMG SRC=../images/endcaution.gif> Fastest access to 
		<em>raw elements/pixels</em> for 2D images.
		@return Returns a reference to the requested element/pixel.
		*/
		T IML_REF raw_fa(IML_LINT Axis0Index, IML_LINT Axis1Index) {
			unsigned long Index =	Subspace.Axes[0].Stride*Axis0Index +
									Subspace.Axes[1].Stride*Axis1Index;
			return( pData[ Index ] );
			};
	
		/** <IMG SRC=../images/caution.gif><p>
		Fastest access to raw elements/pixels <b>without built-in index-bounds
		checking</b>. By <em>raw</em> I mean index ranges start at zero. This function 
		assumes the image is at least three dimensional.  If your image is <em>less</em>
		than three dimensions, then you might be stomping
		on memory and your program <em>will</em> can do unexpected things.
		
		@memo  <IMG SRC=../images/endcaution.gif> Fastest access to 
		<em>raw elements/pixels</em> for 3D images.
		@return Returns a reference to the requested element/pixel.
		*/
		T IML_REF raw_fa(IML_LINT Axis0Index, IML_LINT Axis1Index, IML_LINT Axis2Index) {
			unsigned long Index =	Subspace.Axes[0].Stride*Axis0Index + 
									Subspace.Axes[1].Stride*Axis1Index + 
									Subspace.Axes[2].Stride*Axis2Index;
			return( pData[ Index ] );
			};

		/** <IMG SRC=../images/caution.gif><p>
		Fastest access to raw elements/pixels <b>without built-in index-bounds
		checking</b>.  By <em>raw</em> I mean index ranges start at zero. This function 
		assumes the image is at least four dimensional.  If your image is <em>less</em>
		than four dimensions, then you might be stomping
		on memory and your program <em>will</em> can do unexpected things.

		@memo  <IMG SRC=../images/endcaution.gif> Fastest access to <em>raw elements/pixels</em> for 4D images.
		@return Returns a reference to the requested element/pixel.
		*/
		T IML_REF raw_fa(IML_LINT Axis0Index, IML_LINT Axis1Index, IML_LINT Axis2Index, IML_LINT Axis3Index) {
			unsigned long Index =	Subspace.Axes[0].Stride*Axis0Index + 
									Subspace.Axes[1].Stride*Axis1Index + 
									Subspace.Axes[2].Stride*Axis2Index + 
									Subspace.Axes[3].Stride*Axis3Index;
			return( pData[ Index ] );
			};

		/** <IMG SRC=../images/caution.gif><p>
		Use for accessing raw elements/pixels within a 
		<em>hyperplane</em> in the same spirit as [],
		<b>without built-in index-bounds checking</b>. 
		Hyperplane element(pixel) indices go from zero to 
		{N1*N2*...*Nn - 1} for an <em>n</em> dimensional hyperplane, where
		the Ni are the sizes of the dimensions.<b>Use carefully!</b>

		@memo  <IMG SRC=../images/endcaution.gif> Use for accessing raw elements/pixels 
		within a <em>hyperplane</em> in the same spirit as [].
		@return Returns a reference to the requested element/pixel.
		*/
//		T IML_REF operator()(IML_ULINT Index) 
//			{ return((pData+Subspace.get_offset())[Index]); };

		T IML_REF hp_raw_fa(IML_ULINT Index) 
			{ return((pData+Subspace.get_offset())[Index]); };

		/** <IMG SRC=../images/caution.gif><p>
		Use for <em>hyperplane</em> array access, in the same spirit as raw_fa(), <b>without built-in index-bounds
		checking</b>. Hyperplane element(pixel) indices follow index regions,
		so negative indices are acceptable. <b>Use carefully!</b>
		This function has a variable length argument list, so if you use the wrong number
		of indices your program <em>will</em> do unexpected things.

		@memo  <IMG SRC=../images/endcaution.gif> Use for <em>hyperplane</em> array access,
		in the same spirit as raw_fa(). 
		@return Returns a reference to the requested element/pixel.
		*/
		T IML_REF  hp_fa(IML_LINT Axis0Index, ...);

		/** <IMG SRC=../images/caution.gif><p>
		This method provides <em>image/hyperplane element</em> access. 
		<b>Does index-bounds checking</b>, and obeys the index domain.  
		This means that negative
		indices can be used if the "corner" of the hyperplane is not at the "origin."
		Use this one if you want to be safe, but <b>use carefully!</b>
		This function has a variable length argument list, so if you use the wrong number
		of indices your program <em>will</em> do unexpected things.

		@memo  <IMG SRC=../images/endcaution.gif> Provides <em>image/hyperplane element</em> access.
		@return Returns a reference to the requested element/pixel.
		*/
		T IML_REF operator()(IML_LINT Axis0Index, ...);
//		T IML_REF operator()(IML_LINT Axis0Index, IML_LINT Axis1Index, ...);

		/** Use an <em>iml_index</em> object for the <em>hyperplane</em> access. Has no 
		index-bounds checking.  See the <em>iml_index</em> documentation for it's use.
		@return Returns a reference to the requested element/pixel.
		*/
		T IML_REF operator()(iml_index IML_REF ImIndex) 
			{ return( (pData+Subspace.get_offset())[ImIndex.indexize()] ); };

}; // end class iml_typed_image

// DISABLE 'possible loss of precision' WARNINGS FROM VC++
#pragma warning (push)
#pragma warning( disable : 4244 ) 

//=======================================================================
/** A template image class for scalar images.  Provides
functions for intstance classes, particularly various basic math operations and manipulations.
See the documentation for <b>class iml_typed_image</b> for details about these.
*/
//=======================================================================
template <class T, iml_image_type IType> class iml_scalar_image :
							public iml_typed_image<T,IType> {
	protected:
		IML_DOUBLE MaxMag;
		IML_DOUBLE MinMag;

		//	FIND THE MINIMUM AND MAXIMUM MAXIMUM ELEMENT SIZES
		void find_min_max_elements(void);

	public:

		// DEFINE ASSIGNMENT OPERATORS BECAUSE THEY ARE NOT ALL INHERITED/OVERLOADED
   	/* Fill the pixels of this image with pixel values specified by a 
   	read-only function. */
		void operator=(iml_rofn IML_REF R2DFn)
			{ iml_image::operator=(R2DFn); };
   	/* Fill the pixels of this image with pixel values specified by another, 
   	aribitrary-type image. */
		void operator=(iml_image IML_REF R2DFn)
			{ iml_image::operator=(R2DFn); };
   	/* Fill the pixels of this image with pixel values specified by another (like-type) 
   	image. */
		void operator=( iml_scalar_image<T,IType> IML_REF LikeImage)
			{ iml_typed_image<T,IType>::operator=(LikeImage); }


   	/* Fill the pixels of this image with a constant. */
		void operator=(IML_DOUBLE TVal);
   	/* Increment the pixels of this image by a constant. */
		void operator+=(IML_DOUBLE TVal);
   	/* Decrement the pixels of this image by a constant. */
		void operator-=(IML_DOUBLE TVal);
   	/* Multiply the pixels of this image by a constant. */
		void operator*=(IML_DOUBLE TVal);

   	/* Increment this image's pixels (in register) by the given (like-type) 
   	image's pixels. */
		void operator+=( iml_scalar_image<T,IType> IML_REF LikeImage)
			{ iml_typed_image<T,IType>::operator+=(LikeImage); }
   	/* Decrement this image's pixels (in register) by the given (like-type) 
   	image's pixels. */
		void operator-=( iml_scalar_image<T,IType> IML_REF LikeImage)
			{ iml_typed_image<T,IType>::operator-=(LikeImage); }
   	/* Multply this image's pixels (in register) by the given (like-type) 
   	image's pixels. */
		void operator*=( iml_scalar_image<T,IType> IML_REF LikeImage)
			{ iml_typed_image<T,IType>::operator*=(LikeImage); }

   	/* Increment this image's pixels by the given read-only function. */
		void operator+=(iml_rofn  IML_REF R2DFn)
			{iml_image::operator+=(R2DFn); };
   	/* Decrement this image's pixels by the given read-only function. */
		void operator-=(iml_rofn  IML_REF R2DFn)
			{iml_image::operator-=(R2DFn); };
   	/* Multiply this image's pixels by the given read-only function. */
		void operator*=(iml_rofn  IML_REF R2DFn)
			{iml_image::operator*=(R2DFn); };

   	/* Increment this image's pixels (in register) by the given (different-type-image) 
   	image's pixels. */
		void operator+=(iml_image IML_REF R2DFn)
			{iml_image::operator+=(R2DFn); };
   	/* Decrement this image's pixels (in register) by the given (different-type-image) 
   	image's pixels. */
		void operator-=(iml_image  IML_REF R2DFn)
			{iml_image::operator-=(R2DFn); };
   	/* Multiply this image's pixels (in register) by the given (different-type-image) 
   	image's pixels. */
		void operator*=(iml_image  IML_REF R2DFn)
			{iml_image::operator*=(R2DFn); };


		/** Simple 1-D image constructor. */
		iml_scalar_image(IML_USINT XSize) 
			: iml_typed_image<T,IType>((iml_index&)iml_ld_index(XSize)) { 
			MaxMag = MinMag = 0.0;
			};

		/** Simple 2-D image constructor. */
		iml_scalar_image(IML_USINT XSize, IML_USINT YSize) 
			: iml_typed_image<T,IType>((iml_index&)iml_ld_index(XSize, YSize)) { 
			MaxMag = MinMag = 0.0;
			};

		/** Simple 3-D image constructor. */
		iml_scalar_image(IML_USINT XSize, IML_USINT YSize, IML_USINT TSize) 
			: iml_typed_image<T,IType>((iml_index&)iml_ld_index(XSize, YSize, TSize)) { 
			MaxMag = MinMag = 0.0;
			};

		/** Simple 4-D image constructor. */
		iml_scalar_image(IML_USINT X0, IML_USINT X1, IML_USINT X2, IML_USINT X3) 
			: iml_typed_image<T,IType>((iml_index&)iml_ld_index(X0, X1, X2, X3)) { 
			MaxMag = MinMag = 0.0;
			};

		/** General, n-D image constructor. */
		iml_scalar_image(const iml_index IML_REF Dimensions) 
			: iml_typed_image<T,IType>(Dimensions) { MaxMag = MinMag = 0.0; };
			
		/** <IMG SRC=../images/caution.gif><p>
		A constructor that implements image aliasing.  That is, it creates
		a an image that <em> shares the same memory segment</em> as the image
		in the argument of the constructor.  The memory chunk is deleted when <em> the original
		image is deleted!</em>  Use this when you want to use part of
		a larger image as scratch space, or want to access different hyperplanes of
		one image simultaneously.   Don't forget that
		an alias image points to another image's raw memory!
		
		@memo <IMG SRC=../images/endcaution.gif> A constructor that implements image aliasing
		such that the new image <em> shares the same memory segment</em> as the image
		in the argument of the constructor.
		*/
		iml_scalar_image(const iml_scalar_image<T,IType> IML_REF TrueImage) 
			: iml_typed_image<T,IType>(TrueImage) { MaxMag = MinMag = 0.0; };
			
		/* Simple image destructor. */
		~iml_scalar_image() { };

		/** Find the min and max values of the elements of the image. 
		*/
		void min_max_element(IML_DOUBLE &Min, IML_DOUBLE &Max) 
			{ find_min_max_elements(); Min = MinMag; Max = MaxMag; };

		/** Calculate the mean and standard deviation of the image elements. 
		*/
		void mean_and_sdev(IML_DOUBLE &Mean, IML_DOUBLE &SDev);

		/** Scale the elements of the image to be from Min to Max (inclusive). 
		*/
		void scale_elements(IML_DOUBLE Min, IML_DOUBLE Max);

		// THESE TWO VIRTUAL FUNCTIONS ARE TO SATISFY OUR OBLIGATION TO THE
		// iml_image PARENT CLASS
		virtual IML_DOUBLE get_value( iml_index IML_REF Index) 
			{ return( operator() (Index) ); };


		virtual void set_value( iml_index IML_REF Index,IML_DOUBLE Val) 
			{ operator() (Index) = Val; };
			
		virtual void increment_value( iml_index IML_REF Index,IML_DOUBLE Val) 
			{ operator() (Index) += Val; };
			
		virtual void multiply_value( iml_index IML_REF Index,IML_DOUBLE Val) 
			{ operator()(Index) *= Val; };



}; // end class iml_scalar_image

// RE-ENABLE 'possible loss of precision' WARNINGS FROM VC++
#pragma warning( pop ) 


//===============================================================================
// HERE WE DEFINE A SUB-CLASS, SIMILAR TO SCALAR_IMAGE, FOR USE IN A
// GENERAL FORM FOR VECTOR-TYPE IMAGES (I.E. EACH ELEMENT IS A VECTOR)
//=======================================================================
/** A template image class for vector images.  Provides an interface for
assignments and other operations on vector components.
*/
//===============================================================================
template <class V, iml_image_type VType> class iml_vector_image :
							public iml_typed_image<V,VType> {
	protected:
		iml_vector_part VectorPart;
		// THEN THE DERIVED CLASS USES THE VALUE OF VectorPart TO DECIDE
		// WHAT ACTION TO TAKE WITH CALLS TO get_2d_value AND set_2d_value.
		// METHODS FOR SETTING VectorPart SHOULD BE PROVIDED BY THE DERIVED CLASS.

	public:
		void set_part(iml_vector_part VPart) {VectorPart = VPart;};
		iml_vector_image(const iml_index IML_REF Dimensions) 
			: iml_typed_image<V,VType>(Dimensions) 
			{ VectorPart = IML_DEFAULT_VECTOR_PART; };
		~iml_vector_image() {};
		
		// for image aliasing
		iml_vector_image(const iml_vector_image<V,VType> IML_REF TrueImage) 
			: iml_typed_image<V,VType>(TrueImage)
			{ VectorPart = IML_DEFAULT_VECTOR_PART; };


		void operator=(V IML_REF Vect);
		void operator+=(V IML_REF Vect);
		void operator-=(V IML_REF Vect);
		void operator*=(V IML_REF Vect);


}; // end class template




//THE COMPLEX IMAGE CLASS REQUIRES A LITTLE MORE WORK BECAUSE THERE
//ARE SOME THINGS THAT ARE DIFFERENT ABOUT IT.  FOR EXAMPLE, THE ELEMENTS OF
//THE IMAGE ARE NOT SCALAR VALUES SO WE HAVE TO DEFINE THE get_2d_value
//AND set_2d_value FUNCTIONS.
//=======================================================================
/** A template image class for complex images.  Provides various basic mathematical
operations and manipulations.
See the documentation for <b>class iml_typed_image</b> for details about these.
This class also includes forward and inverse FFT's.
*/
//===============================================================================
template <class CC, class CT, iml_image_type IType>
class iml_complex_image : public iml_vector_image<CC,IType> {

	protected:
		IML_DOUBLE MaxMag;
		IML_DOUBLE MinMag;
		IML_BYTE FourierDomain;

		// THIS INFO IS CALCULATED FOR USIN IN FAST MODULATION OF IMAGE BEFORE FFT.
		IML_UBYTE SignByte, SignMask1, SignMask2, MLSByte;
		void initialize_modulation_info(void); // INITIALIZE THE ABOVE
		enum {N_IN, N_OUT, ROOT_N};
		IML_BYTE ScalingAction;
		
		//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
		// THESE ARE FOR Mark Olesen's ADAPTATION OF R.C. SINGLETON'S FORTRAN ROUTINE FOR FFT
		int rcs_fftradix( CT *Re, CT *Im, size_t nTotal, size_t nPass, size_t nSpan,
					int iSign, int maxFactors, int maxPerm);

		int rcs_fftn( int ndim, const int *dims, CT *Re, CT *Im, int iSign);
		int rcs_factorize (int nPass, int * kt);
		void rcs_fft_free (void);
   	CT * Rtmp;		/* temp space for real part*/
   	CT * Itmp;		/* temp space for imaginary part */
   	CT * Cos;		/* Cosine values */
   	CT * Sin;		/* Sine values */
   	int * Perm;		/* Permutation vector */
		//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

		// CONVERT FREQUENCY INDICES INTO IMAGE INDICES
		IML_LINT translate_from_freq(IML_LINT Axis, IML_LINT V) 
			{ IML_LINT Width = Subspace.size(Axis); return((V + Width)  % Width); };

		void set_indices(void);

		//	FIND THE MINIMUM AND MAXIMUM MAXIMUM ELEMENT SIZES;  ASSUMES THAT
		// VectorPart IS SET CORRECTLY
		void find_min_max_elements(void);

		typedef union { CT RealValue; IML_UBYTE Bytes[sizeof(CT)]; } bg; // bit grabber 
	
		void sign_change(bg &BG) 
			{ BG.Bytes[SignByte] = BG.Bytes[SignByte]&SignMask1 | (~BG.Bytes[SignByte])&SignMask2; };      
		IML_UBYTE mlsb(IML_FLOAT RealValue) { return( ((IML_LINT)RealValue) & 0x01 ); };  // mantissa lsb     

	public:

		// DEFINE ASSIGNMENT OPERATORS BECAUSE THEY ARE NOT INHERITED
		/* Fill the requested part of this image with values specified by a 
		read-only function. 
		*/
		void operator=(iml_rofn IML_REF R2DFn)
			{ iml_image::operator=(R2DFn); VectorPart = IML_DEFAULT_VECTOR_PART;};
	   	/* Fill the requested part of this image with values specified by another, 
	   	aribitrary-type image. 
	   	*/
		void operator=(iml_image IML_REF R2DFn)
			{ iml_image::operator=(R2DFn); VectorPart = IML_DEFAULT_VECTOR_PART; };
	   	/* Fill the requested part of this image with values specified by another (like-type) 
	   	image. 
	   	*/
		void operator=( iml_typed_image<CC,IType> IML_REF LikeImage)
			{ iml_typed_image<CC,IType>::operator=(LikeImage); VectorPart = IML_DEFAULT_VECTOR_PART; }
	   	/* Fill the requested part of this image with a constant double value. 
	   	*/
		void operator=(IML_DOUBLE Constant)
			{ iml_image::operator=(Constant); VectorPart = IML_DEFAULT_VECTOR_PART; };

		// DEFINE ASSIGNMENT OPERATORS BECAUSE THEY ARE NOT ALL INHERITED/OVERLOADED
	   	/* Fill this image with a constant complex value. 
	   	*/
		void operator=(CC IML_REF CVal)
			{ iml_vector_image<CC,IType>::operator=(CVal); VectorPart = IML_DEFAULT_VECTOR_PART; };
	   	/* Increment this image by a constant complex value. 
	   	*/
		void operator+=(CC IML_REF CVal)
			{ iml_vector_image<CC,IType>::operator+=(CVal); VectorPart = IML_DEFAULT_VECTOR_PART; };
	   	/* Decrement this image by a constant complex value. 
	   	*/
		void operator-=(CC IML_REF CVal)
			{ iml_vector_image<CC,IType>::operator-=(CVal); VectorPart = IML_DEFAULT_VECTOR_PART; };
	   	/* Multiply this image by a constant complex value. 
	   	*/
		void operator*=(CC IML_REF CVal)
			{ iml_vector_image<CC,IType>::operator*=(CVal); VectorPart = IML_DEFAULT_VECTOR_PART; };

	   	/* Increment this image's pixels (in register) by the given (like-type) 
	   	image's pixels. 
	   	*/
		void operator+=( iml_typed_image<CC,IType> IML_REF LikeImage)
			{ iml_typed_image<CC,IType>::operator+=(LikeImage); VectorPart = IML_DEFAULT_VECTOR_PART; }
	   	/* Decrement this image's pixels (in register) by the given (like-type) 
	   	image's pixels. 
	   	*/
		void operator-=( iml_typed_image<CC,IType> IML_REF LikeImage)
			{ iml_typed_image<CC,IType>::operator-=(LikeImage); VectorPart = IML_DEFAULT_VECTOR_PART; }
   	/* Multply this image's pixels (in register) by the given (like-type) 
   	image's pixels. 
   	*/
		void operator*=( iml_typed_image<CC,IType> IML_REF LikeImage)
			{ iml_typed_image<CC,IType>::operator*=(LikeImage); VectorPart = IML_DEFAULT_VECTOR_PART; }

   	/* Increment this image by the given read-only function. 
   	*/
		void operator+=(iml_rofn IML_REF R2DFn)
			{iml_image::operator+=(R2DFn); VectorPart = IML_DEFAULT_VECTOR_PART; };
   	/* Decrement this image by the given read-only function. 
   	*/
		void operator-=(iml_rofn  IML_REF R2DFn)
			{iml_image::operator-=(R2DFn); VectorPart = IML_DEFAULT_VECTOR_PART; };
   	/* Multply this image by the given read-only function. 
   	*/
		void operator*=(iml_rofn  IML_REF R2DFn)
			{iml_image::operator*=(R2DFn); VectorPart = IML_DEFAULT_VECTOR_PART; };

   	/* Increment this image's pixels (in register) by the given (different-type-image) 
   	image's pixels. 
   	*/
		void operator+=(iml_image IML_REF R2DFn)
			{iml_image::operator+=(R2DFn); VectorPart = IML_DEFAULT_VECTOR_PART; };
   	/* Decrement this image's pixels (in register) by the given (different-type-image) 
   	image's pixels. 
   	*/
		void operator-=(iml_image  IML_REF R2DFn)
			{iml_image::operator-=(R2DFn); VectorPart = IML_DEFAULT_VECTOR_PART; };
   	/* Multply this image's pixels (in register) by the given (different-type-image) 
   	image's pixels. 
   	*/
		void operator*=(iml_image  IML_REF R2DFn)
			{iml_image::operator*=(R2DFn); VectorPart = IML_DEFAULT_VECTOR_PART; };

   	/* Increment the requested part of this image by a constant. 
   	*/
		void operator+=(IML_DOUBLE Constant)
			{iml_image::operator+=(Constant); VectorPart = IML_DEFAULT_VECTOR_PART; };
   	/* Decrement the requested part of this image by a constant. 
   	*/
		void operator-=(IML_DOUBLE Constant)
			{iml_image::operator-=(Constant); VectorPart = IML_DEFAULT_VECTOR_PART; };
   	/* Multiply the requested part of this image by a constant. 
   	*/
		void operator*=(IML_DOUBLE Constant)
			{iml_image::operator*=(Constant); VectorPart = IML_DEFAULT_VECTOR_PART; };

		/** Simple 1-D image constructor. 
		*/
		iml_complex_image(IML_USINT XSize) 
			: iml_vector_image<CC,IType>((iml_index&)iml_ld_index(XSize)) { 
			complex_init_code();
			};

		/** Simple 2-D image constructor. 
		*/
		iml_complex_image(IML_USINT XSize, IML_USINT YSize) 
			: iml_vector_image<CC,IType>((iml_index&)iml_ld_index(XSize, YSize)) { 
			complex_init_code();
			};

		/** Simple 3-D image constructor. 
		*/
		iml_complex_image(IML_USINT XSize, IML_USINT YSize, IML_USINT TSize) 
			: iml_vector_image<CC,IType>((iml_index&)iml_ld_index(XSize, YSize, TSize)) { 
			complex_init_code();
			};

		/** Simple 4-D image constructor. 
		*/
		iml_complex_image(IML_USINT X0, IML_USINT X1, IML_USINT X2, IML_USINT X3) 
			: iml_vector_image<CC,IType>((iml_index&)iml_ld_index(X0, X1, X2, X3)) { 
			complex_init_code();
			};

		/** General, n-D image constructor. 
		*/
		iml_complex_image(const iml_index IML_REF Dimensions) 
				: iml_vector_image<CC,IType>(Dimensions)  {
			complex_init_code();
			};

		/** A constructor that implements image aliasing.  That is, it creates
		a an image that <em> shares the same memory segment</em> as the image
		in the argument of the constructor. Use this when you want to use part of
		a larger image as scratch space, or want to access different hyperplanes of
		one image simultaneously.
		*/
		iml_complex_image(const iml_complex_image<CC,CT,IType>  IML_REF TrueImage) 
			: iml_vector_image<CC, IType>(TrueImage) {
			complex_init_code();
			};

		void complex_init_code(void) {
			FourierDomain = FALSE;
			ScalingAction = ROOT_N;
			set_indices(); 
			initialize_modulation_info();
			Rtmp = NULL;		/* temp space for real part*/
			Itmp = NULL;		/* temp space for imaginary part */
			Cos = NULL;			/* Cosine values */
			Sin = NULL;			/* Sine values */
			Perm = NULL;		/* Permutation values */
			};
			
		/* The only destructor. */
		~iml_complex_image() { rcs_fft_free(); };

		/** Tell the complex image that it is in the Fourier (frequency) domain. 
		*/
		void fourier_domain() 		{FourierDomain = TRUE;  set_indices();};
		
		/** Tell the complex image that it is in the space-time (whatever ) domain. 
		*/
		void non_fourier_domain()	{FourierDomain = FALSE; set_indices();};
		
		/** Multply this image by the highest frequency sinusoid. Do this *before* 
		<em>forward_FFT()</em> to shift the image so that the origin is in the center of the
		image.  Can then be use after <em>inverse_FFT()</em> to put the image back to normal.
		This is done by <em>forward_FFT()</em> <em>and inverse_FFT()</em> when the input argument is TRUE. 
		*/
		void modulate(void);
		
		void goto_polar(void);
		void goto_cartesian(void);
		
		/** Divide by N, where N is the number of pixels, on forward FFT and
		do no division on inverse FFT. 
		*/
		void use_n_in_scaling(void) { ScalingAction = N_IN; };
		
		/** Divide by N, where N is the number of pixels, on inverse FFT and
		do no division on forward FFT. 
		*/
		void use_n_out_scaling(void) { ScalingAction = N_OUT; };
		
		/** Divide by sqrt(N), where N is the number of pixels, on both forward
		and inverse FFT. 
		*/
		void use_root_n_scaling(void) { ScalingAction = ROOT_N; };
		
		/** Do a forward FFT on the complex image.
		If <em>RemapFlag</em> is true, then <em>modulate()</em> is used to shift 
		zero-frequency to the center of the image, and image indexing is changed to be
		from (-N/2) to (N/2)-1 for all hyperplane dimensions. 
		@return Returns FALSE on error, TRUE on successful completion.
		*/
		IML_BYTE forward_FFT(IML_SINT RemapFlag = TRUE);
		
		/** Do an inverse FFT on the complex image.
		If <em>RemapFlag</em> is true, then <em>modulate()</em> is used to shift 
		zero-frequency from the center of the image, and image indexing is changed to be
		from 0 to N-1 for all hyperplane dimensions. 
		@return Returns FALSE on error, TRUE on successful completion.
		*/
		IML_BYTE inverse_FFT(IML_SINT RemapFlag = TRUE);

		virtual IML_DOUBLE get_value( iml_index IML_REF Index);
		virtual void set_value( iml_index IML_REF Index,IML_DOUBLE Val);
		virtual void increment_value( iml_index IML_REF Index,IML_DOUBLE Val);
		virtual void multiply_value( iml_index IML_REF Index,IML_DOUBLE Val);


		//=======================================================================
		// FOR CONVENIENCE, ADD A VARIABLE ARGUMENT LIST ACCESSOR
		// NOTE THAT A ONE-ARGUMENT INDEXER IS ALREADY PROVIDED FOR LINEAR
		// ACCESS OF THE SUBSPACE ELEMENTS.  THIS ONE IS FOR SIMPLIFIED SYNTACTICAL
		// ACCESS. YOU DON'T HAVE TO WRITE THE set_axis_index CALLS YOURSELF.
		//
		// !!!!!WE WILL ASSUME THAT THE ARGUMENT LIST LENGTH IS THE SAME
		// AS THE CURRENT SUBSPACE DIMENSIONALITY (AT LEAST 2).  "CAVEAT USER!"
		//=======================================================================
		/**  <IMG SRC=../images/caution.gif><p>
		Index the complex image by "frequency" coordinates,
		assuming that the image is in the Fourier domain.   If not, an error message is printed and 
		no action is taken. The function also assumes that the index range has
		not been changed by <em>forward_FFT(TRUE)</em>; that is, index ranges are [0..N-1]
		for all dimensions (where N is the dimension size).  You can use this, for example, to create a Fourier
		domain complex image, fill it's elements with a Fourier representation,
		then use <em>inverse_FFT(FALSE)</em> to produce a "space-domain" image.<p>

		@memo  <IMG SRC=../images/endcaution.gif> Index the complex image by "frequency" coordinates.
		@return  A reference to the complex number requested.
		*/
		CC IML_REF frequency(IML_LINT Axis0Index, ...);

		/** Signals that the next image operation will involve the real part. */
		iml_complex_image<CC,CT,IType>  IML_REF real(void)
			{ VectorPart = IML_COMPLEX_REAL; 	return(*this); };

		/** Signals that the next image operation will involve the imaginary part. */
		iml_complex_image<CC,CT,IType>  IML_REF imaginary(void)
			{ VectorPart = IML_COMPLEX_IMAG; 	return(*this); };

		//void mean_and_sdev(IML_DOUBLE &Mean, IML_DOUBLE &SDev);

		/** Scale the requested part of the image to be from Min to Max (inclusive). 
		*/
		void scale_elements(IML_DOUBLE Min, IML_DOUBLE Max);

		/** Find the min and max values of the requested part of the image. 
		@return Min and Max hold the appropriate values when the function returns.
		*/
		void min_max_element(IML_DOUBLE &Min, IML_DOUBLE &Max) 
			{ find_min_max_elements(); Min = MinMag; Max = MaxMag; };

}; // end class iml_complex_image


/** A signed-byte instance of a scalar-valued-pixel image. */
typedef iml_scalar_image<IML_BYTE,			IML_BYTE_IMAGE>						iml_byte_image;
/** An unsigned-byte instance of a scalar-valued-pixel image. */
typedef iml_scalar_image<IML_UBYTE,			IML_UBYTE_IMAGE>						iml_ubyte_image;
/** A signed 16-bit-integer instance of a scalar-valued-pixel image. */
typedef iml_scalar_image<IML_SINT,			IML_SINT_IMAGE>						iml_sint_image;
/** An unsigned 16-bit-integer instance of a scalar-valued-pixel image. */
typedef iml_scalar_image<IML_USINT,			IML_USINT_IMAGE>						iml_usint_image;
/** A signed 32-bit-integer instance of a scalar-valued-pixel image. */
typedef iml_scalar_image<IML_LINT,			IML_LINT_IMAGE>						iml_lint_image;
/** An unsigned 32-bit-integer instance of a scalar-valued-pixel image. */
typedef iml_scalar_image<IML_ULINT,			IML_ULINT_IMAGE>						iml_ulint_image;
/** A 32-bit floating-point instance of a scalar-valued-pixel image. */
typedef iml_scalar_image<IML_FLOAT,			IML_FLOAT_IMAGE>						iml_float_image;
/** An 64-bit floating-point instance of a scalar-valued-pixel image. */
typedef iml_scalar_image<IML_DOUBLE,		IML_DOUBLE_IMAGE>						iml_double_image;
/** An 80-bit floating-point instance of a scalar-valued-pixel image. */
typedef iml_scalar_image<IML_LDOUBLE,		IML_LDOUBLE_IMAGE>					iml_ldouble_image;


/** A 32-bit floating-point, complex valued instance of a vector-valued-pixel 
image. */
typedef iml_complex_image<iml_fcomplex,	IML_FLOAT,		IML_FCOMPLEX_IMAGE>	iml_fcomplex_image;

/** An 64-bit floating-point, complex valued instance of a vector-valued-pixel 
image. */
typedef iml_complex_image<iml_dcomplex,	IML_DOUBLE,		IML_DCOMPLEX_IMAGE>	iml_dcomplex_image;
typedef iml_complex_image<iml_ldcomplex,	IML_LDOUBLE,	IML_LDCOMPLEX_IMAGE>	iml_ldcomplex_image;

// THESE ARE REQUIRED HERE BECAUSE CODE-WARRIOR DOES TEMPLATE
// CODE GENERATION DIFFERENTLY THAN BORLAND DOES. . . WHAT A PAIN.
#ifndef __I_DO_WINDOWS__
#ifndef __IML_NDIO_CPP__
#include "iml_ndi0.cpp"
#endif
#endif

#endif //__IML_NDIM_H__
