/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1993-99 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "iml_rofn.h"



//=========================================================================
IML_DOUBLE iml_rofn::operator()(IML_DOUBLE X0, IML_DOUBLE X1, IML_DOUBLE X2, ...) {

	// WE WILL ASSUME THAT THE ARGUMENT LIST LENGTH IS THE SAME
	// AS THE CURRENT SUBSPACE DIMENSIONALITY.  "CAVEAT USER!"
	Domain.set_ns(0, X0); Domain.set_ns(1, X1); Domain.set_ns(2, X2);
	// DECLARE A "VARIABLE-ARGUMENT-LIST ARGUMENT POINTER
	va_list ArgPtr;  
	// SET IT TO THE ADDRESS OF THE LAST OBJECT IN THE KNOWN PART OF THE LIST.
	va_start(ArgPtr,X2);
	for(IML_USINT I=3; I<Domain.Dimensionality; I++) {
		// REMEMBER THAT ALL chars, shorts AND ints ARE STORED ON THE STACK AS
		// THE LARGEST INTEGER:  4-BYTE INTEGER
		Domain.set_ns(I, va_arg(ArgPtr, IML_DOUBLE) );
		}

	va_end(ArgPtr);	// THIS REALLY DOES NOTHING, BUT WE KEEP IT FOR STYLE POINTS.
	
	return(get_indexed_value());
} // end 


//=========================================================================
IML_DOUBLE iml_rofn::get_value( iml_index IML_REF Index) {

	// DO SOMETHING SENSIBLE WITH ANY MISMATCH OF DIMENSIONALITY
	IML_ULINT DimCnt = Domain.Dimensionality;
	if(Index.Dimensionality < DimCnt) 
		DimCnt = Index.Dimensionality;
	for(IML_USINT I=0; I<DimCnt; I++) {
		Domain.set_ns(I, Index.axis(I) );
		}
	
	return(get_indexed_value());
} // end



//=========================================================================
void inverse_distance_rofn::set_lnorm(IML_DOUBLE LNormV) {
	if(LNormV < __IML_FNGEN_MINPOWER__) {
		iml_warning("inverse_distance_rofn", "Norm exponent is too small.  Defaulting to 2.");
		LNormV = 2.0;
      } // end if
	if(LNormV > __IML_FNGEN_MAXPOWER__) {
		iml_warning("inverse_distance_rofn", "Norm exponent is too large.  Defaulting to 2.");
		LNormV = 2.0;
      } // end if
	LNorm = LNormV;
} // end


//=========================================================================
IML_DOUBLE inverse_distance_rofn::get_indexed_value( void ) {

	// DO THE TRANSFORMATION
	iml_xlate_scale::xform(Domain);
		
	IML_DOUBLE Temp = 0.0;
	for(IML_USINT I=0; I< Domain.Dimensionality; I++)
		Temp += iml_pow(Domain.axis(I), LNorm);
		
	Temp = iml_pow(Temp, 1.0/LNorm);
   // CALCULATED THIS WAY TO AVOID DIVISION BY ZERO
	return( Magnitude / (1.0 + Magnitude * Temp)  );
} // end


//=========================================================================
// copy constructor, just in case
iml_trig_rofn::iml_trig_rofn(iml_trig_rofn IML_REF TFrofn) 
				: iml_rofn(TFrofn.Domain.Dimensionality) {
	initialize();
}
			
//=========================================================================
iml_trig_rofn::iml_trig_rofn(IML_ULINT DimCnt) : iml_rofn(DimCnt) {
	initialize();
}


//=========================================================================
// put this here so we remember how to delete correctly with []
//=========================================================================
iml_trig_rofn::~iml_trig_rofn() { 
	if(Freq) delete [] Freq; 
	};

//=========================================================================
void iml_trig_rofn::initialize(void) {

	long DimCnt = Domain.Dimensionality;
	
	Freq = new IML_DOUBLE[DimCnt];
	if( !Freq ) {
		iml_error("iml_trig_rofn","Can't allocate frequency array.");
		return;
		} // end if
	for(IML_USINT I=0; I< DimCnt; I++) {
		set_center(I, 0); set_period(I, 10);
		set_scale(I, 1.0/IML_2PI); // convert to radians on axes
		} // end for
	set_phase(0);
} // end


//=====================================================================
IML_DOUBLE iml_cos_rofn::get_indexed_value( void ) {
		
	// DO THE TRANSFORMATION
	iml_xlate_scale::xform(Domain);
		
	IML_DOUBLE Temp = Phase;
	for(IML_USINT I=0; I< Domain.Dimensionality; I++) 
		Temp += Domain.axis(I)*Freq[I];
		
	Temp = Magnitude * iml_cos(Temp);
	return( Temp );
} // end


//=====================================================================
IML_DOUBLE iml_polar_cos_rofn::get_indexed_value( void ) {

	// DO THE TRANSFORMATION
	iml_xlate::xform(Domain);
	iml_polarizer::xform(Domain);
	iml_scale::xform(Domain);
		
	IML_DOUBLE Temp = Phase;
	for(IML_USINT I=0; I< Domain.Dimensionality; I++)
		Temp += Domain.axis(I)*Freq[I];

	return( Magnitude * iml_cos(Temp) );
} // END



//=====================================================================
IML_DOUBLE iml_gaussian_rofn::get_indexed_value( void ) {

	// DO THE TRANSFORMATION
	iml_xlate_scale::xform(Domain);
		
	IML_DOUBLE Temp = 0.0;
	for(IML_USINT I=0; I< Domain.Dimensionality; I++)
		Temp += -Domain.axis(I)*Domain.axis(I);

	Temp = Magnitude * iml_exp( Temp );
	return( Temp );
} // END

//=====================================================================
IML_DOUBLE iml_gabor_rofn::get_indexed_value( void ) {

	// GET THE COSINUSOIDAL VALUE
	for(IML_USINT I=0; I< Domain.Dimensionality; I++)
		Carrier.Domain.set_ns(I, Domain.axis(I));
	IML_DOUBLE CosValue = Carrier.get_indexed_value();
	
	// REMEMBER THAT iml_gassian_rofn::get_indexed_value() MULTS BY Magnitude
	IML_DOUBLE Temp = CosValue * iml_gaussian_rofn::get_indexed_value();
	// RETURN THE REST
	return( Temp );
} // end


//=====================================================================
void iml_symmetric_gaussian_rofn::set_sdev(IML_DOUBLE SDev) { 
	for(IML_USINT I=0; I<Domain.Dimensionality; I++) set_scale(I,SDev);
} // end


//=====================================================================
IML_DOUBLE iml_raised_gaussian_rofn::get_indexed_value( void ) {
      		
	for(IML_USINT I=0; I< Domain.Dimensionality; I++) {
		IML_DOUBLE Mean = get_mean(I);
		IML_DOUBLE RadT = Domain.axis(I)-Mean;
		if(RadT < Radii[I]) Domain.set_ns(I, Mean);
		else Domain.set_ns(I, RadT);
		} // end for
		
	return( iml_gaussian_rofn::get_indexed_value() );
} // end


//=====================================================================
IML_DOUBLE iml_raised_cos_rofn::get_indexed_value( void ) {
      		
	for(IML_USINT I=0; I< Domain.Dimensionality; I++) {
		IML_DOUBLE Center = get_center(I);
		IML_DOUBLE RadT = Domain.axis(I)-Center;
		if(RadT < Radii[I]) Domain.set_ns(I, Center);
		else Domain.set_ns(I, RadT);
		} // end for
		
	return( iml_cos_rofn::get_indexed_value() );
} // end


//=====================================================================
IML_DOUBLE iml_dgaussian_rofn
			::calc_derivative_const(IML_DOUBLE SD, IML_ULINT DOrder) {
	if( SD == 0.0 && DOrder != 0 ) return(1.0); // Does not exist! default val...
	return( iml_pow(IML_SQRT2/SD, DOrder) );
} // end


//=====================================================================
iml_dgaussian_rofn::iml_dgaussian_rofn( IML_ULINT DimCnt )  
	: iml_gaussian_rofn( DimCnt ) {
	initialize();
}

//=====================================================================
iml_dgaussian_rofn::iml_dgaussian_rofn( iml_dgaussian_rofn IML_REF DG )  
	: iml_gaussian_rofn( DG.Domain.Dimensionality ) {
	initialize();
}
//=====================================================================
void iml_dgaussian_rofn::initialize( void ) {
	long DimCnt = Domain.Dimensionality;
	
	Order   = new IML_ULINT[DimCnt];
	DScaler = new IML_DOUBLE[DimCnt];
	Poly    = new hermite_polynomial*[DimCnt];
	if( !Order || !DScaler || !Poly) {
		iml_error("iml_dgaussian_rofn","Coult not allocate arrays.");
		return;
		} // end if
	IML_ULINT I;
	for( I=0; I<Domain.Dimensionality; I++) {Poly[I] = NULL; }
	for( I=0; I<DimCnt; I++) {
		Order[I] = 2;
		DScaler[I] = calc_derivative_const(I, Order[I]);
		Poly[I] = new hermite_polynomial(Order[I], IML_SQRT2 );
		}
} // end


//=====================================================================
iml_dgaussian_rofn::~iml_dgaussian_rofn() {
	if(Order) delete [] Order; 
	if(DScaler) delete [] DScaler;
	for(IML_USINT I=0; I<Domain.Dimensionality; I++) {
		if(Poly[I]) delete Poly[I]; 
		}
	if(Poly) delete [] Poly; 
} // end


//=====================================================================
void iml_dgaussian_rofn::set_order(IML_ULINT Dim, IML_ULINT Der) {
	if(Poly && Poly[Dim]) delete Poly[Dim];
	Poly[Dim] = new hermite_polynomial(Der, IML_SQRT2 );
	Order[Dim] = Der;
	IML_DOUBLE SD = iml_gaussian_rofn::get_sdev(Dim);
	DScaler[Dim] = calc_derivative_const(SD, Der);
} // end


//=====================================================================
void iml_dgaussian_rofn::set_sdev(IML_ULINT Dim, IML_DOUBLE SDev) {
	set_scale(Dim, SDev);
	set_order(Dim, Order[Dim]);
} // end


//=====================================================================
IML_DOUBLE iml_dgaussian_rofn::get_indexed_value( void ) {

	// DO THE TRANSFORMATION
	iml_xlate_scale::xform(Domain);
		
	IML_DOUBLE Temp1 = 0.0, Temp2 = 1.0;
	for(IML_USINT I=0; I< Domain.Dimensionality; I++) {
		Temp1 += -Domain.axis(I)*Domain.axis(I);
		Temp2 *=  DScaler[I]*Poly[I]->operator()(Domain.axis(I));
		} // end for
		
	return( Magnitude * iml_exp( Temp1 ) * Temp2 );
} // end


//=====================================================================
IML_DOUBLE iml_rect_rofn::get_indexed_value( void ) {

	// DO THE TRANSFORMATION
	iml_xlate_scale::xform(Domain);
		
	IML_SINT Center = TRUE;
	for(IML_USINT I=0; I< Domain.Dimensionality; I++)
		Center = Center && (Domain.axis(I)<=0.5);
		
	return( (Center) ? Magnitude : 0 );
} // end


//=====================================================================
IML_DOUBLE iml_sinc_rofn::get_indexed_value( void ) {

	// DO THE TRANSFORMATION
	iml_xlate_scale::xform(Domain);
		
	IML_DOUBLE Temp = 1.0;
	for(IML_USINT I=0; I< Domain.Dimensionality; I++)
		Temp *= sinc( Domain.axis(I) );
		
	return( Magnitude * Temp );
} // end


