/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/


#include "iml_seq.h"
#include "iml_math.h"


//=====================================================================
//
//=====================================================================
template <class Type> void iml_sequence<Type>
					::operator=(iml_sequence<Type> IML_REF OtherSeq) {
	// COPY WITH REPEAT AND WRAP
	for(IML_USINT I=0; I< Length; I++) 
		Sequence[I] = OtherSeq[ (I % OtherSeq.Length) ];
		
} 
// end operator=(iml_sequence<Type> IML_REF OtherSeq)


//=====================================================================
//
//=====================================================================
//template <class Type> iml_sequence<Type>::operator=(scalar_rv RV) {
//
//	for(IML_USINT I=0; I< Length; I++) Sequence[I] = RV.sample();
//} // end operator=(scalar_rv RV)


//=====================================================================
//
//=====================================================================
template <class Type> void iml_sequence<Type>::rotate(IML_SINT RotateAmount) {

	Type IML_PTR Buffer = new Type[Length];
	if(!Buffer) {
		iml_error("iml_sequence","Can't allocate rotation buffer.");
		return;
		} // end if
	// DON'T ROTATE MULTIPLES OF THE LENGTH
	RotateAmount = RotateAmount%Length;
	
	// TURNS A NEGATIVE ROTATION INTO A POSITIVE ONE
	RotateAmount = (RotateAmount + Length) % Length;
	
	for(IML_SINT I=0; I<Length; I++) {
		// WRAP VALUES AROUND TO THE OTHER END OF THE ARRAY
		IML_SINT Src = (I + RotateAmount) % Length;
		Buffer[I] = Sequence[Src];
		} // end for
	delete [] Sequence;
	Sequence = Buffer;

} // end rotate

//=====================================================================
//
//=====================================================================
template <class Type> void iml_sequence<Type>::reverse(void) {

	IML_USINT Middle = 0.5+Length/2.0;
	for(IML_USINT I=0; I<=Middle; I++) {
		Type Temp = Sequence[I];
		Sequence[I] = Sequence[Length-I-1];
		Sequence[Length-I-1] = Temp;
		} // end for

} // end iml_sequence<Type>::reverse


//=====================================================================
//
//=====================================================================
template <class Type> void iml_sequence<Type>::shuffle(IML_USINT Times) {

	// NUMBER OF TIMES TO SHUFFLE	
	for(IML_USINT J=0; J<Times; J++) {
		// MAKE A PASS THROUGH THE LIST, EXCHANGING EACH MEMBER WITH ANOTHER,
		// RANDOMLY CHOSEN MEMBER IN THE LIST.
		for(IML_USINT I=0; I<Length; I++) {
			IML_USINT RndPos = Shuffler.sample();	// GENERATE A RANDOM POSITION
			Type Temp = Sequence[RndPos];				// SAVE CONTENTS OF THAT POSITION
			Sequence[RndPos] = Sequence[I];			// STORE CURRENT MEMBER IN THAT POSITION
			Sequence[I] = Temp;							// STORE RANDOMLY CHOSEN MEMBER HERE
			} //end for
		} // end for

} // end iml_sequence<Type>::shuffle


//=====================================================================
//
//=====================================================================
template <class Type> void iml_sequence<Type>
			::sin(IML_DOUBLE Mean, IML_DOUBLE Amplitude, IML_DOUBLE Period, IML_DOUBLE PhaseDeg){

	for(IML_USINT I=0; I<Length; I++) {
		Sequence[I] = Mean + Amplitude*iml_sin(IML_2PI*(I/Period + PhaseDeg/360.0) );
		} // end for

} // end iml_sequence<Type>::sin


//=====================================================================
//
//=====================================================================
template <class Type> iml_sequence<Type>
	::iml_sequence(iml_sequence<Type> & Seq) {
		
	Length = Seq.Length;
	Sequence = new Type[Length];
	if( !Sequence ) { 
		iml_error("iml_sequence", "Can't allocate memory");
		return;
		}
	operator=(Seq);
}

//=====================================================================
//
//=====================================================================
template <class Type> iml_sequence<Type>
	::iml_sequence(Type Start, Type End, IML_USINT Len, iml_sequence_type SeqType) {

	Length = Len;
	switch(SeqType) {
			
		default:
		case LINEAR: {
			if (Length==0) {
				iml_error("iml_sequence","Minimum linear sequence length is 1");
				Length = 1;
				} // end if
			} break;
			
		case LOGARITHMIC: {
			if( Length < 2 ) {
				iml_error("iml_sequence","Minimum logarithmic sequence length is 2");
				Length = 2;
				} // end if
				
			if( Start == 0) {
				iml_error("iml_sequence","Cant use a zero starting point with logarithmic sequences.");
				Start = 1; End = 10;
				} // end if
			} break;
		} // end switch
	
	Sequence = new Type[Length];
	if( !Sequence ) { 
		iml_error("iml_sequence", "Can't allocate memory");
		return;
		}
		
	switch (SeqType) {
		default:
		case LINEAR:
			generate_linear_sequence( Start, (End-Start)/(Length-1) );
			break;
			
		case LOGARITHMIC:
			generate_logarithmic_sequence( Start, pow(End/Start, 1.0/(Length-1.0)) );
			break;
		} // end switch
	
	// INITIALIZE SHUFFLER CORRECTLY
	Shuffler.set_min_max(0,Length-0.000000001);
	Shuffler.use_ran4();
}; // end constructor
	
//=====================================================================
//
//=====================================================================
template <class Type> void iml_sequence<Type>
		::generate_linear_sequence(Type Start, Type Delta) {
	
	// DO IT THIS WAY SO THAT Type ONLY HAS TO SUPPORT "=" AND "+="
	Type AddAmt = 0;
	for(IML_USINT I=0; I<Length; I++) {
		Sequence[I] = Start;
		Sequence[I] += AddAmt;
		AddAmt += Delta;
		} // END FOR
} // END generate_linear_sequence

//=====================================================================
//
//=====================================================================
template <class Type> void iml_sequence<Type>
		::generate_logarithmic_sequence(Type Start, IML_DOUBLE Delta) {
		
	// DO IT THIS WAY SO THAT Type ONLY HAS TO SUPPORT "=" AND "*="
	for(IML_USINT I=0; I<Length; I++) {
		Sequence[I] = Start;
		IML_DOUBLE Multiplyer =  pow(Delta,I);
		Sequence[I] *= Multiplyer;
		}// END FOR
} // END generate_logarithmic_sequence

//=====================================================================
//
//=====================================================================
template <class Type> iml_sequence<Type>::iml_sequence(IML_USINT Len) {
	Length = (Len == 0) ? 1 : Len;

	Sequence = new Type[Length];
	if( !Sequence ) { 
		iml_error("iml_sequence", "Can't allocate memory");
		return;
		} // end if
	for(IML_USINT I=0; I<Len; I++) Sequence[I]=0;

	// INITIALIZE SHUFFLER CORRECTLY
	Shuffler.set_min_max(0,Length-0.000000001);
//	Shuffler.use_ran4();

} // end constructor
	
/*************************
	// THE NUMBER OF VALUES IS DETERMINED INTERNALLY
template <class Type> iml_sequence<Type>::iml_sequence(Type Start, Type End, IML_DOUBLE Delta, iml_sequence_type SeqType) {
										
	
	switch (SeqType) {
	
		default:
		case LINEAR: {
			if(Delta == 0) {
				iml_error("iml_sequence", "For constant linear sequences use another constructor");
				Delta = 1;
				} // end if
			Length = iml_abs((Start-End)/Delta);
			if (Length == 0) {
				iml_error("iml_sequence","Minimum linear sequence length is 1");
				Length = 1;
				} // end if
			} break;
	
		case LOGARITHMIC: {
			if(Delta == 1) {
				iml_error("iml_sequence", "For constant logarithmic sequences use another constructor");
				Delta = 1.1;
				} // end if
			if( End == 0 || Start == 0) {
				iml_error("iml_sequence","Cant use a zero endpoint with log-spacing in this constructor");
				Start = 1; End = 10;
				} // end if
				
			Length = 1.0 + iml_log10(End/Start)/iml_log10(Delta);
			if( Length < 2 ) {
				iml_error("iml_sequence","Minimum logarithmic sequence length is 2");
				Length = 2;
				} // end if
			} break;
			
		} // end switch


	Sequence = new Type[Length];
	if( !Sequence ) { 
		iml_error("iml_sequence", "Can't allocate memory");
		return;
		} // end if

	switch (SeqType) {
		default:
		case LINEAR:
			generate_linear_sequence( Start, Delta );
			break;
			
		case LOGARITHMIC:
			generate_logarithmic_sequence( Start, Delta );
		} // end switch

	Shuffler.set_min_max(0,Length-0.000000001);
	Shuffler.use_ran4();
};
***********************/

