#ifndef __IML_XFRM_H__
#define __IML_XFRM_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "iml_indx.h"
#include "iml_err.h"

//=========================================================================
//
//=========================================================================
class iml_space_transformer {
	protected:
		IML_ULINT Dimensionality;
	public:
		virtual void xform(iml_index IML_REF Index) {
			iml_error("iml_xform","Base class function xform() should be replaced.");
			} // end transform
			
		 iml_space_transformer( IML_ULINT DimCnt) : Dimensionality(DimCnt) {};

};



//=========================================================================
//
//=========================================================================
class iml_scale : public iml_space_transformer {
	protected:
		IML_DOUBLE *Scale;
		
	public:

		//  DO TRANSLATION, SCALING,  IN THAT ORDER. 
		POLYMORPHIC void xform(iml_index IML_REF Index) 
			{ for(IML_USINT I=0; I<Dimensionality; I++) Index.set_ns(I, Index.axis(I)*Scale[I] ); };

		void set_scale(IML_ULINT Dim, IML_DOUBLE SC) {
			if( SC < __IML_FNGEN_ALMOST_ZERO__ ) {
				iml_error("iml_scale", "X-scale is too small.  Defaulting to ALMOST_ZERO.");
				SC = __IML_FNGEN_ALMOST_ZERO__;
				} // end if
			Scale[Dim] = 1.0/SC;
			};

		IML_DOUBLE get_scale(IML_ULINT Dim) { return(1.0/Scale[Dim]); };

			// SIMPLE CONSTRUCTOR
		iml_scale(IML_ULINT DimCnt) : iml_space_transformer(DimCnt) {
			Scale   = new IML_DOUBLE[DimCnt];
			if( !Scale ) { iml_error("iml_scale","Can't allocate arrays."); return; } // end if
			// SET SOME ARBITRARY DEFAULTS
			for(IML_USINT I=0; I< DimCnt; I++) { set_scale(I, 1.0); } // end for
			};

		~iml_scale() { 
			if(Scale) delete [] Scale; 
			};
			

}; // end class iml_scale

//=========================================================================
//
//=========================================================================
class iml_xlate : public iml_space_transformer {
	protected:
		IML_DOUBLE *Xlate;
		
	public:

		//  DO TRANSLATION, SCALING,  IN THAT ORDER. 
		POLYMORPHIC void xform(iml_index IML_REF Index) 
			{ for(IML_USINT I=0; I<Dimensionality; I++) Index.set_ns(I, Index.axis(I)+ Xlate[I] ); };

		void set_center(IML_ULINT Dim, IML_DOUBLE C) { Xlate[Dim] = -C; };

		IML_DOUBLE get_center( IML_LINT Dim) { return(Xlate[Dim]); };

			// SIMPLE CONSTRUCTOR
		iml_xlate(IML_ULINT DimCnt) : iml_space_transformer(DimCnt) {
			Xlate   = new IML_DOUBLE[DimCnt];
			if( !Xlate ) { iml_error("iml_xlate","Can't allocate arrays."); return; } // end if
			// SET SOME ARBITRARY DEFAULTS
			for(IML_USINT I=0; I< DimCnt; I++)  set_center(I, 0.0);
			};

		~iml_xlate() {  if(Xlate) delete [] Xlate;  };

}; // end class iml_xlate


//=========================================================================
//
//=========================================================================
class iml_xlate_scale : public iml_xlate, public iml_scale {
		
	public:

		//  DO TRANSLATION, SCALING,  IN THAT ORDER. 
		POLYMORPHIC void xform(iml_index IML_REF Index) { 
			for(IML_USINT I=0; I<iml_xlate::Dimensionality; I++) 
				Index.set_ns(I, (Index.axis(I) + Xlate[I])*Scale[I] ); 
			};

		//  DO TRANSLATION, SCALING,  IN THAT ORDER. 
//		POLYMORPHIC void xform(iml_index IML_REF Index) {
//			iml_xlate::xform(Index);
//			iml_scale::xform(Index);	
//			};

			// SIMPLE CONSTRUCTOR
		iml_xlate_scale(IML_ULINT DimCnt) : iml_xlate(DimCnt), iml_scale(DimCnt) { };

		~iml_xlate_scale() { };
			

}; // end class iml_xlate_scale


//=========================================================================
//
//=========================================================================
class iml_rotate_in_xy : public iml_space_transformer {
	protected:
		
		IML_DOUBLE Theta; // ASSUMED TO BE IN DEGREES
		// TEMPORARARY, INTERMEDIATE, ETC.
		IML_DOUBLE CosTheta, SinTheta;
		// XFORM MATRIX ELEMENTS
		IML_DOUBLE XForm11, XForm12, XForm21, XForm22;

		// TO ACTUALLY TO THE TRANSFORM, WE INVERSE TRANSFORM THE COORDINATES.
		// SEE SOME SIMPLE GRAPHICS TEXT FOR HOW THIS IS DONE.
		// NOTE THAT THE MATRIX HOLDS THE ROTATION AND SCALING.
		POLYMORPHIC void set_xform_matrix(void) {
			XForm11 = 		  CosTheta;
			XForm21 = -1.0 * SinTheta;
			XForm12 = 		  SinTheta;
			XForm22 = 		  CosTheta;
			};

	public:

		//  DO ROTATION IN X-Y PLANE (ABOUT Z-AXIS).
		POLYMORPHIC void xform(iml_index IML_REF Index) {
			IML_AXIS_TYPE IML_REF XC = Index.Axes[IML_2D_X].AxisV;
			IML_AXIS_TYPE IML_REF YR = Index.Axes[IML_2D_Y].AxisV;
			IML_AXIS_TYPE TempXC = XC;
			XC = XC * XForm11 	 + YR * XForm21;
			YR = TempXC * XForm12 + YR * XForm22;
			};

		void set_rotation(IML_DOUBLE ThetaIn) {
			Theta = -ThetaIn;
			CosTheta = iml_cos(Theta * IML_RAD_PER_DEG );
			SinTheta = iml_sin(Theta * IML_RAD_PER_DEG );
			set_xform_matrix();
			};

		IML_DOUBLE get_rotation(void) { return( -Theta ); };

			// SIMPLE CONSTRUCTOR
		iml_rotate_in_xy(IML_FLOAT Rotation=0.0)  : iml_space_transformer(2) 
			{ set_rotation(Rotation); };

		~iml_rotate_in_xy() { };
			

}; // end class iml_rotate_in_xy

//=========================================================================
//
//=========================================================================
class iml_polarizer : public iml_space_transformer {

	public:

		//  CONVERT TO R, THETA, PHI.
		POLYMORPHIC void xform(iml_index IML_REF Index) {
			switch(Index.Dimensionality) {
				case 1: {
					IML_FLOAT X = (IML_FLOAT)Index.axis(0);
					Index.set_ns(0, ((X < 0) ? -X : X) ); 
					} break;
				case 2: {
					IML_FLOAT X = (IML_FLOAT)Index.axis(0);
					IML_FLOAT Y = (IML_FLOAT)Index.axis(1);
					Index.set_ns(0, iml_sqrt(X*X+Y*Y) );
					Index.set_ns(1, iml_atan2(Y,X));
					} break;
				default: {
					IML_FLOAT X = (IML_FLOAT)Index.axis(0);
					IML_FLOAT Y = (IML_FLOAT)Index.axis(1);
					IML_FLOAT Z = (IML_FLOAT)Index.axis(2);
					Index.set_ns(0, iml_sqrt(X*X+Y*Y+Z*Z));
					Index.set_ns(1, iml_atan2(Y,X));
					Index.set_ns(2, iml_atan2(Z, iml_sqrt(X*X+Y*Y)) );
					}  break;
				} // end switch
			};
			
		iml_polarizer(void)  : iml_space_transformer(3) {};


}; // end class iml_polarizer


#endif // __IML_XFRM_H__
