/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/


#include <stdio.h>

#include "iml_err.h"
#include "iml_rand.h"

// DISABLE 'loss of precision' warning from VC++ for this third-party code
#pragma warning (push)
#pragma warning (disable : 4244)


//=============================================================================
/* A C-program for MT19937B: Real number version */
/* genrand() generate one pseudorandom number with double precision   */
/* which is uniformly distributed on [0,1]-interval for each call.    */
/* sgenrand(seed) set initial values to the working area of 624 words.*/
/* sgenrand(seed) must be called once before calling genrand()        */
/* (seed is any integer except 0).                                    */
//=============================================================================

/*
 MERSENNE TWISTER --
 A MONSTROUS RANDOM NUMBER GENERATOR (May 1997) 
	 Creating a generator with period 2^800 was not enough
	 for Makoto Matsumoto. The Mersenne Twister, a new
	 variant of TGFSR by Matsumoto and Nishimura, sets new
	 standards for the period, quality and speed of random
	 number generators. The incredible period is 2^19937 - 1, a
	 number with about 6000 decimal digits; the 32-bit random
	 numbers exhibit best possible equidistribution properties in
	 dimensions up to 623; and it's fast, very fast. But even on a
	 DEC Alpha Station, the period is so huge that producing
	 all the numbers would require about 10^5983 millenia.
	 The integer version and the floating point version are
	 available in C. A paper on the Mersenne Twister is
	 submitted to ACM TOMACS.
*/

/* Period parameters */  
//#define matsumoto_N 624
static const unsigned long matsumoto_M=397;
static const unsigned long MATRIX_A=0x9908b0df;   /* constant vector a */
static const unsigned long UPPER_MASK=0x80000000; /* most significant w-r bits */
static const unsigned long LOWER_MASK=0x7fffffff; /* least significant r bits */

/* for tempering */   
#define TEMPERING_MASK_B 0x9d2c5680
#define TEMPERING_MASK_C 0xefc60000
#define TEMPERING_SHIFT_U(y)  (y >> 11)
#define TEMPERING_SHIFT_S(y)  (y << 7)
#define TEMPERING_SHIFT_T(y)  (y << 15)
#define TEMPERING_SHIFT_L(y)  (y >> 18)

//static unsigned long ptgfsr[matsumoto_N]; /* set initial seeds: matsumoto_N = 624 words */

void matsumoto_MT19937B::sgenrand(unsigned long seed) { /* seed should not be 0 */

//	int k;
	if(!seed) seed = 4357;
	/* setting initial seeds to ptgfsr[matsumoto_N] using     */
	/* the generator Line 25 of Table 1 in          */
	/* [KNUTH 1981, The Art of Computer Programming */
	/*    Vol. 2 (2nd Ed.), pp102]                  */
	
	ptgfsr[0]= seed & 0xffffffff;
	for (Ctr=1; Ctr<matsumoto_N; Ctr++)
		ptgfsr[Ctr] = (69069 * ptgfsr[Ctr-1]) & 0xffffffff;
}

// THIS WAS ORIGINALLY CALLED genrand()
double matsumoto_MT19937B::sample(void) {
  unsigned long y;
//  static int k = matsumoto_N;
  static unsigned long mag01[2]={0x0, MATRIX_A};
  /* mag01[x] = x * MATRIX_A  for x=0,1 */
  
	if(Ctr == matsumoto_N){ /* generate matsumoto_N words at one time */
		int kk;
		for (kk=0;kk<matsumoto_N-matsumoto_M;kk++) {
			y = (ptgfsr[kk]&UPPER_MASK)|(ptgfsr[kk+1]&LOWER_MASK);
			ptgfsr[kk] = ptgfsr[kk+matsumoto_M] ^ (y >> 1) ^ mag01[y & 0x1];
			}
		for (;kk<matsumoto_N-1;kk++) {
			y = (ptgfsr[kk]&UPPER_MASK)|(ptgfsr[kk+1]&LOWER_MASK);
			ptgfsr[kk] = ptgfsr[kk+(matsumoto_M-matsumoto_N)] ^ (y >> 1) ^ mag01[y & 0x1];
			}
		y = (ptgfsr[matsumoto_N-1]&UPPER_MASK)|(ptgfsr[0]&LOWER_MASK);
		ptgfsr[matsumoto_N-1] = ptgfsr[matsumoto_M-1] ^ (y >> 1) ^ mag01[y & 0x1];
		
		Ctr = 0;
		}
  
	y = ptgfsr[Ctr++];
	y ^= TEMPERING_SHIFT_U(y);
	y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
	y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
	//  y &= 0xffffffff; /* you may delete this line if word size = 32 */
	y ^= TEMPERING_SHIFT_L(y);
	
	//  return ( (double)y / (unsigned long)0xffffffff );
	return ( (double)y *2.3283064370807974e-10 );
}


//==============================================================================
/* 
FSU - ULTRA	The greatest random number generator that ever was
		or ever will be.  Way beyond Super-Duper.
		(Just kidding, but we think its a good one.)

Authors:	Arif Zaman (arif@stat.fsu.edu) and
		George Marsaglia (geo@stat.fsu.edu).

Date:		27 May 1992

Version:	1.05

Copyright:	To obtain permission to incorporate this program into
		any commercial product, please contact the authors at
		the e-mail address given above or at

		Department of Statistics and
		Supercomputer Computations Research Institute
		Florida State University
		Tallahassee, FL 32306.

See Also:	README		for a brief description
		ULTRA.DOC	for a detailed description
*/ 
//==============================================================================
/*
   File: ULTRA.C

   This is the ULTRA random number generator written entirely in C.

   This may serve as a model for an assembler version of this routine.
   The programmer should avoid simply duplicating and instead use the
   usual assembler features to increase the speed of this routine.

   Especially the subroutine SWB should be replaced by the one
   machine instruction (usually called subtract-with-borrow) that
   is available in almost every hardware.

   For people not familiar with 8086 assembler, it may help to
   consult this when reading the assembler code. This program should
   be a dropin replacement for the assembler versions, but is about
   half as fast.
*/

#define FSULTRA_N2 24           /* The shorter lag      */

long  FSULTRA_DEFAULT_swbseed[FSULTRA_N] = { 0x0D45D69A,	/* The seed for the SWB generator	*/
  0x9DB73B1A, 0xA84604E8, 0x7C5F0CA5, 0xBC0196CE, 0xFF4CB42E, 0x7ACA6BE3,
  0xA9ED2A5A, 0x6405A8F7, 0xAC00D4F8, 0xBDD1FC77, 0x064F9DC5, 0xF10AB737,
  0x781293D3, 0x8410C2D2, 0x1C6587DB, 0x7D8F8F0F, 0xF3DCC4EA, 0xB965F99F,
  0x9A0094D1, 0x65976D1C, 0x09173DC1, 0x8E38B992, 0x84701D44, 0x14F0E1B9,
  0xE8A1EC5F, 0x1E925A12, 0xE77A0B5B, 0xCDB5926E, 0xD16260C8, 0xC917E806,
  0x519076AA, 0x7BF6C21C, 0x808C0C90, 0x3E93C7B8, 0x707D6EA0, 0xF1DB698D};

/*************************************************************************
 * For each of swb32, swb16, swb8 and swb1, there is an array, a counter *
 * and a pointer. The counters count down, the pointers go up. When the  *
 * counter reaches zero, a fill routine is called to refill the array.   *
 ************************************************************************/



fsultra::fsultra(void) {

	for(long int K=0;K<FSULTRA_N;K++)   swbseed[K]=FSULTRA_DEFAULT_swbseed[K];
	flags=0;		/* The carry flag for the SWB generator	*/
	congx = 0x1C3a128F;    /* Seed for congruential generator	*/

	swb32n=0;	/* A counter and a	*/
	swb16n=0;	/* pointer is kept for	*/
	swb8n=0;	/* each array.		*/

	swb1n=0;	/* The swb1 is slightly different.		*/


} // end constructor


unsigned long	fsultra::bit_reverse( unsigned long n ) {
	n = ((n >>  1) & 0x55555555) | ((n <<  1) & 0xAAAAAAAA);
	n = ((n >>  2) & 0x33333333) | ((n <<  2) & 0xCCCCCCCC);
	n = ((n >>  4) & 0x0F0F0F0F) | ((n <<  4) & 0xF0F0F0F0);
	n = ((n >>  8) & 0x00FF00FF) | ((n <<  8) & 0xFF00FF00);
	n = ((n >> 16) & 0x0000FFFF) | ((n << 16) & 0xFFFF0000);
	return(n);
	};


/* rinit initializes the constants and fills the seed array one bit at
   a time by taking the leading bit of the xor of a shift register
   and a congruential sequence. The same congruential generator continues
   to be used as a mixing generator for the Subtract-with-borrow generator
   to produce the `ultra' random numbers

   Since this is called just once, speed doesn't matter much and it might
   be fine to leave this subroutine coded just as it is.

   PS:	there are quick and easy ways to fill this, but since random number
	generators are really "randomness amplifiers", it is important to
	start off on the right foot. This is why we take such care here.
*/

void fsultra::rinit(long unsigned congy,long unsigned shrgx)
{ short i,j;
  unsigned long  tidbits=0;
  congx=congy*2+1;
  for (i=0;i<FSULTRA_N;i++)
  { for (j=32;j>0;j--)
    { congx = congx * 69069;
      shrgx = shrgx ^ (shrgx >> 15);
      shrgx = shrgx ^ (shrgx << 17);
      tidbits = (tidbits>>1) | (0x80000000 & (congx^shrgx));
    }
    swbseed[i] = tidbits;
  }
  swb32n = swb16n = swb8n = swb1n = 0;
  flags = 0;
}

/* SWB is the subtract-with-borrow operation which should be one line
   in assembler code. This should be done by using the hardware s-w-b
   operation in the SWBfill routine.

   What has been done here is to look at the msb of x, y and z=x-y-c.
   Using these three bits, one can determine if a borrow bit is needed
   or not according to the following table:

	msbz=0  msby=0  msby=1          msbz=1  msby=0  msby=1

	msbx=0  0       1               msbx=0  1       1
	msbx=1  0       0               msbx=1  0       1

   PS: note that the definition is very carefully written because the
   calls to SWB have y and z as the same memory location, so y must
   be tested before z is assigned a value.
*/
#define FSULTRA_SWB(c,x,y,z) \
c = (y<0) ? (((z=x-y-c) < 0) || (x>=0)) : (((z=x-y-c) < 0) && (x>=0));

void fsultra::SWBfill(long *x)
{ short i;
/*
   The following two lines are the heart of the system and should be
   written is assembler to be as fast as possible. It may even make sense
   to unravel the loop and simply wirte 37 consecutive SWB operations!
*/
  for (i=0;  i<FSULTRA_N2; i++) FSULTRA_SWB(flags,swbseed[i+FSULTRA_N-FSULTRA_N2],swbseed[i],swbseed[i]);
  for (i=FSULTRA_N2; i<FSULTRA_N;  i++) FSULTRA_SWB(flags,swbseed[i  -FSULTRA_N2],swbseed[i],swbseed[i]);
#ifndef fast
  for (i=0;  i<FSULTRA_N;  i++) *(x++) = swbseed[i] ^ (congx = congx * 69069);
#else
  for (i=0;  i<FSULTRA_N;  i++) *(x++) = swbseed[i];
#endif
}

long fsultra::swb32fill(void)
{ swb32p = &swb32[0];
  SWBfill(&swb32[0]);
  swb32n = FSULTRA_N-1;
  return *(swb32p++);
}

short fsultra::swb16fill(void)
{ swb16p = (short *) &swb16[0];
  SWBfill((long *) swb16p);
  swb16n = 2*FSULTRA_N-1;
  return *(swb16p++);
}

char fsultra::swb8fill(void)
{ swb8p = (char *) &swb8[0];
  SWBfill((long *) swb8p);
  swb8n = 4*FSULTRA_N-1;
  return *(swb8p++);
}

char fsultra::swb1fill(void)
{ unsigned long i,j;
  swb1p = &swb1[0];
  swb1n = 31;
  i = i32bit();
  for (j=0;j<32;j++) { swb1[j] = i&1; i=i>>1; }
  return *(swb1p++);
}

#define FSULTRA_two2neg31  ( (2.0/0x10000) / 0x10000 )
#define FSULTRA_two2neg32  ( (1.0/0x10000) / 0x10000 )

float fsultra::vni(void)
{ long temp;
  temp = i32bit();
  if (temp & 0xFF000000) { return temp * FSULTRA_two2neg31; }
  return (temp + i32bit() * FSULTRA_two2neg32) * FSULTRA_two2neg31;
}

float fsultra::uni(void)
{ long temp;
  temp = i31bit();
  if (temp & 0xFF000000) { return temp * FSULTRA_two2neg31; }
  return (temp + i32bit() * FSULTRA_two2neg32) * FSULTRA_two2neg31;
}

double fsultra::dvni(void)
{ unsigned long temp;
  temp = i32bit();
  return ( (long) i32bit() + temp * FSULTRA_two2neg32) * FSULTRA_two2neg31; }

double fsultra::duni(void)
{ unsigned long temp;
  temp = i32bit();
  return ( i31bit() + temp * FSULTRA_two2neg32) * FSULTRA_two2neg31; }


//RE-ENABLE PREVIOUS WARNING LEVELS
#pragma warning (pop)
