#ifndef __VECTOR_2d_H__
#define __VECTOR_2d_H__

#include "iml_math.h"

//=====================================================================
/**  An auxiliary class used for various uses.  It provides some simplified
manipulation routines.  These include shifting by another position vector
(an x,y offset), getting magnitude and phase, etc.
*/
//=====================================================================
class vector_2d {
	public:
	/** @name A: Publicly manipulable variables. 
	It was simplest to allow direct manipulation of the X and Y positions.
	Be warned that negative values are invalid and will be clipped by
	the MAC-OS display software.  That error will be silent.
	*/
	//@{
		float X, Y;
	//@}
	
		long xi(void) {return(X);}
		long yi(void) {return(Y);}
		long rxi(void) {return(X+0.5);}
		long ryi(void) {return(Y+0.5);}
	
	/** @name B: Constructors.  */
	//@{
		/// The zero-argument constructor for array allocations; XPos=YPos=0.
		vector_2d(void) : X(0), Y(0) {};
		/// The X, Y constructor.
		vector_2d(short x, short y) : X(x), Y(y) {};
		/// The "make a copy" constructor.
		vector_2d(vector_2d const & P) : X(P.X), Y(P.Y) {};
	
		
	//@}
	
	/** @name C: Useful member functions.  */
	//@{
	
		/** Set the position to a vector with <em>Length</em> and <em>Angle</em>. */
		vector_2d & set_polar(double Length, double Angle) { 
//			X = 0.5+Length*cos(IML_RAD_PER_DEG*Angle); 
//			Y = 0.5+Length*sin(IML_RAD_PER_DEG*Angle); 
			X = Length*cos(IML_RAD_PER_DEG*Angle); 
			Y = Length*sin(IML_RAD_PER_DEG*Angle); 
			return(*this); 
			};

		/** Set the X,Y positions. */
		void set(short x, short y) { X=x; Y=y; };

		float magnitude(void) { return(sqrt(X*X+Y*Y)); };

		float phase(void) { return( atan2(X,Y) ); };

		/** Scale the vector by a constant. */
		vector_2d & operator*=(double Value) 
			{ X *= Value; Y *= Value; return(*this); };
		
		// NEED THIS ONE FOR comparison PURPOSES.
		bool operator==(vector_2d const & Other) 
			{ return(X == Other.X && Y == Other.Y); };
			
		// NEED THIS ONE FOR INITIALIZATION PURPOSES.
		vector_2d & operator=(double Value) 
			{ X = Value; Y = Value; return(*this); };
			
		/** Add a vector. */
		vector_2d & operator+=(vector_2d & Delta) 
			{ X += Delta.X; Y += Delta.Y; return(*this); };
			
		/** Subtract a vector. */
		vector_2d & operator-=(vector_2d & Delta) 
			{ X -= Delta.X; Y -= Delta.Y; return(*this); };
	//@}

};
#endif //__VECTOR_2d_H__
