/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/


// WE NEED TO DEFINE "MAC_C" TO A VALUE OF '1' IF WE WANT TO USE THE
// EXTENDED C FUNCTIONS WHICH THE MAC TOOL BOXES USE.  THIS ALLOWS
// STACK EXTENSION, ETC.  
#define MAC_C 1

// THESE ARE C ROUTINES, AND WE MUST TELL THE COMPILER SO THAT WE CAN LINK
// TO THE C-COMPILED LIBRARY.  OTHERWISE THE NAMES ARE MUNGED AND THE LINKER
// LOOKS FOR THE WRONG SYMBOLS.  
extern "C" {
	#include "Videotoolbox.h"
	#include "Luminance.h"
}

#ifdef HUGE_VAL
#undef HUGE_VAL
#endif

#include "mvl_macv.h"

#include "mvl_util.h"



//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%   mvl_temporal_window   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


//=====================================================================
//
//=====================================================================
mvl_temporal_window::~mvl_temporal_window() {
	if( MVLColorSpecTables )  delete [] MVLColorSpecTables;
} // end destructor

//=====================================================================
//
//=====================================================================
mvl_temporal_window::mvl_temporal_window(mvl_temporal_window IML_REF ACK)
			// INITALIZE THE MEMORY STUFF WITH THIS INITIALIZATION
			: iml_float_image(iml_ld_index(1)) {
	iml_error("mvl_temporal_window",
				"BAD PERSON! Never use a copy constructor "
				"with a mvl_temporal_window!  I'm exiting now . . ."
				);
	exit(0);

} // end copy constructor catcher

//=====================================================================
//
//=====================================================================
mvl_temporal_window::mvl_temporal_window(IML_USINT Len) 
			// INITALIZE THE MEMORY STUFF WITH THIS INITIALIZATION
			: iml_float_image(iml_ld_index(Len)) {
	
	if( !pData )  {
		iml_error("mvl_temporal_window","Could not allocate float profile memory.");
		return;
		} // end if
	Length = Len;
	MVLColorSpecTables = new ColorSpecTable[Length];
	if( !MVLColorSpecTables ) {
		iml_error("mvl_temporal_window","Could not allocate ColorSpecTables.");
		} // end if

} // end constructor




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%   mvl_clut   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

mvl_clut::mvl_clut(IML_UBYTE DBits, IML_USINT CSize) {
	DacSize = DBits;
	CLUTSize = CSize;
	LeftShift = sizeof(CLUT[0].rgb.red)*8 - DacSize;
	MaxDac = (0x0001 << DacSize) - 1;
	CLUT = new mvl_ColorSpec[CLUTSize];
	if(CLUT) {
		// DEFAULT INITIALIZATION IS NO GAMMA CORRECTION.
		for(IML_SINT I=0;I<CLUTSize;I++) {
			mvl_ColorSpec & C = CLUT[I];
			C.rgb.red = C.rgb.blue = C.rgb.green = (I<<LeftShift);
			C.value = I;
			} // end for
		} // end if
	};

mvl_clut::mvl_clut(mvl_clut IML_REF Example) {
	DacSize = Example.DacSize;
	CLUTSize = Example.CLUTSize;
	LeftShift = Example.LeftShift;
	MaxDac = Example.MaxDac;
	CLUT = new mvl_ColorSpec[CLUTSize];
	if(CLUT) {
		// DEFAULT INITIALIZATION IS NO GAMMA CORRECTION.
		for(IML_SINT I=0;I<CLUTSize;I++)  CLUT[I] = Example.CLUT[I];
		} // end if
	};


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%   mvl_video_manager   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


//=====================================================================
//
//=====================================================================
mvl_video_manager::mvl_video_manager(mvl_video_manager IML_REF Copy) {
	iml_error("mvl_video_manager",
				"BAD PERSON! Never use a copy constructor "
				"with a video manager!  I'm exiting now . . ."
				);
	exit(0);

} // end copy constructor catcher

//=====================================================================
//
//=====================================================================
mvl_video_manager::mvl_video_manager(
					IML_USINT MaxImageSequences, IML_USINT MaxBlitSequences, 
					IML_USINT MaxBlitGroups, IML_ULINT ExitOnErr) {

	set_up_SIOUX();

// THIS IS ONE OF PELLI'S FUNCTION CALLS, AND APPARENTLY
// *MUST* BE CALLED BECAUSE IT SETS UP SOME INTERNAL
// FLAGS USED BY OTHER OF HIS ROUTINES. . .
// ...BUT I HAVE DUPLICATED, BELOW, ALL THE CHECKS Require() DOES 
// (THAT WE NEED?)
	Require(gestalt8BitQD);

	// USE THIS FLAG LATER TO DECIDE WHETHER TO EXIT
	ExitOnError = ExitOnErr;

	// ALTERNATIVE WINDOW-FILLING METHOD THAT GETS PAST APPLE'S SHENANIGANS.
	FillWindowID = -1;
	FillerImage = NULL;

	// SET UP VIDEO INFORMATION STRUCTURES
	IML_CLIPIT(MaxImageSequences, 1, MaxImageSequences, MAX_SEQUENCES);
	mvl_mac_video_information *VideoInfo = 
					new mvl_mac_video_information(MaxImageSequences,MaxBlitSequences);
	if( !VideoInfo->successful() ) {
		iml_error("mvl_video_manager","Failed to allocate VideoInfo stucture.");
		if(ExitOnError) exit(1);
		return;
		} // end if

	// SAVE A POINTER TO THE INFO FOR LATER ACCESS.
	MacVideoInfo = (void *) VideoInfo;

	GroupCount = MaxBlitGroups;
	BlitGroups = new mvl_blit_sequence_group*[GroupCount];
	if( !BlitGroups ) {
		iml_error("mvl_video_manager","Failed to allocate BlitGroups array.");
		if(ExitOnError) exit(1);
		return;
		} // end if

	// FOR SAFETY, NULL ALL THE POINTERS
	for(IML_SINT I=0; I< GroupCount; I++) BlitGroups[I] = NULL;
	
	// ALLOCATE EACH BLIT GROUP
	for(IML_SINT I=0; I< GroupCount; I++) {
		BlitGroups[I] = new mvl_blit_sequence_group(MaxBlitSequences);
		if( !BlitGroups[I]->successful() ) {
			iml_error("mvl_video_manager","Failed to allocate a BlitGroup.");
			if(ExitOnError) exit(1);
			return;
			} // end if
		BlitGroups[I]->clear_group();
		} // end for

	// SET UP SOME DEFAULT VALUES
	DisplayNumber = -1;
	CLUT_Sequence = NULL;
	CalibrationPolynomial = NULL;
	MeanLuminance = 50;
	MeanLFraction = 0.5;
	XResolution=0, YResolution=0;
	HasGestalt=SystemOSVersion=IsSystemSeven=IsPowerPC=SystemArchitecture=0;
	FPUType=CPUType=0;
	QuickDrawVersion=Has8BitQD=NewPaletteManagerPresent=DeviceIsColor=InitialPixelSize=0;
	
		// StackSpace() IS A MAC-OS CALL IN Memory.h
	StackSize = StackSpace();

	if( StackSize < 10000L ) {
		extend_stack(10000); // ask to extend it to 10,000 bytes
		if( StackSize < 10000 )
			iml_warning("main","Could not make loads of stack space");
		} // end if
	
	HasGestalt = TRUE;
	if( Gestalt(gestaltFPUType, &FPUType) ) {
		iml_error("mvl_video_manager","System call Gestalt() is not supported. Your System is too old.");
		HasGestalt = FALSE;
		if(ExitOnError) exit(1);
		return;
		} // end if
		
	Gestalt(gestaltNativeCPUtype,&CPUType);
// THESE ARE FROM Gestalt.h
//	gestaltNativeCPUtype		= 'cput',						/* Native CPU type 									  *//
//	gestaltCPU68000				= 1,							/* Various 68k CPUs... 	*/
//	gestaltCPU68010				= 2,
//	gestaltCPU68020				= 3,
//	gestaltCPU68030				= 4,
//	gestaltCPU68040				= 5,
//	gestaltCPU601				= 0x101,						/* IBM 601 												*/
//	gestaltCPU603				= 0x103,
//	gestaltCPU604				= 0x104

	Gestalt(gestaltSysArchitecture,&SystemArchitecture);
//	gestaltSysArchitecture		= 'sysa',						/* Native System Architecture */
//	gestalt68k					= 1,							/* Motorola MC68k architecture */
//	gestaltPowerPC				= 2								/* IBM PowerPC architecture */

	IsPowerPC = (SystemArchitecture == gestaltPowerPC);
		
	Gestalt(gestaltQuickdrawVersion, &QuickDrawVersion);
	Has8BitQD  = (QuickDrawVersion >= gestalt8BitQD);
//	gestaltQuickdrawVersion		= 'qd  ',						/* quickdraw version */
//	gestaltOriginalQD			= 0x000,						/* original 1-bit QD */
//	gestalt8BitQD				= 0x100,						/* 8-bit color QD */
//	gestalt32BitQD				= 0x200,						/* 32-bit color QD */
//	gestalt32BitQD11			= 0x201,						/* 32-bit color QDv1.1 */
//	gestalt32BitQD12			= 0x220,						/* 32-bit color QDv1.2 */
//	gestalt32BitQD13			= 0x230							/* 32-bit color QDv1.3 */
	
	if( !Has8BitQD ) {
		iml_error("mvl_video_manager","Must have at least 8 bit Quickdraw");
		if(ExitOnError) exit(1);
		} // end if
	
	// GRAB THE SYSTEM VERSION
	Gestalt(gestaltSystemVersion, &SystemOSVersion);
		// CHECK THE RESULT
	NewPaletteManagerPresent = (SystemOSVersion >= 0x605);
	IsSystemSeven = (SystemOSVersion >= 0x700);

	if( !IsSystemSeven ) {
		// for Microseconds() timing
		iml_error("mvl_video_manager","Timing requires at least System 7."); 
		if(ExitOnError) exit(1);
		} // end if

	// REMARKS AND FUNCTION CALL DIRECTLY FROM Pelli's RestoreCluts.c FILE:
	//		Somewhere near the beginning of your program insert the line:

	//AtExitToShell(RestoreCluts);

	//		This requests that RestoreCluts() be run when you quit, even if
	//		it's an abnormal exit, e.g. escape to shell from MacsBug.
	//		(AtExitToShell is equivalent to the non-ANSI _atexit() routine provided
	//		by the THINK C environment.)
} // end constructor


//=====================================================================
//
//=====================================================================
void mvl_video_manager::set_mvl_calibration(IML_FLOAT B, IML_FLOAT G,
					IML_FLOAT MinL, IML_FLOAT MaxL) {

	IML_CLIP_LO_NW(B,0.0);
	IML_CLIPIT_NW( G, 0.1, G, 10.0);
	set_mvl_calibration( mvl_calibration_polynomial(log(B), 1.0/G ), MinL, MaxL);
} // end set_mvl_calibration

	
//=====================================================================
//
//=====================================================================
void mvl_video_manager::set_mvl_calibration(polynomial IML_REF CalPoly,
					IML_FLOAT MinL, IML_FLOAT MaxL) {

	if(CalibrationPolynomial) delete CalibrationPolynomial;
	CalibrationPolynomial = new mvl_calibration_polynomial(CalPoly);
	if(!CalibrationPolynomial) {
		iml_error("mvl_video_manager","Could not allocate mvl_calibration_polynomial!");
		return;
		} // end if
	MinLuminance = MinL;
	MaxLuminance = MaxL;
//	set_mean_luminance(0.5);
//	set_max_contrast(1.0);

} // end set_mvl_calibration

//=====================================================================
//
//=====================================================================
void mvl_video_manager::best_b_plus_r_match(mvl_rgb & Dest, IML_FLOAT Fractional) {
	if(RStepsPerBDAC > 1.0) {
		Dest.blue  = ((IML_USINT)Fractional) << LeftShift;
		Dest.red   = ((IML_USINT)((Fractional-((IML_USINT)Fractional))*RStepsPerBDAC + 0.5)) << LeftShift;
		Dest.green = 0;
	} else {
		Dest.blue  = ((IML_USINT)(Fractional + 0.5)) << LeftShift;
		Dest.red   = ((IML_USINT)(Fractional + 0.5)) << LeftShift;
		Dest.green = ((IML_USINT)(Fractional + 0.5)) << LeftShift;
		}// end if
}

//=====================================================================
		// !!!! NOTE THAT THIS GUY WANTS FIRST-ENTRY, LENGTH AS ARGS!@!!@!#$$%$^
//=====================================================================
void mvl_video_manager::fill_dac(mvl_ColorSpec *CSpecs, IML_USINT Start, IML_USINT Length, 
				IML_FLOAT MeanL, IML_FLOAT Contrast) {

	polynomial &CalPoly = *CalibrationPolynomial;
	IML_FLOAT HalfLuminanceRange = 0.5 - iml_abs(0.5-MeanL);
	IML_FLOAT StartL = MeanL - Contrast*HalfLuminanceRange;
	IML_FLOAT DeltaL = 2*Contrast*HalfLuminanceRange/(Length-1.0);
	for(IML_USINT K=0; K < Length; K++) {
		IML_FLOAT TargetL = StartL + K * DeltaL;
//		IML_FLOAT Target_DAC_Value = Beta*pow(TargetL, GammaInv);
		IML_FLOAT Target_DAC_Value = (TargetL!=0.0) ? exp(CalPoly(log(TargetL))) : 0;
		mvl_rgb &RGB = CSpecs[Start + K].rgb;
		best_b_plus_r_match(RGB, Target_DAC_Value);
		} // end for

} // end fill_dac

//=====================================================================
//
//=====================================================================
void mvl_video_manager::flush_all_events(void) {
	FlushEvents(everyEvent,0);											// clear out the crap
} // end flush_all_events

//=====================================================================
//
//=====================================================================
mvl_video_manager::~mvl_video_manager() {

	// BE NICE AND CLEAR OUT EXTRANEOUS KEY PRESSES, ETC.
	flush_all_events();
	// A NICE FUNCTION SUPPLIED BY PELLI
	RestoreCluts();
	
	if( MacVideoInfo ) {
		// GRAB MAC VIDEO INFO STRUCTURE
		mvl_mac_video_information IML_PTR MVI = (mvl_mac_video_information*)MacVideoInfo;

		SetPort(MVI->OldPort);

		delete MVI;
		} // end if

} // end destructor


//=====================================================================
//
//=====================================================================
IML_LINT mvl_video_manager::set_dgp_calibration(const char * FileName) {
char NameBuffer[64];	// TEMPORARY SPACE, IF NEEDED
int NullFlag=FALSE;

	if(DisplayNumber < 0) {
		iml_error("mvl_video_manager","Please initialize a display device before loading a luminance record.");
		return(FALSE);
		} // end if

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;
	
	LuminanceRecord *LRPtr = new LuminanceRecord;			// details of luminance calibration
	if( !LRPtr ) {
		iml_error("mvl_video_manager","Cannot allocate the luminance record.");
		if(ExitOnError) exit(1);
		return(FALSE);
		} // end if

	// FIRST LOAD WITH DEFAULT . . .
	{ // BEGIN BLOCK
	LuminanceRecord IML_REF LR = *LRPtr;
	#include "default_dgp_calibration.h"	// a file containing monitor details and calibration 
	} // END BLOCK

	if(!FileName) {
		iml_warning("mvl_video_manager","Using default (compile time) luminance record.");
	} else if(  ReadLuminanceRecord( (char *)FileName, LRPtr, 0) <= 0 ) {
		iml_warning("mvl_video_manager","Cannot load the requested luminance record. Sticking with default.");
		//if(ExitOnError) exit(1);
		} // end else if

	// SAVE THE INFORMATION FOR LATER USE
	MVI.Calibration = LRPtr;

	if( RStepsPerBDAC == 0.0 ) {
		// IN THIS CASE WE WANT TO SET THE VIDEO-GUN GAINS SO THAT THE RED AND BLUE GUNS
		// REMAIN UNUSED. D. Pelli's STUFF USES THESE VALUES WHEN CONVERTING
		// REQUESTED LUMINANCE VALUES INTO LOOK-UP-TABLE VALUES (FOR USE WITH
		// THE ATTENUATOR).
		LRPtr->r=0.0;
		LRPtr->g=1.0;
		LRPtr->b=0.0;
		} // end if


	MinLuminance = LRPtr->LMin;
	MaxLuminance = LRPtr->LMax;
	
	set_mean_luminance(0.5);
//	set_max_contrast(0.0);

	// SetLuminances IS ONE OF Pelli's FUNCTIONS.
	// AS DEFAULT:
	// set all entries in current LUT to mean luminance, 
	// except for entry 255 and 0 which remain black and white
	SetLuminances(MVI.DisplayDevice, LRPtr, 1, CLUTSize-2, MeanLuminance, MeanLuminance);

	return(TRUE);

} // end set_dgp_calibration



//=====================================================================
//
//=====================================================================
void mvl_video_manager::set_mean_luminance(IML_DOUBLE MeanL) {

	IML_CLIPIT(MeanL, 0.0, MeanL, 1.0);
	MeanLFraction = MeanL;
	MeanLuminance = MinLuminance + MeanL*(MaxLuminance-MinLuminance);
} // end set_mean_luminance


//=====================================================================
//
//=====================================================================
void mvl_video_manager::set_default_clut_segment(IML_USINT StartC, IML_USINT CLength) {

	if( (StartC > CLUTSize-1) || (StartC+CLength) > CLUTSize) {
		iml_error("set_default_clut_segment","Invalid destination CLUT segment.");
		
		if(ExitOnError) {
			exit(1);
		} else {
			iml_error("set_default_clut_segment","Clipping the segment to be valid.");
			StartC = (StartC>CLUTSize-1) ? CLUTSize-1 : StartC;
			CLength = ((StartC+CLength) > CLUTSize) ? CLUTSize-StartC : CLength;
			} // end if
		
		} // end if

	StartCLUTSegment = StartC;
	CLUTSegmentLength = CLength;

}; 
		

//=====================================================================
//
//=====================================================================
void mvl_video_manager::linearize_clut(IML_DOUBLE MeanL, IML_DOUBLE MaxC,
	IML_USINT StartC, IML_USINT CLength) {
	
	if( (StartC > CLUTSize-1) || (StartC+CLength) > CLUTSize) {
		iml_error("linearize_clut","Invalid destination CLUT segment.");
		if(ExitOnError) exit(1);
		return;
		} // end if

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.Calibration || !MVI.DisplayDevice ) {
		iml_error("linearize_clut","Null ptr in MVI.");
		if(ExitOnError) exit(1);
		return;
		} // end if
		
	// GRAB THE LUMINANCE RECORD STRUCTURE
	LuminanceRecord * LRPtr = MVI.Calibration;
	
	switch(CalibrationType) {
	
		case DGP_CALIBRATION: {

			// CALCULATE THE START AND END LUMINANCE VALUES
			IML_DOUBLE DeltaLum = MeanL * MaxC;	
			IML_DOUBLE StartL   = MeanL-DeltaLum;
			IML_DOUBLE EndL     = MeanL+DeltaLum;
			
				// STORE THE CLUT TABLE IN THE LOCAL LUMINANCE RECORD
			SetLuminancesAndRange(NULL, LRPtr, StartC, CLength, StartL, EndL, LRPtr->LMin, LRPtr->LMax);
			}; break;
		
		default:
		case MVL_CALIBRATION: {
		// !!!! NOTE THAT THIS GUY WANTS FIRST-ENTRY, LENGTH AS ARGS!@!!@!#$$%$^
			fill_dac((mvl_ColorSpec *)LRPtr->table, StartC, CLength, MeanL, MaxC);
			}; break;
			
		} // end switch
	
	if(MVI.ThunderCard) {
		MVI.ThunderCard->wait_for_VBL();
		MVI.ThunderCard->load_dacs(LRPtr->table, StartC, CLength);
	} else {
		// SET THE DISPLAY DEVICE CLUT FROM THE ONE JUST GENERATED
		// !!!! NOTE THAT THIS GUY WANTS FIRST-ENTRY, LAST-ENTRY AS ARGS!@!!@!#$$%$^
		LoadLuminances(MVI.DisplayDevice, LRPtr, StartC, StartC+CLength-1);
		} // end else
} // end linearize_clut


//=====================================================================
//
//=====================================================================
mvl_clut IML_PTR mvl_video_manager::get_clut_copy(void) {

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.DisplayDevice ) {
		iml_error("mvl_video_manager","Null ptr in get_clut_copy.");
		if(ExitOnError) exit(1);
		return(NULL);
		} // end if

	mvl_clut *CLUTable = new mvl_clut(DacSize, CLUTSize);
//	mvl_ColorSpec *CSTable = CLUTable->iml_ColorSpecTable();
	mvl_ColorSpec *CSTable = &((*CLUTable)[0]);
	// !!! THIS GUY *claims to want* FIRST-ENTRY, COUNT, BUT COUNT HAS TO
	// BE LESS 1 IN ORDER TO AVOID AN ERROR MESSAGE!! AS ARGS!!@#$%@$^%
	long Err = GDGetEntries(MVI.DisplayDevice, 0, CLUTable->clut_size()-1, (ColorSpec*)CSTable);
	if(Err) {
		iml_error("mvl_video_manager","Error message from GDGetEntries.");
		if(ExitOnError) exit(1);
		return(NULL);
		} // end if
	
	return(CLUTable);
} // end get_clut_copy



	
//=====================================================================
//
//=====================================================================
void mvl_video_manager::set_default_clut(mvl_clut IML_REF CLUTable) {

	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.DisplayDevice ) {
		iml_error("mvl_video_manager","Null ptr in load_clut.");
		if(ExitOnError) exit(1);
		return;
		} // end if

	mvl_ColorSpec *CSTable = &CLUTable[0];
	
		// GRAB THE LUMINANCE RECORD STRUCTURE
	mvl_ColorSpec * LRPtr = (mvl_ColorSpec *)MVI.Calibration;
	
	// COPY THE INPUT CLUT TO THE ONE USED DURING DISPLAY.
	for(IML_LINT K = 0; K<CLUTSize; K++) LRPtr[K] = CSTable[K];

	load_clut(CLUTable);

} // end make_default_clut

//=====================================================================
//
//=====================================================================
void mvl_video_manager::load_clut(mvl_clut IML_REF CLUTable) {

	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.DisplayDevice ) {
		iml_error("mvl_video_manager","Null ptr in load_clut.");
		if(ExitOnError) exit(1);
		return;
		} // end if

	mvl_ColorSpec *CSTable = &CLUTable[0];
	
//	LoadLuminances(MVI.DisplayDevice, (LuminanceRecord *)CSTable, 0, CLUTable.clut_size() - 1);
	if(MVI.ThunderCard) {
		MVI.ThunderCard->wait_for_VBL();
		MVI.ThunderCard->load_dacs((ColorSpec *)CSTable, 0, CLUTable.clut_size() - 1);
	} else {
		// SET THE DISPLAY DEVICE CLUT FROM THE ONE JUST GENERATED
		// !!!! NOTE THAT THIS GUY WANTS FIRST-ENTRY, LAST-ENTRY AS ARGS!@!!@!#$$%$^
		LoadLuminances(MVI.DisplayDevice, (LuminanceRecord *)CSTable, 0, CLUTable.clut_size() - 1);
		} // end else

} // end load_clut


//=====================================================================
//
//=====================================================================
void mvl_video_manager::set_clut_entry(IML_SINT Entry, IML_SINT R, IML_SINT G, IML_SINT B) { 
	
	IML_CLIPIT(Entry,0,Entry,CLUTSize-1);
	
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.Calibration || !MVI.DisplayDevice ) {
		iml_error("mvl_video_manager","Null ptr in set_clut_entry.");
		if(ExitOnError) exit(1);
		return;
		} // end if

	mvl_ColorSpec *CSTable = (mvl_ColorSpec*)MVI.Calibration;
	
	CSTable[Entry].rgb.red		= (R << LeftShift);
	CSTable[Entry].rgb.green	= (G << LeftShift);
	CSTable[Entry].rgb.blue		= (B << LeftShift);
	
	if(MVI.ThunderCard) {
		MVI.ThunderCard->wait_for_VBL();
		MVI.ThunderCard->load_dacs((ColorSpec *)CSTable, Entry, Entry);
	} else {
		// SET THE DISPLAY DEVICE CLUT FROM THE ONE JUST GENERATED
		// !!!! NOTE THAT THIS GUY WANTS FIRST-ENTRY, LAST-ENTRY AS ARGS!@!!@!#$$%$^
		LoadLuminances(MVI.DisplayDevice, (LuminanceRecord *)CSTable, Entry, Entry);
		} // end else

} // end load_clut_entry

//=====================================================================
//
//=====================================================================
void mvl_video_manager::load_clut_entry(mvl_clut IML_REF CLUTable, IML_SINT Entry) { 
	
	IML_CLIPIT(Entry,0,Entry,CLUTSize-1);
	
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.DisplayDevice ) {
		iml_error("mvl_video_manager","Null ptr in load_clut_entry.");
		if(ExitOnError) exit(1);
		return;
		} // end if

	mvl_ColorSpec *CSTable = &CLUTable[0]; // access the table via operator[]
	if(MVI.ThunderCard) {
		MVI.ThunderCard->wait_for_VBL();
		MVI.ThunderCard->load_dacs((ColorSpec *)CSTable, Entry, Entry);
	} else {
		// SET THE DISPLAY DEVICE CLUT FROM THE ONE JUST GENERATED
		// !!!! NOTE THAT THIS GUY WANTS FIRST-ENTRY, LAST-ENTRY AS ARGS!@!!@!#$$%$^
		LoadLuminances(MVI.DisplayDevice, (LuminanceRecord *)CSTable, Entry, Entry);
		} // end else

} // end load_clut_entry


//=====================================================================
//
//=====================================================================
void mvl_video_manager::identity_clut(IML_UBYTE RorGorB,
	IML_USINT StartC, IML_USINT CLength, IML_FLOAT StartDac, IML_FLOAT DeltaDac) {

	if( (StartC > CLUTSize-1) || (StartC+CLength) > CLUTSize) {
		iml_error("identity_clut","Invalid destination CLUT segment.");
		if(ExitOnError) exit(1);
		return;
		} // end if

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.Calibration || !MVI.DisplayDevice ) {
		iml_error("mvl_video_manager","Null ptr in identity_clut.");
		if(ExitOnError) exit(1);
		return;
		} // end if
		
	// GRAB THE LUMINANCE RECORD STRUCTURE
	LuminanceRecord * LRPtr = MVI.Calibration;

	// FIRST CLUT IS WHITE, LAST IS BLACK (OR VISA-VERSA) AND WE LEAVE THEM BE.
	IML_LINT StartClut = StartC;
	IML_LINT EndClut = StartC+CLength-1;
	
	mvl_ColorSpec *CLUT = (mvl_ColorSpec*)LRPtr->table;
	// FIRST DAC VALUE FOR FIRST CLUT ENTRY
	double DacValue = StartDac;
	for(IML_LINT CEntry=StartClut; CEntry<=EndClut; 
		DacValue += DeltaDac, CEntry++ ) {
	
		// ROUND AND DISCRETIZE THE DAC VALUE
		IML_USINT IDacValue = long(DacValue+0.5);
		// CREATE THE APPROPRIATELY SHIFTED ENTRY
		mvl_rgb RGBEntry(	(RorGorB & RED_GUN)   ? (IDacValue<<LeftShift) : 0,
  							(RorGorB & GREEN_GUN) ? (IDacValue<<LeftShift) : 0,
  							(RorGorB & BLUE_GUN)  ? (IDacValue<<LeftShift) : 0
  					);
		// COPY OVER
		CLUT[CEntry].rgb = RGBEntry;			
		} // end for

	if(MVI.ThunderCard) {
		MVI.ThunderCard->wait_for_VBL();
		MVI.ThunderCard->load_dacs(LRPtr->table, StartClut, EndClut);
	} else {
		// SET THE DISPLAY DEVICE CLUT FROM THE ONE JUST GENERATED
		LoadLuminances(MVI.DisplayDevice, LRPtr, StartClut, EndClut);
		} // end else
} // end identity_clut


//=====================================================================
//
//=====================================================================
void mvl_video_manager::fill_window_relative(IML_FLOAT Luminance) {

	IML_CLIPIT(Luminance, 0.0, Luminance, 1.0);

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	// CLUT ENTRIES GO FROM 0 TO CLUTSize-1
	IML_LINT CLUT_Entry = (CLUTSize-1) * Luminance + 0.5; // MAKE SURE TO ROUND
	
	fill_window_absolute( CLUT_Entry );
	
} // end fill_window_relative
	
//=====================================================================
//
//=====================================================================
void mvl_video_manager::use_mvl_window_filling(IML_USINT Speed) {

	if(FillWindowID >= 0) return;
	
	IML_ULINT XSize, YSize;
	IML_CLIPIT_NW(XSize, 4, Speed, XResolution);
	IML_CLIPIT_NW(YSize, 4, Speed, YResolution);
	
	FillerImage = new iml_ubyte_image(XSize,YSize);
	if( !FillerImage->valid_memory() ) {
		iml_error("mvl_video_manager","Out of memory in use_mvl_window_clearing.");
		delete FillerImage;
		FillerImage = NULL;
		return;
		} // end if
	iml_ubyte_image & Filler = *FillerImage;
	memset( &Filler(0,0), (unsigned char)127, XSize*YSize);
	Filler = 127;
	FillWindowID = store_image_sequence(Filler);
	if( FillWindowID < 0 ) {
		iml_error("mvl_video_manager","Could not store filler image in use_mvl_window_clearing.");
		delete FillerImage;
		FillerImage = NULL;
		} // end if
		
	
} // end use_mvl_window_filling
	
//=====================================================================
//
//=====================================================================
void mvl_video_manager::fill_window_absolute(IML_SINT CLUT_Entry) {

	// CLUT ENTRIES GO FROM 0 TO CLUTSize-1
	IML_CLIPIT(CLUT_Entry, 0, CLUT_Entry, CLUTSize-1);

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if(FillWindowID < 0) {
		//  set to requested luminance
		PmBackColor( CLUT_Entry );

		// EraseRect IS A MAC-OS CALL
		// wipe display clean, removing any open windows etc.
		EraseRect( &(((WindowPtr)MVI.DisplayWindow)->portRect) );
	} else {

		iml_ubyte_image & Filler = *FillerImage;
		IML_ULINT XSize = Filler.hyperplane_dimension_size(0);
		IML_ULINT YSize = Filler.hyperplane_dimension_size(1);
		memset( &Filler(0,0), (IML_UBYTE)CLUT_Entry, XSize*YSize);
		long FillerID = update_image_sequence(Filler, FillWindowID);

		if(FillerID != FillWindowID) return;

		long XCnt = XResolution/XSize + ((XResolution%XSize) ? 1 : 0);
		long YCnt = YResolution/YSize + ((YResolution%YSize) ? 1 : 0);
		long Count = XCnt*YCnt;
		iml_ulint_sequence BlitOrder(Count);
		BlitOrder.generate_linear_sequence(0,1);
		BlitOrder.shuffle();
		point_2d BlitPos;
		for(long K=0; K<Count; K++) {
			BlitPos.X = XSize*(BlitOrder(K)%XCnt);
			BlitPos.Y = YSize*(BlitOrder(K)/XCnt);
			display_frame(FillerID, 0, BlitPos);
			} // end for
		}// end else

} // end fill_window_absolute


//=====================================================================
//
//=====================================================================
void mvl_video_manager::set_attenuator_red_steps( IML_FLOAT RedSteps) {

	if(RedSteps < 0) RedSteps = 0;
	if(RedSteps > MaxDac) RedSteps = MaxDac;
	RStepsPerBDAC = RedSteps;
	
} // end set_attenuator_red_steps



//=====================================================================
//
//=====================================================================
void mvl_video_manager::initialize_screen(IML_SINT DispNumber,  IML_SINT ThunderCardVideo) {

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	// SAVE FOR FUTURE REFERENCE
	DisplayNumber = DispNumber;
	GDHandle GDevice = GetScreenDevice(DisplayNumber);
	
	if( GDevice == NULL ) {
		iml_error("mvl_video_manager","Sorry, can't grab the requested video device.");
		if(ExitOnError) exit(1);
		} // end if

	// SAVE THIS AS A struct ** SO THE HEADER FILE DOES NOT INCLUDE ALL THE VIDEO CRAP.
	// TO USE IT LATE WE WILL TYPE-CAST.
	MVI.DisplayDevice = GDevice;

	// CREATE THE APPROPRIATE INFO
	if(ThunderCardVideo) MVI.ThunderCard = new pci_thunder_30_1600_card(GDevice);

	// A RETURN OF 0 MEANS A MONOCHROME DEVICE; 1 MEANS COLOR.
	DeviceIsColor = TestDeviceAttribute(GDevice, gdDevType);
	// GRAB THE CURRENT PIXEL SIZE FOR THE SCREEN
	InitialPixelSize=(*(*GDevice)->gdPMap)->pixelSize;
	
	IML_LINT WhichFlagBit = 1<<gdDevType;
	
	IML_LINT ModeNum = HasDepth(GDevice, 8, WhichFlagBit, 1);
	if( !ModeNum ) {
		iml_error("mvl_video_manager","Requested device cannot do 8 bits/pixel.");
		if(ExitOnError) exit(1);
		} // end if

	// NOTE THAT IF WE ARE ALREADY IN 8 PIXEL MODE THEN WE DON'T DO ANYTHING.
	// THIS TEST OF OldIsColor LOOKS WRONG.  ASK PETE ABOUT IT. . .
//	if(NewPaletteManager && (OldPixelSize!=8 || !OldIsColor)) {
	if(NewPaletteManager && ( (InitialPixelSize!=8) || !DeviceIsColor) ) {
		// SET DEPTH IS A MAC-OS CALL DESCRIBED IN Pallete.h
		// PARAMS ARE (1) GDevice HANDLE, 
		// (2) PIXEL DEPTH (8) 
		// === NB:  USING MODENUMBER GOTTEN FROM HasDepth CAUSES A CRASH, EVEN THOUGH IT SHOULD
		// ===      BE ALLOWED!  THE CRASH OCCURS WHEN TRYING TO ALLOCATE THE WINDOW BEL0W.
		// (3) THE MASK FOR THE BIT SPECIFYING COLOR/MONOCHROME, 
		// AND (4) REQUESTED MODE (1==COLOR, 0 == MONOCHROME)
// THIS IS PETE'S CALL:  SetDepth(GDevice,8,1<<gdDevType, 1);
//	CAUSES CRASH:: SetDepth(GDevice, ModeNum, WhichFlagBit, 1);
		IML_LINT Failure = SetDepth(GDevice, 8 /*pixel bit-depth*/, WhichFlagBit, 1);
		if( Failure ) {
			iml_error("mvl_video_manager","Sorry, failure in SetDepth() call.");
			if(ExitOnError) exit(1);
			return; // avoid a crash when creating a window
			} // end if
		} // end if
	
	// TEST IF THE REQUEST WAS SUCCESSFUL.
	if( (*(*GDevice)->gdPMap)->pixelSize!=8 ) {
		iml_error("mvl_video_manager","Sorry, can't seem to get 8 bits/pixel.");
		if(ExitOnError) exit(1);
		} // end if

	if(ThunderCardVideo) {
		DacSize = 10;
	} else {
		DacSize = GDDacSize(GDevice);
		} // end if
	
	// NOW WE KNOW WHAT THE MAX DAC VALUE IS
	MaxDac = (0x0001<<DacSize)-1;
	
	// ASK THE HARDWARE FOR THE SIZE OF THE CLUT (COLOR LOOK-UP TABLE).
	CLUTSize=GDClutSize(GDevice);

	// THE DEFAULT IS NOT TO OVERWRITE THE FIRST AND LAST CLUT
	// ENTRIES BECAUSE THEY HOLD BLACK AND WHITE FOR TEXT DISPLAY.
	set_default_clut_segment(1, CLUTSize-2);
	
	// DAC VALUES ARE LEFT JUSTIFIED IN THE BIT REPRESENTATION.
	// THIS CALCULATION ASSUMES R,G,B ARE SAME SIZE.
	LeftShift = (sizeof(mvl_rgb)/3)*8 - DacSize;
	
	// IN SYSTEM 7.6 THE DEFAULT IS NOT TO WAIT FOR VBL; SET IT BACK TO DEFAULT
	if(SystemOSVersion >= 0x0760) {
		Boolean DontWaitForVBL;
		double NanoSeconds;
		//OSErr GDSetDelay(GDHandle device,Boolean dontWaitForVBL,double nanoseconds);
		//OSErr GDGetDelay(GDHandle device,Boolean *dontWaitForVBLPtr,double *nanosecondsPtr);
		GDGetDelay(GDevice, &DontWaitForVBL, &NanoSeconds);
		GDSetDelay(GDevice, FALSE, NanoSeconds);
		} // end if


	// FUNCTION CALL (GDOpenWindow) IS ONE OF PELLI'S ROUTINES (GDOpenWindow.c).
	// THE GDOpenWindow OPENS A WINDOW THE SIZE OF THE GDevice DISPLAY.  THAT WINDOW
	// HAS AN INDEPENDENT COLOR MAP CREATED FOR IT RATHER THAN SHARING THAT OF THE
	// ACTUAL DEVICE (BACKGROUND DISPLAY?).  THAT CLUT IS SET TO A GRAY MAP.  THIS
	// CWindow IS A POINTER TO "GRAPHICS CONTEXT" INFORMATION DEFINED BY Apple.
	// WHEN THIS IS WRITTEN GDOpenWindow ALSO KILLED THE VIDEO CARD GAMMA
	// CORRECTION TABLE; I.E. LINEARIZED IT WITH A CALL TO GDUncorrectedGamma(device);
	CWindowPtr CWindow = GDOpenWindow(GDevice);
	
	// RETURN TOP-LEFT CORNER IN GLOBAL SYSTEM COORDINATES
	Rect RR = (*GDevice)->gdRect;
	GlobalWindowCoords.X = RR.left;
	GlobalWindowCoords.Y = RR.top;

	// SAVE FOR LATER USE
	MVI.DisplayWindow = CWindow;

	// SAVE THE WINDOW RESOLUTION IN THE X AND Y DIMENSIONS
	XResolution = MVI.DisplayWindow->portRect.right;
	YResolution = MVI.DisplayWindow->portRect.bottom;

	// GRAB OFFSCREEN PIXMAP ADDRESS
	PixMap **PixMapHandle = ((CGrafPtr)CWindow)->portPixMap;	

	// SAVE FOR LATER USE
	MVI.DisplayPixMapHandle = PixMapHandle;
	
	// MoveHHi IS A MAC-OS CALL
	// MOVE THE RELOCATABLE MEMORY BLOCK TO THE TOP OF THE APPLICATION HEAP SO THAT
	// WHEN WE LOCK IT DOWN (BELOW) IT WILL CAUSE LESS (HOPEFULLY) MEMORY
	// FRAGMENTATION PROBLEMS.
	MoveHHi((Handle) PixMapHandle);

	// HLock IS A MAC-OS CALL
	// HLock LOCKS DOWN THE RELOCATABLE BLOCK SO THAT WE CAN READ/WRITE IT DIRECTLY.
	HLock((Handle) PixMapHandle);

	// SAVE THE ORIGINAL GRAPHICS PORT FOR RESTORATION IN THE DESTRUCTOR
	WindowPtr OldPort;
	GetPort(&OldPort);
	MVI.OldPort = OldPort;

	// SetPort IS A MAC-OS CALL
	// DO EVERYTHING IN THE DISPLAY WINDOW WE CREATED
	SetPort((WindowPtr)CWindow);

	// FOR NOW, USE A DEFAULT CALIBRATION
	set_dgp_calibration(NULL); //use PELLI default UNTIL USER DECIDES TO LOAD SOMETHING ELSE
	set_mvl_calibration( MaxDac, 2.0, 0, 100);
	set_attenuator_red_steps( 0.0 ); // 0 means no attenuator
	
	// BUT DEFAULT TO MVL
	use_mvl_calibration();


} // end initialize_screen



//=====================================================================
//
//=====================================================================
void mvl_video_manager::set_up_SIOUX(void) {

	SIOUXSettings.toppixel=LMGetMBarHeight()+20;
	SIOUXSettings.leftpixel=10;
	SIOUXSettings.rows=10;
	SIOUXSettings.autocloseonquit=0;
	SIOUXSettings.showstatusline=0;
	SIOUXSettings.asktosaveonclose=0;
	printf("\n");
	
} // end set_up_SIOUX


//=====================================================================
//
//=====================================================================
IML_LINT mvl_video_manager::extend_stack(IML_ULINT Extension) {

	IML_LINT OldStackSize = StackSpace();
	SetApplLimit( GetApplLimit()-Extension );
	StackSize = StackSpace();
	
	return(StackSize-OldStackSize);
} // end extend_stack



//=====================================================================
//
//=====================================================================
void mvl_video_manager::initialize_CLUT_tables(mvl_temporal_window IML_REF TWin) {

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !TWin.MVLColorSpecTables || !MVI.Calibration ) {
		iml_error("mvl_video_manager","Null ptr in initialize_CLUT_tables.");
		if(ExitOnError) exit(1);
		return;
		} // end if
		
	ColorSpecTable *CLUT_Tables = (ColorSpecTable *)TWin.MVLColorSpecTables;
		// STORE THE CLUT TABLE IN THE LOCAL LUMINANCE RECORD
	LuminanceRecord * LRPtr = MVI.Calibration;

	// CALCULATE A SEQUENCE OF CLUT TABLES WHICH CORRESPOND TO THE TEMPORAL WINDOW
	// REPRESENTED BY THE 1-D FLOAT IMAGE IN TWin
	for(IML_USINT I=0; I< TWin.Length; I++) { 	

		// DO THIS EXPLICITLY FOR CLARITY
		IML_DOUBLE WinContrast = TWin[I];
		switch(CalibrationType) {
		
			case DGP_CALIBRATION: {

				IML_DOUBLE DeltaLum = MeanLuminance * WinContrast;	
				IML_DOUBLE StartL   = MeanLuminance-DeltaLum;
				IML_DOUBLE EndL     = MeanLuminance+DeltaLum;
				// use 254 (256 minus 1st and last) entries to set the optimal 
				// brightness range for require contrast (pseudo-14 bits)
				
				SetLuminancesAndRange(NULL, LRPtr, StartCLUTSegment, 
							CLUTSegmentLength, StartL, EndL, LRPtr->LMin, LRPtr->LMax);
				}; break;
			
			default:
			case MVL_CALIBRATION: {
		// !!!! NOTE THAT THIS GUY WANTS FIRST-ENTRY, LENGTH AS ARGS!@!!@!#$$%$^
				fill_dac((mvl_ColorSpec *)LRPtr->table, StartCLUTSegment, 
							CLUTSegmentLength, MeanLFraction, WinContrast);
				}; break;
				
			} // end switch

			CLUT_Tables[I] = *((ColorSpecTable *)LRPtr->table);

		}// end for i
		
} // end initialize_CLUT_tables



//=====================================================================
//
//=====================================================================
void mvl_video_manager::draw_fixation_marker(point_2d IML_REF Pos, IML_LINT Size){
// draw fixation cross centred on x, y, of size
		MoveTo(Pos.X-Size/2,Pos.Y);
		LineTo(Pos.X+Size/2,Pos.Y);
		MoveTo(Pos.X,Pos.Y-Size/2);
		LineTo(Pos.X,Pos.Y+Size/2);
} // end draw_fixation_point




//=====================================================================
//
//=====================================================================
IML_SINT mvl_video_manager::store_image_sequence(iml_ubyte_image IML_REF Image) {

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.video_storage_left() ) {
		iml_error("mvl_video_manager","Image storage is full already.");
		return(-1);
		} // end if
		
	// GRAB A COPY OF THE SUBSPACE DEFINITION
	iml_index IML_REF SubSpace = Image.get_indexer(); 
	if( (SubSpace.Dimensionality < 2) || (SubSpace.Dimensionality > 3) ) {
		iml_error("mvl_video_manager","Stored image must be 2 or 3 dimensional.");
		return(-1);
		} // end if
		
	IML_USINT ColCount   = SubSpace.size(IML_2ST_X);
	IML_USINT RowCount   = SubSpace.size(IML_2ST_Y);
	IML_USINT FrameCount = ((SubSpace.Dimensionality == 2) ? 1 : SubSpace.size(IML_2ST_T));
	
	// ALLOCATE THE OFFSCREEN VIDEO MEMORY
	mvl_video_image_sequence IML_PTR NewSeq = 
					new mvl_video_image_sequence( ColCount, RowCount, FrameCount);
	if( !NewSeq || !NewSeq->successful() ) {
		if( NewSeq ) delete NewSeq;
		return(-1);
		} // end if
	
	// GRAB FIRST BYTE OF THE IMAGE SUBSPACE AND STORE IT IN THE OFFSCREEN VIDEO MEMORY
	IML_SINT Success = NewSeq->store_all_bytes( &Image[SubSpace.indexize()] );
	delete &SubSpace; // DON'T NEED THIS ANY MORE
	if( !Success ) { delete NewSeq; return(-1); }  // end if

	IML_SINT ImSeqID = MVI.store_image_sequence(NewSeq);
	return( ImSeqID );

} // end store_image_sequence


//=====================================================================
//
//=====================================================================
IML_SINT mvl_video_manager::update_image_sequence(iml_ubyte_image IML_REF Image, IML_USINT ImSeqID) {

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.valid_video_id(ImSeqID) ) {
		iml_error("mvl_video_manager","Invalid ImageID in update_image_sequence.");
		return(-1);
		} // end if
		
	// GRAB A COPY OF THE SUBSPACE DEFINITION
	iml_index IML_REF SubSpace = Image.get_indexer(); 
	if( (SubSpace.Dimensionality < 2) || (SubSpace.Dimensionality > 3) ) {
		iml_error("mvl_video_manager","Stored image must be 2 or 3 dimensional.");
		return(-1);
		} // end if
		
	IML_USINT XSize   = SubSpace.size(IML_2ST_X);
	IML_USINT YSize   = SubSpace.size(IML_2ST_Y);
	IML_USINT FrameCount = ((SubSpace.Dimensionality == 2) ? 1 : SubSpace.size(IML_2ST_T));
	
	// ALLOCATE THE OFFSCREEN VIDEO MEMORY
	mvl_video_image_sequence IML_PTR ImageSequence = 
			(mvl_video_image_sequence IML_PTR) MVI.video(ImSeqID);
	if( !ImageSequence ) return(-1);
		
	if( ImageSequence->FramePositions[0].right  != XSize ||
		ImageSequence->FramePositions[0].bottom != YSize ||
		ImageSequence->FrameCount != FrameCount		) {
		iml_error("mvl_video_manager","Mismatched image dimensions in update_image_sequence.");
		return(-1);
		}// end if
	
	// GRAB FIRST BYTE OF THE IMAGE SUBSPACE AND STORE IT IN THE OFFSCREEN VIDEO MEMORY
	IML_SINT Success = ImageSequence->store_all_bytes( &Image[SubSpace.indexize()] );
	delete &SubSpace; // DON'T NEED THIS ANY MORE
	if( !Success ) { 
		iml_error("mvl_video_manager","Could not update image properly in update_image_sequence.");
		return(-1);
		}  // end if

	return( ImSeqID );

} // end update_image_sequence


//=====================================================================
//
//=====================================================================
void mvl_video_manager::clear_blit_sequences(void) {
	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;
	MVI.clear_blit_sequences();
} // end clear_blit_sequences

//=====================================================================
// FOR MAKING GROUPS OF SEQUENCES
//=====================================================================
IML_SINT mvl_video_manager::add_to_group(IML_USINT BlitSeqID, IML_USINT GroupID) {
	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;
	if( !MVI.valid_blit_id(BlitSeqID) ) {
		iml_error("add_to_group","Invalid blit sequence id.");
		return(FALSE);
		} // end if

	if(GroupID >= GroupCount) {
		iml_error("add_to_group","GroupID is out of range.");
		return(FALSE);
		} // end if
	
	// FOR CONVENIENCE
	mvl_blit_sequence_group IML_REF BGroup = *BlitGroups[GroupID];
	
	if(BGroup.ElementCnt >= BGroup.MaxElementCnt) {
		iml_error("add_to_group","GroupID is already full!");
		return(FALSE);
		} // end if
		
	BGroup.blit(BGroup.ElementCnt) = BlitSeqID;
	BGroup.ElementCnt++;
	return(TRUE);
} // end add_to_group


//=====================================================================
// FOR POSITIONS OF IMAGES ON THE SCREEN
//=====================================================================
IML_USINT mvl_video_manager::get_display_array_size(iml_image IML_REF Image, IML_USINT Spacing) {
	IML_USINT XSize = Image.dimension_size(IML_2D_X);
	IML_USINT YSize = (Image.dimensionality() < 2) ? 1 : Image.dimension_size(IML_2D_Y);
	return( get_display_array_size(XSize, YSize, Spacing) );
};

IML_USINT mvl_video_manager::get_display_array_size(
		IML_USINT XSize, IML_USINT YSize, IML_USINT Spacing) {

	long XS=XResolution-Spacing, XD=XSize+Spacing;
	long YS=YResolution-Spacing, YD=YSize+Spacing;
	long XCnt=0,YCnt=0; 
	while(XS>=XD) {XCnt++;XS-=XD;}
	while(YS>=YD) {YCnt++;YS-=YD;}
	if(XCnt==0)XCnt=1; if(YCnt==0)YCnt=1;
	return(XCnt*YCnt);
}; // end get_display_array_size



//=====================================================================
// FOR GETTING POSITIONS OF IMAGES ON THE SCREEN
//=====================================================================
point_2d mvl_video_manager::get_grid_size(iml_image IML_REF Image, IML_USINT Spacing) {
	IML_USINT XSize = Image.dimension_size(IML_2D_X);
	IML_USINT YSize = (Image.dimensionality() < 2) ? 1 : Image.dimension_size(IML_2D_Y);
	return( get_grid_size(XSize, YSize, Spacing) );
};

point_2d mvl_video_manager::
		get_grid_size(IML_USINT XSize, IML_USINT YSize, IML_USINT Spacing) {
static point_2d P;
	long XS=XResolution-Spacing, XD=XSize+Spacing;
	long YS=YResolution-Spacing, YD=YSize+Spacing;
	long XCnt=0,YCnt=0; 
	while(XS>=XD) {XCnt++;XS-=XD;}
	while(YS>=YD) {YCnt++;YS-=YD;}
	if(XCnt==0)XCnt=1; if(YCnt==0)YCnt=1;
	P=point_2d(XCnt,YCnt);
	return(P);
}

point_2d mvl_video_manager::
		get_screen_position(IML_SINT ScreenPosition, iml_image IML_REF Image, IML_USINT Spacing) {
	IML_USINT XSize = Image.dimension_size(IML_2D_X);
	IML_USINT YSize = (Image.dimensionality() < 2) ? 1 : Image.dimension_size(IML_2D_Y);
	return( get_screen_position(ScreenPosition, XSize, YSize, Spacing) );
};

point_2d mvl_video_manager::get_screen_position(
		IML_SINT ScreenPosition, IML_USINT XSize, IML_USINT YSize, IML_USINT Spacing) {
static point_2d P;

	switch(ScreenPosition) {
		case CENTER: {
			P=point_2d( (XResolution-XSize)/2, (YResolution-YSize)/2 );
			}; break;
		case UPPER_LEFT: {
			P=point_2d( 0, 0);
			}; break;
		case UPPER_RIGHT: {
			P=point_2d( XResolution-XSize, 0);
			}; break;
		case LOWER_LEFT: {
			P=point_2d( 0,  YResolution-YSize);
			}; break;
		case LOWER_RIGHT: {
			P=point_2d( XResolution-XSize,  YResolution-YSize);
			}; break;
		default: {
			P=  get_grid_size(XSize,YSize,Spacing);
			long YP=ScreenPosition/P.X, XP=ScreenPosition-YP*P.X;
			P = point_2d(Spacing*(XP+1)+XP*XSize, Spacing*(YP+1)+YP*YSize);
			}; break;
		} // end switch
		
	if(P.X < 0) P.X=0;
	if(P.X >= XResolution) P.X = 0;
	if(P.Y < 0) P.Y=0;
	if(P.Y >= YResolution) P.Y = 0;
	return(P);
}; // end get_screen_position


point_2d mvl_video_manager::
		get_screen_position(point_2d & XY, iml_image IML_REF Image, IML_USINT Spacing) {
	IML_USINT XSize = Image.dimension_size(IML_2D_X);
	IML_USINT YSize = (Image.dimensionality() < 2) ? 1 : Image.dimension_size(IML_2D_Y);
	return( get_screen_position(XY, XSize, YSize, Spacing) );
};

point_2d mvl_video_manager::get_screen_position(
		point_2d & XY, IML_USINT XSize, IML_USINT YSize, IML_USINT Spacing) {
static point_2d P;

	P = get_grid_size(XSize,YSize,Spacing);
	P = point_2d(Spacing*(XY.X+1)+XY.X*XSize, Spacing*(XY.Y+1)+XY.Y*YSize);

	if(P.X < 0) P.X=0;
	if(P.X >= XResolution) P.X = 0;
	if(P.Y < 0) P.Y=0;
	if(P.Y >= YResolution) P.Y = 0;
	return(P);
}; // end get_screen_position


//=====================================================================
// FOR CLEARING GROUPS OF SEQUENCES
//=====================================================================
IML_SINT mvl_video_manager::clear_group(IML_SINT GroupID) {
	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if(GroupID >= GroupCount) {
		iml_error("add_to_group","GroupID is out of range.");
		return(FALSE);
		} // end if
	
	if( GroupID < 0 ) { // CLEAR ALL GROUPS
		for( long G=0; G<GroupCount; G++)  BlitGroups[G]->clear_group();
	} else {
		BlitGroups[GroupID]->clear_group();
		} // end else
	
	return(TRUE);
} // end clear_group


//=====================================================================
//
//=====================================================================
void mvl_video_manager::clear_image_sequences(void) {
	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;
	MVI.clear_video_sequences();
} // end clear_image_sequences


//=====================================================================
//
//=====================================================================
IML_SINT mvl_video_manager::create_blit_sequence(IML_USINT ImSeqID, 
			point_2d IML_REF DestPos, iml_imageframe_sequence IML_REF FrameSeq) {

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.blit_storage_left() ) {
		iml_error("mvl_video_manager","Blit storage is full already.");
		return(-1);
		} // end if

	if( !MVI.valid_video_id(ImSeqID) ) {
		iml_error("mvl_video_manager","Invalid image id in create_blit_sequence.");
		return(-1);
		} // end if
	
	mvl_blit_sequence *NewSeq = 
		(mvl_blit_sequence *)allocate_blit_sequence( ImSeqID, DestPos, FrameSeq);

	if( !NewSeq ) return (-1);
	
	IML_SINT BltSeqID = MVI.store_blit_sequence(NewSeq);
	
	return( BltSeqID );
	
} // end create_blit_sequence

//=====================================================================
//
//=====================================================================
void * mvl_video_manager::allocate_blit_sequence(IML_USINT ImSeqID, 
			point_2d IML_REF DestPos, iml_imageframe_sequence IML_REF FrameSeq) {

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	// GET THE SEQUENCE LENGTH
	IML_USINT Length = FrameSeq.length();

	if(Length < 1 || Length > MAX_BLITS) {
		iml_warning("mvl_video_manager","Sequence length out of range [1,MAX_BLITS]; clipping.");
		IML_CLIPIT(Length, 1,Length, MAX_BLITS);
		} // end if
	
	// GRAB THE VIDEO SEQUENCE POINTER
	const mvl_video_image_sequence IML_REF VidSeq = *MVI.video(ImSeqID);
	
	// ALLOCATE A NEW BLIT SEQUENCE OBJECT OF THE APPROPRIATE LENGTH
	mvl_blit_sequence IML_PTR NewSeq = //      DST PIXMAP          SRC PIXMAP
				new mvl_blit_sequence(Length, MVI.DisplayPixMapHandle, VidSeq.SequencePixMapHandle);
	if( !NewSeq || !NewSeq->successful() ) {
		if(NewSeq) delete NewSeq;
		return(NULL);
		} // end if
		
	Rect * SrcRects = NewSeq->SourceRect;
	Rect * DstRects = NewSeq->DestinationRect;
	
	// FIRST FRAME RECT HAS CORRECT SIZE FOR FRAME
	Rect Dest = VidSeq.FramePositions[0];
	Dest.top  = DestPos.Y; 
	Dest.left = DestPos.X;
	Dest.bottom += DestPos.Y;
	Dest.right  += DestPos.X;
	
	// USE THE BlitPtr->Length BECAUSE IT HAS BEEN CLIPPED TO BOUNDS
	for(IML_USINT I=0; I< NewSeq->Length; I++) {
		// FOR NOW, THE DEST IS ALWAYS THE SAME
		DstRects[I] = Dest;
		// THE SOURCE RECTS ARE THE PERMUTED FRAME POSITIONS.
		SrcRects[I] = VidSeq.FramePositions[FrameSeq[I]];
		} // end for

	return(NewSeq);

} // end allocate_blit_sequence

//=====================================================================
//
//=====================================================================
IML_SINT mvl_video_manager::display_frame(IML_USINT ImSeqID, IML_USINT Frame, 
												point_2d IML_REF DestPos) {

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;
	if( !MVI.valid_video_id(ImSeqID) ) {
		iml_error("mvl_video_manager","Invalid image id in display_frame.");
		return(FALSE);
		} // end if

	// GRAB THE VIDEO SEQUENCE POINTER
	const mvl_video_image_sequence IML_REF VidSeq = *MVI.video(ImSeqID);
	
	if( Frame >= VidSeq.FrameCount ) {
		iml_error("mvl_video_manager","Invalid frame id in display_frame.");
		return(FALSE);
		} // end if
	
	// FIRST FRAME RECT HAS CORRECT SIZE FOR FRAME
	Rect DestRect = VidSeq.FramePositions[0];
	DestRect.top  = DestPos.Y; 
	DestRect.left = DestPos.X;
	DestRect.bottom += DestPos.Y;
	DestRect.right  += DestPos.X;

		// copy from the source PixMap to the destination window, 
		// the source rect, to the destination area 
		// [without interpolation (re-sizing) if the rects are the same size]
	CopyBitsQuickly(
			(BitMap *)(*VidSeq.SequencePixMapHandle),		(BitMap *)*MVI.DisplayPixMapHandle,
			&VidSeq.FramePositions[Frame],					&DestRect,
			srcCopy, NULL
			);

	return(TRUE);
} // end display_frame


//=====================================================================
//
//=====================================================================
void *mvl_video_manager::calculate_CLUT_ptrs(IML_USINT PlayLength) {

	// ALLOCATE SOME MEMORY
	ColorSpecTable **CLUTPtrs = new ColorSpecTable *[PlayLength];
	
	// CHECK FOR ALLOCATION ERROR
	if(!CLUTPtrs) return(NULL);

	if( CLUT_Sequence ) {

		if( PlayLength != CLUT_Sequence->Length ) {
			iml_warning("mvl_video_manager","Play length and temporal window length differ.  REPEATING!");
//			if( CLUT_Sequence->Length < BlitSeq.Length ) PlayLength = CLUT_Sequence->Length;
			} // end if

		ColorSpecTable *CLUTSeq = (ColorSpecTable*)CLUT_Sequence->MVLColorSpecTables;
		IML_ULINT Wrap = CLUT_Sequence->Length;
		for(IML_ULINT I=0; I<PlayLength; I++) 
			CLUTPtrs[I] = &CLUTSeq[ I%Wrap ];
		
	} else { // DEFAULT TO THE CURRENTLY SET CLUT
		// GRAB MAC VIDEO INFO STRUCTURE
		mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;
		LuminanceRecord * DefaultLRPtr = MVI.Calibration;
		for(IML_ULINT I=0; I<PlayLength; I++) 
			CLUTPtrs[I] = (ColorSpecTable*)DefaultLRPtr->table;	
		} // end else
		
	return(CLUTPtrs);
} // end calculate_CLUT_ptrs


//=====================================================================
//
//=====================================================================
IML_FLOAT mvl_video_manager::play_blit_sequence(IML_USINT BlitSeqID) {


	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;
	if( !MVI.valid_blit_id(BlitSeqID) ) {
		iml_error("play_blit_sequence","Invalid blit sequence id in play_blit_sequence.");
		return(0.0);
		} // end if

	// GRAB THE BLIT SEQUENCE INFORMATION AND PLAY IT
	return(play_blit_sequence( (void*)MVI.blit(BlitSeqID) ) );

} // end play_blit_sequence




//=====================================================================
//
//=====================================================================
IML_FLOAT mvl_video_manager::play_blit_sequence(void *BS) {

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	mvl_blit_sequence & BlitSeq = *(mvl_blit_sequence*)BS;
	IML_USINT PlayLength = BlitSeq.Length;

	ColorSpecTable **CLUTPtrs = (ColorSpecTable **)calculate_CLUT_ptrs(PlayLength);

	if( !CLUTPtrs ) {
		iml_error("play_blit_sequence","Cannot allocate CLUT ptr memory.");
		return(0.0);
		} // end if

	// LOCK DOWN THE DESTINATION PIXELS IN MEMORY
//	LockPixels(BlitSeq.SourcePM); 
//	LockPixels(BlitSeq.DestinationPM); 
				
	// UnSignedWide IS A STRUCT WITH TWO UNSIGNED LONG INTS. SEE Types.h (MAC-OS)
	UnsignedWide StartTicks, EndTicks;

	// synch for last image frame time
//	LoadLuminances(MVI.DisplayDevice, (LuminanceRecord*)CLUTPtrs[0], 1, CLUTSize-2);

	// A MAC-OS CALL  --   TO TIME THE PRESENTATION
	Microseconds(&StartTicks); 
	
	for(IML_USINT I=0; I < PlayLength; I++) {			

		if(MVI.ThunderCard) {
			MVI.ThunderCard->wait_for_VBL();
			MVI.ThunderCard->load_dacs((ColorSpec*)CLUTPtrs[I], 1, CLUTSize-2);
		} else {
			// synch image copying with LUT update and change LUT every frame for contrast windowing
			LoadLuminances(MVI.DisplayDevice, (LuminanceRecord*)CLUTPtrs[I], 1, CLUTSize-2);
			}

		// copy from the source PixMap to the destination window, 
		// the source rect, to the destination area 
		// [without interpolation (re-sizing) if the rects are the same size]
		CopyBitsQuickly(
			(BitMap *)*BlitSeq.SourcePM,		(BitMap *)*BlitSeq.DestinationPM,
			&BlitSeq.SourceRect[I],				&BlitSeq.DestinationRect[I],
			srcCopy, NULL
			);
			
		} // end for

//	// synch for last image frame time
//	LoadLuminances(MVI.DisplayDevice, (LuminanceRecord*)CLUTPtrs[PlayLength-1], 1, CLUTSize-2);

	// A MAC-OS CALL  --   TO END PRESENTATION TIMING
	Microseconds(&EndTicks);

	// UNLOCK THE DESTINATION PIXELS IN MEMORY
//	UnlockPixels(BlitSeq.SourcePM); 
//	UnlockPixels(BlitSeq.DestinationPM); 

	delete [] CLUTPtrs;

	// CALCULATE THE TOTAL TIME TAKEN
	IML_DOUBLE TotalTime = (EndTicks.lo-StartTicks.lo);
	// 4294967296.0 IS 2^32 IN DOUBLE FORMAT; *MUST* BE A DOUBLE :)
	TotalTime += 4294967296.0*(EndTicks.hi-StartTicks.hi);
	TotalTime /= 1e6;

	// report the time taken to present images
	// this allows the user to compare it with expected time to catch
	// "missed" frames (if any)
	return((IML_FLOAT)TotalTime);


} // end play_blit_sequence


//=====================================================================
//
//=====================================================================
IML_FLOAT mvl_video_manager::play_blit_sequence(IML_USINT ImSeqID, 
			point_2d IML_REF DestPos, iml_imageframe_sequence IML_REF FrameSeq) {

	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;

	if( !MVI.blit_storage_left() ) {
		iml_error("mvl_video_manager","Blit storage is full already.");
		return(-1);
		} // end if

	if( !MVI.valid_video_id(ImSeqID) ) {
		iml_error("mvl_video_manager","Invalid image id in create_blit_sequence.");
		return(-1);
		} // end if
	
	void *NewSeq = allocate_blit_sequence( ImSeqID, DestPos, FrameSeq);

	if( !NewSeq ) return (-1);
	
	IML_FLOAT Time =  play_blit_sequence( NewSeq );
	
	delete NewSeq;
	
	return( Time );
	
} // end create_blit_sequence


//=====================================================================
//
//=====================================================================
IML_FLOAT mvl_video_manager::play_blit_group(IML_SINT GroupID) {


	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;
//	if( !MVI.valid_blit_id(BlitSeqID) ) {
//		iml_error("play_blit_sequences","Invalid blit sequence id in play_blit_sequence.");
//		return(0.0);
//		} // end if

	IML_LINT SeqCount;
	
	if(GroupID < 0) {
		SeqCount = MVI.BlitSequenceCount;
	} else {
		
		if(GroupID >= GroupCount) {
			iml_error("play_blit_group","GroupID is out of range.");
			return(0.0);
			} // end if
		
		// FOR CONVENIENCE
		mvl_blit_sequence_group IML_REF BGroup = *BlitGroups[GroupID];
		
		if(BGroup.ElementCnt == 0) {
			iml_error("play_blit_group","Group has no elements!");
			return(0.0);
			} // end if
			
		SeqCount = BGroup.ElementCnt;
		} // end if
	
	const mvl_blit_sequence ** BlitSequences = new mvl_blit_sequence*[SeqCount];
	if(!BlitSequences) {
		iml_error("play_blit_group","Cannot allocate temp storage!");
		return(0.0);
		} // end if

	// GRAB POINTERS TO ALL STRUCTURES AND GET MINIMUM PLAYLENGTH
	IML_USINT PlayLength = 32000;
	for(IML_LINT I=0; I<SeqCount; I++) {
	
		// FOR THE GIVEN CASE, GRAB THE NEXT BLIT SEQUENCE POINTER
		if(GroupID < 0) { BlitSequences[I] = MVI.blit(I); }
		else {
			mvl_blit_sequence_group IML_REF BGroup = *BlitGroups[GroupID];
			BlitSequences[I] = MVI.blit(BGroup.blit(I));
			} // end else
		
		if( BlitSequences[I]->Length < PlayLength) 
			PlayLength = BlitSequences[I]->Length;
		} // end for
		
	// CHECK FOR ERROR CONDITION (SHOULD NOT OCCUR BUT CHECK ANYWAY)
	if(PlayLength < 1) {
		iml_error("play_blit_group","Zero legth blit sequence!");
		return(0.0);
		} // end if
	
	// MAKE A COLOR SPEC TABLE POINTER LIST
	ColorSpecTable **CLUTPtrs = (ColorSpecTable **)calculate_CLUT_ptrs(PlayLength);

	if( !CLUTPtrs ) {
		iml_error("play_blit_group","Cannot allocate CLUT ptr memory.");
		return(0.0);
		} // end if

	// LOCK DOWN THE DESTINATION PIXELS IN MEMORY
//	LockPixels(BlitSeq.SourcePM); 
//	LockPixels(BlitSeq.DestinationPM); 
				
	// UnSignedWide IS A STRUCT WITH TWO UNSIGNED LONG INTS. SEE Types.h (MAC-OS)
	UnsignedWide StartTicks, EndTicks;

	// synch for last image frame time
//	LoadLuminances(MVI.DisplayDevice, (LuminanceRecord*)CLUTPtrs[0], 1, CLUTSize-2);

	// A MAC-OS CALL  --   TO TIME THE PRESENTATION
	Microseconds(&StartTicks); 
	
	for(IML_USINT I=0; I < PlayLength; I++) {			

		if(MVI.ThunderCard) {
			MVI.ThunderCard->wait_for_VBL();
			MVI.ThunderCard->load_dacs((ColorSpec*)CLUTPtrs[I], 1, CLUTSize-2);
		} else {
			// synch image copying with LUT update and change LUT every frame for contrast windowing
			LoadLuminances(MVI.DisplayDevice, (LuminanceRecord*)CLUTPtrs[I], 1, CLUTSize-2);
			}

		// BLIT ONE FRAME FROM EACH SEQUENCE
		for(IML_LINT J=0; J<SeqCount;J++) {
		
			// MAKE A TEMPORARY GRAB ONTO THIS INFO
			const mvl_blit_sequence IML_REF BlitSeq = *BlitSequences[J];

		// copy from the source PixMap to the destination window, 
		// the source rect, to the destination area 
		// [without interpolation (re-sizing) if the rects are the same size]
			CopyBitsQuickly(
				(BitMap *)*BlitSeq.SourcePM,		(BitMap *)*BlitSeq.DestinationPM,
				&BlitSeq.SourceRect[I],				&BlitSeq.DestinationRect[I],
				srcCopy, NULL
				);
			} // end for
			
		} // end for

//	// synch for last image frame time
//	LoadLuminances(MVI.DisplayDevice, (LuminanceRecord*)CLUTPtrs[PlayLength-1], 1, CLUTSize-2);

	// A MAC-OS CALL  --   TO END PRESENTATION TIMING
	Microseconds(&EndTicks);

	// UNLOCK THE DESTINATION PIXELS IN MEMORY
//	UnlockPixels(BlitSeq.SourcePM); 
//	UnlockPixels(BlitSeq.DestinationPM); 

	delete [] CLUTPtrs;
	delete [] BlitSequences;
	
	// CALCULATE THE TOTAL TIME TAKEN
	IML_DOUBLE TotalTime = (EndTicks.lo-StartTicks.lo);
	// 4294967296.0 IS 2^32 IN DOUBLE FORMAT; *MUST* BE A DOUBLE :)
	TotalTime += 4294967296.0*(EndTicks.hi-StartTicks.hi);
	TotalTime /= 1e6;

	// report the time taken to present images
	// this allows the user to compare it with expected time to catch
	// "missed" frames (if any)
	return((IML_FLOAT)TotalTime);


} // end play_blit_sequences



//=====================================================================
//
//=====================================================================
IML_SINT mvl_video_manager::display_image(  iml_ubyte_image IML_REF Image, 
							 const point_2d IML_REF DestPos) {

	if(Image.Subspace.Dimensionality != 2) {
		iml_error("mvl_video_manager::display_image","Image is not two-dimensional!");
		return(FALSE);
		} // end if


	// GRAB MAC VIDEO INFO STRUCTURE
	mvl_mac_video_information IML_REF MVI = *(mvl_mac_video_information*)MacVideoInfo;
	if( !MVI.DisplayPixMapHandle ) {
		iml_error("mvl_video_image_sequence","Bad DisplayPixMapHandle in display_image!");
		return(FALSE);
		} // end if
	
	// GRAB THE IMAGE FRAME SIZE.
	IML_ULINT RowWidth = Image.Subspace.size(IML_2D_X);
	IML_ULINT RowCount = Image.Subspace.size(IML_2D_Y);
	
	// ALLOCATE A BUFFER ARRAY (BECAUSE OF PELLI'S IMPLEMENTATION)
	IML_ULINT IML_PTR ULArray = new IML_ULINT[RowWidth];
	if( !ULArray ) {
		iml_error("mvl_video_image_sequence","Could not allocate buffer!");
		return(FALSE);
		} // end if
		
	 // USE hp_raw_fa BECAUSE WE WANT TO USE THE IMAGE SUBSPACE SYSTEM CORRECTLY.
	 register IML_UBYTE IML_PTR BytePtr = &Image.hp_raw_fa(0); 

	// LOCK DOWN THE DESTINATION PIXELS IN MEMORY
	LockPixels(MVI.DisplayPixMapHandle); 			

	// STORE THE IMAGE STRIP INTO A LONG ARRAY, ONE ROW AT A TIME, AND STORE THAT ROW.
	// THIS IS REQUIRED BECAUSE OF PELLI'S IMPLEMENTATION . . .
	for(IML_LINT YRow=0; YRow<RowCount; YRow++){
		// STORE A ROW IN THE UNSIGNED LONG BUFFER. USE UGLY register PTRS TO SPEED THINGS UP.
//		register IML_ULINT IML_PTR LongPtr = ULArray;
//		for(register IML_LINT XCol=0; XCol<RowWidth; XCol++) *LongPtr++ = *BytePtr++;
		 IML_ULINT IML_PTR LongPtr = ULArray;
		for( IML_LINT XCol=0; XCol<RowWidth; XCol++) {
			*LongPtr++ = *BytePtr++;
			} // end for
		// STORE THE ROW IN THE OFFSCREEN PIX MAP MEMORY
		//                              destination
		SetPixmapPixelsQuickly( *MVI.DisplayPixMapHandle, 
		//            x           y               source   count
					DestPos.X, DestPos.Y+YRow, ULArray, RowWidth);  
		} // end for 

	// UNLOCK THE PIXELS FOR MEMORY MANAGEMENT REASONS
	UnlockPixels(MVI.DisplayPixMapHandle);

	// CLEAN UP THE BUFFER
	delete [] ULArray;
	
	return(TRUE);
} // end display_image

