#ifndef __MVL_MAC_VIDEO_MANAGER__
#define __MVL_MAC_VIDEO_MANAGER__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "mvl_mclk.h"

#include "iml_ndim.h"

#include "iml_seq.h"

#include "iml_poly.h"

#include "vector_2d.h"

#include <string.h>

//=====================================================================
/**  Use the auxiliary 2D vector class for specifying the upper-left 
corner of an image.
*/
//=====================================================================
typedef vector_2d point_2d;


//============================================================================
//
//============================================================================
//vector_2d & operator=(vector_2d & V, scalar_rv & Rnd);
//vector_2d & operator=(vector_2d & V, scalar_rv & Rnd) { 
//V.set(Rnd.sample(), Rnd.sample()); return(V); 
//};


//=====================================================================
/**  An auxiliary class used for specifying monitor calibration.  
We adopt here a gamma correcting function of the form:
<pre>

	DAC value = exp(p);

</pre>
where <b>exp(p)</b> is the <em>Natural Exponential</em> function, <b>p</b> is 
a polynomial in ln(LF), and <b>ln()</b> is the <em>Natural Logarithm</em>. 
The nth order polynomial <b>p</b> has parameters {a0, a1, ... an}.
A first order polynomial produces a two parameter a0*L**(1/a1).  This is equivalent
to the historical gamma function: Beta*L**Gamma where Beta=a0 and Gamma=(1/a1).
*/
//=====================================================================
class mvl_calibration_polynomial : public polynomial {
	protected:

	public:
	
	mvl_calibration_polynomial(	polynomial IML_REF CalPoly)
		: polynomial(CalPoly) {};

	mvl_calibration_polynomial(
		IML_DOUBLE A0, IML_DOUBLE A1
		)  : polynomial(1) {
		Coefficients[0] = A0;	Coefficients[1] = A1;
		};
		
	mvl_calibration_polynomial(
		IML_DOUBLE A0, IML_DOUBLE A1, IML_DOUBLE A2
		) : polynomial(2) {
		Coefficients[0] = A0;	Coefficients[1] = A1;	Coefficients[2] = A2;
		};
		
	mvl_calibration_polynomial(
		IML_DOUBLE A0, IML_DOUBLE A1, IML_DOUBLE A2, IML_DOUBLE A3
		) : polynomial(3) {
		Coefficients[0] = A0;	Coefficients[1] = A1;
		Coefficients[2] = A2;	Coefficients[3] = A3;	
		};
	
	mvl_calibration_polynomial(
		IML_DOUBLE A0, IML_DOUBLE A1, IML_DOUBLE A2, IML_DOUBLE A3, IML_DOUBLE A4
		) : polynomial(4) {
		Coefficients[0] = A0;	Coefficients[1] = A1;
		Coefficients[2] = A2;	Coefficients[3] = A3;	Coefficients[4] = A4;	
		};
	
	mvl_calibration_polynomial(
		IML_DOUBLE A0, IML_DOUBLE A1, IML_DOUBLE A2, IML_DOUBLE A3,
		IML_DOUBLE A4, IML_DOUBLE A5
		) : polynomial(5) {
		Coefficients[0] = A0;	Coefficients[1] = A1;	Coefficients[2] = A2;
		Coefficients[3] = A3;	Coefficients[4] = A4;	Coefficients[5] = A5;
		};
	
}; // end mvl_calibration_polynomial




//=====================================================================
/** An auxiliary class used in specifying the entries in a Color Lookup 
Table (CLUT). NOTE: THE DATA STRUCTURE IS A DIRECT RIP-OFF OF RGBColor 
in QuickDraw.h!
The only real difference is in the naming.  If (should I say, when?) 
Apple decides to change this, then so will I.  And if you decide to
manipulate the values directly (you should not have to), then so will you. :-)
*/
//=====================================================================
class mvl_rgb {
	public:
		/** The actual red DAC value. */
		IML_USINT red;
		/** The actual green DAC value. */
		IML_USINT green;
		/** The actual blue DAC value. */
		IML_USINT blue;
		
		/** Zero argument constructor for arrays.  */
		mvl_rgb(void) { red = 0; green = 0; blue = 0; };
		/** Simple constructor for users.  */
		mvl_rgb(IML_SINT R, IML_SINT G, IML_SINT B)
			{ red = R; green = G; blue = B; };
		
		
		/** Simple assignment; copy the contents of another mvl_rgb object.  */
		mvl_rgb & operator=(mvl_rgb & V) 
			{ red = V.red; green = V.green; blue = V.blue; return(*this); }
			
		/** Simple scaling to allow variable step sizes.  */
		mvl_rgb & operator*=(IML_FLOAT F) 
			{ red *= F; green *= F; blue *= F; return(*this); }

		/** Simple addition to allow rgb spacing by a delta.  */
		mvl_rgb & operator+=(mvl_rgb & V) 
			{ red += V.red; green += V.green; blue += V.blue; return(*this); }
			
		/** Simple subtraction to allow rgb spacing by a delta.  */
		mvl_rgb & operator-=(mvl_rgb & V) 
			{ red -= V.red; green -= V.green; blue -= V.blue; return(*this); }
		mvl_rgb & operator=(IML_FLOAT F) 
			{ red = F; green = F; blue = F; return(*this); }
	};		// true color



//=====================================================================
/** An auxiliary class used in specifying the entries in a Color Lookup 
Table (CLUT). NOTE: THE DATA STRUCTURE IS A DIRECT RIP-OFF OF ColorSpec 
in QuickDraw.h!
The only real difference is in the naming.  If (should I say, when?) 
Apple decides to change this, then so will I.  And if you decide to
manipulate the values directly (you should not have to), then so will you. :-)
*/
//=====================================================================
class mvl_ColorSpec { 
	public:
		/** Index or other value; it goes unused in my code. */
		IML_USINT	value;
		/** The actual red, green and blue (unsigned short int) values. */
		mvl_rgb 	rgb;		// rgb color
		/** Copy the values from another mvl_ColorSpec object. */
		mvl_ColorSpec & operator=(mvl_ColorSpec & CS) 
			{ value = CS.value; rgb = CS.rgb; return(*this); };
		/** Copy the rgb values from another mvl_rgb object. */
		mvl_ColorSpec & operator=(mvl_rgb & V) 
			{ rgb = V; return(*this); }
};



//=====================================================================
/**  An auxiliary class used for manipulating the Color Lookup Table (CLUT)
of the video display.  Generally you will grab a copy of the current CLUT 
by asking the mvl_video_manager object for a copy.  Then you can 
make some changes and load it back into the hardware.
*/
//=====================================================================
class mvl_clut {

	protected:
		mvl_ColorSpec *CLUT;
		IML_USINT CLUTSize;
		IML_USINT DacSize;
		IML_USINT MaxDac;
		IML_USINT LeftShift;
		
	public:


	/** @name A: Constructors, destructors and status. 
	There is only one constructor, and one destructor that knows how to.
	*/
	//@{

		/** A simple constructor.  The input values are the width of the DAC in bits
		and the CLUT length (entry count).
		*/
		mvl_clut(IML_UBYTE DBits, IML_USINT CSize);
		
		
		/** A copy constructor.  The input is just a clut for which you want
		a duplicate.  This function creates a new internal table and copies
		the CLUT entry values from the exemplar.
		*/
		mvl_clut(mvl_clut IML_REF Example);

		/** The destructor.  The only required action is to clean up the table if it was
		allocated.
		*/
		~mvl_clut() { if(CLUT) delete [] CLUT; };
		
		/** This function checks to see if the memory allocation for the CLUT was successful.
		It returns TRUE for success, FALSE for failure.
		*/
		IML_SINT success(void) { return( CLUT!=NULL ); };
	//@}


	/** @name B: Public information. 
	These values are publicly available but cannot be manipulated
	directly.  They are made accessible for your convenience.  These are identical to
	values contained in the mvl_video_manager class structure.
	*/
	//@{

		/// Number of bits in the D-to-A converter (that is, resolution of the R, G, & B guns).
		IML_USINT dac_size(void) { return(DacSize); };
		/// Maximum Dac Value of 2**dacbits - 1, where '**' denotes exponentiation.
		IML_USINT max_dac(void) { return(MaxDac); };
		/// Number of entries in the CLUT.
		IML_USINT clut_size(void) { return(CLUTSize); };
	//@}

	/** @name C: Public manipulation and access functions. 
	These functions allow the user to read and write DAC entry values.  Note that
	actualy red, green and blue DAC entries are "left justified" in the MAC world
	(don't ask me why --- I think it's weird) so if you decide to set those values
	yourself, you are responsible for getting that right.  Otherwise it is done
	for you if you use the access functions here.
	*/
	//@{

		/** Provides direct access to a CLUT entry.  See the description of mvl_ColorSpec
		for details of the contents of an entry.
		*/
		mvl_ColorSpec & operator[](IML_USINT I) 
			{ return(CLUT[I&MaxDac]); };// Bit mask is a fast mod operation.
		
		/** Read the red gun value from a CLUT entry. */
		short red(IML_USINT I)   { return(CLUT[I].rgb.red >> LeftShift);   };
		/** Read the green gun value from a CLUT entry. */
		short green(IML_USINT I) { return(CLUT[I].rgb.green >> LeftShift); };
		/** Read the blue gun value from a CLUT entry. */
		short blue(IML_USINT I)  { return(CLUT[I].rgb.blue >> LeftShift);  };
		
		/** Write the red gun value to a CLUT entry. */
		void red(IML_USINT I, IML_USINT D)   { CLUT[I].rgb.red = (D << LeftShift); };
		/** Write the green gun value to a CLUT entry. */
		void green(IML_USINT I, IML_USINT D) { CLUT[I].rgb.green = (D << LeftShift);  };
		/** Write the blue gun value to a CLUT entry. */
		void blue(IML_USINT I, IML_USINT D)  { CLUT[I].rgb.blue = (D << LeftShift);  };

		/** Write the red, green and blue values to a CLUT entry all at once.  The call
		assumes that the  R,G, & B values are NOT left justified.  
		</p>  */
		void set_clut_entry(IML_SINT I, IML_SINT R, IML_SINT G, IML_SINT B) 
			{ red(I,R);  green(I,G);  blue(I,B); };
		/** Write the red, green and blue values to a CLUT entry all at once from an mvl_ColorSpec
		structure. The call assumes that the R,G, & B values are NOT left justified.  
		</p>  */
		void set_clut_entry(IML_SINT I, mvl_ColorSpec & Color) 
			{ red(I,Color.rgb.red); green(I,Color.rgb.green); blue(I,Color.rgb.blue);  };

}; // end mvl_clut

// USED LOCALLY; THIS IS A RIP-OFF OF PELLI'S AND APPLE'S DEFINITION.
//typedef  ColorSpec ColorSpecTable[255];
typedef struct{
	mvl_ColorSpec table[256];
}ColorSpecTable;

//=====================================================================
// THIS IS A SPECIALIZED HELPER-TYPE IMAGE CLASS.  USE IT TO GENERATE CLUT SEQUENCES
//=====================================================================
class mvl_temporal_window : public iml_float_image {

	friend class mvl_video_manager;
	protected:
		ColorSpecTable * MVLColorSpecTables; // REALLY A POINTER TO ARRAYS OF ColorSpec[255]

	public:	
	// USER SHOULD USE THESE, RATHER THAN THE TRADITIONAL iml_float_image VARIABLES
	// OTHERWISE NORMAL IMAGE OPERATIONS ARE ALLOWED. SEE imlimabs.h FOR DETAILS.
		IML_USINT Length;
		mvl_temporal_window(IML_USINT Len);
		mvl_temporal_window(mvl_temporal_window IML_REF ACK);
		~mvl_temporal_window();
		
		// =========== HOUSE KEEPING FOR INHERITANCE PURPOSES ======================
		// DEFINE ASSIGNMENT OPERATORS BECAUSE THEY ARE NOT ALL INHERITED/OVERLOADED
		void operator=(iml_rofn &R2DFn)
			{ iml_image::operator=(R2DFn); };
		void operator=(iml_image &R2DFn)
			{ iml_image::operator=(R2DFn); };
		void operator=( iml_float_image IML_REF LikeImage)
			{ iml_float_image::operator=(LikeImage); }


		void operator=(IML_DOUBLE TVal)
			{ iml_float_image::operator=(TVal); }
		void operator+=(IML_DOUBLE TVal)
			{ iml_float_image::operator+=(TVal); }
		void operator-=(IML_DOUBLE TVal)
			{ iml_float_image::operator-=(TVal); }
		void operator*=(IML_DOUBLE TVal)
			{ iml_float_image::operator*=(TVal); }

		void operator+=( iml_float_image IML_REF LikeImage)
			{ iml_float_image::operator+=(LikeImage); }
		void operator-=( iml_float_image IML_REF LikeImage)
			{ iml_float_image::operator-=(LikeImage); }
		void operator*=( iml_float_image IML_REF LikeImage)
			{ iml_float_image::operator*=(LikeImage); }

		void operator+=(iml_rofn  IML_REF R2DFn)
			{iml_image::operator+=(R2DFn); };
		void operator-=(iml_rofn  IML_REF R2DFn)
			{iml_image::operator-=(R2DFn); };
		void operator*=(iml_rofn  IML_REF R2DFn)
			{iml_image::operator*=(R2DFn); };

		void operator+=(iml_image IML_REF R2DFn)
			{iml_image::operator+=(R2DFn); };
		void operator-=(iml_image  IML_REF R2DFn)
			{iml_image::operator-=(R2DFn); };
		void operator*=(iml_image  IML_REF R2DFn)
			{iml_image::operator*=(R2DFn); };
		
}; // end class mvl_temporal_window



//=====================================================================
// USED TO STORE GROUPS OF BLIT SEQUENCES WHICH WILL BE PLAYED TOGETHER.
//=====================================================================
class mvl_blit_sequence_group {
	public:
		IML_UBYTE MaxElementCnt;
		IML_UBYTE ElementCnt;
		IML_BYTE *Group;
	
		mvl_blit_sequence_group(IML_UBYTE MaxCnt) 
			{ Group = new IML_BYTE[MaxCnt]; MaxElementCnt = MaxCnt; ElementCnt=0;}
		~mvl_blit_sequence_group() { if(Group) delete [] Group; }
		IML_BYTE IML_REF blit(IML_USINT Index) { return Group[Index]; };		
		void clear_group(void) { ElementCnt=0;memset(Group, -1, MaxElementCnt); };
		IML_SINT successful(void) { return(Group != 0); };
}; // end class mvl_blit_sequence_group

//=====================================================================
/**  A class for displaying 2-dimensional images either singly or in
a movie sequence.  The class allows manipulation of the color 
lookup table (CLUT) and provides easy access to information
about the operating system and the video characteristics.
<p>
An important requirment for image display in psychophysical
experiments is fine control of pixel luminance.  To this
end the mvl_video_manager class provides two methods for
CLUT linearization. First, you can use the linearization method provided
for in Denis Pelli's VideoToolbox.  All you need is the
file created by that calibration program. That data includes
information regarding your ISR video-attenuator, if you have one.

<p>
Alternatively, you can use the linearization method provided by
this package.  The method also allows for use of a resistive network
(ISR Video Attenuator or otherwise) for increasing luminance resolution 
but requires neither a photometer nor a voltmeter. Instead the monitor-luminance
and attenuator characteristics are measured psychophysically.

<p>
<b> The MVL Linearization Method</b>
<p>
Our task in linearization is to produce a sequence of DAC values
(perhaps RGB, or reg/green/blue DAC-value triplets) that produce
a linear change of perceived luminance on the target display with
linear changes of pixel value.  That is, we want to to have
<pre>

	Perceived_Luminance(Pixel_Value) = MinLuminance + Konstant*Pixel_Value

</pre>
The relationship of the function that maps a video DAC value to a screen luminance
in cathode ray tubes(CRT's) 
<pre>

	L = f(DAC value)

</pre>
is generally nonlinear.
This is the function normally measured using a photometer.  In practice
we use the inverse of <b>f()</b>, call it <b>g()</b>, to compute the DAC value required
for the requested luminance.  The linearization method provided by
this package <b>g()</b> takes the form
<pre>

	DAC value = g(LF)

</pre>
where LF is a fraction of the luminance range.  If a DAC value of 0
produces LMin, and a maximum dac value (that is, (2**DacBits-1), where ** denotes
exponentiation) produces LMax, then LF = (L-LMin)/(LMax-LMin).  
<p>
Historically this process (that is, use of g()) has been galled gamma 
correction because f() was often taken
to be of the form alpha+beta*(L**gamma).  This is not really
adequate for good linearization of any monitor, so we adopt
a form for g() of:
<pre>

	DAC value = exp(p)  

</pre>
where <b>exp(p)</b> is the <em>Natural Exponential</em> function, <b>p</b> is 
a polynomial in ln(LF), and <b>ln()</b> is the <em>Natural Logarithm</em>. 
The nth order polynomial <b>p</b> has parameters {a0, a1, ... an}.
A first order polynomial produces a two parameter a0*(L**(1/a1)).  This is equivalent
to the historical gamma function: Beta*L**Gamma where Beta=a0 and Gamma=(1/a1).
Since a first order polynomial is not 
generally adequate for all hardware we allow higher order polynomials in our
calibration.


<p>
<b> The Attenuator Characteristics</b>
<p>
Generally, the resistive network (e.g. Video Attenuator) produces a (linear) weighted
sum of the voltages from the red, blue and green guns and the result is output onto
the green channel:
<pre>

	Vgreen_out = Ar*Vred + Ag*Vgreen + Ab*Vblue 
	
		where Ar, Ag, and Ab are constants that satisfy
	
	Ar + Ag + Ab = 1.0
	
		and where (at least, for the Video Attenuator)
	
	Ab > Ag > Ar

</pre>
For the Video Attenuator, Ar > Ab/256.  Assuming this to be true
for any attanuator used with this software, we can use the red
gun (alone) to interpolate between luminances produced by integral values
of the blue gun.  To measure this, the calibration program included
with this software finds the red gun dac value that makes
<pre>

	BlueDacValue + RedDacValue == BlueDacValue+Delta

</pre>
true for some Delta.  The number of red gun steps per blue
step is all we need to know, assuming that the gains for each
of the red, green and blue guns
are independent. (This is unlikely, but the assumption works
well enough.)  This 'step-count' is used in the linearization 
process.


*/
//=====================================================================
class mvl_video_manager {

	protected:
		// THIS CONTAINS A POINTER TO AN OBJECT THE USER DOES NOT HAVE
		// TO DEAL WITH.  IT IS DEFINED COMPLETELY IN macvideo.cpp AND USED
		// THERE ONLY.  THIS ISOLATES THE USER (IF SHE SO DESIRES) FROM
		// ALL THAT MAC GRAPHIC SLUDGE, AND FROM HAVING TO UNDERSTAND
		// PELLI'S VIDEO LIBRARY (OR INCLUDE HIS HEADER FILES).
		void * MacVideoInfo;			// REALLY  mvl_mac_video_information *
		IML_SINT FillWindowID;
		iml_ubyte_image IML_PTR FillerImage;
	
		enum {DGP_CALIBRATION, MVL_CALIBRATION};
		IML_SINT CalibrationType;
		mvl_calibration_polynomial IML_PTR CalibrationPolynomial;
//		IML_FLOAT Beta, GammaInv;
		IML_FLOAT RStepsPerBDAC;
		void best_b_plus_r_match(mvl_rgb & Dest, IML_FLOAT Fractional);

		// THIS IS A POINTER TO AN *EXTERNALLY* MAINTAINED CLUT SEQUENCE HOLDER.
		mvl_temporal_window IML_PTR CLUT_Sequence;

		// USED INTERNALLY WHEN DISPLAYING BLIT SEQUENCES
		void *calculate_CLUT_ptrs(IML_USINT PlayLength);
		
		IML_SINT GroupCount;
		mvl_blit_sequence_group **BlitGroups;

		mvl_clock Clock;

		// USE PETE'S DEFAULT CONFIGURATION FOR CODE-WARRIOR WINDOW.
		void set_up_SIOUX(void);
		
		/* Pelli's library needs a minimum stack size. Users of
		this Mac Video Library should not have to deal with this call.
		And it might therefore be made private instead of public. . . 
		*/
		IML_LINT extend_stack(IML_ULINT Extension);
		
	public:
		point_2d GlobalWindowCoords;

		/** @name A: Informational Variables
		These variables are intended to provide information to the knowledgable
		user.  They are not protected, because the knowledgeable user knows
		that changing them is either useless, or a bad idea. :-)  
		In other words, you can copy them or print them out, but don't change them.
		*/
		//@{

		/** @name 1: Operating system and hardware information. */
		//@{
		
		/// Keeps track of which monitor is used for display; 0, 1, etc.
		IML_LINT DisplayNumber;	
		
		/// True if this MAC-OS version is new enough for Gestalt()
		IML_LINT HasGestalt;		
		
		/// Tells the MAC-OS version in hex. (e.g. 0x710 is 7.1.0)
		IML_LINT SystemOSVersion;		
		
		// True if (SystemOSVersion >= 0x700) is true.
		IML_LINT IsSystemSeven;
		
		/** Tells what cpu architecture is used in this pc.
		Possible MAC-OS macro names and integer values are:
		<pre>
		
			gestalt68k:     1  -- Motorola MC68k 
			gestaltPowerPC: 2  -- IBM PowerPC
			
		</pre> */
		IML_LINT SystemArchitecture;

		/** Tells the cpu type in more detail than SystemArchitecture.
		Possible MAC-OS macro names and integer values are:
		<pre>
		
			gestaltCPU68000: 1	-- Various 68k CPUs...
			gestaltCPU68010: 2
			gestaltCPU68020: 3
			gestaltCPU68030: 4
			gestaltCPU68040: 5
			gestaltCPU601:   0x101,		-- IBM 601, etc.
			gestaltCPU603:   0x103,
			gestaltCPU604:   0x104
		
		</pre>
		*/
		IML_LINT CPUType;
		
		/// True if CPU is one of (601,603,604.... i.e. SystemArchitecture == 2)
		IML_LINT IsPowerPC;	
		
		/** Tells what kind of FPU is present if SystemArchitecture == 1.
		Possible MAC-OS macro names and integer values are:
		<pre>
		
			gestaltNoFPU:    0  -- no FPU 
			gestalt68881:    1  -- 68881 FPU 
			gestalt68882:    2  -- 68882 FPU 
			gestalt68040FPU: 3  -- 68040 built-in FPU
		
		</pre>
		*/
		IML_LINT FPUType;
									
		
		/** Tells what version of QuickDraw is used on this computer.
		Possible MAC-OS macro names and integer values are:
		<pre>
		
			gestaltOriginalQD: 0x000 -- original 1-bit QD 
			gestalt8BitQD:     0x100 -- 8-bit color QD 
			gestalt32BitQD:    0x200 -- 32-bit color QD 
			gestalt32BitQD11:  0x201 -- 32-bit color QDv1.1 
			gestalt32BitQD12:  0x220 -- 32-bit color QDv1.2 
			gestalt32BitQD13:  0x230 -- 32-bit color QDv1.3 
		
		</pre>
		*/
		IML_LINT QuickDrawVersion;

		///  True if (QuickDrawVersion >= gestalt8BitQD) is true.
		IML_LINT Has8BitQD;
		
		/// True if (SystemOSVersion >= 0x605) is true.
		IML_LINT NewPaletteManagerPresent;
		
			/// Seems archaic, but verifies that the monitor is a color one.
		IML_LINT DeviceIsColor;	
						
		/// Tells the current stack size, including after calling extend_stack().
		IML_LINT StackSize;
		
		//@}

		/** @name 2: Monitor specific information.  
		This information gives details about using the specified monitor in the
		256 color video mode. These are set when the user calls initalize_screen().
		*/
		//@{
		
		/// The width of a D-to-A (DAC) converter slot in bits. Usually it's 8.
		IML_LINT DacSize;
		/// The the last DAC entry address:  2**DacSize - 1, where '**' denotes exponentiation.
		IML_LINT MaxDac;
		/// Used for loading values into the system DAC; LeftShift = 16 - DacSize.
		IML_LINT LeftShift;
		/// How long is the clut; not necesarily the same as 2**DacSize (where '**' denotes exponentiation).
		IML_LINT CLUTSize;
		/// Pixel resolution across the screen.
		IML_LINT XResolution;
		/// Pixel resolution down the screen.
		IML_LINT YResolution;
		/// Pixel depth (in bits) before we changed it to 8.
		IML_LINT InitialPixelSize;
		/// This is actually supplied by the user, since the software does not know.
		IML_LINT UsingAttenuator;			
		//@}
		
		//@}
		
		/** @name B: Behavioral Variables. 
		These variables are set by the user in order to change the way 
		the images are displayed, or what to do in case of errors.  *DONT*
		change these yourself!  Go through the provided system calls. 
		*/
		//@{
		
		/// Exit on various bad errors, if and when they occur.
		IML_LINT ExitOnError;		
		
		/** Starting place for linearization procedure. Together with CLUTSegmentLength this allows
		you to store more than one linearized sequence, or to use all or only a 
		segment of the CLUT for a linear sequence. 
		*/
		IML_USINT StartCLUTSegment;
		
		/** Number of CLUT entries to use in the linearization procedure. Together with 
		StartCLUTSegment this allows you to store more than one linearized sequence, or 
		to use all or only a segment of the CLUT for a linear sequence. */
		IML_USINT CLUTSegmentLength;
		
		/// Current max luminance.
		IML_DOUBLE MaxLuminance;
		/// Current min luminance.
		IML_DOUBLE MinLuminance;
		/** Current mean luminance as a fraction of the maximum range.
		*/
		IML_DOUBLE MeanLFraction;
		/** Current mean luminance.  This is equal to
		<pre>
			MinLuminance + MeanLFraction*(MaxLuminance+MinLuminance).
		</pre> 
		*/
		IML_DOUBLE MeanLuminance;
//		/// Current maximum contrast (determined by the CLUT entries) as a fraction of mean luminance.
//		IML_DOUBLE MaxContrast;

		//@}

		/** @name C: Constructors. */
		//@{
		
		/**
		There is only one constructor with all arguments having defaults.
		The MaxImageSequences, MaxBlitSequences, and MaxBlitGroups are
		currently required because I was lazy in the design phase. :-)
		Later on I plan to make the data structures whose sizes depend on these
		inputs dynamically allocated.  Then the only required input will be
		whether to ExitOnErr.
		*/
		mvl_video_manager(IML_USINT MaxImageSequences = 8, 
					IML_USINT MaxBlitSequences = 64, 
					IML_USINT MaxBlitGroups = 32,
					IML_ULINT ExitOnErr = TRUE);
		
		/* This catches when people do bad things . . */
		mvl_video_manager(mvl_video_manager IML_REF CopyConstruction);
					
		// The only destructor.
		~mvl_video_manager();
		
		//@}


		// ============ INITIALIZATION ROUTINES ========================
		/** @name D: Initialization routines. */
		//@{
		
		/** Tells the mac video manager about the display.  The input
		arguments are which video display
		to use, whether the video hardware is a Radius PCI Thunder30 card. 
		*/
		void initialize_screen(IML_SINT DispNumber, IML_SINT ThunderCardVideo=FALSE);
		
		/** This call trys to load display linearization information (DG Pelli's 
		library format) from a file.  A NULL input represents a request for a
		default file.  You take your chances with that, since I don't know if the
		compiled-in information is any good for a particular display.
		*/
		IML_LINT set_dgp_calibration(const char * FileName=NULL);


		/** This call tells the video manager how many luminance steps the red gun 
		provides relative to one step of the blue gun when using a resistive network 
		for combining red, green and blue guns of your CRT.  
		If RedSteps == 0.0 indicates that an attenuator is not used.  If
		RedSteps is > 1.0 then linear cluts are calculated assuming that the blue and
		red guns are combined using the attenuator, and that the red gun provides the least
		significant (small step) bits.

		<b> The Attenuator Characteristics</b>
		<p>
		Generally, the resistive network (e.g. ISR Video Attenuator) produces a 
		linear, weighted sum of the voltages from the red, blue and green CRT
		guns, and the resulting voltage is output to the green channel:
		<pre>

			Vgreen_out = Ar*Vred + Ag*Vgreen + Ab*Vblue 
			
				where Ar, Ag, and Ab are constants that satisfy
			
			Ar + Ag + Ab ~ 1.0
			
				and where (at least, for the Video Attenuator)
			
			Ab > Ag > Ar

		</pre>
		For the Video Attenuator, Ar > Ab/(MaxDac+1).  Assuming this to be true
		for any attanuator used with this software, we can use the red
		gun (alone) to interpolate between luminances produced by integral values
		of the blue gun.  To measure this, the calibration program included
		with this software finds the red gun dac value that makes
		<pre>

			BlueDacValue + RedDacValue == BlueDacValue+Delta

		</pre>
		true for some Delta.  The number of red gun steps per blue
		step is all we need to know, assuming that the gain for each
		of the red, green and blue guns
		are independent. (This is unlikely, but the assumption works
		well enough.)  This 'step-count' is used in the linearization 
		process.
		*/
		void set_attenuator_red_steps( IML_FLOAT RedSteps );
		
		/** This call suplies the video manager with linearization data as obtained
		from the linearization program supplied with this software.
		The model in this case is
		<pre>
		
			DacValue = B*L**(1/G)
		
		</pre>
		where '**' denotes exponentiation.
		B should usually be set to (or be close to) MaxDac.  MinL and MaxL are 
		the minimum and maximum luminance values as measured with a photometer.  MinL 
		and MaxL values are not necessary for the linearization.  
		L is a luminance as a fraction of maximum; L=0.0 at MinL and L=1.0 at MaxL.
		<p>*/
		void set_mvl_calibration(IML_FLOAT B, IML_FLOAT G, IML_FLOAT MinL, IML_FLOAT MaxL);


		/** This call suplies the video manager with linearization data as obtained
		from the linearization program supplied with this software.
		The model in this case is
		<pre>

			DAC value = exp(p)  

		</pre>
		where <b>exp(p)</b> is the <em>Natural Exponential</em> function, <b>p</b> is 
		a polynomial in ln(LF), and <b>ln()</b> is the <em>Natural Logarithm</em>. 
		The nth order polynomial <b>p</b> has parameters {a0, a1, ... an}.
		MinL and MaxL are  the minimum and maximum luminance values as measured with
		a photometer.  L is a luminance as a fraction of maximum; L=0.0 at MinL 
		and L=1.0 at MaxL.   Accurate values of MinL and MaxL are not necessary for
		linearization, only for your publications ;).
		*/
		void set_mvl_calibration(polynomial IML_REF CalibrationInfo, IML_FLOAT MinL, IML_FLOAT MaxL);

		/// FYI -- Especially if calibration was loaded from a file.
		IML_DOUBLE get_min_luminance(void) { return( MinLuminance ); };
		/// FYI -- Especially if calibration was loaded from a file.
		IML_DOUBLE get_max_luminance(void) { return( MaxLuminance ); };
		
		/** Tells the video manager to use DGP's linearization methods.  
		'DGP' is D.G. Pelli (cf. <em>use_mvl_calibration()</em>). 
		You should also call set_dgp_calibration() with the file name created
		by DGP's calibration program before creating a linear CLUT.
		*/
		void use_dgp_calibration(void)
			{ CalibrationType = mvl_video_manager::DGP_CALIBRATION; };

		/** Tells the video manager to use MVL gamma correction; this is the
		default (cf. <em>use_dgp_calibration()</em>). You should also 
		call one of the <em>set_mvl_calibration()</em> before creating a linear CLUT.
		*/
		void use_mvl_calibration(void)
			{ CalibrationType = mvl_video_manager::MVL_CALIBRATION; };
		
		/** Tells the video manager to use <em>MVL style</em> window filling.
		This is necessary only in the cases where you would like to fill the
		window with CLUT values of 0 or MaxDac.  This occurs because, according
		to Apple's manuals on the Color Manager, the Color Manager 
		is not normally guaranteed to honor that palette color N will be stored as N in
		each pixel.  For example, the calls
		<pre>
			//  set to requested luminance
			PmBackColor( CLUT_Entry );

			// EraseRect IS A MAC-OS CALL
			// Fill rect with CLUT index.
			EraseRect( &Rect );
		</pre>
		with CLUT_Entry value of 0 or 255 (for 8-bit DAC video hardware)
		will often fail because the Color manager strongly objects to meddling
		with the first and last CLUT entries.  If you do this <b>after</b> filling
		the CLUT with a linearized sequence, the Color Manager will likely
		re-fill the first and last entries with white and black (full and no
		luminance) respectively.  
		<p>
		<b> THIS IS A ROYAL PAIN. </b>
		<p>
		So, this method does window filling (clearing) using a pre-set image
		of size Speed x Speed pixels. :-)
		<p></p>*/
		void use_mvl_window_filling(IML_USINT Speed = 64);
			
		//@}


		// THESE TWO ARE CURRENTLY ONLY USED INTERNALLY.
		void fill_dac(mvl_ColorSpec *, IML_USINT Start, IML_USINT Length, 
					IML_FLOAT MeanL, IML_FLOAT Contrast);
					
		void fill_dac(mvl_ColorSpec *CSpecs,  IML_FLOAT MeanL, IML_FLOAT Contrast)
			{ fill_dac(CSpecs, 1, CLUTSize-2, MeanL, Contrast); };

		// =========== USE THESE FOR DOING YOUR WORK =========================



		/** @name E: Color Lookup Table (CLUT) manipulation routines. */
		//@{
		
		/** Set the two behavioral variables, StartCLUTSegment and CLUTSegmentLength.  Their
		values are used when loading the hardware CLUT with a linearized sequence  
		by a call to set_display_CLUT(IML_DOUBLE MeanL, IML_DOUBLE MaxC).
		*/
		void set_default_clut_segment(IML_USINT StartC, IML_USINT CLength); 
		
		/** Set the hardware CLUT to the requested mean luminance (MeanL) and 
		maximum contrast (MaxC) using the linearization information from
		intialization calls.  The linearized sequence is stored in
		the CLUT segment given by two behavioral variables, <em>StartCLUTSegment</em> 
		and <em>CLUTSegmentLength</em>. These are set with set_default_clut_segment().
		The linearize_clut() call stores a copy for use in movie
		(blit sequence) displays.
		*/
		void linearize_clut(IML_DOUBLE MeanL, IML_DOUBLE MaxC) {
			linearize_clut(MeanL, MaxC, StartCLUTSegment, CLUTSegmentLength);
			};
		
		void set_display_CLUT(IML_DOUBLE MeanL, IML_DOUBLE MaxC) {
		fprintf(stderr,"\007set_display_CLUT() should be replaced with linearize_clut()\n");
		fflush(stderr);
			linearize_clut(MeanL, MaxC, StartCLUTSegment, CLUTSegmentLength);
			};
		
		
		/** Set the hardware CLUT to the requested mean luminance (MeanL) and 
		maximum contrast (MaxC) using the linearization information previously
		defined in intialization calls.  The linearized sequence is stored in
		the CLUT segment given by Start and Length.
		The linearize_clut() call stores a copy for use in movie (blit sequence) displays.
		*/
		void linearize_clut(IML_DOUBLE MeanL, IML_DOUBLE MaxC,
				IML_USINT Start, IML_USINT Length);
		
		void set_display_CLUT(IML_DOUBLE MeanL, IML_DOUBLE MaxC,
				IML_USINT Start, IML_USINT Length) {
		fprintf(stderr,"\007set_display_CLUT() should be replaced with linearize_clut()\n");
		fflush(stderr);
			linearize_clut(MeanL, MaxC, StartCLUTSegment, CLUTSegmentLength);
			};
		
		/** Set a single entry in the hardware CLUT to the requested (R,G,B) combination.  
		The entry is also stored in the default CLUT and will be used in movie sequences.
		*/
		void set_clut_entry(IML_SINT Entry, IML_SINT R, IML_SINT G, IML_SINT B);

		/** Set a single entry in the hardware CLUT to the requested (R,G,B) combination.  
		The entry is also stored in the default CLUT and will be used in movie sequences.
		*/
		void set_clut_entry(IML_SINT Entry, mvl_rgb IML_REF RGB)
			{ set_clut_entry(Entry, RGB.red, RGB.green, RGB.blue); }

		
//======== CURRENTLY USED INTERNALLY ==================================

		enum {RED_GUN=0x01, GREEN_GUN=0x02, BLUE_GUN=0x04, ALL_GUNS=0x07};
		/** Fill the hardware CLUT with an <em>un-gamma-corrected, linear</em> sequence
		of DAC values.  The sequence is stored in the CLUT segment given by Start and Length.
		This call also stores a copy for use in movie (blit sequence) displays.
		The DAC values stored in the CLUT are 
		<pre>
		
		(short int)(0.5 + StartDac + (CLUTEntry-StartCLUT)*DeltaDac)
		
		</pre>
		for CLUTEntry going from StartCLUT to StartCLUT+CLUTLength-1.  The values are
		stored in any combination of red, green and blue DAC entries as designated by
		the RorGorB parameter.  RorGorB should be one of
		<pre>
		
		mvl_video_manager::RED_GUN
		mvl_video_manager::GREEN_GUN
		mvl_video_manager::BLUE_GUN
		mvl_video_manager::ALL_GUNS
		
		</pre>
		
		or any bit-wise OR'ed combination.   ALL_GUNS is equivalent to
		<pre>
		
		mvl_video_manager::RED_GUN | mvl_video_manager::GREEN_GUN | mvl_video_manager::BLUE_GUN
		
		</pre>
		*/
		void identity_clut(IML_UBYTE RorGorB, IML_USINT StartCLUTEntry, IML_USINT Length,
			IML_FLOAT StartDacValue, IML_FLOAT DeltaDac);
			
		void identity_clut(IML_UBYTE RorGorB ) {
			identity_clut(RorGorB, StartCLUTSegment,
				CLUTSegmentLength, StartCLUTSegment, 1 );
			};

		// THESE VALUES ARE STORED FOR USE WHEN CALLING initialize_CLUT_tables
		IML_DOUBLE get_mean_luminance(void) { return( MeanLFraction ); }
		void set_mean_luminance(IML_DOUBLE MeanL); //(0.0 <= MeanL <= 1.0)

//		IML_DOUBLE get_max_contrast(void) { return(MaxContrast); };
//		void set_max_contrast(IML_DOUBLE MaxC); // (-1.0 <= MaxC <= 1.0)
		
		
		/** Get a copy of the current CLUT (that is, the one currently displayed by the
		video hardware). The call returns a pointer but I recommend attaching
		it to a reference, that is 
		<pre>
		
			mvl_clut & CLUT_Reference = *MacVideo.get_clut_copy();
			
		<p></pre>
		Then you can alter the copy of the clut to your heart's content, and load
		it back into the hardware later.  But remember to delete the
		CLUT copy when you're done!  That is:
		<pre>
		
			delete &CLUT_Reference;
			
		</pre> 
		*/		
		mvl_clut IML_PTR get_clut_copy(void);
		
		/** Load the given CLUT information into the video hardware and <em>copy it into
		the default CLUT</em>.  The default CLUT is used in movie (blit sequence) displays.
		*/		
		void set_default_clut(mvl_clut IML_REF CLUT);
		
		void make_default_clut(mvl_clut IML_REF CLUT) {		
			fprintf(stderr,"\make_default_clut() should be replaced with set_default_clut()\n");
			fflush(stderr);
			set_default_clut(CLUT);
			};
		
		/** Load all entries from the given CLUT into the video hardware but 
		<em>DON'T</em> copy it into the default CLUT. Movie (blit sequence) displays 
		will use the current default instead.
		*/		
		void load_clut(mvl_clut IML_REF CLUT);
				
		/** Load a single entry from the given CLUT into the video hardware
		but <em>DON'T</em> copy it into the default CLUT.  Movie (blit sequence) displays will 
		use the current default instead
		*/		
		void load_clut_entry(mvl_clut IML_REF CLUT, IML_SINT Entry);

		
		// GIVEN A TEMPORAL WINDOW *CONTRAST* PROFILE REPRESENTATION, GENERATE A SEQUENCE OF
		// CLUT TABLES FOR DISPLAYING SEQUENCES. NOTE THAT THE PROFILE IS NOT CHECKED
		// OR CLIPPED DIRECTLY, BUT PELLI'S CLUT ROUTINE DOES THAT WITH THE LUMINANCE
		// VALUES.  THERE IS NO ERROR MESSAGE FOR OUT-OF-BOUNDS (< -1, > 1.0) VALUES.
		void initialize_CLUT_tables(mvl_temporal_window IML_REF TWin);
		
		// USE THIS TEMPORAL WINDOW FOR NEXT MOVIE.  SHOULD HAVE BEEN
		// INITIALIZED WITH initialize_CLUT_tables
		void use_CLUT_sequence(mvl_temporal_window IML_REF TWindow) {
				CLUT_Sequence = &TWindow;
				};
		
		
		//@}


		/** @name F: Image storage and display-preparation routines. */
		//@{
		
		/** Fill the display window with a mean luminance value as described by the
		current calibration information.
		*/		
		void clear_window(void) { fill_window_relative(0.5); };
		
		/** Fill the display window with a luminance value as described by the
		calibration information.  Luminance is in the range 0.0 to 1.0 and produces
		a luminance of
		<pre>
		
			MinLuminance + Luminance*(MaxLuminance-MinLuminance).
		
		</pre>
		*/		
		void fill_window_relative(IML_FLOAT Luminance); 

		/** Fill the display window with a "color" given in the requested CLUT entry.
		CLUT_Index is restricted to the range from 0 to CLUTSize-1.
		*/		
		void fill_window_absolute(IML_SINT CLUT_Index);

		
		/** Draw a fixation marker at the requested position, of the requested size, in pixels.
		*/
		void draw_fixation_marker(point_2d IML_REF DestPos, IML_LINT Size);

		/* Draw a sequence of lines.
		*/
//		void draw_lines(mvl_screen_lines IML_REF DestPos, IML_LINT ConnectLastWithFirst=TRUE);

		/** Store one or more image frames so that they can be displayed quickly
		using DGP's blitting code.	 Images are stored in off-screen graphics memory.
		The memory type is assumed to be unsigned byte.
		
		@return On successful completion the function returns an image sequence 
		identity.  The identity is used in other function calls, such as 
		<em>create_blit_sequence()</em> or <em>display_frame()</em>.  If the request
		could not be completed for some reason, an error message is printed to the
		console window and a value of -1 is returned.
		*/		
		IML_SINT store_image_sequence(iml_ubyte_image IML_REF Image);

		/** Same as above, but expects signed byte images.
		*/		
		IML_SINT store_image_sequence(iml_byte_image IML_REF Image)
			{ return(store_image_sequence((iml_ubyte_image IML_REF)Image)); };

		/** Update the pixels of an  image sequence previously stored
		with <em>store_image_sequence()</em>. 
		The memory type is assumed to be unsigned byte.
		
		@return On successful completion the function returns the image sequence 
		identity passed in.  If the request
		could not be completed for some reason, an error message is printed to the
		console window and a value of -1 is returned.
		*/		
		IML_SINT update_image_sequence(iml_ubyte_image IML_REF Image, IML_USINT ImSeqID);

		/** Clear out all previously stored image sequences. 
		*/
		void clear_image_sequences(void);
				
		/** Create an internal blit sequence for use in showing a movie. ImSeqID is the
		value returned from store_image_sequence().  FrameSeq is the iml_imageframe_sequence
		object which dictates the index sequence of image frames to blit.

		@return On successful completion the function returns a blit sequence 
		identity.  The identity is used in other function calls, such as 
		<em>add_to_group()</em> or <em>play_blit_sequence()</em>.  If the request
		could not be completed for some reason, an error message is printed to the
		console and a value of -1 is returned.
		*/		
		IML_SINT create_blit_sequence(IML_USINT ImSeqID, point_2d IML_REF DestPos, 
												iml_imageframe_sequence IML_REF FrameSeq);


		void * allocate_blit_sequence(IML_USINT ImSeqID, 
			point_2d IML_REF DestPos, iml_imageframe_sequence IML_REF FrameSeq);

		/** Clear out all the blit sequences.  This sets the sequence count
		to zero and invalidates any previously held blit sequence ID's.  
		*/
		void clear_blit_sequences(void);
		
		/** Add a blit sequence to a target group that will be played together (simultaneously).
		That is, all blit sequences in a blit-group will be played in frame synchronization.
		Note that if portions of the image sequences overlap spatially, the overlap region will
		show the image added most recently to the group.  GroupID defaults to group identity zero
		if no group identity is supplied.

		@return On successful completion the function returns TRUE.  If the request
		could not be completed for some reason, an error message is printed to the
		console and FALSE is returned.
		*/		
		IML_SINT add_to_group(IML_USINT BlitSeqID, IML_USINT GroupID=0);

		/** Clear out one or all blit sequence groups.  This 
		invalidates the previously held group ID(s).  Passing in a negative
		number requests clearing of all groups.

		@return On successful completion the function returns TRUE.  If the request
		could not be completed for some reason, an error message is printed to the
		console and FALSE is returned. GroupID defaults to group identity -1
		if no group identity is supplied.  A negative-valued group identity requests
		clearing of <em>all</em> groups.
		*/
		IML_SINT clear_group(IML_SINT GroupID = -1);
		
		//@}


		/** @name G: Movie sequence and 2D image/movie-frame display routines. */
		//@{
		
		/** Play the requested blit sequence as a movie using the current temporal window
		setup.  BlitSeqID is the value returned by <em>create_blit_sequence()</em>. Use the
		return value to check that video synchronization is working.
		
		@return On successful completion the function returns the total playing time
		for the blit sequence, in seconds.  If the request
		could not be completed for some reason, an error message is printed to the
		console and a value of 0.0 seconds is returned.
		*/		
		IML_FLOAT play_blit_sequence(IML_USINT BlitSeqID);

IML_FLOAT play_blit_sequence(void *BS);
IML_FLOAT play_blit_sequence(IML_USINT ImSeqID, 
			point_2d IML_REF DestPos, iml_imageframe_sequence IML_REF FrameSeq);

		/** Play all the blit sequences in the given group simultaneously. The return value
		is the playing time in seconds so you can make sure the frame rate was right.
		*/		
		IML_FLOAT play_blit_group(IML_SINT GroupID = -1);

		/** Show a single frame from a previously stored image frame sequence.
		ImSeqID is the value returned from <em>store_image_sequence()</em>, Frame is the frame
		number (zero to FrameCount-1) and DestPos indicates the position in which to place the
		upper left corner of the frame.  GroupID defaults to group identity -1
		if no group identity is supplied.  A negative-valued group identity requests
		playing of <em>all</em> groups, hence <em>all</em> blit sequences, simultaneously.
		Blit sequence identities that occur multiple times will be played multiple times.
		*/		
		IML_SINT display_frame(IML_USINT ImSeqID, IML_USINT Frame, 
												point_2d IML_REF DestPos);

		/** Copy the requested image directly to the display without having to copy it to
		off screen memory and make a blit sequence.  There is no guarantee of timing
		or syncing to CLUT loading with this call.  Image is assumed to be unsigned byte.

		@return On successful completion the function returns TRUE.  If the request
		could not be completed for some reason, an error message is printed to the
		console and FALSE is returned.
		*/		
		IML_SINT display_image(  iml_ubyte_image IML_REF Image, 
									point_2d const IML_REF DestPos);		
		
		/** Same as above, but for *signed* byte images.
		*/		
		IML_SINT display_image(  iml_byte_image IML_REF Image, 
									point_2d const IML_REF DestPos) 
			{ return(display_image((iml_ubyte_image IML_REF)Image,DestPos)); }
		
		//@}


//		See  <A HREF="../doc_mvl_overview.html">the general overview</A> for a general
//		description of how to select where to put your image on the screen.
		/** @name H: Easy methods for choosing image presentation locations. 
		*/
		//@{
		
		
		/** Calculates the number of images what will fit on the display, given
		XSize, YSize, and Spacing.  For a 400x400 display,
		for example, with XSize=100, Ysize=100, and Spacing=0, we have a 4x4
		array of image positions, or an array size of 16.
		*/
		IML_USINT get_display_array_size(IML_USINT XSize, IML_USINT YSize, IML_USINT Spacing=0);
		/** Same as above, but gets image XSize and YSize from the image passed in.
		*/
		IML_USINT get_display_array_size(iml_image IML_REF Image, IML_USINT Spacing=0);

		/** Just like the calls above, this call calculates the number of images what will fit 
		on the display, given XSize, YSize, and Spacing.  The difference is that the
		screen space is treated as a two-dimensional array rather than as a linear array.
		For a 400x400 display,
		for example, with XSize=100, Ysize=100, and Spacing=0, we have a 4x4
		array of image positions, or an array size of 16.
		*/
		point_2d get_grid_size(IML_USINT XSize, IML_USINT YSize, IML_USINT Spacing=0);
		/** Same as above, but gets image XSize and YSize from the image passed in.
		*/
		point_2d get_grid_size(iml_image IML_REF Image, IML_USINT Spacing=0);
		
//		See  <A HREF="../doc_mvl_overview.html">the general overview</A> for a general
//		description of how to select where to put your image on the screen.
			/** Calculate an point_2d that will place an image of the
		specified size into the requested location.  
		
		XSize, YSize and Spacing
		together with the display resolution (see <em>XResolution</em>, 
		<em>YResolution</em> variables) determine a sequence of positions 
		on the display in an array format.  For a 400x400 display,
		for example, with XSize=100, Ysize=100, and Spacing=0, we have a 4x4
		array of image positions.  For a 400x400 display with  XSize=100, 
		Ysize=100, and Spacing=10, we only have a 3x3 array of image positions with 
		some leftover space.  ImageIndex is an integer index into the array 
		and is in the range [0,rows*cols-1].  Position value increments from left 
		to right, top to bottom.  For example, in a 3x3 image matrix positions are
		designated by:
		</p><pre> 
		0   1    2
		3   4    5
		6   7    8
		</pre>
		The second method allows for special positions on the display to 
		be selected. ImageIndex can be passed the values:
		<pre>
		mvl_video_manager::CENTER
		mvl_video_manager::UPPER_LEFT
		mvl_video_manager::UPPER_RIGHT
		mvl_video_manager::LOWER_LEFT
		mvl_video_manager::LOWER_RIGHT
		</pre>
		This selects the obvious centered or justified image positions.
		*/
		point_2d get_screen_position(IML_SINT ImageIndex, 
				IML_USINT XSize, IML_USINT YSize, IML_USINT Spacing=0);
		/** Same as above, but gets image XSize and YSize from the image passed in.
		*/
		point_2d get_screen_position(IML_SINT ImageIndex, 
				iml_image IML_REF Image, IML_USINT Spacing=0);

enum {CENTER=-1,UPPER_LEFT=-2, UPPER_RIGHT=-3, LOWER_LEFT=-4, LOWER_RIGHT=-5};

		/** Just like the calls above, this call calculates an point_2d 
		that will place an image of the specified size into the requested location.  
		The difference is that the
		screen space is treated as a two-dimensional array rather than as a linear array.
		XY and Spacing together 
		with the display resolution (see <em>XResolution</em>, <em>YResolution</em>
		variables) determine an array of positions on the display.  For a 400x400 display,
		for example, and XSize=100, Ysize=100, and Spacing=0, we have a 4x4
		array of image positions.  For a 400x400 display with  XSize=100, 
		Ysize=100, and Spacing=10, we have a 3x3 array of image positions with 
		some leftover space.  XY is a 2-dimensional index into the array 
		and is in the range [0,cols-1; 0,rows-1].  Position value increments from left 
		to right, top to bottom.  For example, in a 3x3 image matrix, positions are
		designated by:
		</p><pre> 
		0,0   0,1    0,2
		1,0   1,2    1,2
		2,0   2,1    2,2
		</pre>
		*/
		point_2d get_screen_position(point_2d & XY, 
				IML_USINT XSize, IML_USINT YSize, IML_USINT Spacing=0);
		/** Same as above, but gets image XSize and YSize from the image passed in.
		*/
		point_2d get_screen_position(point_2d & XY, 
				iml_image IML_REF Image, IML_USINT Spacing=0);
		
		//@}


		void flush_all_events(void);


}; // end class



#endif // __MVL_MAC_VIDEO_MANAGER__
