#ifndef __MVL_MCLK_H__
#define __MVL_MCLK_H__

/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/



#include<Timer.h>
#include "iml_type.h"
#include "iml_err.h"

//=====================================================================
/** This is an abstract base class that defines the interface for 
system clocks.  It is provided here in the anticipation that the
MVL interface can be ported to other platforms with minimal change.
*/
//=====================================================================
class iml_clock {
	public:
		// THIS IS A GENERIC INTERFACE TO BE USED ACROSS MACHINES
		void start(void)
			{ iml_warning("iml_clock:start","Hey, use a real clock!"); return; };
		IML_FLOAT time(void) 
			{ iml_warning("iml_clock:time","Hey, use a real clock!"); return(0); };
		IML_ULINT seed(void)
			{ iml_warning("iml_clock:seed","Hey, use a real clock!"); return(-1); };
		void wait(IML_FLOAT WaitTime)
			{ iml_warning("iml_clock:wait","Hey, use a real clock!"); return; };
}; // end class iml_clock

//=====================================================================
/** This class allows simple access to the MAC-OS provided microsecond
resolution timer.  Only a minimal and, I hope, useful set of functions
are provided.
*/
//=====================================================================
class mvl_clock : public iml_clock {
	protected:
		// LOCAL VARIABLES FOR TIMING PURPOSES;
		UnsignedWide ST, ET;
		UnsignedWide ClockSeed;
		// SO THE USER CAN DO TIMING.  ON THE MAC, THIS USES THE Microseconds() CALL.
	public:
	
		mvl_clock(void) {start();};
		/** Start the timer.  This really resets the time
		to zero because the actual timer is running continuously.  Use this in
		conjunction with time().
		*/
		void start(void);
		
		/** Get the amount of time (in seconds) that has passed since the last start() call.
		Use this in conjunction with start().
		*/
		IML_FLOAT time(void);
		
		/** Use this call if you just want to pause your program for a 
		specified amount of time (in seconds).
		*/
		void wait(IML_FLOAT WaitTime);
		
		/** This returns a seed from the low 32 bits of the microsecond timer
		so you can use it as a seed for your favorite (or is that favourite?)
		random number generator.
		*/
		IML_ULINT seed(void);
};

#endif // __MVL_MCLK_H__
