/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

// WE NEED TO DEFINE "MAC_C" TO A VALUE OF '1' IF WE WANT TO USE THE
// EXTENDED C FUNCTIONS WHICH THE MAC TOOL BOXES USE.  THIS ALLOWS
// STACK EXTENSION, ETC.  
#define MAC_C 1

// THESE ARE C ROUTINES, AND WE MUST TELL THE COMPILER SO THAT WE CAN LINK
// TO THE C-COMPILED LIBRARY.  OTHERWISE THE NAMES ARE MUNGED AND THE LINKER
// LOOKS FOR THE WRONG SYMBOLS.  
extern "C" {
	#include "Videotoolbox.h"
	#include "Luminance.h"
}

#ifdef HUGE_VAL
#undef HUGE_VAL
#endif


#include "iml_type.h"
#include "iml_math.h"
#include "iml_err.h"

#include "mvl_util.h"


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   mvl_blit_sequence   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//=====================================================================
//
//=====================================================================
mvl_blit_sequence::~mvl_blit_sequence() {
	if(DestinationRect)	delete [] DestinationRect;
	if(SourceRect) 		delete [] SourceRect;
} // end destructor

//=====================================================================
//
//=====================================================================
mvl_blit_sequence::mvl_blit_sequence(IML_USINT Len, PixMap ** DstPM, PixMap **SrcPM) {

	Length = Len;
	DestinationRect = new Rect[Length];
	if( !DestinationRect ) return;
	SourceRect 		 = new Rect[Length];
	if( !SourceRect ) return;
	
	DestinationPM	= DstPM;
	SourcePM			= SrcPM;
} // end constructor


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%   mvl_video_image_sequence   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//=====================================================================
//
//=====================================================================
IML_SINT mvl_video_image_sequence::store_all_bytes(register IML_UBYTE IML_PTR BytePtr) {

	if( !SequencePixMapHandle ) {
		iml_error("mvl_video_image_sequence","Must allocate PixMap before storing data!");
		return(FALSE);
		} // end if
	
	// THE FIRST FRAME INFO IS THE SAME AS THE FRAME SIZE.
	Rect IML_REF FrameSize = FramePositions[0];
	
	// TEMPORARY VARIABLES FOR CLARITY
	IML_ULINT RowCount = FrameSize.bottom * FrameCount;
	IML_ULINT RowWidth = FrameSize.right;
	
	// ALLOCATE A BUFFER ARRAY (BECAUSE OF PELLI'S IMPLEMENTATION)
	IML_ULINT IML_PTR ULArray = new IML_ULINT[RowWidth];
	if( !ULArray ) {
		iml_error("mvl_video_image_sequence","Could not allocate buffer!");
		return(FALSE);
		} // end if

	// LOCK DOWN THE DESTINATION PIXELS IN MEMORY
	LockPixels(SequencePixMapHandle); 			

	// STORE THE IMAGE STRIP INTO A LONG ARRAY, ONE ROW AT A TIME, AND STORE THAT ROW.
	// THIS IS REQURIED BECAUSE OF PELLI'S IMPLEMENTATION . . .
	for(IML_LINT YRow=0; YRow<RowCount; YRow++){
		// STORE A ROW IN THE UNSIGNED LONG BUFFER. USE UGLY register PTRS TO SPEED THINGS UP.
		register IML_ULINT IML_PTR LongPtr = ULArray;
		for(register IML_LINT XCol=0; XCol<RowWidth; XCol++) *LongPtr++ = *BytePtr++;
		// STORE THE ROW IN THE OFFSCREEN PIX MAP MEMORY
		//                              destination    x   y     source   count
		SetPixmapPixelsQuickly( *SequencePixMapHandle, 0, YRow, ULArray, RowWidth);  
		} // end for 

	// UNLOCK THE PIXELS FOR MEMORY MANAGEMENT REASONS
	UnlockPixels(SequencePixMapHandle);

	// CLEAN UP THE BUFFER
	delete [] ULArray;
	
	return(TRUE);
} // end store_sequence


//=====================================================================
//
//=====================================================================
mvl_video_image_sequence::~mvl_video_image_sequence() {
	if(OffScreenGWorld) DisposeGWorld(OffScreenGWorld);
	if(FramePositions) delete [] FramePositions;
} // end destructor

//=====================================================================
//
//=====================================================================
mvl_video_image_sequence::mvl_video_image_sequence(
					IML_USINT ColCount, IML_USINT RowCount, IML_USINT FCount) {
	
	FramePositions = NULL;  OffScreenGWorld = 0;

	IML_CLIPIT(FrameCount, 1,FCount,MAX_FRAMES);
	Rect FrameSize;// = {0,0,RowCount,ColCount};
	FrameSize.top = FrameSize.left = 0;
	FrameSize.bottom = RowCount;  FrameSize.right = ColCount;
	Rect SequenceRect = FrameSize;
	SequenceRect.bottom *= FrameCount;
	// allocate an offscreen memory for the images big enough for all the adjacent movie frames
	IML_SINT Error=NewGWorld(&OffScreenGWorld, 8, &SequenceRect,NULL,NULL,keepLocal+useTempMem);
	if(Error)Error=NewGWorld(&OffScreenGWorld, 8, &SequenceRect,NULL,NULL,keepLocal);
//	IML_SINT Error=NewGWorld(&OffScreenGWorld, 8, &SequenceRect,NULL,NULL,useTempMem);
//	if(Error)Error=NewGWorld(&OffScreenGWorld, 8, &SequenceRect,NULL,NULL,NULL);
	if(Error) {
		iml_error("mvl_video_image_sequence","Error allocating PixMap for image.");
		return;
		} // end if
		
	// define the offscreen world as a pixmap
	SequencePixMapHandle = GetGWorldPixMap(OffScreenGWorld);
	
	FramePositions = new Rect[FrameCount];
	if( !FramePositions )  {
		iml_error("mvl_video_image_sequence","Error allocating FramePositions.");
		return;
		} // end if
		
	// INITIALIZE ALL THE RECTS TO POINT TO EACH FRAME
	for(IML_USINT I=0; I<FrameCount; I++) {
		FramePositions[I] = FrameSize;
		FramePositions[I].top += I*RowCount;
		FramePositions[I].bottom = FramePositions[I].top+RowCount;
		} // end for
} // end constructor


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%   mvl_mac_video_information   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//=====================================================================
//
//=====================================================================
IML_SINT mvl_mac_video_information::store_blit_sequence(
								mvl_blit_sequence IML_PTR NewSeq) {
	if( !blit_storage_left() ) return(-1);
	// STORE THE FINAL DATA STRUCTURE
	BlitSequences[BlitSequenceCount] = NewSeq;
	// GRAB THAT BLIT'S ID FOR USER USE.
	IML_SINT BltSeqID = BlitSequenceCount;
	// INCREMENT THE STORAGE COUNTER
	BlitSequenceCount++;
	// RETURN THE IMAGE ID
	return(BltSeqID);
} // end store_blit_sequence


//=====================================================================
//
//=====================================================================
IML_SINT mvl_mac_video_information::store_image_sequence(
								mvl_video_image_sequence IML_PTR NewSeq) {
	if( !video_storage_left() ) return(-1);
	// STORE THE FINAL DATA STRUCTURE
	VideoSequences[VideoSequenceCount] = NewSeq;
	// GRAB THAT IMAGE'S ID FOR USER USE.
	IML_SINT ImSeqID = VideoSequenceCount;
	// INCREMENT THE STORAGE COUNTER
	VideoSequenceCount++;
	// RETURN THE IMAGE ID
	return(ImSeqID);
} // end store_image_sequence



//=====================================================================
//
//=====================================================================
void mvl_mac_video_information::clear_blit_sequences(void) {

	if(BlitSequences) { // FREE UP EACH MEMORY OBJECT
		for(IML_USINT I=0; I<BlitSequenceCount; I++) {
			delete BlitSequences[I];
			BlitSequences[I] = NULL;
			} // end for
		} // end if
	BlitSequenceCount = 0;
	
} // end clear_blit_sequences


//=====================================================================
//
//=====================================================================
void mvl_mac_video_information::clear_video_sequences(void) {

	if(VideoSequences) { 
		for(IML_USINT I=0; I<VideoSequenceCount; I++) {
			if(VideoSequences[I]) delete VideoSequences[I];
			VideoSequences[I] = NULL;
			} // end for
		} // end if
	VideoSequenceCount = 0;
		
	
} // end clear_blit_sequences


//=====================================================================
//
//=====================================================================
mvl_mac_video_information::~mvl_mac_video_information() { 
	// DELETE OUR DISPLAY WINDOW (ON SCREEN MEMORY SPACE)
	if(DisplayWindow) GDDisposeWindow(DisplayWindow);
	
	if(VideoSequences) {
		// FREE UP EACH REGISTERED OFFSCREEN VIDEO OBJECT
		for(IML_USINT I=0; I<VideoSequenceCount; I++) delete VideoSequences[I];
		// DELETE THE ARRAY SPACE
		delete [] VideoSequences;
		} // end if 
		
	if(BlitSequences) {
		// FREE UP EACH REGISTERED MEMORY OBJECT
		for(IML_USINT I=0; I<BlitSequenceCount; I++) delete BlitSequences[I];
		// DELETE THE ARRAY SPACE
		delete [] BlitSequences;
		} // end if
		
} // end destructor

//=====================================================================
//
//=====================================================================
mvl_mac_video_information::
			mvl_mac_video_information(IML_USINT MaxSeqCnt, IML_USINT MaxBltCnt) {
		
	Calibration				=	NULL;
	DisplayDevice			=	NULL;
	DisplayWindow			=	NULL;
	DisplayPixMapHandle	=	NULL;
 	ThunderCard					=	NULL;
	VideoSequenceCount 	=	0;
	VideoSequences			=	new mvl_video_image_sequence*[MaxSeqCnt];
	if(VideoSequences) {
		MaxVideoSequence = MaxSeqCnt;
		for(IML_USINT I=0; I<MaxVideoSequence; I++) {
			VideoSequences[I]=NULL;
			}// end for
	} else {
		MaxVideoSequence = 0;
		} // end else

		
	BlitSequenceCount 	=	0;
	BlitSequences			=	new mvl_blit_sequence*[MaxBltCnt];
	if(VideoSequences) {
		MaxBlitSequence = MaxBltCnt;
		for(IML_USINT I=0; I<MaxBlitSequence; I++) {
			BlitSequences[I]=NULL;
			}// end for
	} else {
		MaxBlitSequence = 0;
		} // end else

} // end constructor
