#ifndef __MVL_UTIL_H__
#define __MVL_UTIL_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "PCIThunder.h"

#define MAX_FRAMES		512
#define MAX_SEQUENCES	512
#define MAX_BLITS			8192

//=====================================================================
// CONTAINER CLASS FOR USE IN mvl_video_manager OPERATIONS
//=====================================================================
class mvl_blit_sequence {
	public:
		IML_USINT	Length;
		Rect		*	DestinationRect;
		Rect		*	SourceRect;
		PixMap	**	DestinationPM;
		PixMap	**	SourcePM;
		
		mvl_blit_sequence(IML_USINT Len, PixMap **DstPM, PixMap **SrcPM);
		~mvl_blit_sequence();
		
		IML_SINT successful(void) { return( (DestinationRect && SourceRect) ? TRUE : FALSE ); };

}; // end class


//=====================================================================
// CONTAINER CLASS FOR USE IN mvl_video_manager OPERATIONS
//=====================================================================
class mvl_video_image_sequence {
	public:
		GWorldPtr 	OffScreenGWorld;			
		PixMap 	**	SequencePixMapHandle;	
		IML_USINT	FrameCount;
		Rect 		*	FramePositions;

		mvl_video_image_sequence(IML_USINT ColCount, IML_USINT RowCount, IML_USINT FCount);	
		~mvl_video_image_sequence();
		IML_SINT successful(void) { return( (OffScreenGWorld && SequencePixMapHandle) ? TRUE : FALSE ); };
		IML_SINT store_all_bytes(IML_UBYTE IML_PTR BytePtr);
		//IML_SINT store_frame(IML_UBYTE IML_PTR BytePtr, IML_USINT FrameNumber);
		
}; // end class mvl_video_image_sequence


//=====================================================================
//	THIS CLASS CONTAINS ALL THE GRAPHICS INFORMATION THAT PELLI'S LIBRARY
// CODE NEEDS FROM TIME TO TIME.
//=====================================================================
class mvl_mac_video_information {
	public:
		LuminanceRecord				*	Calibration;
		GDHandle 							DisplayDevice;
		CWindowPtr 							DisplayWindow;
		PixMap 							**	DisplayPixMapHandle;
		WindowPtr 							OldPort;
		pci_thunder_30_1600_card *		ThunderCard;
		
		// STORE IMAGE SEQUENCES IN OFF-SCREEN MEMORY
		IML_USINT							MaxVideoSequence;  	// INITIALIZED ARRAY SIZE
		IML_USINT							VideoSequenceCount;	// CURRENT COUNT
		mvl_video_image_sequence	**	VideoSequences;		// POINTERS TO CONTAINERS
		
		// SEPARATELY STORE BLIT SEQUENCES; THIS ALLOWS US TO HAVE MULTIPLE
		// BLIT SEQUENCES FOR A SINGLE IMAGE, ALLOWING, FOR EXAMPLE, DUAL 
		// MOTION DIRECTION, OR MULTIPLE NOISE (BLIT) SEQUENCES FROM ONE 
		// IMAGE SEQUENCE.
		IML_USINT							MaxBlitSequence;  	// INITIALIZED ARRAY SIZE
		IML_USINT							BlitSequenceCount;	// CURRENT COUNT
		mvl_blit_sequence				**	BlitSequences;			// POINTERS TO CONTAINERS
		
	// CONSTRUCTOR AND DESTRUCTOR
	mvl_mac_video_information(IML_USINT MaxSeqCnt, IML_USINT MaxBltCnt);
	~mvl_mac_video_information();
	
	// RETURNS TRUE IF ALL THE MEMORY WAS ALLOCATED
	IML_SINT successful(void) { return( (VideoSequences && BlitSequences) ? TRUE : FALSE ); };
	
	// VALIDATE BLIT AND IMAGE/VIDEO SEQUENCE ID NUMBERS
	IML_USINT valid_video_id(IML_USINT VID){ return((VID < VideoSequenceCount) ? TRUE : FALSE); };
	IML_USINT valid_blit_id(IML_USINT BID) { return((BID < BlitSequenceCount)  ? TRUE : FALSE); };

	// RETURN A POINTER TO THE REQUESTED CONTAINER
	const mvl_video_image_sequence IML_PTR video(IML_USINT VID) { return(VideoSequences[VID]);};
	const mvl_blit_sequence IML_PTR blit(IML_USINT BID) { return(BlitSequences[BID]);};
	
	// CHECK IF ROOM IS LEFT IN THE CONTAINER ARRAYS
	IML_USINT video_storage_left(void) { return(VideoSequenceCount<MaxVideoSequence); };
	IML_USINT blit_storage_left(void) { return(BlitSequenceCount<MaxBlitSequence); };
	
	// IMPORTANT STORAGE AND CLEARAGE OPERATIONS.
	IML_SINT store_image_sequence( mvl_video_image_sequence IML_PTR NewSeq);
	IML_SINT store_blit_sequence( mvl_blit_sequence IML_PTR NewSeq);
	void clear_blit_sequences(void);
	void clear_video_sequences(void);

}; // end class mvl_mac_video_information

#endif // __MVL_UTIL_H__
