/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/


#include "PCIThunder.h"
#include <stdio.h>
const IML_ULINT pci_thunder_30_1600_card::Thunder30VBLStatus = 0x00C40004L;		//offset from card base address
const IML_ULINT pci_thunder_30_1600_card::Thunder30IndexReg  = 0x00D00002L;		//offset from card base address
const IML_ULINT pci_thunder_30_1600_card::Thunder30DataReg   = 0x00D00006L;		//offset from card base address
const IML_ULINT pci_thunder_30_1600_card::Thunder30VBL_Bit   = 0x00D00002L;		//offset from card base address

//====================================================================
//
//====================================================================
pci_thunder_30_1600_card::pci_thunder_30_1600_card(GDHandle device) :
	// class member initialization list

	//David Moe's way of getting base address of card's frame buffer
	CardBase( (IML_ULINT) (**(**device).gdPMap).baseAddr),	
	
	// initialize references for card register addresses
	ClutIndexReg( *( (IML_USINT*)( ( CardBase & 0xFF000000 ) + Thunder30IndexReg ) ) ),
	ClutDataReg(  *( (IML_USINT*)( ( CardBase & 0xFF000000 ) + Thunder30DataReg  ) ) ),
   VBLStatusReg( *( (IML_ULINT*)( ( CardBase & 0xFF000000 ) + Thunder30VBLStatus) ) )
  	// end class member initialization
  	{ 
  		
  	// nothing else to do in the constructor

} // end constructor


//====================================================================
//
//====================================================================
OSErr pci_thunder_30_1600_card::load_dacs(ColorSpec *table, IML_USINT Start, IML_USINT CLength) {

	register IML_USINT & ClutIndexRegRef = ClutIndexReg;
	register IML_USINT & ClutDataRegRef  = ClutDataReg;
	
	register IML_USINT *red   = &table[0].rgb.red;
	register IML_USINT *green = &table[0].rgb.green;
	register IML_USINT *blue  = &table[0].rgb.blue;
		
//IML_LINT Kludge = 512;
	for(register IML_USINT ClutIndex = Start; ClutIndex<(Start+CLength); ClutIndex++) {
		//assumes that the Clut Index Register counts up automatically
		//or expects three writes to the Clut Data Register
		ClutIndexRegRef = ClutIndex;

/*		ClutDataRegRef  = Kludge;//*red   >> 6;
		ClutDataRegRef  = Kludge;//*green >> 6;
		ClutDataRegRef  = Kludge;//*blue  >> 6;
*/
		ClutDataRegRef  = *red   >> 6;
		ClutDataRegRef  = *green >> 6;
		ClutDataRegRef  = *blue  >> 6;

		red   += 4;  	// advance by 4 shorts: r,g,b,alpha
		green += 4;  	// advance by 4 shorts: g,b,alpha,r
		blue  += 4;  	// advance by 4 shorts: b,alpha,r,g
//Kludge++;
		} // end for
		
	// for now, no error conditions are checked
	return(0);
} // end load_dacs


//====================================================================
//
//====================================================================
OSErr pci_thunder_30_1600_card::read_dacs(ColorSpec *table, IML_USINT Start, IML_USINT CLength) {

	register IML_USINT & ClutIndexRegRef = ClutIndexReg;
	register IML_USINT & ClutDataRegRef  = ClutDataReg;
	
	register IML_USINT *red   = &table[0].rgb.red;
	register IML_USINT *green = &table[0].rgb.green;
	register IML_USINT *blue  = &table[0].rgb.blue;

	for(IML_USINT ClutIndex = Start; ClutIndex<(Start+CLength); ClutIndex++) {
		//assumes that the Clut Index Register counts up automatically
		//or expects three writes to the Clut Data Register
		ClutIndexRegRef = ClutIndex;
		  *red   = ClutDataRegRef;		
		  *green = ClutDataRegRef;		
		  *blue  = ClutDataRegRef;		
		  red   += 4;  	// advance by 4 shorts: r,g,b,alpha
		  green += 4;  	// advance by 4 shorts: g,b,alpha,r
		  blue  += 4;  	// advance by 4 shorts: b,alpha,r,g
		} // end for
		
	// for now, no error conditions are checked
	return(0);
} // end read_dacs


//====================================================================
// Waits for the beginning of the next vertical blanking interval.
//====================================================================
OSErr pci_thunder_30_1600_card::wait_for_VBL(void) {

	register IML_ULINT & VBLStatusRegRef = VBLStatusReg;
	register IML_ULINT   VBL_Bit_Mask = Thunder30VBL_Bit;

//IML_ULINT Temp = 3000;
//while(Temp-- && (VBLStatusReg & VBL_Bit_Mask)) {}
	//while (  VBLStatusReg & VBL_Bit_Mask );  // if we're already blanking, wait till end.
//fprintf(stderr,"%asdfasdfasdfasdfasdfasdfasdfasdf\n");
/****
	do  {
 //Temp = VBLStatusReg;
		fprintf(stderr,"%d\n",VBLStatusRegRef);
		} while (  Temp-- && (VBLStatusRegRef & VBL_Bit_Mask) );  // if we're already blanking, wait till end.
	while ( !(VBLStatusReg & VBL_Bit_Mask) ) {
		};  // wait until beginning of blanking interval.
****/
	// for now, no errors are check for
	return(0);
} // end wait_for_VBL
