
//=========================================================================
// 
//   Source File: imlwin.cpp
//   Description: Source for imlwin application
//   Date:        Thu Jan 06 03:48:22 1994
//
//=========================================================================


#include "iml_win.h"
#include "iml_snds.h"
#include <string.h>
#include <mmsystem.h>

// Defines for menu item IDs
#define   IDM_RUNEXPERIMENT			101
#define   IDM_TEST_MOUSE				102
#define   IDM_QUIT						103
#define   IDM_ABOUTEXPERIMENT			104


// ASSIGN THE GLOBAL CLASS NAME
char iml_winapp_shell::szAppName[__IML__MAXNAME] = "IMLWIN";


// GLOBAL OBJECT FOR WINDOWS SHELL APPLICATIONS
iml_winapp_shell IMLWinApp("IML Experiment");


//=========================================================================
// WinMain PROCEDURE TO SATISFY THE WINDOWS STUFF
//=========================================================================
int PASCAL WinMain(HINSTANCE hInstance,      // Application Instance Handle
                   HINSTANCE hPrevInstance,  // Previous Instance Handle
                   LPSTR  lpszCmdLine,       // Pointer to Command Line
						 int    nCmdShow)  {        // Show Window Option

	int Status = IMLWinApp.initialize(hInstance,hPrevInstance,lpszCmdLine,nCmdShow);

	if(!Status) return(FALSE);

	IMLWinApp.CmdLineString = lpszCmdLine;

	return(IMLWinApp.run());

} // end WinMain


//=========================================================================
//  WINDOWS FUNCTION CALL BACK WRAPPER
//=========================================================================
LRESULT CALLBACK MainWndProc( HWND hWnd, UINT wMessage, WPARAM wParam, LPARAM lParam) {
	return(IMLWinApp.handle_main_message(hWnd,wMessage,wParam,lParam));
};


//=========================================================================
// WINDOWS FUNCTION CALL BACK WRAPPER
//=========================================================================
BOOL CALLBACK fnAboutBoxDlgProc( HWND hDlg, UINT wMessage,
			WPARAM wParam, LPARAM lParam) {

	return(IMLWinApp.handle_dlg_message(hDlg,wMessage,wParam,lParam));

} // end fnAboutBoxDlgProc



//=========================================================================
//  PRINT A STRING TO THE PSEUDO TERMINAL
//=========================================================================
void iml_winapp_shell::print(char *String) {

	static char *Controls = "\t\r\n";
	char *Here = String;

	while( *Here ) {
		IML_LINT ChunkSize = strcspn( Here, Controls );

		HDC DC = GetDC(hWndMain);

		WORD TextWidth = LOWORD(GetTextExtent(DC, Here, ChunkSize));

		if(ChunkSize) TextOut(DC, CurrentChar, CurrentLine * CharHeight , Here, ChunkSize);
		if( hDiskFile ) _lwrite(hDiskFile, Here, (UINT)ChunkSize);

		CurrentChar += TextWidth;
		Here += ChunkSize;
		switch(*Here) {
			case '\t':
				CurrentChar += __IML_TABSIZE__*CharWidth;
				Here++;
				if( hDiskFile ) _lwrite(hDiskFile, "\t", (UINT) 1);
				break;
			case '\n':
				// CONVERT \n TO CR/LF
				CurrentChar =  __IML_STARTCHAR__*CharWidth;
				CurrentLine++;
				Here++;
				if( hDiskFile ) _lwrite(hDiskFile, "\r\n", (UINT) 2);
				break;
			case '\r':
				CurrentChar =  __IML_STARTCHAR__*CharWidth;
				Here++;
				if( hDiskFile ) _lwrite(hDiskFile, "\r", (UINT) 1);
				break;
			} // end switch

		ReleaseDC(hWndMain,DC);

		while(CurrentLine >= MaxLine) {
			ScrollWindow(hWndMain, 0, -CharHeight, NULL, NULL);
			UpdateWindow(hWndMain); // TO PREVENT CLEARING OF THE LINE
			CurrentLine--;
			} // end if

		} // end while

} // end print


//=========================================================================
//  STRING PRINTER FOR ANY POSITION
//=========================================================================
void iml_winapp_shell::print(IML_USINT Row, IML_USINT Col, char *String) {
	HDC DC = GetDC(hWndMain);
	TextOut(DC, Col * CharWidth, Row * CharHeight, String, strlen(String));
	ReleaseDC(hWndMain,DC);
} // end print


//=========================================================================
// CLEAR THE PSEUDO TERMINAL
//=========================================================================
void iml_winapp_shell::clear(void) {

	InvalidateRect(hWndMain,NULL,TRUE);
	UpdateWindow(hWndMain);
	CurrentLine = __IML_STARTLINE__;
	CurrentChar = __IML_STARTCHAR__*CharWidth;

} // end clear (screen)


//=========================================================================
// DO-LITTLE CONSTRUCTOR FOR GLOBAL VARIABLE
//=========================================================================
iml_winapp_shell::iml_winapp_shell(const char *InstanceName) {

	// POINT GLOBAL VARIABLE AT 'this'
	IMLAppShell = this;

	for(int I = 0; *InstanceName && (I < __IML__MAXNAME); I++)
		 szInstanceName[I] = InstanceName[I];
	szInstanceName[__IML__MAXNAME-1] = 0;

}; // end constructor



//=========================================================================
// CREATE OR APPEND TO FILE FileName
//=========================================================================
void iml_winapp_shell::copy_output_to(const char *FileName) {

	// JUST IN CASE WE HAVE SOMETHING ELSE OPEN; I.E. WE CAN ONLY HAVE ONE
   // FILE OPEN AT A TIME.
	end_copy_output();

	UINT OpenFileMode;
	OFSTRUCT OFStruct;
	OFStruct.cBytes = sizeof(OFSTRUCT);

   // TEST FOR FILE EXISTENCE
	hDiskFile = OpenFile((LPCSTR)FileName, &OFStruct, OF_EXIST);

   // DETERMINE FILE OPENING MODE
	if(hDiskFile == HFILE_ERROR) 	// THE FILE DOES NOT EXIST
		OpenFileMode = OF_CREATE;
	else									// FILE DOES EXIST
		OpenFileMode = OF_WRITE;

   // OPEN THE FILE
	hDiskFile = OpenFile((LPCSTR)FileName, &OFStruct, OpenFileMode);
	if(hDiskFile == HFILE_ERROR) {
		iml_error("iml_winapp_shell::Error on opening file <%s>",(char *)FileName);

      // CLEAN UP THE MESS
      hDiskFile = NULL;
		} // end if

	// GO TO ZERO BYTES FROM THE END (2) OF THE FILE;  I.E. ALWAYS APPENDS
	_llseek(hDiskFile, 0, 2);

};


//=========================================================================
// CLOSE THE FILE IF IT IS OPEN
//=========================================================================
void iml_winapp_shell::end_copy_output(void) {

	if(hDiskFile) {

		if( _lclose( hDiskFile ) == HFILE_ERROR ) {
			print("iml_winapp_shell::Error on closing file");
			} // end if

      //  CLEAN UP
      hDiskFile = NULL;
		} // end if
};



//=========================================================================
// PROVIDE SOME SOUNDS
//=========================================================================
void iml_winapp_shell::sound(IML_SINT SoundNumber) {

	if( !SoundsEnabled && (SoundNumber != IML_DEFAULT_SOUND) ) return;

	switch(SoundNumber) {
		case 1:	MessageBeep(MB_ICONASTERISK);		break;

		case 2:	MessageBeep(MB_ICONQUESTION);		break;

		case 3:	MessageBeep(MB_ICONEXCLAMATION);	break;

		case 4:	MessageBeep(MB_ICONHAND);			break;

		case 5:	MessageBeep(MB_OK);					break;

		default: play_wave(SoundNumber);					break;
		} // end case

} // end sound



//=========================================================================
// PROVIDE SOME SOUNDS FROM .WAV FILES
//=========================================================================
void iml_winapp_shell::play_wave(IML_SINT SoundID) {

	HANDLE hResInfo = NULL;
	HANDLE hWave = NULL;
	LPSTR  lpWave = NULL;
   long int ResourceID = (0x0L | SoundID);
	if((hResInfo	= FindResource( hInst, (char*)ResourceID, "WAVE" )) != NULL &&
		(hWave		= LoadResource( hInst, hResInfo )) != NULL		&&
		(lpWave		= LockResource( hWave )) != NULL     ) {

      // IF WE SUCCEEDED, THEN PLAY THE THING, AND UNLOCK AND FREE IT
		sndPlaySound( lpWave, SND_MEMORY | SND_SYNC );
	} else {

		// DEFAULT WAS ASKED FOR, OR THE FIND/LOAD/LOCK SEQUENCY FAILED
		MessageBeep(-1);
      } // end else


	if(hWave) {
		UnlockResource( hWave );
		FreeResource( hWave );
      } //e nd if

} // end playwave



//=========================================================================
//  REAL INIT TAKES PLACE WHEN WinMain IS CALLED
//=========================================================================
#pragma argsused
int iml_winapp_shell::initialize(
			HINSTANCE hInstance,
			HINSTANCE hPrevInstance,
			LPSTR  lpszCmdLine,
			int    nCmdShow) {


	// NULL OUT FILE POINTER FOR LATER USE
	hDiskFile = NULL;


	WNDCLASS wndclass;

	if(!hPrevInstance) {
		wndclass.style           = CS_HREDRAW | CS_VREDRAW;
      wndclass.lpfnWndProc     = MainWndProc;
      wndclass.cbClsExtra      = 0;
      wndclass.cbWndExtra      = 0;
      wndclass.hInstance       = hInstance;
      wndclass.hCursor         = LoadCursor(NULL, IDC_ARROW);
		wndclass.hIcon           = 0;//LoadIcon(hInstance, "DEMO");
      wndclass.hbrBackground   = CreateSolidBrush((LONG)16777215L);
      wndclass.lpszMenuName    = "imlwin";
      wndclass.lpszClassName   = szAppName;

		if(!RegisterClass(&wndclass)) {
			MessageBox(NULL, "Unable to register window class", "System Error",
                   MB_SYSTEMMODAL | MB_ICONHAND | MB_OK);

			return(FALSE);
			} // end if
		} // end if

	// SAVE INSTANCE INFO
	hInst = hInstance;


	HDC DThdc = GetDC(NULL);  // NULL MEANS SCREEN DC
	HorzRes = GetDeviceCaps(DThdc,HORZRES);	// SAVE HORIZONTAL SCREEN SIZE
	VertRes = GetDeviceCaps(DThdc,VERTRES);   // SAVE VERTICAL SCREEN SIZE
	ReleaseDC(NULL,DThdc); // USE NULL FOR HWND PARM BECAUSE IT'S THE SCREEN

   // MAKE A WINDOW THAT FILLS THE ENTIRE SCREEN
   if(!(hWndMain = CreateWindow(szAppName,
							  szInstanceName,
							  WS_CAPTION ,
							  0,0,
							  HorzRes, VertRes,
                       NULL, (HMENU)NULL, hInstance, NULL))) {
		  MessageBox(NULL, "Unable to create main window", "System Error",
                   MB_SYSTEMMODAL | MB_ICONHAND | MB_OK);
		  return(FALSE);
        } // end if

	HDC MWhdc = GetDC(hWndMain);  // GET MAIN WINDOW INFORMATION
	DWORD TextInfo = GetTextExtent(MWhdc, "TESTTESTTESTTESTTEST", 20);
	CharHeight = HIWORD(TextInfo);
	CharWidth  = LOWORD(TextInfo) / 20.0;
	ReleaseDC(hWndMain,MWhdc); // USE NULL FOR HWND PARM BECAUSE IT'S THE SCREEN

   // MAKE SURE THAT WE HAVE A MINIMUM CHARACTER HEIGHT
	CharHeight = (CharHeight > 4) ? CharHeight : 4;

	// ADD SOME VERTICAL SPACING BETWEEN LINES
	CharHeight += 3;

   // COUNT NUMBER OF LINES WE HAVE ON THE SCREEN
	MaxLine = VertRes / CharHeight - 5;
	CurrentLine = __IML_STARTLINE__;
	CurrentChar = __IML_STARTCHAR__*CharWidth;


	ShowWindow(hWndMain, nCmdShow);
	UpdateWindow(hWndMain);

	return(TRUE);

} // end pseudo constructor


//========================================================================
//  MOUSE BUTTON STATE GETTER;  WAITS FOR A COMPLETE BUTTON PRESS, BUT
//  DOES NOT HANDLE DOUBLE CLICKS AT THIS TIME;
//========================================================================
button_info iml_winapp_shell::get_button_press(void) {
MSG      Msg;

	// GRAB ALL OF THE MOUSE EVENTS FOR OURSELVES!!
	// this crashes the system!!! SetCapture(hWndMain);

	while(
		PeekMessage(&Msg, NULL, WM_MOUSEFIRST, WM_MOUSELAST, PM_NOYIELD | PM_REMOVE)
		) { /* FLUSH MOUSE MESSAGES */ }

	while(
		PeekMessage(&Msg, NULL, WM_KEYFIRST, WM_KEYLAST, PM_NOYIELD | PM_REMOVE)
		) { /* FLUSH KEYBOARD MESSAGES */ }


	unsigned short int DnButton = NO_BUTTON;
	do { // GET A DOWN BUTTON OF SOME KIND
		GetMessage(&Msg, NULL, 0, 0 );
		TranslateMessage(&Msg);
		switch(Msg.message) {
			case WM_LBUTTONDOWN:	
			case WM_RBUTTONDOWN:
			case WM_MBUTTONDOWN:
				DnButton = Msg.message;
				break;
			} // end switch

		} while(DnButton == NO_BUTTON);


	button_info B(NO_BUTTON, NO_BUTTON);
	do { // GET THE CORRECT UP BUTTON
		GetMessage(&Msg, NULL, 0, 0 );
		TranslateMessage(&Msg);
		switch(Msg.message) {
			case WM_LBUTTONUP:
				if(DnButton == WM_LBUTTONDOWN) B.Clicked = LEFT_BUTTON;
				break;
			case WM_RBUTTONUP:
				if(DnButton == WM_RBUTTONDOWN) B.Clicked = RIGHT_BUTTON;
				break;
			case WM_MBUTTONUP:
				if(DnButton == WM_MBUTTONDOWN) B.Clicked = MIDDLE_BUTTON;
				break;
			case WM_LBUTTONDOWN:
			case WM_RBUTTONDOWN:
			case WM_MBUTTONDOWN:
			case WM_MOUSEMOVE:     // STUFF TO BE THROWN AWAY
			default:
				break;

			} // end switch

		} while( B.Clicked == NO_BUTTON );

	// ADD IN OTHER INFORMATION; SIMULTANEOUSLY PRESSED BUTTONS AND KEYS
	if(Msg.wParam & MK_SHIFT)   B.Modifiers |= SHIFT_KEY;
	if(Msg.wParam & MK_CONTROL) B.Modifiers |= CONTROL_KEY;
	if(Msg.wParam & MK_MBUTTON) B.Modifiers |= MIDDLE_BUTTON;
	if(Msg.wParam & MK_LBUTTON) B.Modifiers |= LEFT_BUTTON;
	if(Msg.wParam & MK_RBUTTON) B.Modifiers |= RIGHT_BUTTON;


	if( MouseClickEnabled ) sound(MouseSound);
	return(B);
} // end run

//========================================================================
//
//========================================================================
int iml_winapp_shell::run(void) {
MSG      Msg;

	while(GetMessage(&Msg, NULL, 0, 0)) {
		TranslateMessage(&Msg);
		DispatchMessage(&Msg);
		} // end while


	return(Msg.wParam);
} // end run


//=========================================================================
//
//=========================================================================
int iml_winapp_shell::handle_main_message (
					HWND hWnd,
					UINT wMessage,
					WPARAM wParam,
					LPARAM lParam) {

	switch(wMessage) {

		case WM_INITDIALOG :
			break;

		case WM_COMMAND :
			if(!LOWORD(lParam)) {         // Process Menu Commands
				switch(wParam) {
					case IDM_RUNEXPERIMENT: {
						run_experiment(CmdLineString);
						end_copy_output(); // IN CASE FILE REDIRECTION WAS USED...
						} // end case
						break;

					case IDM_TEST_MOUSE: {
               	
						int Flag = FALSE;
						do {

							button_info Button = IMLWinApp.get_button_press();
							if		 (Button.Clicked == LEFT_BUTTON)	print("LEFT_WBUTTON    ");
							else if(Button.Clicked == RIGHT_BUTTON)	print("RIGHT_WBUTTON   ");
							else if(Button.Clicked == MIDDLE_BUTTON)	print("MIDDLE_WBUTTON  ");
							else Flag = TRUE;

							if(Button.Modifiers & SHIFT_KEY)			print("    M-SHIFT     ");
							if(Button.Modifiers & CONTROL_KEY)		print("    M-CONTROL   ");
							if(Button.Modifiers & LEFT_BUTTON)		print("    M-LEFT      ");
							if(Button.Modifiers & RIGHT_BUTTON)		print("    M-RIGHT     ");
							if(Button.Modifiers & MIDDLE_BUTTON)	print("    M-MIDDLE    ");

							if(Button.Clicked == MIDDLE_BUTTON) Flag = TRUE;

                  	} while ( !Flag );

						} // end case
               	break;

				 	case IDM_QUIT: {
						// RELEASE THE MOUSE
						//ReleaseCapture();
						PostQuitMessage(0);
						} // end case
				 		break;

					case IDM_ABOUTEXPERIMENT: {
					 	if(fnAboutBox(hWnd)) {
							} // end if
                	} // end case
					 	break;

					} // end switch
				} else {

					switch(wParam) {
						default: break;

						} // end switch

          		} // end else
           
			 break; // WM_COMMAND

       case WM_DESTROY :

			 PostQuitMessage(0);
          break;

       default :
          return DefWindowProc(hWnd, wMessage, wParam, lParam);
	} // end switch

	return(0L);
} // end handle_main_message


//=========================================================================
//
//=========================================================================
#pragma argsused
int iml_winapp_shell::fnAboutBox(HWND hParentWnd) {
   int   RetCode;

   DLGPROC  lpfnfnAboutBoxDlgProc;


	lpfnfnAboutBoxDlgProc = (DLGPROC)MakeProcInstance((FARPROC)fnAboutBoxDlgProc, hInst);


	if((RetCode = DialogBox(hInst,
									(LPCSTR)"AboutBox",
									hParentWnd,
									lpfnfnAboutBoxDlgProc)) == -1) {
        MessageBox(NULL, "Unable to display dialog", "System Error",
                   MB_SYSTEMMODAL | MB_ICONHAND | MB_OK);
		  return FALSE;
   	} // end if

	FreeProcInstance((FARPROC)lpfnfnAboutBoxDlgProc);
   return(RetCode);
}


//=========================================================================
//
//=========================================================================
#pragma argsused
int iml_winapp_shell::handle_dlg_message(
			HWND hDlg,
			UINT wMessage,
			WPARAM wParam,
			LPARAM lParam) {


	switch(wMessage) {

       case WM_INITDIALOG :


            return TRUE;

       case WM_COMMAND : 
				switch(wParam) {
               case IDOK :

					  EndDialog(hDlg, TRUE);
                 return TRUE;

					default :
                  return FALSE;

             } // end switch

       case WM_SYSCOMMAND :

				switch(wParam & 0xFFF0) {
                case SC_CLOSE :
                  EndDialog(hDlg, FALSE);
                  return TRUE;
            }// end switch
	} // end switch

	return FALSE;

} // end handle_dlg_message

