#include <stdio.h>

#include "iml_rand.h"

#include "iml_qest.h"

#include "iml_app.h"


// FOR BUILDING MESSAGES IN OUR SAMPLE CODE
char QuestMsg[200];


//=====================================================================
// USE THIS FOR TESTING THINGS
//=====================================================================
class random_weibull_user_input : public user_input {
	protected:
		adjustable_parameter IML_REF ThresholdVariable;

	public:
		IML_FLOAT Alpha;
		IML_FLOAT Beta;
      IML_UBYTE MAFC;
		// RANDOM VARIABLES
		scalar_rv IML_REF RVar;

		random_weibull_user_input(adjustable_parameter IML_REF TV, scalar_rv IML_REF RV) :
			ThresholdVariable(TV), RVar(RV) {
			Alpha = 0.8;
			Beta = 2.0;
			MAFC = 2;    srand(12);
			};

		IML_BYTE response(void) {
//			IML_FLOAT Sample = random(100); //RVar.sample();
			IML_FLOAT Sample = RVar.sample();
			IML_FLOAT Prob = 100*Weibull(ThresholdVariable, Alpha, Beta, MAFC);
			sprintf(QuestMsg,"random user tv<%f> sample<%f> prob<%f> <%d>\n",
				ThresholdVariable, Sample, Prob, (Sample <= Prob) );
			IMLAppShell->print(QuestMsg);
			return( Sample <= Prob );
			};

}; // end class


//=====================================================================
// FOR TESTING
//=====================================================================
class test_stimulus : public adjustable_stimulus {
	public:
		adjustable_parameter ParameterValue;

		test_stimulus(IML_DOUBLE InitParm) : ParameterValue(InitParm) {};

		adjustable_parameter IML_REF threshold_variable(void) {
			return(ParameterValue);
			};

		void present(void) { /* do nothing */ };

}; // end class


//=======================
#ifdef _Windows
int run_experiment(void) {
#else
int main () {
#endif


	// RANDOM VARIABLES
	gaussian_rv G1(0, 1);

	exponential_rv E1(1);

	uniform_rv U1(0, 1);


	const IML_FLOAT DBStep 		= 0.08;	// STEP SIZE; 0.2 dB IS A FACTOR OF 1.047
	const IML_FLOAT BetaValue	= 2.0;	// USE DEFAULT BETA (FOR WEIBULL) OF 2.0
   const IML_UBYTE AFCCount	= 2;		// ASSUME AFC OF 2
	const IML_UBYTE PLife		= 15;		// HOW MANY TRIALS TO USE THE PRIOR INFO
	const IML_UBYTE Trials		= 50;		// TOTAL TRAIL COUNT
	const IML_UBYTE Samples		= 20;		// NUMBER OF CONTRAST POINTS TO SAMPLE (INITIALLY)
	const IML_FLOAT PWeight		= 0.3;	// WEIGHTING FACTOR OF PRIOR INFORMATION

	IML_DOUBLE RealAlpha = 0.8;
	test_stimulus TestStim(0.5);  // INITIALIZE THRESHOLD VARIABLE

	U1.seed(-4314);
	random_weibull_user_input Subject( TestStim.threshold_variable(), U1);

	Subject.Alpha = RealAlpha; // GENERATE GUESSES FOLLOWING "TRUE" ALPHA
	Subject.Beta = BetaValue;  // ASSUME BETA IS CORRECT
   Subject.MAFC = AFCCount;   // USE SAME AFC COUNT

	wm_quest_experiment Experiment( TestStim, Subject);

	//	 CONFIGURE THE QUEST EXPERIMENT
	Experiment.configure(		DBStep,
										BetaValue, AFCCount,
										PLife, Trials,
										Samples, PWeight
										);


	// SET PARAMETER TO BEST ESTIMATE ACCORDING TO PRIOR ALPHA/BETA;
	// ALPHA IS THE INIT VALUE FOR THE STIMULUS PARAMETER
	Experiment.wm_quest_adjust(BEST_ESTIMATE);

sprintf(QuestMsg,"best contrast is <%f>\n",(float)TestStim.ParameterValue);
IMLAppShell->print(QuestMsg);


	// LOOP UNTIL WE ARE THROUGH
	while( !Experiment.staircase_complete() ) {

		Experiment.make_a_step();
sprintf(QuestMsg,"next contrast is <%f>\n",(float)TestStim.ParameterValue);
IMLAppShell->print(QuestMsg);
//getch();
	} // end while

sprintf(QuestMsg,"trial number <%d>\n",Experiment.TrialCount);
IMLAppShell->print(QuestMsg);

	for(const wm_quest_item IML_PTR WMQPtr = Experiment.WMQMgr.first();
		  WMQPtr != NULL; WMQPtr = Experiment.WMQMgr.next() ) {
		sprintf(QuestMsg,"t<%.4f> seen<%5.5ld> notseen<%5.5ld>\n",
				WMQPtr->ParameterValue, WMQPtr->Seen, WMQPtr->NotSeen);
      IMLAppShell->print(QuestMsg);

		} // end for
	sprintf(QuestMsg,"datumcount = <%ld>\n",(long)Experiment.WMQMgr.DatumCount);
   IMLAppShell->print(QuestMsg);
	sprintf(QuestMsg,"bestest = <%f>\n",(float)Experiment.wm_quest_adjust(BEST_ESTIMATE));
   IMLAppShell->print(QuestMsg);

	// MAKE THE COMPILER HAPPY BY RETURNING A VALUE
	return(0);

} // end main
