//=====================================================================
// THESE ARE CLASSES INTENDED TO MAKE DESIGNING AND CONTROLLING
// EXPERIMENTS EASIER. :-)
//=====================================================================

// THIS PRAGMA OPTION CAUSES TEMPLATE FUNCTION BODIES TO BE GENERATTED IN THIS FILE;
// OTHE FILES SHOULD BE COMPILED "TEMPLATE GENERATION EXTERNAL" IN "C++ Options"

#ifdef _Windows
#pragma option -Jgd
#endif
#include "iml_exps.h"

//#include "imlapp.h"
IML_BYTE adjustable_parameter::Buf[100] = "0.0";

//=====================================================================
// A MANAGER FOR GENERIC LISTS OF DOUBLY LINKED ITEMS
//=====================================================================
template <class DLLType>
DLLType IML_REF dl_list_manager<DLLType>::operator[](IML_LINT Index) {

	if( (Index < 0) || (Index >= DatumCount) ) {
		iml_error("dl_list_manager","Index out of bounds!!!");
		// RETURN USELESS, UNDEFINED VALUE;  BUT MAKE IT A VALID OBJECT
		return(Safety);
		} // end if

	// SEARCH FOR N'TH POSITION
	for(dl_list_item IML_PTR Here = (DLLType IML_PTR) Head.Next;
		Index > 0;   // THIS IS '>' RATHER THAN '>=' ON PURPOSE
		Index--) {
		Here = Here->Next;
		} // end for

	return((DLLType IML_REF) (*Here));
} // end operator[]


//========================================================================
//
//========================================================================
template <class DLLType>
void dl_list_manager<DLLType>::clear_data(void) {

	DLLType IML_PTR Here = (DLLType IML_PTR) Head.Next;
	while( Here != (fos_item IML_PTR) &Tail) {
		Here->remove();
		delete Here;
		Here = (DLLType IML_PTR) Head.Next;
		} // end while
	DatumCount = 0;

} // end clear_data


//========================================================================
//
//========================================================================
template <class DLLType>
DLLType IML_PTR dl_list_manager<DLLType>::find_or_add_item(IML_DOUBLE ParmValue) {

	// SEARCH FOR PROPER POSITION
	dl_list_item IML_PTR Here = &Head;   // HEAD/TAIL PARM VALUES ENSURE END TO LOOP :-)
	while( Here->sort_value() < ParmValue ) { Here = Here->Next; }

	// AT THIS POINT WE EITHER FOUND AN EQUAL PARM VALUE OR WE FOUND
	// AN INSERTION POINT;  WE ARE POINTING TO ONE THAT IS GREATER THAN
	// OR EQUAL TO THE ADDED DATUM.

	// IF WE NEED TO INSERT A NEW ONE;
	// (OTHERWISE WE TAKE THE ONE THAT Here POINTS TO)
	if( Here->sort_value() > ParmValue ) {
		// FIRST CREATE NEW ONE WITH PROPER POINTERS;

		Here = new DLLType(ParmValue, Here, Here->Prev);
		if( !Here ) {
			iml_error("dl_list_manager", "Could not allocate new data point in add_datum1!!");
			return(&Safety);
			} // end if

		DatumCount++;


		// THEN RE-POINT THE POINTERS OF THE GUYS HE GETS INSERTED BETWEEN;
		Here->Next->Prev = Here;
		Here->Prev->Next = Here;
		} // end if

	return( (DLLType IML_PTR) Here);
}  // end find_or_add_item




//========================================================================
//  CHECK TO SEE IF ALL EXPERIMENTS ARE COMPLETED
//========================================================================
IML_BYTE block_staircase_experiment::block_complete(void) {
	IML_BYTE Flag = TRUE;

	for(int I = 0; I < Count; I++)
		Flag *= Block[I]->staircase_complete();

	return(Flag);
} // end block_complete


//========================================================================
// ONE PRESENTATION, ONE STIMULUS, CHOSEN RANDOMLY
//========================================================================
void block_staircase_experiment::make_a_random_step(void) {

	if( block_complete() )  return;

	IML_UBYTE Index = random_index();
	Block[Index]->make_a_step();

} // end make_a_random_step



//========================================================================
// ONE TRIAL FOR EACH STIMULUS, RANDOMIZED ORDER
//========================================================================
void block_staircase_experiment::make_a_block_step(void) {
IML_SINT I;

	if( block_complete() )  return;

	// FILL FIXED ARRAY WITH INITIAL VALUES;  COMPLETED EXPERIMENTS NOT INCLUDED;
	for( I = 0; I < Count; I++) {
   	IML_UBYTE Index = I;
		if( Block[Index]->staircase_complete() ) Index = -1;
		Indices[I] = Index;
		} // end for

	// NOW RANDOMLY EXCHANGE ELEMENTS max(10,Count) TIMES
   IML_SINT ExchangeCount = (Count > 10 ? Count : 10);
	for( I = 0; I < ExchangeCount; I++) {
		IML_UBYTE Index1= random_index();
		IML_UBYTE Index2= random_index();
		IML_SINT Temp = Indices[Index1];
		Indices[Index1] = Indices[Index2];
      Indices[Index2] = Temp;
		} // end for

	// NOW DO THE PRESENTATIONS
	for(I = 0; I < Count; I++) {
      IML_SINT Index = Indices[I];
		if(Index >= 0) Block[Index]->make_a_step();
      if(Block[Index]->abort_requested()) return;
		} // end for

} //end make_a_block_step



//========================================================================
// A BLOCK OF N PRESENTATIONS, STIMULI CHOSEN AT RANDOM
//========================================================================
void block_staircase_experiment::make_a_random_block_step(IML_UBYTE Size) {
IML_SINT I;

	if( block_complete() )  return;

	// ADD UP ESTIMATED STEPS REMAINING AND STORE IN INDICES[] TEMPORARILY
   IML_FLOAT Total = 0;
	for(I=0; I < Count; I++) {
		Indices[I] = Block[I]->estimated_steps();
		Total += Indices[I];
      } // end for

   // CREATE THE DISTRIBUTION OF ESTIMATED STEPS REMAINING
	IML_FLOAT Level = 0.0;
	for(I=0; I < Count; I++) {
		Distribution[I] = Level + Indices[I]/Total;
//IMLAppShell->printf("Distribution[%d]=%lf",I,(double)Distribution[I]);
		Level = Distribution[I];
		} // end for

	// SELECT Size INDICES ACCORDING TO THE PROBABILITY DISTRIBUTION
	for(I=0; I < Size; I++) {

   	// GENERATE A UNIFORM RANDOM NUMBER
		IML_FLOAT Prob = UniformRV.sample(); // VALUE BETWEEN 0.0001 AND 0.9999

		// INDEX INTO DISTRIBUTION TO SELECT AN INDEX
		IML_UBYTE Index;
		for(Index=0; (Prob > Distribution[Index]) && (Index < Count); ) {
      	Index += 1;
			} // end for

//IMLAppShell->printf("choose index <%d> via <%lf>", Index, (double) Prob);
		// PREVIOUS INTERVAL WAS THE ONE ??
		if(Index == Count) {  // FALLING OFF THE END SHOULD NOT HAPPEN IN AN IDEAL WORLD
			Index = Count - 1; // strangeness????
			iml_error("make_a_random_block_step","WEIRDNESS!!!!");
			} // end if

		Indices[I] = Index;
		} // end for
	
	// NOW DO THE PRESENTATIONS
	for(I = 0; I < Size; I++) {
		IML_SINT Index = Indices[I];
		if(Index >= 0) Block[Index]->make_a_step();
		if(Block[Index]->abort_requested()) return;
		} // end for


} // end make_a_random_block_step



