// THIS IS THE FILE CONTAINING THE CODE FOR WILLIAM MACILHAGGA'S
// QUEST PROCEDURE.  THE CLASS DEFINITIION IS IN wmquest.h

#include "iml_type.h"
#include "iml_math.h"
#include "iml_qest.h"


//#include "imlapp.h"

//============================================================================
// Weibull APPROXIMATOR FOR THE PSYCHOMETRIC FUNCTION;
// GIVES THE PROBABILITY OF BEING CORRECT AT A GIVEN CONTRAST FOR AN
// M-AFC WEIBULL HAVING PARAMETERS ALPHA, BETA
//============================================================================
IML_DOUBLE Weibull(IML_DOUBLE contrast, IML_DOUBLE alpha,
						 IML_DOUBLE beta, 	 IML_UBYTE m) {
IML_DOUBLE p,e,pr;

	p = iml_pow(contrast/alpha,beta);
	e = iml_exp(-(IML_DOUBLE)p);
	pr = 1.0 - e/(IML_DOUBLE)m;
	return(pr);

} // end Weibull


//============================================================================
// Quest initialisation routine; MUST BE CALLED *AFTER* OTHER STUFF IS INITED.
//============================================================================
void wm_quest_experiment::init_quest(IML_UBYTE InitCount, IML_FLOAT Weight) {
IML_LINT I;
IML_DOUBLE Prob, LogQ;
wm_quest_item IML_PTR Item;

	//  work out the contrasts/alphas to be used, based on StepSize
	//  the number which, when multiplied to ThresholdProb, raises the
	//  probability by an amount = StepSize 
	IML_DOUBLE ThresholdProb = Weibull( Alpha, Alpha, Beta, AFCCount);
	//IML_DOUBLE Ratio = (1.0-ThresholdProb-StepSize) / (1.0-ThresholdProb);
	//IML_DOUBLE ContrastStep = iml_pow( iml_sqrt(1.0-iml_log(Ratio)), 1.0/Beta);

	// SET FOR DECIBEL STEP SIZES
	IML_DOUBLE ContrastStep = iml_pow(10.0, -StepSize/20.0);

   adjustable_parameter Value;
	for (I=0; I<InitCount; I++) {
		Value = iml_pow(ContrastStep,((IML_FLOAT)I)-(InitCount/2))*Alpha;
		WMQMgr.find_or_add_item( Value );
		} // end for

	// work out the values for the prior; create dummy data 
	for (Item = WMQMgr.first(); Item ; Item = WMQMgr.next() )
		Item->Temp = Weibull(Item->ParameterValue,Alpha,Beta,AFCCount);

	// the expected # correct assuming alpha beta are right on 
	for (I=0; I<InitCount; I++) {

		// FOR EACH ONE IN THE LIST
		wm_quest_item IML_REF Cur = WMQMgr[I];

		for (Item = WMQMgr.first(); Item ; Item = WMQMgr.next() ) {

			Prob = Weibull(Item->ParameterValue, Cur.ParameterValue, Beta, AFCCount);

			// BE CAREFUL OF OVERFLOW
			LogQ = (Prob > 0.9999999) ? -16.1 : iml_log(1.0-Prob);

			Cur.PriorL += Weight*
				(Item->Temp*iml_log(Prob) + (1.0 - Item->Temp)*LogQ);
			} // end for
		} // end for

}; // end init_quest


//============================================================================
//
//============================================================================
void wm_quest_experiment::clear_prior(void) {

	wm_quest_item IML_PTR Item = WMQMgr.first();
	while(Item) {
		Item->PriorL = 0.0;
		Item = WMQMgr.next();
		} // end while

}; // end clear_prior


//============================================================================
//
//============================================================================
void wm_quest_experiment::modify_prior(IML_FLOAT Factor) {

	wm_quest_item IML_PTR Item = WMQMgr.first();
	while(Item) {
		Item->PriorL *= Factor;
		Item = WMQMgr.next();
		} // end while

}; // end modify_prior


//============================================================================
//	Computes new posterior probs for each contrast/alpha, returns the best
//	contrast/alpha to use. The routine has been modified to extend the range of
//	contrasts/alphas that it returns.  This is in practice probably not useful because
//	Quest will rarely work properly if the true contrast differs much from
//	the prior.
//	Note the following convention for response:
//	NOT_SEEN - incorrect response == FALSE;   SEEN - correct response == TRUE;
//	BEST_ESTIMATE - no response, just want to know current best alpha estimate.
//============================================================================
IML_DOUBLE wm_quest_experiment::wm_quest_adjust(quest_response Response) {
IML_FLOAT BestP, Prob;
wm_quest_item IML_PTR Item;
wm_quest_item IML_PTR BestI;

	  // MUST UPDATE THE FOS DATA
	if (Response != BEST_ESTIMATE) {

//  IMLAppShell->printf("add resp <%d> for alpha <%lf>",Response,(double)Alpha);
		// SEARCH FOR CORRECT CONTRAST/ALPHA "BIN" VALUE
		// IF ALPHA IS NOT IN THE LIST, THEN ADD TO LIST
		WMQMgr.add_datum(Alpha, Response);

		// UPDATE ALL THE LOG LIKELIHOODS 
		Item = WMQMgr.first();
		while( Item ) {

			Prob = Weibull(Alpha, Item->ParameterValue, Beta, AFCCount);

			IML_FLOAT Temp = 	( Response == SEEN) ? iml_log(Prob) :
									((Prob>0.9999999)   ? -16.2 : iml_log(1.0-Prob) );

			Item->SumLogL += Temp;

         // CHECK NEXT ONE
			Item = WMQMgr.next();
			} // end while

   } // end if

	// SELECT THE BEST ESTIMATE: USE THE MAXIMUM LIKELIHOOD WITH PRIOR LIKELIHOOD
	IML_SINT Index = 0;
	Item = WMQMgr.first();
	BestP = Item->SumLogL + Item->PriorL;
	//BestI = Item;
	IML_SINT BestIndex = Index;
	for(Item = WMQMgr.next(); Item; Item = WMQMgr.next()) {
		Index++; // Index WILL BE THE POSITION IN THE LIST
		IML_FLOAT Temp = Item->SumLogL + Item->PriorL;
		if( BestP < Temp ) {
			BestP = Temp;
			//BestI = Item;
         BestIndex = Index;
			} // end if
      } // end for

//  IMLAppShell->printf("best est is <%d>",BestIndex);
  // ADD A RANDOM JITTER BIASED TOWARD VISIBILITY IF THIS IS THE EXPERIMENT
	if (Response != BEST_ESTIMATE) {
		BestIndex += int( 3*UniformRV.sample() ) - 1;  // ADD -1, 0, OR 1
		if(BestIndex < 0)
			Alpha = WMQMgr[0].ParameterValue;// / ContrastStep;
		else if (BestIndex >= WMQMgr.DatumCount)
			Alpha = WMQMgr[WMQMgr.DatumCount-1].ParameterValue;// ContrastStep;
		else {
			// if best item
			// Alpha = BestI->ParameterValue

			// IF RANDOM JITTER
			Alpha =  WMQMgr[BestIndex].ParameterValue;
         } // end else
		} // end if

//  IMLAppShell->printf("randomxized is <%d> for new alpha <%lf>",BestIndex,(double)Alpha);


	//Alpha = BestI->ParameterValue * Jitter;

   return(Alpha);
} // end wm_quest_adjust


//=====================================================================
// REPLACING VIRTUAL FUNCTION IN THE BASE CLASS;  MAKE A STEP IN THE STAIRCASE;
//=====================================================================
void wm_quest_experiment::make_a_step(IML_SINT InFlags) {

	// REMOVE PRIOR AT PROPER TIME
//	if(TrialCount >= PriorLife) clear_prior();
	if(TrialCount >= PriorLife) {

		// THIS METHOD REDUCES THE PRIOR WEIGHT OVER ABOUT 20 STEPS.
		// AFTER ABOUT 10 STEPS IT IS DOWN TO ABOUT 1/5 OF THE ORIGIONAL WEIGHT.
		// AFTER ABOUT 20 STEPS IT IS ABOUT 1/25 OF THE ORIGINAL WEIGHT.
   	IML_USINT Step = TrialCount - PriorLife + 1;
		modify_prior( iml_abs( iml_cos(IML_2PI * Step/45) ) );
      } // end if

	// PRESENT STIMULUS HERE
	IML_USINT StimFlags = Stimulus.present(InFlags);

	// GET THE SUBJECT'S RESPONSE
	IML_LINT Answer = Subject.response(StimFlags);

//	if( Subject.repeat_requested() || Subject.abort_requested() ) return;

   // ADJUST PARAMETER VALUE ACCORDING TO THE ANSWER 
	wm_quest_adjust((quest_response)Answer);

	TrialCount++;

} // end make_a_step


/* test routines to ensure that quest works */
/*
void main(void)
{ float al,bet,contrast,p;
  int i,n,response;
  float *con,*correct,*trials;

  randomize();
  printf("Give prior alpha, beta: "); scanf("%f %f",&al,&bet);
  InitQuest(al,bet,0.08,2,0.3);
  printf("Give real alpha, beta: "); scanf("%f %f",&al,&bet);
  // do 20 trials 
  contrast = Quest(0.0,-1);
  for (i=0; i<50; i++)
   { p = 100*Weibull(contrast,al,bet,2);
     if (random(100)<p) response = 1; else response = 0;
     printf("%d  contrast %f  response %d\n",i,contrast,response);
     contrast = Quest(contrast,response);
     //getchar();
     if (i==15) RemovePrior();
   }

   printf("info: "); getchar();
   QuestInfo(&con,&correct,&trials,&n);
   for (i=0; i<n; i++)
     { if (trials[i]==0.0) correct[i]=0.0; else correct[i] /= trials[i];
     printf("%f  %f %f  (%d) (%f)\n",con[i],correct[i],
	     Weibull(con[i],al,bet,2),(int)trials[i], (float)Q_Prob[i]); }
  printf("hit key: "); getchar();
}
*/
