#ifndef __IML_QEST_H__
#define __IML_QEST_H__

#include <stdlib.h>
#include <stdio.h>
#include "iml_math.h"
#include "iml_rand.h"
#include "iml_exps.h"


IML_DOUBLE Weibull(IML_DOUBLE contrast, IML_DOUBLE alpha,
						 IML_DOUBLE beta, 	 IML_UBYTE m);

//=====================================================================
//  A SINGLE STAIRCASE EXPERIMENT
//=====================================================================
typedef enum {BEST_ESTIMATE = -1, NOT_SEEN = 0, SEEN = 1} quest_response;

class wm_quest_experiment  : public staircase_experiment {
	uniform_rv UniformRV;  // USED TO RANDOMIZE THE BEST GUESS

	protected:
		// A STIMULUS OBJECT SHOULD USE THE adjustable_parameter TYPE FOR ALL
		// STIMULUS PARAMETERS THAT MIGHT BE USED FOR MEASURING A THRESHOLD.
		adjustable_parameter IML_REF Alpha;

		// the value of beta used in Weibull functions in wm_quest
		IML_FLOAT Beta;

		// STEP SIZES FOR INCEASING OR DECREASING VISIBILITY
		IML_FLOAT	StepSize;

      // HOW LONG TO USE THE PRIOR VALUES
		IML_UBYTE PriorLife;

      // THE TOTAL NUMBER OF TRIALS TO USE
		IML_UBYTE MaxTrials;

		// N-alternative-forced-choice
		IML_UBYTE AFCCount;

	public:

		// TRIAL COUNTER
      IML_UBYTE TrialCount;

		// REFERENCE TO POINT TO STIMULUS OBJECT
		adjustable_stimulus IML_REF Stimulus;

      // REFERENCE TO POINT TO SUBJECT RESPONSE GETTER
      user_input IML_REF Subject;

		// WE WANT TO KEEP RESPONCE DATA FOR EACH THRESHOLD PARAMETER VALUE
		wm_quest_data_manager WMQMgr;

		// CONSTRUCTOR
		wm_quest_experiment(
					adjustable_stimulus IML_REF Stim,
					user_input IML_REF UserIn) :
							Stimulus(Stim),
							Subject(UserIn),
							Alpha(Stim.threshold_variable()) {

			configure(
					1.0,			// 1 DECIBEL steps for the estimated psych. func
					2.0, 2,		// Beta = 2.0; AFC Count = 2;
					15, 50,		// KILL PRIOR AFTER 15 STEPS;  DO 50 TRIALS
					20, 0.3     // START WITH 20 SAMPLES; WEIGHT OF PRIOR = 0.3
					);				

			}; // end constructor

      // REQUIRED BY
		IML_SINT estimated_steps(void) { return(MaxTrials - TrialCount); };

		// SEE THE SUBROUTINE CALLS...
		void configure(IML_FLOAT StepS= 1.0,   // DEFAULT 1 DECIBEL STEP SIZE
							IML_FLOAT BetaV = 2, IML_UBYTE AFCC = 2,
							IML_UBYTE PLife = 15, IML_UBYTE Trials = 50,
							IML_UBYTE InitCount = 20, IML_FLOAT Weight = 0.3) {
			StepSize = StepS;
			Beta = BetaV;
         AFCCount = AFCC;
			PriorLife = PLife;
			MaxTrials = Trials;
			reset();
			init_quest(InitCount,Weight);
      	};

		void init_quest(IML_UBYTE InitCount, IML_FLOAT Weight);

		void clear_prior(void);

		void modify_prior(IML_FLOAT Factor);

		void reset(void) { TrialCount = 0; WMQMgr.clear_data(); };

		IML_DOUBLE wm_quest_adjust(quest_response Response);

		// REPLACING VIRTUAL FUNCTION OF BASE CLASS; TEST FOR COMPLETE;
		IML_BYTE staircase_complete() { return(TrialCount >= MaxTrials); };

		IML_BYTE abort_requested(void) {return(Subject.abort_requested()); };

		// REPLACING VIRTUAL FUNCTION IN THE BASE CLASS;  MAKE A STEP IN THE STAIRCASE;
		void make_a_step(IML_SINT InFlags = 0);

}; // end class


#endif // __IML_QEST_H__
