//=====================================================================
// THESE ARE CLASSES INTENDED TO MAKE DESIGNING AND CONTROLLING
// EXPERIMENTS EASIER. :-)
//=====================================================================

#include "iml_udsc.h"


//========================================================================
// GIVE SUBJECT'S ANSWER TO STAIRCASE MANAGER AND ASK FOR NEXT ACTION;  SEE ENUM ABOVE;
//========================================================================
staircase_response staircase_manager::next_action(IML_SINT YesCorrectTrue) {
			// SUBJECT WAS CORRECT/SAID YES ==> NON-ZERO;  INCORRECT/NO == ZERO

	staircase_response Resp;

	// FIND OUT WHICH DIRECTION WE ARE SUPPOSED TO CHANGE IN
	if( !YesCorrectTrue ) { // GOT A NO/FALSE/INCORRECT
		TotalYeses = 0;
		Resp.Action = INCREASE_VISIBILITY;

		// ELSE WE GOT A YES
		}  else if( (TotalYeses++) == RequiredYeses) {
			TotalYeses = 0;
			Resp.Action = DECREASE_VISIBILITY;

		}  else { // IF NOT ENOUGH YESES, THEN SAME AS BEFORE
			Resp.Action = CONSTANT_VISIBILITY;
         } // end else

		// DOES THIS REPRESENT A REVERSAL?
		if( (LastChange == NO_REVERSAL) ||    // CASE 1:  IN CASE OF START-UP
			(Resp.Action == CONSTANT_VISIBILITY) || // CASE 2:  NOT ENOUGH YESES YET
			(LastChange == Resp.Action)  ) {        // CASE 3:  CHANGE IS SAME AS LAST TIME
			Resp.Reversal = NO_REVERSAL;
		} else {
			Resp.Reversal = REVERSAL;
			TotalReversals++;
         } // end if

      // SAVE THE CURRENT RESPONSE FOR THE NEXT COMPARISON
		if(Resp.Action != CONSTANT_VISIBILITY) LastChange = Resp.Action;

      // SEE IF THE STAIR CASE IS COMPLETE
		if(TotalReversals == RequiredReversals)
			Resp.Action = STAIRCASE_COMPLETE;

		// RETURN THE RESPONSE AS A RETURN VALUE
		return(Resp);

} // end next_action


//========================================================================
// 
//========================================================================
void fixed_staircase_experiment::reset(void) {

	SCMgr.reset();  // RESET THE STAIR CASE MANAGER

	FOSMgr.clear_data(); // REMOVE FOS DATA

	Accumulator = 0;// ZERO THE ACCUMULATOR

	if(ReversalPoints) { // CLEAR THE STORED REVERSAL POINTS
		for(int I = 0; I < SCMgr.RequiredReversals; I++) ReversalPoints[I] = 0.0;
		} // end if
} // end reset


//=====================================================================
//  A SINGLE STAIRCASE EXPERIMENT
//=====================================================================
fixed_staircase_experiment::fixed_staircase_experiment(
		adjustable_stimulus IML_REF Stim,
		user_input IML_REF UserIn) :
				Stimulus(Stim),
				ThresholdVariable(Stim.threshold_variable()),
				Subject(UserIn),
				ReversalPoints(NULL) {

	// SET SOME DEFAULT VALUES
	init_stepsizes(1.0, -1.0, DB_STEP);

	configure(10, 3, 6);

	reset();
} // end fixed_staircase_experiment constructor



//=====================================================================
// 
//=====================================================================
void fixed_staircase_experiment::init_stepsizes(IML_FLOAT IncV, IML_FLOAT DecV,
											fixed_step_type ChangeInDB) {
	IncVisibility = IncV;
	DecVisibility = DecV;
	DBFlag = ChangeInDB;
} // init_stepsizes


//=====================================================================
// 
//=====================================================================
void fixed_staircase_experiment::configure(
						IML_UBYTE Reversals, IML_UBYTE Yeses, IML_UBYTE AverageCnt) {


	SCMgr.configure(Reversals,Yeses);

	NAveragePts = AverageCnt;

	if(ReversalPoints) delete [] ReversalPoints;

	ReversalPoints = new IML_DOUBLE(Reversals);
	if(!ReversalPoints) {
		iml_error("fixed_staircase_experiment",
						"Could not allocate reversal point memory! Better QUIT!!!");
      } // end if
} // end configure





//=====================================================================
// REPLACING VIRTUAL FUNCTION IN THE BASE CLASS;  MAKE A STEP IN THE STAIRCASE;
//=====================================================================
void fixed_staircase_experiment::make_a_step(IML_SINT InFlags) {

	// PRESENT STIMULUS HERE
	IML_USINT StimFlags = Stimulus.present(InFlags);

	// GET THE SUBJECT'S RESPONSE
	IML_LINT Answer = Subject.response(StimFlags);

//	if( Subject.repeat_requested() || Subject.abort_requested() ) return;

	// STORE THE FOS DATA
	FOSMgr.add_datum(ThresholdVariable, Answer);

	// FIND OUT THE NEXT ACTION TO TAKE;
	// Answer COMES BACK CONTAINING A REVERSAL FLAG
	staircase_response Response =  SCMgr.next_action(Answer);

	// IF WE DID A REVERSAL, THEN WE MIGHT WANT TO ACCUMULATE...
	if(Response.Reversal == REVERSAL) { // STORE/PROCESS REVERSAL VALUES

		// STORE REVERSAL POINT VALUE
		ReversalPoints[ SCMgr.reversals() - 1 ] = ThresholdVariable;

		if(SCMgr.reversals() > (SCMgr.RequiredReversals - NAveragePts) )
			Accumulator += ThresholdVariable;
		} // end if


	// USE A SWITCH TO PROCESS THE STAIRCASE ACTION
	switch( Response.Action ) {

		case STAIRCASE_COMPLETE:  break;		// STAIR CASE WILL END NOW...

		case CONSTANT_VISIBILITY: break;    // WAITING FOR ENOUGH YESES...

		case DECREASE_VISIBILITY:           // DECREMENT THE VISIBILITY
			if(DBFlag)	ThresholdVariable.inc_dBV(DecVisibility);
			else			ThresholdVariable = ThresholdVariable + DecVisibility;
			break;

		case INCREASE_VISIBILITY:           // INCREMENT THE VISIBILITY
			if(DBFlag)	ThresholdVariable.inc_dBV(IncVisibility);
			else			ThresholdVariable = ThresholdVariable + IncVisibility;
			break;  

		default: iml_message("Experiment",	// IN CASE OF ERROR
						"Hey, got a strange output from the staircase manager");
      	break;
		} // end switch

} // end make_a_step


