//=====================================================================
// THESE ARE CLASSES INTENDED TO MAKE DESIGNING AND CONTROLLING
// EXPERIMENTS EASIER. :-)
//=====================================================================

#ifndef __IML_UDSC_H__
#define __IML_UDSC_H__

#include "iml_exps.h"


//=====================================================================
// THIS CLASS IS INTENDED TO MANAGE THE UPS AND DOWNS OF PSYCHOPHYSICAL STAIRCASE
// PROCEDRUES. :-)  IT IS ASSUMED THAT THE STAIRCASE IS ON A SINGLE PARAMETER, AND
// THAT WE ARE ADJUSTING IT ACCORDING AN 'N-ANSWERS-CORRECT' STRATEGY.  N-ANSWERS
// CORRECT CHANGES THE PARAMETER IN ONE DIRECTION WHILE ANY OTHER SEQUENCY CHANGES
// THE ANSWER IN THE OTHER DIRECTION.
//=====================================================================

#define	STAIRCASE_COMPLETE		0x0001
#define	INCREASE_VISIBILITY		0x0002
#define	DECREASE_VISIBILITY		0x0004
#define	CONSTANT_VISIBILITY		0x0008
#define	REVERSAL						0x0010
#define	NO_REVERSAL					0x0020


class staircase_response {
	public:
		IML_USINT Action;
		IML_USINT Reversal;
}; // end class


class staircase_manager {

	friend class fixed_staircase_experiment;

	protected:
   	IML_USINT LastChange;
   	
		IML_UBYTE RequiredReversals;
		IML_UBYTE TotalReversals;

		IML_UBYTE RequiredYeses;
		IML_UBYTE TotalYeses;

	public:

		// SINGLE CONSTRUCTOR
		staircase_manager(void) {
			configure(10, 3);    // DEFAULT TO 10 REVERSALS AND 3 YESES (79% correct)
			reset();
			};

      // CONFIGURE STAIRCASE PARAMETERS
		void configure(IML_UBYTE RevCnt, IML_UBYTE YesCnt) {
			RequiredReversals = RevCnt;
			RequiredYeses		= YesCnt;
			};


		// RESET COUNTS TO RESTART THE STAIRCASE PROCEDURE
		void reset(void) { TotalReversals = TotalYeses = 0; LastChange = NO_REVERSAL; };


		// RETURN REVERSAL COUNT
		IML_UBYTE reversals(void) { return(TotalReversals); };

		// BOOLEAN TEST OF COMPLETION
		IML_BYTE staircase_complete(void) { return(TotalReversals == RequiredReversals); };


		// GIVE SUBJECT'S ANSWER TO STAIRCASE MANAGER AND ASK FOR NEXT ACTION;  SEE ENUM ABOVE;
		staircase_response next_action(IML_SINT YesCorrectTrue);

}; // end class



//=====================================================================
//  A SINGLE STAIRCASE EXPERIMENT
//=====================================================================
typedef enum {LINEAR_STEP, DB_STEP} fixed_step_type;

class fixed_staircase_experiment : public staircase_experiment {

	protected:
		// A STIMULUS OBJECT SHOULD USE THE adjustable_parameter TYPE FOR ALL
		// STIMULUS PARAMETERS THAT MIGHT BE USED FOR MEASURING A THRESHOLD.
		adjustable_parameter IML_REF ThresholdVariable;

		// SET TO TRUE IF VARIATION IS IN DB RATHER THAN LINEAR INCREMENTS
		fixed_step_type		DBFlag;

		// STEP SIZES FOR INCEASING OR DECREASING VISIBILITY
		IML_FLOAT	IncVisibility;
		IML_FLOAT	DecVisibility;


	public:
		// REFERENCE TO POINT TO STIMULUS OBJECT
		adjustable_stimulus IML_REF Stimulus;

      // REFERENCE TO POINT TO SUBJECT RESPONSE-GETTER
      user_input IML_REF Subject;

		// WE WANT TO KEEP TRACK OF 'FREQUENCY OF SEEING'
		// FOR EACH THRESHOLD PARAMETER VALUE
		fos_data_manager FOSMgr;

		// ARRAY PTR FOR STORING REVERSAL POINTS
		IML_DOUBLE	IML_PTR ReversalPoints;

      // ACCUMULATOR FOR AVERAGING REVERSAL POINTS
		IML_DOUBLE	Accumulator;

      // AVERAGE THE FINAL 'N' POINTS
		IML_UBYTE	NAveragePts;

		// MANAGE THE STAIRCASE REVERSALS, DIRECTIONS, ETC.
		staircase_manager SCMgr;

		// CONSTRUCTOR
		fixed_staircase_experiment(
					adjustable_stimulus IML_REF Stim, user_input IML_REF UserIn);

		void init_stepsizes(IML_FLOAT IncV, IML_FLOAT DecV,
											fixed_step_type ChangeInDB = DB_STEP);

		void configure(IML_UBYTE Reversals, IML_UBYTE Yeses, IML_UBYTE AverageCnt);

		void reset(void);

      // TEST FOR STAIRCASE CMOMPLTE;  REPLACING VIRTUAL FUNCTION OF BASE CLASS
      IML_BYTE staircase_complete() { return(SCMgr.staircase_complete()); };


		IML_BYTE abort_requested(void) {return(Subject.abort_requested()); };

		// REPLACING VIRTUAL FUNCTION IN THE BASE CLASS;  MAKE A STEP IN THE STAIRCASE;
		void make_a_step(IML_SINT InFlags = 0);

		// RETURN THE PERCENT CORRECT TRACKING LEVEL FOR THE GIVEN NUMBER OF YESES
		IML_FLOAT tracking_level(void) {
			return(0.0); // UNTIL I GET A GENERAL METHOD FOR ASSYMETRIC INC/DEC...
			//return( iml_pow(2.0, -1.0/((IML_DOUBLE)SCMgr.RequiredYeses)) );
			};

}; // end class



#endif  // __IML_UDSC_H__


