
/*
 * simple test program
 */
/* $Id: test.c,v 1.6 1997/04/24 02:40:05 fftw Exp $ */
#include <fftw.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#define drand() (((double) rand()) / 31415)

void check_leaks(void)
{
     extern int fftw_malloc_cnt, 
	  fftw_node_cnt, fftw_plan_cnt, fftw_twiddle_size;
     if (fftw_malloc_cnt || 
	 fftw_node_cnt || fftw_plan_cnt || fftw_twiddle_size) {
     printf("MEMORY LEAK!!! fftw_malloc = %d node=%d plan=%d twiddle=%d\n", 
	    fftw_malloc_cnt,
	    fftw_node_cnt, fftw_plan_cnt, fftw_twiddle_size);
     }
}

void test_correctness(int n, int istride, int ostride, int howmany, fftw_direction dir)
{
     FFTW_COMPLEX *in1, *in2, *out1, *out2;
     fftw_plan plan;
     int i, j;
     double error, a;

     printf("testing n = %d  istride = %d   ostride = %d  howmany = %d dir = %s\n",
	    n, istride, ostride, howmany, dir == FFTW_FORWARD ? "direct" : "inverse");

     in1 = fftw_malloc(istride * n * sizeof(FFTW_COMPLEX) * howmany);
     in2 = fftw_malloc(n * sizeof(FFTW_COMPLEX) * howmany);
     out1 = fftw_malloc(ostride * n * sizeof(FFTW_COMPLEX) * howmany);
     out2 = fftw_malloc(n * sizeof(FFTW_COMPLEX) * howmany);

     /* generate random inputs */
     for (i = 0; i < n * howmany; ++i) {
	  c_re(in1[i * istride]) = c_re(in2[i]) = drand();
	  c_im(in1[i * istride]) = c_im(in2[i]) = drand();
     }

     /* 
      * fill in other positions of the array, to make sure that
      * fftw doesn't overwrite them 
      */
     for (j = 1; j < istride; ++j)
	  for (i = 0; i < n * howmany; ++i) {
	       c_re(in1[i * istride + j]) = i * istride + j;
	       c_im(in1[i * istride + j]) = i * istride - j;
	  }

     for (j = 1; j < ostride; ++j)
	  for (i = 0; i < n * howmany; ++i) {
	       c_re(out1[i * ostride + j]) = -i * ostride + j;
	       c_im(out1[i * ostride + j]) = -i * ostride - j;
	  }

     plan = fftw_create_plan(n, dir, FFTW_ESTIMATE);
     if (plan == NULL)
	  fftw_die("Can't create plan\n");

     fftw_print_plan(plan);

     /* fft-ize */
     fftw(plan, howmany, in1, istride, n * istride, out1, ostride,
	  n * ostride);

     fftw_destroy_plan(plan);

     /* check for overwriting */
     for (j = 1; j < istride; ++j)
	  for (i = 0; i < n * howmany; ++i)
	       if (c_re(in1[i * istride + j]) != i * istride + j ||
		   c_im(in1[i * istride + j]) != i * istride - j) {
		    printf("INNOCENT INPUT HAS BEEN OVERWRITTEN !\n");
		    exit(-1);
	       }
     for (j = 1; j < ostride; ++j)
	  for (i = 0; i < n * howmany; ++i)
	       if (c_re(out1[i * ostride + j]) != -i * ostride + j ||
		   c_im(out1[i * ostride + j]) != -i * ostride - j) {
		    printf("INNOCENT OUTPUT HAS BEEN OVERWRITTEN !\n");
		    exit(-1);
	       }

     for (i = 0; i < howmany; ++i) {
	  if (dir == FFTW_FORWARD) 
	       fftw_naive(n, in2 + n * i, out2 + n * i);
	  else
	       fftwi_naive(n, in2 + n * i, out2 + n * i);
     }

     /* compute the relative error */
     error = 0.0;
     for (i = 0; i < n * howmany; ++i) {
	  a = sqrt((c_re(out1[i * ostride]) - c_re(out2[i])) *
		   (c_re(out1[i * ostride]) - c_re(out2[i])) +
		   (c_im(out1[i * ostride]) - c_im(out2[i])) *
		   (c_im(out1[i * ostride]) - c_im(out2[i])));
	  a /= sqrt(c_re(out2[i]) * c_re(out2[i]) + c_im(out2[i])
		    * c_im(out2[i]));
	  if (a > error)
	       error = a;
     }
     if (error > 1e-3) {
	  printf("n=%d error=%e\n", n, error);
	  printf("SHOULD BE:\n");
	  for (i = 0; i < n; ++i)
	       printf("%f + %fi\n", c_re(out2[i]), c_im(out2[i]));
	  printf("INSTEAD OF:\n");
	  for (i = 0; i < n; ++i)
	       printf("%f + %fi\n", c_re(out1[i * ostride]), c_im(out1[i * ostride]));
	  exit(-1);
     } else {
	  printf("OK\n");
     }

     fftw_free(in1);  
     fftw_free(in2);
     fftw_free(out1);
     fftw_free(out2);
     check_leaks();
}

void test_correctness_inplace(int n, int istride,int howmany, fftw_direction dir)
{
     FFTW_COMPLEX *in1, *in2, *out2;
     fftw_plan plan;
     int i, j;
     double error, a;

     printf("testing in place n = %d  istride = %d    howmany = %d dir = %s\n",
	    n, istride,  howmany, dir == FFTW_FORWARD ? "direct" : "inverse");

     in1 = fftw_malloc(istride * n * sizeof(FFTW_COMPLEX) * howmany);
     in2 = fftw_malloc(n * sizeof(FFTW_COMPLEX) * howmany);
     out2 = fftw_malloc(n * sizeof(FFTW_COMPLEX) * howmany);

     /* generate random inputs */
     for (i = 0; i < n * howmany; ++i) {
	  c_re(in1[i * istride]) = c_re(in2[i]) = drand();
	  c_im(in1[i * istride]) = c_im(in2[i]) = drand();
     }

     /* 
      * fill in other positions of the array, to make sure that
      * fftw doesn't overwrite them 
      */
     for (j = 1; j < istride; ++j)
	  for (i = 0; i < n * howmany; ++i) {
	       c_re(in1[i * istride + j]) = i * istride + j;
	       c_im(in1[i * istride + j]) = i * istride - j;
	  }

     plan = fftw_create_plan(n, dir, FFTW_ESTIMATE | FFTW_IN_PLACE);
     if (plan == NULL)
	  fftw_die("Can't create plan\n");

     fftw_print_plan(plan);

     /* fft-ize */
     fftw(plan, howmany, in1, istride, n * istride, (FFTW_COMPLEX *)0, 0,
	  0);

     fftw_destroy_plan(plan);

     /* check for overwriting */
     for (j = 1; j < istride; ++j)
	  for (i = 0; i < n * howmany; ++i)
	       if (c_re(in1[i * istride + j]) != i * istride + j ||
		   c_im(in1[i * istride + j]) != i * istride - j) {
		    printf("INNOCENT INPUT HAS BEEN OVERWRITTEN !\n");
		    exit(-1);
	       }

     for (i = 0; i < howmany; ++i) {
	  if (dir == FFTW_FORWARD) 
	       fftw_naive(n, in2 + n * i, out2 + n * i);
	  else
	       fftwi_naive(n, in2 + n * i, out2 + n * i);
     }

     /* compute the relative error */
     error = 0.0;
     for (i = 0; i < n * howmany; ++i) {
	  a = sqrt((c_re(in1[i * istride]) - c_re(out2[i])) *
		   (c_re(in1[i * istride]) - c_re(out2[i])) +
		   (c_im(in1[i * istride]) - c_im(out2[i])) *
		   (c_im(in1[i * istride]) - c_im(out2[i])));
	  a /= sqrt(c_re(out2[i]) * c_re(out2[i]) + c_im(out2[i])
		    * c_im(out2[i]));
	  if (a > error)
	       error = a;
     }
     if (error > 1e-3) {
	  printf("n=%d error=%e\n", n, error);
	  printf("SHOULD BE:\n");
	  for (i = 0; i < n; ++i)
	       printf("%f + %fi\n", c_re(out2[i]), c_im(out2[i]));
	  printf("INSTEAD OF:\n");
	  for (i = 0; i < n; ++i)
	       printf("%f + %fi\n", c_re(in1[i * istride]), c_im(in1[i * istride]));
	  exit(-1);
     } else {
	  printf("OK\n");
     }

     fftw_free(in1);  
     fftw_free(in2);
     fftw_free(out2);

     check_leaks();
}

#define ITERATIONS 25000000

void test_speed(int n)
{
     FFTW_COMPLEX *in, *out;
     fftw_plan plan;
     int i, iter;
     double t;
     fftw_time begin, end;

     if (n == 1)
	  iter = ITERATIONS;
     else
	  iter = (int) ((double) ITERATIONS /
			((double) n * log(2.0 * (double) n)));

     if (iter == 0)
	  iter = 1;

     in = fftw_malloc(n * sizeof(FFTW_COMPLEX));
     out = fftw_malloc(n * sizeof(FFTW_COMPLEX));

     plan = fftw_create_plan(n, FFTW_FORWARD, FFTW_MEASURE);
     if (plan == NULL)
	  fftw_die("Can't create plan\n");

     fftw_print_plan(plan);

     /* fft-ize */
     printf("\nRunning out of place FFT's for array size N = %d\n", n);
     printf("Number of iterations = %d\n", iter);

     begin = fftw_get_time();
     for (i = 0; i < iter; ++i) {
	  int j;

	  for (j = 0; j < n; ++j) {
	       c_re(in[j]) = 1.0;
	       c_im(in[j]) = 32.432;
	  }

	  fftw(plan, 1, in, 1, 0, out, 1, 0);
     }

     end = fftw_get_time();

     fftw_destroy_plan(plan);

     t = fftw_time_to_sec(end - begin);
     printf("iterations took %f s\n", t);

     t = t / (double) iter;

     printf("one fft took %e s (%e s/point)\n", t, t / n);
     printf("normalized time = %e\n",
	    t / ((double) n * log((double) n)));

     free(in);
     free(out);


     in = fftw_malloc(n * sizeof(FFTW_COMPLEX));

     plan = fftw_create_plan(n, FFTW_FORWARD, FFTW_MEASURE | FFTW_IN_PLACE);
     if (plan == NULL)
	  fftw_die("Can't create plan\n");

     fftw_print_plan(plan);

     /* fft-ize */
     printf("\nRunning in-place FFT's for array size N = %d\n", n);
     printf("Number of iterations = %d\n", iter);

     begin = fftw_get_time();
     for (i = 0; i < iter; ++i) {
	  int j;

	  for (j = 0; j < n; ++j) {
	       c_re(in[j]) = 1.0;
	       c_im(in[j]) = 32.432;
	  }

	  fftw(plan, 1, in, 1, 0, (FFTW_COMPLEX *) 0, 1, 0);
     }


     end = fftw_get_time();

     fftw_destroy_plan(plan);

     t = fftw_time_to_sec(end - begin);
     printf("iterations took %f s\n", t);

     t = t / (double) iter;

     printf("one fft took %e s (%e s/point)\n", t, t / n);
     printf("normalized time = %e\n",
	    t / ((double) n * log((double) n)));

     free(in);
}

#define K 100
void test_planner(void)
{
     /*
      * create and destroy many plans, at random.  Check the
      * garbage-collecting allocator of twiddle factors 
      */
     int i;
     int r, s;
     fftw_plan p[K];

     for (i = 0; i < K; ++i) {
	  p[i] = (fftw_plan) 0;
     }

     for (i = 0; i < K * K; ++i) {
	  r = rand();
	  if (r < 0)
	       r = -r;
	  r = r % K;

	  do {
	       s = rand();
	       if (s < 0)
		    s = -s;
	       s = s % 10000;
	  } while (s == 0);

	  if (p[r]) 
	       fftw_destroy_plan(p[r]);

	  p[r] = fftw_create_plan(s, FFTW_FORWARD, FFTW_ESTIMATE);

	  if (i % (K*K/20) == 0)
	       printf("test planner --- so far so good\n");
     }

     for (i = 0; i < K; ++i) {
	  if (p[i]) 
	       fftw_destroy_plan(p[i]);
     }

     check_leaks();
}

int main(int argc, char *argv[])
{
     int n;

     if (argc <= 1) {
	usage:
	  printf("Usage: fft [-c | -s] n  (correctness or speed)\n");
	  printf("   or: fft -t           (test all (goes on forever))\n");
	  printf("   or: fft -p           (exercise the planner)\n");
	  return 1;
     }
     if (strcmp(argv[1], "-t") == 0) {
	  int n;
	  int is, os, h;

	  for (n = 1; n < 100000; ++n)
	       for (is = 1; is <= 4; ++is) {
			 for (h = 1; h <= 4; ++h) {
			      for (os = 1; os <= 4; ++os) {
				   test_correctness(n, is, os, h, FFTW_FORWARD);
				   test_correctness(n, is, os, h, FFTW_BACKWARD);
			      }
			      test_correctness_inplace(n, is, h, FFTW_FORWARD);
			      test_correctness_inplace(n, is, h, FFTW_BACKWARD);
			 }
	       }
	  return 0;
     }


     if (strcmp(argv[1], "-p") == 0) {
	  test_planner();
	  return 0;
     }

     if (argc <= 2)
	  goto usage;

     n = atoi(argv[2]);

     if (strcmp(argv[1], "-c") == 0) {
	  int is, os, h;
	  for (is = 1; is <= 4; ++is) {
	       for (h = 1; h <= 4; ++h) {
		    for (os = 1; os <= 4; ++os) {
			 test_correctness(n, is, os, h, FFTW_FORWARD);
			 test_correctness(n, is, os, h, FFTW_BACKWARD);
		    }
		    test_correctness_inplace(n, is, h, FFTW_FORWARD);
		    test_correctness_inplace(n, is, h, FFTW_BACKWARD);
	       }
	  }
     } else
	  test_speed(n);

     return 0;
}
